//  -*- Mode: C++ -*-

// explanations.hh

/*
 +--------------------------------------------------------------------------+
 |                                                                          |
 |  COPYRIGHT (C) UNIVERSITY OF SOUTHERN CALIFORNIA, 1997-2003              |
 |  University of Southern California, Information Sciences Institute       |
 |  4676 Admiralty Way                                                      |
 |  Marina Del Rey, California 90292                                        |
 |                                                                          |
 |  This software was developed under the terms and conditions of Contract  |
 |  No. N00014-94-C-0245 between the Defense Advanced Research Projects     |
 |  Agency and the University of Southern California, Information Sciences  | 
 |  Institute.  Use and distribution of this software is further subject    |
 |  to the provisions of that contract and any other agreements developed   |
 |  between the user of the software and the University of Southern         |
 |  California, Information Sciences Institute.  It is supplied "AS IS",    |
 |  without any warranties of any kind.  It is furnished only on the basis  |
 |  that any party who receives it indemnifies and holds harmless the       |
 |  parties who furnish and originate it against any claims, demands, or    |
 |  liabilities connected with using it, furnishing it to others or         |
 |  providing it to a third party.  THIS NOTICE MUST NOT BE REMOVED FROM    |
 |  THE SOFTWARE, AND IN THE EVENT THAT THE SOFTWARE IS DIVIDED, IT SHOULD  |
 |  BE ATTACHED TO EVERY PART.                                              |
 |                                                                          |
 +--------------------------------------------------------------------------+
*/


namespace logic {
  using namespace stella;

// Class definitions:
class ExplainException : public LogicException {
public:
  ExplainException(const std::string& msg) : LogicException(msg) {
}

};

class ExplainNoQueryException : public ExplainException {
public:
  ExplainNoQueryException(const std::string& msg) : ExplainException(msg) {
}

};

class ExplainNoSolutionException : public ExplainException {
public:
  ExplainNoSolutionException(const std::string& msg) : ExplainException(msg) {
}

};

class ExplainNoMoreSolutionsException : public ExplainException {
public:
  ExplainNoMoreSolutionsException(const std::string& msg) : ExplainException(msg) {
}

};

class ExplainNotEnabledException : public ExplainException {
public:
  ExplainNotEnabledException(const std::string& msg) : ExplainException(msg) {
}

};

class ExplainNoSuchLabelException : public ExplainException {
public:
  ExplainNoSuchLabelException(const std::string& msg) : ExplainException(msg) {
}

};

class ExplainQueryTrueException : public ExplainException {
public:
  ExplainQueryTrueException(const std::string& msg) : ExplainException(msg) {
}

};

class ExplanationInfo : public StandardObject {
public:
  char* label;
  int depth;
  boolean explainedP;
public:
  virtual Surrogate* primaryType();
};


// Global declarations:
extern DECLARE_STELLA_SPECIAL(oEXPLANATION_FORMATo, Keyword* );
extern DECLARE_STELLA_SPECIAL(oEXPLANATION_STYLEo, Keyword* );
extern DECLARE_STELLA_SPECIAL(oEXPLANATION_AUDIENCEo, Keyword* );
extern int oDEFAULT_EXPLANATION_DEPTHo;
extern char* oEXPLANATION_TAB_STRINGo;
extern int oMAX_INLINE_LABEL_LENGTHo;
extern char* oEXPLANATION_ASSERTION_MARKERo;
extern char* oEXPLANATION_FAILURE_MARKERo;
extern char* oEXPLANATION_CUTOFF_MARKERo;
extern char* oEXPLANATION_INFERENCE_MARKERo;
extern DECLARE_STELLA_SPECIAL(oEXPLANATION_VOCABULARYo, HashTable* );
extern KeyValueList* oEXPLANATION_VOCABULARIESo;
extern DECLARE_STELLA_SPECIAL(oCURRENTJUSTIFICATIONo, Justification* );
extern DECLARE_STELLA_SPECIAL(oMOST_RECENT_EXPLANATION_MAPPINGo, KeyValueList* );

// Function signatures:
ExplainException* newExplainException(char* message);
ExplainNoQueryException* newExplainNoQueryException(char* message);
ExplainNoSolutionException* newExplainNoSolutionException(char* message);
ExplainNoMoreSolutionsException* newExplainNoMoreSolutionsException(char* message);
ExplainNotEnabledException* newExplainNotEnabledException(char* message);
ExplainNoSuchLabelException* newExplainNoSuchLabelException(char* message);
ExplainQueryTrueException* newExplainQueryTrueException(char* message);
char* explanationTruthMarker(Justification* justification);
void defineExplanationPhrase(Keyword* phrasekey, Keyword* audience, char* phrase, int modifiers, ...);
char* lookupExplanationPhrase(Keyword* phrasekey, Cons* modifiers, Keyword* audience);
ExplanationInfo* newExplanationInfo();
Object* accessExplanationInfoSlotValue(ExplanationInfo* self, Symbol* slotname, Object* value, boolean setvalueP);
ExplanationInfo* registerJustification(Justification* self, KeyValueList* mapping);
ExplanationInfo* getExplanationInfo(Justification* self, KeyValueList* mapping, boolean createP);
ExplanationInfo* explanationInfo(Justification* self, KeyValueList* mapping);
Justification* lookupJustification(KeyValueList* mapping, char* label);
void resetMappingForSubexplanation(KeyValueList* mapping);
KeyValueList* printExplanation(Justification* self, OutputStream* stream, KeyValueList* mapping, int maxdepth, Keyword* audience);
void printExplanationSupport(Justification* self, OutputStream* stream, KeyValueList* mapping, int maxdepth, List* unexplained);
void printOneExplanation(Justification* self, OutputStream* stream, KeyValueList* mapping, int maxdepth, List* unexplained);
boolean printingJustificationP();
void printJustificationPropositionForFormat(Justification* self, OutputStream* stream, int indent);
void printJustificationProposition(Justification* self, OutputStream* stream, int indent);
void printExplanationLabel(OutputStream* stream, char* label, boolean headP);
void printExplanationHeader(Justification* self, OutputStream* stream, KeyValueList* mapping);
void printExplanationText(Justification* self, OutputStream* stream, KeyValueList* mapping);
boolean markAsExplicitAssertionP(Justification* self);
boolean markAsFailedGoalP(Justification* self);
boolean markAsCutoffGoalP(Justification* self);
boolean partiallyFollowsP(Justification* self);
char* makeRuleOriginExplanationPhrase(Justification* self);
KeyValueList* getExplanationSubstitution(Justification* self);
void printOneVariableSubstitution(OutputStream* stream, Object* var, Object* value);
void printExplanationSubstitution(Justification* self, OutputStream* stream, KeyValueList* mapping);
void printExplanationAntecedents(Justification* self, OutputStream* stream, KeyValueList* mapping, int maxdepth, List* unexplained);
Justification* visibleJustification(Justification* self);
List* visibleAntecedents(Justification* self);
void collectVisibleAntecedents(Justification* self, List* visibleantecedents);
void why(Cons* args);
void whyEvaluatorWrapper(Cons* arguments);
void explainWhy(char* label, Keyword* style, int maxdepth, OutputStream* stream);
Justification* getWhyJustification(char* label);
boolean commandOptionEqlP(Object* arg, char* option);
char* parseWhyArguments(Cons* args, Keyword*& _Return1, int& _Return2, boolean& _Return3);
void helpStartupExplanations1();
void helpStartupExplanations2();
void startupExplanations();

// Auxiliary global declarations:
extern Keyword* KWD_EXPLANATIONS_ASCII;
extern Keyword* KWD_EXPLANATIONS_BRIEF;
extern Keyword* KWD_EXPLANATIONS_TECHNICAL;
extern Keyword* KWD_EXPLANATIONS_ASSERTION;
extern Keyword* KWD_EXPLANATIONS_FAILURE;
extern Keyword* KWD_EXPLANATIONS_CUTOFF;
extern Keyword* KWD_EXPLANATIONS_INFERENCE;
extern Keyword* KWD_EXPLANATIONS_HTML;
extern Keyword* KWD_EXPLANATIONS_IMPLIES;
extern Symbol* SYM_EXPLANATIONS_LOGIC_EXPLANATION_VOCABULARY;
extern Keyword* KWD_EXPLANATIONS_UNKNOWN_RULE;
extern Keyword* KWD_EXPLANATIONS_FOLLOWS;
extern Keyword* KWD_EXPLANATIONS_LAY;
extern Keyword* KWD_EXPLANATIONS_PARTIAL;
extern Keyword* KWD_EXPLANATIONS_HOLDS;
extern Keyword* KWD_EXPLANATIONS_DEFINITION;
extern Keyword* KWD_EXPLANATIONS_FAILED;
extern Keyword* KWD_EXPLANATIONS_NOT_ASSERTED;
extern Keyword* KWD_EXPLANATIONS_NO_RULES;
extern Surrogate* SGT_EXPLANATIONS_LOGIC_EXPLANATION_INFO;
extern Symbol* SYM_EXPLANATIONS_STELLA_LABEL;
extern Symbol* SYM_EXPLANATIONS_LOGIC_DEPTH;
extern Symbol* SYM_EXPLANATIONS_LOGIC_EXPLAINEDp;
extern Symbol* SYM_EXPLANATIONS_LOGIC_EXPLANATION_MAPPING;
extern Keyword* KWD_EXPLANATIONS_XML;
extern Keyword* KWD_EXPLANATIONS_CYC_NL;
extern Keyword* KWD_EXPLANATIONS_KIF_ONTOSAURUS;
extern Keyword* KWD_EXPLANATIONS_JAVA_GUI_HTML;
extern Keyword* KWD_EXPLANATIONS_REALISTIC;
extern Symbol* SYM_EXPLANATIONS_STELLA_NOT;
extern Keyword* KWD_EXPLANATIONS_AMPLIFICATION;
extern Keyword* KWD_EXPLANATIONS_REVERSE;
extern Keyword* KWD_EXPLANATIONS_PRIMITIVE_STRATEGY;
extern Keyword* KWD_EXPLANATIONS_SCAN_PROPOSITIONS;
extern Symbol* SYM_EXPLANATIONS_LOGIC_MASTER_PROPOSITION;
extern Keyword* KWD_EXPLANATIONS_SCAN_COLLECTION;
extern Keyword* KWD_EXPLANATIONS_LOOKUP_ASSERTIONS;
extern Keyword* KWD_EXPLANATIONS_EQUIVALENCE;
extern Symbol* SYM_EXPLANATIONS_LOGIC_COMPLEMENT_DESCRIPTION;
extern Surrogate* SGT_EXPLANATIONS_LOGIC_NAMED_DESCRIPTION;
extern Surrogate* SGT_EXPLANATIONS_LOGIC_ALTERNATIVE_BINDINGS_SET;
extern Keyword* KWD_EXPLANATIONS_MODUS_PONENS;
extern Keyword* KWD_EXPLANATIONS_MODUS_TOLLENS;
extern Keyword* KWD_EXPLANATIONS_DISPROOF;
extern Keyword* KWD_EXPLANATIONS_GOAL_COMPLEMENT;
extern Keyword* KWD_EXPLANATIONS_PATTERN;
extern Keyword* KWD_EXPLANATIONS_AND_INTRODUCTION;
extern Symbol* SYM_EXPLANATIONS_LOGIC_JUSTIFICATION;
extern Surrogate* SGT_EXPLANATIONS_STELLA_GENERALIZED_SYMBOL;
extern Keyword* KWD_EXPLANATIONS_DIGIT;
extern Keyword* KWD_EXPLANATIONS_VERBOSE;
extern Keyword* KWD_EXPLANATIONS_DEPTH;
extern Symbol* SYM_EXPLANATIONS_LOGIC_STARTUP_EXPLANATIONS;
extern Symbol* SYM_EXPLANATIONS_STELLA_METHOD_STARTUP_CLASSNAME;


} // end of namespace logic
