//  -*- Mode: C++ -*-

// propositions.cc

/*
 +--------------------------------------------------------------------------+
 |                                                                          |
 |  COPYRIGHT (C) UNIVERSITY OF SOUTHERN CALIFORNIA, 1997-2003              |
 |  University of Southern California, Information Sciences Institute       |
 |  4676 Admiralty Way                                                      |
 |  Marina Del Rey, California 90292                                        |
 |                                                                          |
 |  This software was developed under the terms and conditions of Contract  |
 |  No. N00014-94-C-0245 between the Defense Advanced Research Projects     |
 |  Agency and the University of Southern California, Information Sciences  | 
 |  Institute.  Use and distribution of this software is further subject    |
 |  to the provisions of that contract and any other agreements developed   |
 |  between the user of the software and the University of Southern         |
 |  California, Information Sciences Institute.  It is supplied "AS IS",    |
 |  without any warranties of any kind.  It is furnished only on the basis  |
 |  that any party who receives it indemnifies and holds harmless the       |
 |  parties who furnish and originate it against any claims, demands, or    |
 |  liabilities connected with using it, furnishing it to others or         |
 |  providing it to a third party.  THIS NOTICE MUST NOT BE REMOVED FROM    |
 |  THE SOFTWARE, AND IN THE EVENT THAT THE SOFTWARE IS DIVIDED, IT SHOULD  |
 |  BE ATTACHED TO EVERY PART.                                              |
 |                                                                          |
 +--------------------------------------------------------------------------+
*/

#include "logic/logic-system.hh"

namespace logic {
  using namespace stella;

BacklinksIndex* newBacklinksIndex() {
  { BacklinksIndex* self = NULL;

    self = new BacklinksIndex();
    self->predicatePropositionsTable = NULL;
    self->dependentIsaPropositionsList = NULL;
    self->dependentPropositionsList = NULL;
    return (self);
  }
}

Surrogate* BacklinksIndex::primaryType() {
  { BacklinksIndex* self = this;

    return (SGT_PROPOSITIONS_LOGIC_BACKLINKS_INDEX);
  }
}

Object* accessBacklinksIndexSlotValue(BacklinksIndex* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_PROPOSITIONS_LOGIC_DEPENDENT_PROPOSITIONS_LIST) {
    if (setvalueP) {
      self->dependentPropositionsList = ((SequenceIndex*)(value));
    }
    else {
      value = self->dependentPropositionsList;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_DEPENDENT_ISA_PROPOSITIONS_LIST) {
    if (setvalueP) {
      self->dependentIsaPropositionsList = ((SequenceIndex*)(value));
    }
    else {
      value = self->dependentIsaPropositionsList;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_PREDICATE_PROPOSITIONS_TABLE) {
    if (setvalueP) {
      self->predicatePropositionsTable = ((HashTable*)(value));
    }
    else {
      value = self->predicatePropositionsTable;
    }
  }
  else {
    { OutputStringStream* stream000 = newOutputStringStream();

      *(stream000->nativeStream) << "`" << slotname << "'" << " is not a valid case option";
      throw *newStellaException(stream000->theStringReader());
    }
  }
  return (value);
}

LogicObject* newLogicObject() {
  { LogicObject* self = NULL;

    self = new LogicObject();
    self->dependentPropositionsIndex = NULL;
    self->dynamicSlots = newKeyValueList();
    self->homeContext = oCONTEXTo.get();
    self->surrogateValueInverse = NULL;
    self->variableValueInverse = NULL;
    logLogicObject(self);
    return (self);
  }
}

Surrogate* LogicObject::primaryType() {
  { LogicObject* self = this;

    return (SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT);
  }
}

List* LogicObject::dependentDescriptions_reader() {
  { LogicObject* self = this;

    { List* answer = ((List*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DEPENDENT_DESCRIPTIONS, NULL)));

      if (!((boolean)(answer))) {
        return (NIL_LIST);
      }
      else {
        return (answer);
      }
    }
  }
}

Cons* LogicObject::variableValueInverse_reader() {
  { LogicObject* self = this;

    { Cons* answer = ((Cons*)(accessInContext(self->variableValueInverse, self->homeContext, FALSE)));

      if (!((boolean)(answer))) {
        return (NIL);
      }
      else {
        return (answer);
      }
    }
  }
}

BooleanWrapper* LogicObject::badP_reader() {
  { LogicObject* self = this;

    { BooleanWrapper* answer = ((BooleanWrapper*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_STELLA_BADp, NULL)));

      if (!((boolean)(answer))) {
        return ((FALSE ? TRUE_WRAPPER : FALSE_WRAPPER));
      }
      else {
        return (answer);
      }
    }
  }
}

Object* accessLogicObjectSlotValue(LogicObject* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_PROPOSITIONS_LOGIC_DEPENDENT_DESCRIPTIONS) {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DEPENDENT_DESCRIPTIONS, ((List*)(value)), NULL);
    }
    else {
      value = self->dependentDescriptions_reader();
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_VARIABLE_VALUE_INVERSE) {
    if (setvalueP) {
      { LogicObject* object000 = self;
        Cons* value000 = ((Cons*)(value));
        Object* oldValue000 = object000->variableValueInverse;
        Object* newValue000 = updateInContext(oldValue000, value000, object000->homeContext, FALSE);

        if (!(((boolean)(oldValue000)) &&
            (oldValue000->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
          object000->variableValueInverse = newValue000;
        }
      }
    }
    else {
      value = self->variableValueInverse_reader();
    }
  }
  else if (slotname == SYM_PROPOSITIONS_STELLA_SURROGATE_VALUE_INVERSE) {
    if (setvalueP) {
      self->surrogateValueInverse = ((Surrogate*)(value));
    }
    else {
      value = self->surrogateValueInverse;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_OBJECT_STRINGIFIED_SOURCE) {
    if (setvalueP) {
      stringifiedSourceSetter(self, ((StringWrapper*)(value))->wrapperValue);
    }
    else {
      value = wrapString(stringifiedSource(self));
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_OBJECT_ORIGINATED_PROPOSITIONS) {
    if (setvalueP) {
      originatedPropositionsSetter(self, ((List*)(value)));
    }
    else {
      value = originatedPropositions(self);
    }
  }
  else {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, slotname, value, NULL);
    }
    else {
      value = self->dynamicSlots->lookup(slotname);
    }
  }
  return (value);
}

boolean LogicObject::deletedP() {
  { LogicObject* self = this;

    { BooleanWrapper* deletedP = ((BooleanWrapper*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_STELLA_DELETED_OBJECTp, NULL)));

      if (((boolean)(deletedP))) {
        return (coerceWrappedBooleanToBoolean(deletedP));
      }
      else {
        return (FALSE);
      }
    }
  }
}

boolean LogicObject::deletedPSetter(boolean value) {
  { LogicObject* self = this;

    setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_STELLA_DELETED_OBJECTp, (value ? TRUE_WRAPPER : FALSE_WRAPPER), NULL);
    return (value);
  }
}

void LogicObject::printObject(std::ostream* stream) {
  { LogicObject* self = this;

    printTopLevelObject(self, stream);
  }
}

LogicThing* newLogicThing() {
  { LogicThing* self = NULL;

    self = new LogicThing();
    self->dynamicSlots = newKeyValueList();
    self->surrogateValueInverse = NULL;
    return (self);
  }
}

Surrogate* LogicThing::primaryType() {
  { LogicThing* self = this;

    return (SGT_PROPOSITIONS_LOGIC_LOGIC_THING);
  }
}

Object* accessLogicThingSlotValue(LogicThing* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (TRUE) {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, slotname, value, NULL);
    }
    else {
      value = self->dynamicSlots->lookup(slotname);
    }
  }
  return (value);
}

void logLogicObject(LogicObject* self) {
  if (descriptionModeP() ||
      isaP(self, SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE)) {
    return;
  }
  if (!oLOADINGREGENERABLEOBJECTSpo.get()) {
    locallyConceivedInstances(oCONTEXTo.get())->push(self);
  }
}

Skolem* newSkolem() {
  { Skolem* self = NULL;

    self = new Skolem();
    self->dependentPropositionsIndex = NULL;
    self->dynamicSlots = newKeyValueList();
    self->homeContext = oCONTEXTo.get();
    self->surrogateValueInverse = NULL;
    self->variableValueInverse = NULL;
    self->definingProposition = NULL;
    self->variableValue = NULL;
    self->skolemName = NULL;
    self->skolemType = NULL;
    logLogicObject(self);
    return (self);
  }
}

Surrogate* Skolem::primaryType() {
  { Skolem* self = this;

    return (SGT_PROPOSITIONS_LOGIC_SKOLEM);
  }
}

Object* accessSkolemSlotValue(Skolem* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_PROPOSITIONS_LOGIC_SKOLEM_TYPE) {
    if (setvalueP) {
      self->skolemType = ((Surrogate*)(value));
    }
    else {
      value = self->skolemType;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_SKOLEM_NAME) {
    if (setvalueP) {
      self->skolemName = ((Symbol*)(value));
    }
    else {
      value = self->skolemName;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_VARIABLE_VALUE) {
    if (setvalueP) {
      { Skolem* object000 = self;
        Object* value000 = value;
        Object* oldValue000 = object000->variableValue;
        Object* newValue000 = updateInContext(oldValue000, value000, object000->homeContext, FALSE);

        if (!(((boolean)(oldValue000)) &&
            (oldValue000->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
          object000->variableValue = newValue000;
        }
      }
    }
    else {
      value = ((Object*)(accessInContext(self->variableValue, self->homeContext, FALSE)));
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_DEFINING_PROPOSITION) {
    if (setvalueP) {
      self->definingProposition = ((Proposition*)(value));
    }
    else {
      value = self->definingProposition;
    }
  }
  else {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, slotname, value, NULL);
    }
    else {
      value = self->dynamicSlots->lookup(slotname);
    }
  }
  return (value);
}

PatternVariable* newPatternVariable() {
  { PatternVariable* self = NULL;

    self = new PatternVariable();
    self->dependentPropositionsIndex = NULL;
    self->dynamicSlots = newKeyValueList();
    self->homeContext = oCONTEXTo.get();
    self->surrogateValueInverse = NULL;
    self->variableValueInverse = NULL;
    self->definingProposition = NULL;
    self->variableValue = NULL;
    self->skolemType = NULL;
    self->skolemName = NULL;
    self->boundToOffset = NULL_INTEGER;
    logLogicObject(self);
    return (self);
  }
}

Surrogate* PatternVariable::primaryType() {
  { PatternVariable* self = this;

    return (SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE);
  }
}

Object* accessPatternVariableSlotValue(PatternVariable* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_PROPOSITIONS_LOGIC_BOUND_TO_OFFSET) {
    if (setvalueP) {
      self->boundToOffset = ((IntegerWrapper*)(value))->wrapperValue;
    }
    else {
      value = wrapInteger(self->boundToOffset);
    }
  }
  else if (slotname == SYM_PROPOSITIONS_STELLA_VARIABLE_NAME) {
    if (setvalueP) {
      self->skolemName = ((Symbol*)(value));
    }
    else {
      value = self->skolemName;
    }
  }
  else {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, slotname, value, NULL);
    }
    else {
      value = self->dynamicSlots->lookup(slotname);
    }
  }
  return (value);
}

// Save space by structure-sharing zero-length variable vectors.
Vector* ZERO_VARIABLES_VECTOR = NULL;

ForwardChainingIndex* newForwardChainingIndex() {
  { ForwardChainingIndex* self = NULL;

    self = new ForwardChainingIndex();
    self->consequent = NULL;
    self->cacheId = NULL;
    self->inputBindings = NULL;
    self->queryBody = NULL;
    self->ioVariables = NULL;
    self->masterRule = NULL;
    self->forwardGoal = NULL;
    return (self);
  }
}

Surrogate* ForwardChainingIndex::primaryType() {
  { ForwardChainingIndex* self = this;

    return (SGT_PROPOSITIONS_LOGIC_FORWARD_CHAINING_INDEX);
  }
}

Object* accessForwardChainingIndexSlotValue(ForwardChainingIndex* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_PROPOSITIONS_LOGIC_FORWARD_GOAL) {
    if (setvalueP) {
      self->forwardGoal = ((Proposition*)(value));
    }
    else {
      value = self->forwardGoal;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_MASTER_RULE) {
    if (setvalueP) {
      self->masterRule = ((Proposition*)(value));
    }
    else {
      value = self->masterRule;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_IO_VARIABLES) {
    if (setvalueP) {
      self->ioVariables = ((Cons*)(value));
    }
    else {
      value = self->ioVariables;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_QUERY_BODY) {
    if (setvalueP) {
      self->queryBody = ((Cons*)(value));
    }
    else {
      value = self->queryBody;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_INPUT_BINDINGS) {
    if (setvalueP) {
      self->inputBindings = ((Cons*)(value));
    }
    else {
      value = self->inputBindings;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_CACHE_ID) {
    if (setvalueP) {
      self->cacheId = ((Symbol*)(value));
    }
    else {
      value = self->cacheId;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_CONSEQUENT) {
    if (setvalueP) {
      self->consequent = ((Description*)(value));
    }
    else {
      value = self->consequent;
    }
  }
  else {
    { OutputStringStream* stream000 = newOutputStringStream();

      *(stream000->nativeStream) << "`" << slotname << "'" << " is not a valid case option";
      throw *newStellaException(stream000->theStringReader());
    }
  }
  return (value);
}

Description* newDescription() {
  { Description* self = NULL;

    self = new Description();
    self->dependentPropositionsIndex = NULL;
    self->dynamicSlots = newKeyValueList();
    self->homeContext = oCONTEXTo.get();
    self->surrogateValueInverse = NULL;
    self->variableValueInverse = NULL;
    self->queryPatterns = NULL;
    self->internalVariables = ZERO_VARIABLES_VECTOR;
    self->ioVariables = NULL;
    self->proposition = NULL;
    logLogicObject(self);
    return (self);
  }
}

Surrogate* Description::primaryType() {
  { Description* self = this;

    return (SGT_PROPOSITIONS_LOGIC_DESCRIPTION);
  }
}

Relation* Description::nativeRelation() {
  { Description* self = this;

    return (NULL);
  }
}

List* Description::forwardChainingIndices_reader() {
  { Description* self = this;

    { List* answer = ((List*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_FORWARD_CHAINING_INDICES, NULL)));

      if (!((boolean)(answer))) {
        return (NIL_LIST);
      }
      else {
        return (answer);
      }
    }
  }
}

BooleanWrapper* Description::deferredContrapositivesP_reader() {
  { Description* self = this;

    { BooleanWrapper* answer = ((BooleanWrapper*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DEFERRED_CONTRAPOSITIVESp, NULL)));

      if (!((boolean)(answer))) {
        return ((FALSE ? TRUE_WRAPPER : FALSE_WRAPPER));
      }
      else {
        return (answer);
      }
    }
  }
}

Object* accessDescriptionSlotValue(Description* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_PROPOSITIONS_LOGIC_PROPOSITION) {
    if (setvalueP) {
      self->proposition = ((Proposition*)(value));
    }
    else {
      value = self->proposition;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_IO_VARIABLES) {
    if (setvalueP) {
      self->ioVariables = ((Vector*)(value));
    }
    else {
      value = self->ioVariables;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_INTERNAL_VARIABLES) {
    if (setvalueP) {
      self->internalVariables = ((Vector*)(value));
    }
    else {
      value = self->internalVariables;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_QUERY_PATTERNS) {
    if (setvalueP) {
      self->queryPatterns = ((KeyValueList*)(value));
    }
    else {
      value = self->queryPatterns;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_FORWARD_CHAINING_INDICES) {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_FORWARD_CHAINING_INDICES, ((List*)(value)), NULL);
    }
    else {
      value = self->forwardChainingIndices_reader();
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_DEFERRED_CONTRAPOSITIVESp) {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DEFERRED_CONTRAPOSITIVESp, ((BooleanWrapper*)(value)), NULL);
    }
    else {
      value = self->deferredContrapositivesP_reader();
    }
  }
  else {
    if (slotname == SYM_PROPOSITIONS_LOGIC_RELATIVE_COMPLEMENT) {
      slotname = SYM_PROPOSITIONS_LOGIC_COMPLEMENT_DESCRIPTION;
    }
    else {
    }
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, slotname, value, NULL);
    }
    else {
      value = self->dynamicSlots->lookup(slotname);
    }
  }
  return (value);
}

NamedDescription* newNamedDescription() {
  { NamedDescription* self = NULL;

    self = new NamedDescription();
    self->dependentPropositionsIndex = NULL;
    self->dynamicSlots = newKeyValueList();
    self->homeContext = oCONTEXTo.get();
    self->surrogateValueInverse = NULL;
    self->variableValueInverse = NULL;
    self->queryPatterns = NULL;
    self->internalVariables = ZERO_VARIABLES_VECTOR;
    self->ioVariables = NULL;
    self->proposition = NULL;
    self->objectOriginatedPropositions = NULL;
    self->objectStringifiedSource = NULL;
    self->extension = NULL;
    self->ioVariableTypes = NULL;
    self->ioVariableNames = NULL;
    logLogicObject(self);
    return (self);
  }
}

Surrogate* NamedDescription::primaryType() {
  { NamedDescription* self = this;

    return (SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION);
  }
}

Relation* NamedDescription::nativeRelation() {
  { NamedDescription* self = this;

    return (((Relation*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_NATIVE_RELATION_BACK_POINTER, NULL))));
  }
}

void NamedDescription::nativeRelationSetter(Relation* value) {
  { NamedDescription* self = this;

    setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_NATIVE_RELATION_BACK_POINTER, value, NULL);
  }
}

Object* accessNamedDescriptionSlotValue(NamedDescription* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_PROPOSITIONS_LOGIC_IO_VARIABLE_NAMES) {
    if (setvalueP) {
      self->ioVariableNames = ((List*)(value));
    }
    else {
      value = self->ioVariableNames;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_IO_VARIABLE_TYPES) {
    if (setvalueP) {
      self->ioVariableTypes = ((List*)(value));
    }
    else {
      value = self->ioVariableTypes;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_STELLA_EXTENSION) {
    if (setvalueP) {
      self->extension = ((SequenceIndex*)(value));
    }
    else {
      value = self->extension;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_OBJECT_STRINGIFIED_SOURCE) {
    if (setvalueP) {
      stringifiedSourceSetter(self, ((StringWrapper*)(value))->wrapperValue);
    }
    else {
      value = wrapString(stringifiedSource(self));
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_OBJECT_ORIGINATED_PROPOSITIONS) {
    if (setvalueP) {
      originatedPropositionsSetter(self, ((List*)(value)));
    }
    else {
      value = originatedPropositions(self);
    }
  }
  else {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, slotname, value, NULL);
    }
    else {
      value = self->dynamicSlots->lookup(slotname);
    }
  }
  return (value);
}

TruthValue* newTruthValue() {
  { TruthValue* self = NULL;

    self = new TruthValue();
    self->dependentPropositionsIndex = NULL;
    self->dynamicSlots = newKeyValueList();
    self->homeContext = oCONTEXTo.get();
    self->surrogateValueInverse = NULL;
    self->variableValueInverse = NULL;
    self->positiveScore = 0.0;
    self->strength = NULL;
    self->polarity = KWD_PROPOSITIONS_UNKNOWN;
    logLogicObject(self);
    return (self);
  }
}

Surrogate* TruthValue::primaryType() {
  { TruthValue* self = this;

    return (SGT_PROPOSITIONS_LOGIC_TRUTH_VALUE);
  }
}

Object* accessTruthValueSlotValue(TruthValue* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_PROPOSITIONS_LOGIC_POLARITY) {
    if (setvalueP) {
      self->polarity = ((Keyword*)(value));
    }
    else {
      value = self->polarity;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_STRENGTH) {
    if (setvalueP) {
      self->strength = ((Keyword*)(value));
    }
    else {
      value = self->strength;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_POSITIVE_SCORE) {
    if (setvalueP) {
      self->positiveScore = ((FloatWrapper*)(value))->wrapperValue;
    }
    else {
      value = wrapFloat(self->positiveScore);
    }
  }
  else {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, slotname, value, NULL);
    }
    else {
      value = self->dynamicSlots->lookup(slotname);
    }
  }
  return (value);
}

PartialSupport* newPartialSupport() {
  { PartialSupport* self = NULL;

    self = new PartialSupport();
    self->argumentScores = NULL;
    self->id = NULL_INTEGER;
    self->score = NULL_FLOAT;
    self->axiom = NULL;
    self->fact = NULL;
    return (self);
  }
}

Surrogate* PartialSupport::primaryType() {
  { PartialSupport* self = this;

    return (SGT_PROPOSITIONS_LOGIC_PARTIAL_SUPPORT);
  }
}

Object* accessPartialSupportSlotValue(PartialSupport* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_PROPOSITIONS_LOGIC_FACT) {
    if (setvalueP) {
      self->fact = ((Proposition*)(value));
    }
    else {
      value = self->fact;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_AXIOM) {
    if (setvalueP) {
      self->axiom = ((Proposition*)(value));
    }
    else {
      value = self->axiom;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_SCORE) {
    if (setvalueP) {
      self->score = ((FloatWrapper*)(value))->wrapperValue;
    }
    else {
      value = wrapFloat(self->score);
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_ID) {
    if (setvalueP) {
      self->id = ((IntegerWrapper*)(value))->wrapperValue;
    }
    else {
      value = wrapInteger(self->id);
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_ARGUMENT_SCORES) {
    if (setvalueP) {
      self->argumentScores = ((Cons*)(value));
    }
    else {
      value = self->argumentScores;
    }
  }
  else {
    { OutputStringStream* stream000 = newOutputStringStream();

      *(stream000->nativeStream) << "`" << slotname << "'" << " is not a valid case option";
      throw *newStellaException(stream000->theStringReader());
    }
  }
  return (value);
}

int oPARTIAL_SUPPORT_COUNTERo = 0;

Proposition* newProposition() {
  { Proposition* self = NULL;

    self = new Proposition();
    self->dynamicSlots = newKeyValueList();
    self->homeContext = oCONTEXTo.get();
    self->dependentPropositions = NIL_NON_PAGING_INDEX;
    self->operatoR = NULL;
    self->arguments = NULL;
    self->truthValue = NULL;
    self->kind = NULL;
    return (self);
  }
}

Surrogate* Proposition::primaryType() {
  { Proposition* self = this;

    return (SGT_PROPOSITIONS_LOGIC_PROPOSITION);
  }
}

BooleanWrapper* Proposition::variableTypeP_reader() {
  { Proposition* self = this;

    { BooleanWrapper* answer = ((BooleanWrapper*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_VARIABLE_TYPEp, NULL)));

      if (!((boolean)(answer))) {
        return ((FALSE ? TRUE_WRAPPER : FALSE_WRAPPER));
      }
      else {
        return (answer);
      }
    }
  }
}

List* Proposition::satellitePropositions_reader() {
  { Proposition* self = this;

    { List* answer = ((List*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_SATELLITE_PROPOSITIONS, NULL)));

      if (!((boolean)(answer))) {
        return (NIL_LIST);
      }
      else {
        return (answer);
      }
    }
  }
}

BooleanWrapper* Proposition::unfastenedP_reader() {
  { Proposition* self = this;

    { BooleanWrapper* answer = ((BooleanWrapper*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_UNFASTENEDp, NULL)));

      if (!((boolean)(answer))) {
        return ((FALSE ? TRUE_WRAPPER : FALSE_WRAPPER));
      }
      else {
        return (answer);
      }
    }
  }
}

double Proposition::weightDelta_reader() {
  { Proposition* self = this;

    { double answer = ((FloatWrapper*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_WEIGHT_DELTA, NULL_FLOAT_WRAPPER)))->wrapperValue;

      if (answer == NULL_FLOAT) {
        return (0.0);
      }
      else {
        return (answer);
      }
    }
  }
}

Cons* Proposition::support_reader() {
  { Proposition* self = this;

    { Cons* answer = ((Cons*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_SUPPORT, NULL)));

      if (!((boolean)(answer))) {
        return (NIL);
      }
      else {
        return (answer);
      }
    }
  }
}

BooleanWrapper* Proposition::badP_reader() {
  { Proposition* self = this;

    { BooleanWrapper* answer = ((BooleanWrapper*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_STELLA_BADp, NULL)));

      if (!((boolean)(answer))) {
        return ((FALSE ? TRUE_WRAPPER : FALSE_WRAPPER));
      }
      else {
        return (answer);
      }
    }
  }
}

Object* accessPropositionSlotValue(Proposition* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_PROPOSITIONS_LOGIC_KIND) {
    if (setvalueP) {
      self->kind = ((Keyword*)(value));
    }
    else {
      value = self->kind;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_TRUTH_VALUE) {
    if (setvalueP) {
      { Proposition* object000 = self;
        TruthValue* value000 = ((TruthValue*)(value));
        Object* oldValue000 = object000->truthValue;
        Object* newValue000 = updateInContext(oldValue000, value000, object000->homeContext, FALSE);

        if (!(((boolean)(oldValue000)) &&
            (oldValue000->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
          object000->truthValue = newValue000;
        }
      }
    }
    else {
      value = ((TruthValue*)(accessInContext(self->truthValue, self->homeContext, FALSE)));
    }
  }
  else if (slotname == SYM_PROPOSITIONS_STELLA_ARGUMENTS) {
    if (setvalueP) {
      self->arguments = ((Vector*)(value));
    }
    else {
      value = self->arguments;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_STELLA_OPERATOR) {
    if (setvalueP) {
      self->operatoR = ((GeneralizedSymbol*)(value));
    }
    else {
      value = self->operatoR;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_RELATIONREF) {
    if (setvalueP) {
      self->operatoR = ((Surrogate*)(value));
    }
    else {
      value = ((Surrogate*)(self->operatoR));
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_DEPENDENT_PROPOSITIONS) {
    if (setvalueP) {
      self->dependentPropositions = ((NonPagingIndex*)(value));
    }
    else {
      value = self->dependentPropositions;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_VARIABLE_TYPEp) {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_VARIABLE_TYPEp, ((BooleanWrapper*)(value)), NULL);
    }
    else {
      value = self->variableTypeP_reader();
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_SATELLITE_PROPOSITIONS) {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_SATELLITE_PROPOSITIONS, ((List*)(value)), NULL);
    }
    else {
      value = self->satellitePropositions_reader();
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_PROPOSITION_ORIGINATED_PROPOSITIONS) {
    if (setvalueP) {
      originatedPropositionsSetter(self, ((List*)(value)));
    }
    else {
      value = originatedPropositions(self);
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_PROPOSITION_STRINGIFIED_SOURCE) {
    if (setvalueP) {
      stringifiedSourceSetter(self, ((StringWrapper*)(value))->wrapperValue);
    }
    else {
      value = wrapString(stringifiedSource(self));
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_UNFASTENEDp) {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_UNFASTENEDp, ((BooleanWrapper*)(value)), NULL);
    }
    else {
      value = self->unfastenedP_reader();
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_WEIGHT_DELTA) {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_WEIGHT_DELTA, wrapFloat(((FloatWrapper*)(value))->wrapperValue), NULL_FLOAT_WRAPPER);
    }
    else {
      value = wrapFloat(self->weightDelta_reader());
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_SUPPORT) {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_SUPPORT, ((Cons*)(value)), NULL);
    }
    else {
      value = self->support_reader();
    }
  }
  else {
    if (setvalueP) {
      setDynamicSlotValue(self->dynamicSlots, slotname, value, NULL);
    }
    else {
      value = self->dynamicSlots->lookup(slotname);
    }
  }
  return (value);
}

void Proposition::printObject(std::ostream* stream) {
  { Proposition* self = this;

    printTopLevelObject(self, stream);
  }
}

IntegerLogicWrapper* newIntegerLogicWrapper(int wrapperValue) {
  { IntegerLogicWrapper* self = NULL;

    self = new IntegerLogicWrapper();
    self->wrapperValue = wrapperValue;
    self->dependentPropositionsIndex = NULL;
    return (self);
  }
}

Surrogate* IntegerLogicWrapper::primaryType() {
  { IntegerLogicWrapper* self = this;

    return (SGT_PROPOSITIONS_LOGIC_INTEGER_LOGIC_WRAPPER);
  }
}

void IntegerLogicWrapper::printObject(std::ostream* stream) {
  { IntegerLogicWrapper* self = this;

    if (oPRINTREADABLYpo.get()) {
      *(stream) << self->wrapperValue;
    }
    else {
      *(stream) << "|W|" << self->wrapperValue;
    }
  }
}

FloatLogicWrapper* newFloatLogicWrapper(double wrapperValue) {
  { FloatLogicWrapper* self = NULL;

    self = new FloatLogicWrapper();
    self->wrapperValue = wrapperValue;
    self->dependentPropositionsIndex = NULL;
    return (self);
  }
}

Surrogate* FloatLogicWrapper::primaryType() {
  { FloatLogicWrapper* self = this;

    return (SGT_PROPOSITIONS_LOGIC_FLOAT_LOGIC_WRAPPER);
  }
}

void FloatLogicWrapper::printObject(std::ostream* stream) {
  { FloatLogicWrapper* self = this;

    if (oPRINTREADABLYpo.get()) {
      *(stream) << self->wrapperValue;
    }
    else {
      *(stream) << "|W|" << self->wrapperValue;
    }
  }
}

StringLogicWrapper* newStringLogicWrapper(char* wrapperValue) {
  { StringLogicWrapper* self = NULL;

    self = new StringLogicWrapper();
    self->wrapperValue = wrapperValue;
    self->dependentPropositionsIndex = NULL;
    return (self);
  }
}

Surrogate* StringLogicWrapper::primaryType() {
  { StringLogicWrapper* self = this;

    return (SGT_PROPOSITIONS_LOGIC_STRING_LOGIC_WRAPPER);
  }
}

void StringLogicWrapper::printObject(std::ostream* stream) {
  { StringLogicWrapper* self = this;

    if (oPRINTREADABLYpo.get()) {
      *(stream) << self->wrapperValue;
    }
    else {
      *(stream) << "|W|" << self->wrapperValue;
    }
  }
}

char* stringifiedSource(Object* self) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
      { Object* self000 = self;
        NamedDescription* self = ((NamedDescription*)(self000));

        return (self->objectStringifiedSource);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* self001 = self;
        LogicObject* self = ((LogicObject*)(self001));

        return (((StringWrapper*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_OBJECT_STRINGIFIED_SOURCE, NULL_STRING_WRAPPER)))->wrapperValue);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
      { Object* self002 = self;
        Proposition* self = ((Proposition*)(self002));

        return (((StringWrapper*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_PROPOSITION_STRINGIFIED_SOURCE, NULL_STRING_WRAPPER)))->wrapperValue);
      }
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
        throw *newStellaException(stream000->theStringReader());
      }
    }
  }
}

char* stringifiedSourceSetter(Object* self, char* value) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
      { Object* self000 = self;
        NamedDescription* self = ((NamedDescription*)(self000));

        self->objectStringifiedSource = value;
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* self001 = self;
        LogicObject* self = ((LogicObject*)(self001));

        setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_OBJECT_STRINGIFIED_SOURCE, wrapString(value), NULL_STRING_WRAPPER);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
      { Object* self002 = self;
        Proposition* self = ((Proposition*)(self002));

        setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_PROPOSITION_STRINGIFIED_SOURCE, wrapString(value), NULL_STRING_WRAPPER);
      }
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
        throw *newStellaException(stream000->theStringReader());
      }
    }
  }
  return (value);
}

List* originatedPropositions(Object* self) {
  { Object* value = NULL;

    { Surrogate* testValue000 = safePrimaryType(self);

      if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
        { Object* self000 = self;
          NamedDescription* self = ((NamedDescription*)(self000));

          value = self->objectOriginatedPropositions;
        }
      }
      else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
        { Object* self001 = self;
          LogicObject* self = ((LogicObject*)(self001));

          value = ((List*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_OBJECT_ORIGINATED_PROPOSITIONS, NULL)));
        }
      }
      else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
        { Object* self002 = self;
          Proposition* self = ((Proposition*)(self002));

          value = ((List*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_PROPOSITION_ORIGINATED_PROPOSITIONS, NULL)));
        }
      }
      else {
        { OutputStringStream* stream000 = newOutputStringStream();

          *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
          throw *newStellaException(stream000->theStringReader());
        }
      }
    }
    if (!((boolean)(value))) {
      return (NIL_LIST);
    }
    else {
      return (((List*)(value)));
    }
  }
}

List* originatedPropositionsSetter(Object* self, List* value) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
      { Object* self000 = self;
        NamedDescription* self = ((NamedDescription*)(self000));

        self->objectOriginatedPropositions = value;
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* self001 = self;
        LogicObject* self = ((LogicObject*)(self001));

        setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_OBJECT_ORIGINATED_PROPOSITIONS, value, NULL);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
      { Object* self002 = self;
        Proposition* self = ((Proposition*)(self002));

        setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_PROPOSITION_ORIGINATED_PROPOSITIONS, value, NULL);
      }
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
        throw *newStellaException(stream000->theStringReader());
      }
    }
  }
  return (value);
}

// Maps names of KIF operators to relational surrogates.
PropertyList* oOPERATOR_NAME_TO_SURROGATE_TABLEo = NULL;

Proposition* createProposition(Symbol* kind, int argumentcount) {
  enforceCodeOnly();
  { Proposition* proposition = newProposition();

    proposition->arguments = newVector(argumentcount);
    proposition->kind = internKeyword(kind->symbolName);
    if ((kind == SYM_PROPOSITIONS_STELLA_FUNCTION) ||
        ((kind == SYM_PROPOSITIONS_STELLA_PREDICATE) ||
         (kind == SYM_PROPOSITIONS_STELLA_ISA))) {
    }
    else {
      proposition->operatoR = ((Surrogate*)(oOPERATOR_NAME_TO_SURROGATE_TABLEo->lookup(proposition->kind)));
    }
    if (descriptionModeP()) {
      setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTIVEp, (TRUE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
    }
    setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_UNFASTENEDp, TRUE_WRAPPER, NULL);
    return (proposition);
  }
}

void enforceCodeOnly() {
  if (((BooleanWrapper*)(dynamicSlotValue(oMODULEo.get()->dynamicSlots, SYM_PROPOSITIONS_STELLA_CODE_ONLYp, FALSE_WRAPPER)))->wrapperValue) {
    { OutputStringStream* stream000 = newOutputStringStream();

      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        *(stream000->nativeStream) << "ERROR: " << "Can't create a relation, instance or proposition in module:" << std::endl << "       " << "`" << oMODULEo.get()->moduleName << "'" << std::endl << "    because it is marked as 'code-only?'" << std::endl << "." << std::endl;
        helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
      }
      throw *newPropositionError(stream000->theStringReader());
    }
  }
}

Module* oLOGIC_MODULEo = NULL;

Module* oPL_KERNEL_MODULEo = NULL;

boolean logicModuleP(Module* self) {
  // Return TRUE if 'self' is a logic module, implying
  // that relations defined within it define a knowledge base.  A module
  // is a logic module iff it inherits the module 'PL-KERNEL'.
  { boolean testValue000 = FALSE;

    if (self == oPL_KERNEL_MODULEo) {
      testValue000 = TRUE;
    }
    else {
      {
        { boolean foundP000 = FALSE;

          { Module* m = NULL;
            Cons* iter000 = self->parentModules->theConsList;

            while (!(iter000 == NIL)) {
              m = ((Module*)(iter000->value));
              iter000 = iter000->rest;
              if (logicModuleP(m)) {
                foundP000 = TRUE;
                break;
              }
            }
          }
          testValue000 = foundP000;
        }
        if (!testValue000) {
          { boolean foundP001 = FALSE;

            { Module* m = NULL;
              Cons* iter001 = self->uses->theConsList;

              while (!(iter001 == NIL)) {
                m = ((Module*)(iter001->value));
                iter001 = iter001->rest;
                if (logicModuleP(m)) {
                  foundP001 = TRUE;
                  break;
                }
              }
            }
            testValue000 = foundP001;
          }
        }
      }
    }
    { boolean value000 = testValue000;

      return (value000);
    }
  }
}

SequenceIndex* locallyConceivedPropositions(Module* self) {
  { SequenceIndex* propositions = ((SequenceIndex*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_LOCALLY_CONCEIVED_PROPOSITIONS_INTERNAL, NULL)));

    if (!((boolean)(propositions))) {
      propositions = createSequenceIndex(KWD_PROPOSITIONS_PAGING, cons(KWD_PROPOSITIONS_CONTEXT_PROPOSITIONS, cons(((!((boolean)(NIL))) ? NIL : NIL), cons(self, NIL))));
      setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_LOCALLY_CONCEIVED_PROPOSITIONS_INTERNAL, propositions, NULL);
    }
    return (propositions);
  }
}

void locallyConceivedPropositionsSetter(Module* self, SequenceIndex* value) {
  setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_LOCALLY_CONCEIVED_PROPOSITIONS_INTERNAL, value, NULL);
}

SequenceIndex* locallyConceivedInstances(Context* self) {
  { SequenceIndex* instances = ((SequenceIndex*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_LOCALLY_CONCEIVED_INSTANCES_INTERNAL, NULL)));

    if (!((boolean)(instances))) {
      instances = createSequenceIndex(KWD_PROPOSITIONS_PAGING, cons(KWD_PROPOSITIONS_CONTEXT_INSTANCES, cons(((!((boolean)(NIL))) ? NIL : NIL), cons(self, NIL))));
      setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_LOCALLY_CONCEIVED_INSTANCES_INTERNAL, instances, NULL);
    }
    return (instances);
  }
}

void locallyConceivedInstancesSetter(Context* self, SequenceIndex* value) {
  setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_LOCALLY_CONCEIVED_INSTANCES_INTERNAL, value, NULL);
}

// The NOW time stamp is incremented whenever a series
// of one or more updates is followed by a query.
int oNOW_TIMESTAMPo = 0;

void incrementNowTimestamp() {
  oNOW_TIMESTAMPo = oNOW_TIMESTAMPo + 1;
}

int getNowTimestamp() {
  return (oNOW_TIMESTAMPo);
}

// Records whether the last KB access was a query or
// an update.  Used to determine when to increment the NOW time stamp
// counter.
Keyword* oLAST_KB_ACTIONo = NULL;

void updateNowTimestamp(Keyword* kbaction) {
  if (kbaction == oLAST_KB_ACTIONo) {
    return;
  }
  if (kbaction == KWD_PROPOSITIONS_UPDATE_PROPOSITION) {
    if (descriptionModeP() ||
        ((!(oMODULEo.get() == oCONTEXTo.get())) ||
         oINVISIBLEASSERTIONpo.get())) {
      return;
    }
    incrementNowTimestamp();
  }
  else if (kbaction == KWD_PROPOSITIONS_EXECUTE_QUERY) {
  }
  else {
    { OutputStringStream* stream000 = newOutputStringStream();

      *(stream000->nativeStream) << "`" << kbaction << "'" << " is not a valid case option";
      throw *newStellaException(stream000->theStringReader());
    }
  }
  oLAST_KB_ACTIONo = kbaction;
}

// The undefined individual.  Denotes the non-existence of
// an individual in whatever slot it occupies.
LogicObject* BOTTOM = NULL;

boolean bottomP(Object* self) {
  // Return TRUE if 'self' is the undefined individual BOTTOM.
  return (self == BOTTOM);
}

boolean subrelationOfP(NamedDescription* desc1, NamedDescription* desc2) {
  { 
    BIND_STELLA_SPECIAL(oREVERSEPOLARITYpo, boolean, FALSE);
    return (collectionImpliesCollectionP(desc1, desc2));
  }
}

boolean logicalSubtypeOfP(Surrogate* type1, Surrogate* type2) {
  if (type1 == type2) {
    return (TRUE);
  }
  else {
    { NamedDescription* desc1 = getDescription(type1);
      NamedDescription* desc2 = getDescription(type2);

      if ((!((boolean)(desc1))) ||
          (!((boolean)(desc2)))) {
        return (FALSE);
      }
      return (subrelationOfP(desc1, desc2));
    }
  }
}

boolean logicalSubtypeOfLiteralP(Surrogate* type) {
  { NamedDescription* desc = surrogateToDescription(type);
    NamedDescription* literalclass = surrogateToDescription(SGT_PROPOSITIONS_STELLA_LITERAL);
    Cons* literalsubs = NIL;

    if (!((boolean)(desc))) {
      return (FALSE);
    }
    else if (desc == literalclass) {
      return (TRUE);
    }
    else {
      { 
        BIND_STELLA_SPECIAL(oREVERSEPOLARITYpo, boolean, FALSE);
        BIND_STELLA_SPECIAL(oCONTEXTo, Context*, getPropertyTestContext());
        { MemoizationTable* memoTable000 = ((MemoizationTable*)(SGT_PROPOSITIONS_LOGIC_F_LOGICAL_SUBTYPE_OF_LITERALp_MEMO_TABLE_000->surrogateValue));
          Cons* memoizedEntry000 = NULL;
          Object* memoizedValue000 = NULL;

          if (!((boolean)(memoTable000))) {
            initializeMemoizationTable(SGT_PROPOSITIONS_LOGIC_F_LOGICAL_SUBTYPE_OF_LITERALp_MEMO_TABLE_000, "(:MAX-VALUES 10 :TIMESTAMPS (:META-KB-UPDATE))");
            memoTable000 = ((MemoizationTable*)(SGT_PROPOSITIONS_LOGIC_F_LOGICAL_SUBTYPE_OF_LITERALp_MEMO_TABLE_000->surrogateValue));
          }
          if (oMEMOIZATION_ENABLEDpo) {
            memoizedEntry000 = lookupMruMemoizedValue(((MruMemoizationTable*)(memoTable000)), oCONTEXTo.get(), MEMOIZED_NULL_VALUE, NULL, NULL, -1);
            memoizedValue000 = memoizedEntry000->value;
          }
          if (((boolean)(memoizedValue000))) {
            if (memoizedValue000 == MEMOIZED_NULL_VALUE) {
              memoizedValue000 = NULL;
            }
          }
          else {
            memoizedValue000 = allSubcollections(literalclass)->consify();
            if (oMEMOIZATION_ENABLEDpo) {
              memoizedEntry000->value = ((!((boolean)(memoizedValue000))) ? MEMOIZED_NULL_VALUE : memoizedValue000);
            }
          }
          literalsubs = ((Cons*)(memoizedValue000));
        }
        if (literalsubs->membP(desc)) {
          return (TRUE);
        }
      }
    }
    return (FALSE);
  }
}

boolean literalTypeP(Surrogate* type) {
  { Object* value = type->surrogateValue;

    return (isaP(value, SGT_PROPOSITIONS_STELLA_CLASS) &&
        (subtypeOfP(type, SGT_PROPOSITIONS_STELLA_LITERAL) ||
         subtypeOfP(type, SGT_PROPOSITIONS_STELLA_LITERAL_WRAPPER)));
  }
}

boolean booleanTypeP(Surrogate* self) {
  return ((self == SGT_PROPOSITIONS_STELLA_BOOLEAN) ||
      (self == SGT_PROPOSITIONS_STELLA_BOOLEAN_WRAPPER));
}

boolean classDescriptionP(NamedDescription* self) {
  return (classP(self));
}

boolean functionDescriptionP(NamedDescription* self) {
  return (functionP(self));
}

boolean variableArityP(Description* self) {
  return (testPropertyP(self, SGT_PROPOSITIONS_PL_KERNEL_KB_VARIABLE_ARITY));
}

int NamedDescription::arity() {
  { NamedDescription* self = this;

    if (variableArityP(self)) {
      return (-1);
    }
    else {
      return (self->ioVariableTypes->length());
    }
  }
}

int Description::arity() {
  { Description* self = this;

    return (self->ioVariables->length());
  }
}

boolean monadicP(Object* self) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) {
      { Object* self000 = self;
        Description* self = ((Description*)(self000));

        return (self->arity() == 1);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_COLLECTION)) {
      { Object* self001 = self;
        Collection* self = ((Collection*)(self001));

        return (TRUE);
      }
    }
    else {
      return (FALSE);
    }
  }
}

boolean computedRelationP(NamedDescription* description) {
  return ((((boolean)(description->nativeRelation())) &&
      (!description->nativeRelation()->abstractP)) ||
      ((boolean)(accessBinaryValue(description, SGT_PROPOSITIONS_PL_KERNEL_KB_RELATION_COMPUTATION))));
}

Object* lastArgument(Proposition* proposition) {
  return ((proposition->arguments->theArray)[(proposition->arguments->length() - 1)]);
}

Surrogate* objectSurrogate(Object* self) {
  // Return the surrogate naming the logic object `self'.
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* self000 = self;
        LogicObject* self = ((LogicObject*)(self000));

        return (self->surrogateValueInverse);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
      { Object* self001 = self;
        Proposition* self = ((Proposition*)(self001));

        return (((Surrogate*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_STELLA_SURROGATE_VALUE_INVERSE, NULL))));
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_THING)) {
      { Object* self002 = self;
        Thing* self = ((Thing*)(self002));

        return (self->surrogateValueInverse);
      }
    }
    else {
      return (NULL);
    }
  }
}

Surrogate* objectSurrogateSetter(Object* self, Surrogate* name) {
  // Return the name of the logic object `self' to `name'.
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* self000 = self;
        LogicObject* self = ((LogicObject*)(self000));

        self->surrogateValueInverse = name;
        return (name);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
      { Object* self001 = self;
        Proposition* self = ((Proposition*)(self001));

        setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_STELLA_SURROGATE_VALUE_INVERSE, name, NULL);
        return (name);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_THING)) {
      { Object* self002 = self;
        Thing* self = ((Thing*)(self002));

        self->surrogateValueInverse = name;
        return (name);
      }
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
        throw *newStellaException(stream000->theStringReader());
      }
    }
  }
}

Symbol* objectName(Object* self) {
  // Return the name symbol for the logic object `self'.
  { Surrogate* surrogate = objectSurrogate(self);

    if (((boolean)(surrogate))) {
      return (internSymbolInModule(surrogate->symbolName, ((Module*)(surrogate->homeContext)), TRUE));
    }
    else {
      return (NULL);
    }
  }
}

char* objectNameString(Object* self) {
  // Return the name string for the logic object `self'.
  { Surrogate* surrogate = objectSurrogate(self);

    if (((boolean)(surrogate))) {
      return (surrogate->symbolName);
    }
    else {
      return (NULL);
    }
  }
}

char* objectStringName(Object* self) {
  return (objectNameString(self));
}

// Indicates the context for evaluating a proposition.  One
// of :DESCRIPTION, :INTENSIONAL-ASSERTION, or :EXTENSIONAL-ASSERTION.
DEFINE_STELLA_SPECIAL(oEVALUATIONMODEo, Keyword* , NULL);

boolean descriptionModeP() {
  return (oEVALUATIONMODEo.get() == KWD_PROPOSITIONS_DESCRIPTION);
}

// When enabled, slot-value assertions can be retracted
// by later conflicting assertions.
DEFINE_STELLA_SPECIAL(oCLIPPINGENABLEDpo, boolean , FALSE);

// When enabled, blocks normalizations that significantly
// change the behavior of inference rules.
DEFINE_STELLA_SPECIAL(oNATURALDEDUCTIONMODEpo, boolean , FALSE);

boolean naturalDeductionModeP() {
  // True if normalization is governed by natural
  // deduction semantics.
  return (oNATURALDEDUCTIONMODEpo.get());
}

// Signals that we are performing search across multiple
// contexts.  Used to disable retraction from collections, since that increases
// the overhead of the context mechanism.
boolean oCONTEXT_DEPENDENT_SEARCH_MODEpo = FALSE;

boolean contextDependentSearchModeP() {
  return (oCONTEXT_DEPENDENT_SEARCH_MODEpo);
}

void logNewlyConceivedProposition(Module* self, Proposition* proposition) {
  { boolean foundP000 = FALSE;

    { Object* arg = NULL;
      Vector* vector000 = proposition->arguments;
      int index000 = 0;
      int length000 = vector000->length();

      while (index000 < length000) {
        arg = (vector000->theArray)[index000];
        index000 = index000 + 1;
        if (isaP(arg, SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE)) {
          foundP000 = TRUE;
          break;
        }
      }
    }
    if (foundP000) {
      *(STANDARD_WARNING->nativeStream) << "Warning: " << "LOGGED QUESTIONABLE CONCEPTION: " << std::endl << "   " << "`" << proposition << "'" << std::endl << std::endl;
    }
  }
  locallyConceivedPropositions(self)->insert(proposition);
}

void clipOrClashWithOldPredicateValue(Proposition* proposition) {
  { Vector* arguments = proposition->arguments;
    int countminustwo = arguments->length() - 2;

    { Proposition* p = NULL;
      Iterator* iter000 = allTrueDependentPropositions((arguments->theArray)[0], ((Surrogate*)(proposition->operatoR)), FALSE);

      while (iter000->nextP()) {
        p = ((Proposition*)(iter000->value));
        { boolean alwaysP000 = TRUE;

          { int i = NULL_INTEGER;
            int iter001 = 1;
            int upperBound000 = countminustwo;
            boolean unboundedP000 = upperBound000 == NULL_INTEGER;

            while (unboundedP000 ||
                (iter001 <= upperBound000)) {
              i = iter001;
              iter001 = iter001 + 1;
              if (!eqlP(valueOf((arguments->theArray)[i]), valueOf((p->arguments->theArray)[i]))) {
                alwaysP000 = FALSE;
                break;
              }
            }
          }
          if (alwaysP000) {
            if (eqlP(valueOf((p->arguments->theArray)[(p->arguments->length() - 1)]), valueOf(arguments->last()))) {
            }
            else if (oCLIPPINGENABLEDpo.get() &&
                worldStateP(oCONTEXTo.get())) {
              deassignTruthValue(p, KWD_PROPOSITIONS_RETRACT_TRUE);
            }
            else {
              signalUnificationClash(valueOf((p->arguments->theArray)[(p->arguments->length() - 1)]), valueOf(arguments->last()));
            }
          }
        }
      }
    }
  }
}

void runTruthChangeDemon(Proposition* proposition, Surrogate* truthchangerelation) {
  { NamedDescription* description = getDescription(((Surrogate*)(proposition->operatoR)));
    Object* demoncomputation = accessBinaryValue(description, truthchangerelation);

    if (((boolean)(demoncomputation))) {
      { cpp_function_code functioncode = functionCodeFromProcedure(demoncomputation);

        if (functioncode != NULL) {
          ((void  (*) (Proposition*))functioncode)(proposition);
        }
      }
    }
  }
}

void runGoesTrueDemons(Proposition* proposition) {
  { Keyword* testValue000 = proposition->kind;

    if ((testValue000 == KWD_PROPOSITIONS_FUNCTION) ||
        (testValue000 == KWD_PROPOSITIONS_ISA)) {
      runTruthChangeDemon(proposition, SGT_PROPOSITIONS_PL_KERNEL_KB_GOES_TRUE_DEMON);
    }
    else if (testValue000 == KWD_PROPOSITIONS_PREDICATE) {
      { NamedDescription* description = getDescription(((Surrogate*)(proposition->operatoR)));

        if (singleValuedTermP(description)) {
          clipOrClashWithOldPredicateValue(proposition);
        }
      }
      runTruthChangeDemon(proposition, SGT_PROPOSITIONS_PL_KERNEL_KB_GOES_TRUE_DEMON);
    }
    else if (testValue000 == KWD_PROPOSITIONS_IMPLIES) {
      if ((!isaP((proposition->arguments->theArray)[0], SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) ||
          (!isaP((proposition->arguments->theArray)[1], SGT_PROPOSITIONS_LOGIC_DESCRIPTION))) {
        return;
      }
      if (!descriptionModeP()) {
        addTaxonomyImpliesSubsumesLink(((Description*)((proposition->arguments->theArray)[0])), ((Description*)((proposition->arguments->theArray)[1])));
      }
      reviseEquivalences(proposition, TRUE);
      if (oMODULEo.get() == oCONTEXTo.get()) {
        bumpInferableTimestamp();
      }
      flushInferableDirectSubdescriptionsCache(proposition);
    }
    else if (testValue000 == KWD_PROPOSITIONS_FORALL) {
      if (proposition->satellitePropositions_reader() == NIL_LIST) {
        deriveSatelliteRules(proposition);
      }
      bumpInferableTimestamp();
    }
    else {
    }
  }
  return;
}

void runGoesFalseDemons(Proposition* proposition) {
  { Keyword* testValue000 = proposition->kind;

    if ((testValue000 == KWD_PROPOSITIONS_FUNCTION) ||
        ((testValue000 == KWD_PROPOSITIONS_ISA) ||
         (testValue000 == KWD_PROPOSITIONS_PREDICATE))) {
      runTruthChangeDemon(proposition, SGT_PROPOSITIONS_PL_KERNEL_KB_GOES_FALSE_DEMON);
    }
    else {
    }
  }
  return;
}

void runGoesUnknownDemons(Proposition* proposition) {
  { Keyword* testValue000 = proposition->kind;

    if (testValue000 == KWD_PROPOSITIONS_ISA) {
      { Object* argument = (proposition->arguments->theArray)[0];

        if (subtypeOfP(safePrimaryType(argument), SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
          { Object* argument000 = argument;
            Skolem* argument = ((Skolem*)(argument000));

            { Surrogate* skolemtype = argument->skolemType;

              if ((skolemtype == ((Surrogate*)(proposition->operatoR))) &&
                  (!(skolemtype == SGT_PROPOSITIONS_STELLA_THING))) {
                std::cout << "Can't retract skolem type of " << "`" << argument << "'" << " equaling " << "`" << skolemtype->symbolName << "'" << "." << std::endl << std::endl;
              }
            }
          }
        }
        else {
        }
      }
      runTruthChangeDemon(proposition, SGT_PROPOSITIONS_PL_KERNEL_KB_GOES_UNKNOWN_DEMON);
    }
    else if ((testValue000 == KWD_PROPOSITIONS_PREDICATE) ||
        (testValue000 == KWD_PROPOSITIONS_FUNCTION)) {
      runTruthChangeDemon(proposition, SGT_PROPOSITIONS_PL_KERNEL_KB_GOES_UNKNOWN_DEMON);
    }
    else if (testValue000 == KWD_PROPOSITIONS_IMPLIES) {
      dropTaxonomyImpliesSubsumesLink(((Description*)((proposition->arguments->theArray)[0])), ((Description*)((proposition->arguments->theArray)[1])));
      reviseEquivalences(proposition, FALSE);
      { Proposition* p = ((Proposition*)(dynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_MASTER_PROPOSITION, NULL)));

        if (((boolean)(p)) &&
            trueP(p)) {
          deassignTruthValue(p, KWD_PROPOSITIONS_RETRACT_TRUE);
        }
      }
    }
    else if (testValue000 == KWD_PROPOSITIONS_FORALL) {
      { Proposition* p = NULL;
        Cons* iter000 = proposition->satellitePropositions_reader()->theConsList;

        while (!(iter000 == NIL)) {
          p = ((Proposition*)(iter000->value));
          iter000 = iter000->rest;
          deassignTruthValue(p, KWD_PROPOSITIONS_RETRACT_TRUE);
        }
      }
    }
    else {
    }
  }
  return;
}

void updateLinksAndTimestamps(Proposition* proposition, Keyword* updatemode) {
  if (!((boolean)(proposition))) {
    return;
  }
  updateNowTimestamp(KWD_PROPOSITIONS_UPDATE_PROPOSITION);
  if ((updatemode == KWD_PROPOSITIONS_ASSERT_TRUE) ||
      (updatemode == KWD_PROPOSITIONS_PRESUME_TRUE)) {
    runGoesTrueDemons(proposition);
  }
  else if ((updatemode == KWD_PROPOSITIONS_ASSERT_FALSE) ||
      (updatemode == KWD_PROPOSITIONS_PRESUME_FALSE)) {
    runGoesFalseDemons(proposition);
  }
  else if (updatemode == KWD_PROPOSITIONS_RETRACT_TRUE) {
    runGoesUnknownDemons(proposition);
  }
  else if (updatemode == KWD_PROPOSITIONS_RETRACT_FALSE) {
    runGoesUnknownDemons(proposition);
  }
  else {
    { OutputStringStream* stream000 = newOutputStringStream();

      *(stream000->nativeStream) << "`" << updatemode << "'" << " is not a valid case option";
      throw *newStellaException(stream000->theStringReader());
    }
  }
}

void assignTruthValue(Proposition* self, Object* truthvalue) {
  { TruthValue* oldtruthvalue = ((TruthValue*)(accessInContext(self->truthValue, self->homeContext, FALSE)));
    TruthValue* newtruthvalue = NULL;

    if (truthvalue == TRUE_WRAPPER) {
      newtruthvalue = TRUE_TRUTH_VALUE;
    }
    else if (truthvalue == FALSE_WRAPPER) {
      newtruthvalue = FALSE_TRUTH_VALUE;
    }
    else {
      newtruthvalue = ((TruthValue*)(truthvalue));
    }
    if (oldtruthvalue == newtruthvalue) {
    }
    else if (((boolean)(oldtruthvalue))) {
      if ((self == TRUE_PROPOSITION) ||
          (self == FALSE_PROPOSITION)) {
      }
      else if (oldtruthvalue == INCONSISTENT_TRUTH_VALUE) {
      }
      else if (newtruthvalue == TRUE_TRUTH_VALUE) {
        if (oldtruthvalue == DEFAULT_TRUE_TRUTH_VALUE) {
          { Proposition* object000 = self;
            TruthValue* value000 = NULL;
            Object* oldValue001 = object000->truthValue;
            Object* newValue000 = updateInContext(oldValue001, value000, object000->homeContext, FALSE);

            if (!(((boolean)(oldValue001)) &&
                (oldValue001->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
              object000->truthValue = newValue000;
            }
          }
          assignTruthValue(self, newtruthvalue);
        }
        else if (oldtruthvalue == DEFAULT_FALSE_TRUTH_VALUE) {
          deassignTruthValue(self, KWD_PROPOSITIONS_RETRACT_FALSE);
          assignTruthValue(self, newtruthvalue);
        }
        else if (oldtruthvalue == FALSE_TRUTH_VALUE) {
          signalTruthValueClash(self);
        }
      }
      else if (newtruthvalue == FALSE_TRUTH_VALUE) {
        if (oldtruthvalue == DEFAULT_FALSE_TRUTH_VALUE) {
          { Proposition* object001 = self;
            TruthValue* value001 = NULL;
            Object* oldValue002 = object001->truthValue;
            Object* newValue001 = updateInContext(oldValue002, value001, object001->homeContext, FALSE);

            if (!(((boolean)(oldValue002)) &&
                (oldValue002->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
              object001->truthValue = newValue001;
            }
          }
          assignTruthValue(self, newtruthvalue);
        }
        else if (oldtruthvalue == DEFAULT_TRUE_TRUTH_VALUE) {
          deassignTruthValue(self, KWD_PROPOSITIONS_RETRACT_TRUE);
          assignTruthValue(self, newtruthvalue);
        }
        else if (oldtruthvalue == TRUE_TRUTH_VALUE) {
          signalTruthValueClash(self);
        }
      }
      else if (newtruthvalue == DEFAULT_TRUE_TRUTH_VALUE) {
        if (oDEFERINGDEFAULTFORWARDINFERENCESpo.get()) {
          deferredDefaultPropositionsSetter(oCONTEXTo.get(), cons(self, deferredDefaultPropositions(oCONTEXTo.get())));
          return;
        }
        else if ((oldtruthvalue == TRUE_TRUTH_VALUE) ||
            (oldtruthvalue == FALSE_TRUTH_VALUE)) {
        }
        else if (oldtruthvalue == DEFAULT_FALSE_TRUTH_VALUE) {
          deassignTruthValue(self, KWD_PROPOSITIONS_RETRACT_FALSE);
        }
      }
      else if (newtruthvalue == DEFAULT_FALSE_TRUTH_VALUE) {
        if (oDEFERINGDEFAULTFORWARDINFERENCESpo.get()) {
          deferredDefaultPropositionsSetter(oCONTEXTo.get(), cons(self, deferredDefaultPropositions(oCONTEXTo.get())));
          return;
        }
        else if ((oldtruthvalue == FALSE_TRUTH_VALUE) ||
            (oldtruthvalue == TRUE_TRUTH_VALUE)) {
        }
        else if (oldtruthvalue == DEFAULT_TRUE_TRUTH_VALUE) {
          deassignTruthValue(self, KWD_PROPOSITIONS_RETRACT_TRUE);
        }
      }
      else if (newtruthvalue == INCONSISTENT_TRUTH_VALUE) {
        if ((oldtruthvalue == TRUE_TRUTH_VALUE) ||
            (oldtruthvalue == DEFAULT_TRUE_TRUTH_VALUE)) {
          deassignTruthValue(self, KWD_PROPOSITIONS_RETRACT_TRUE);
        }
        else if ((oldtruthvalue == FALSE_TRUTH_VALUE) ||
            (oldtruthvalue == DEFAULT_FALSE_TRUTH_VALUE)) {
          deassignTruthValue(self, KWD_PROPOSITIONS_RETRACT_FALSE);
        }
        assignTruthValue(self, newtruthvalue);
      }
      else {
        signalTruthValueClash(self);
      }
    }
    else if (self->kind == KWD_PROPOSITIONS_FUNCTION) {
      { Proposition* object002 = self;
        TruthValue* value002 = TRUE_TRUTH_VALUE;
        Object* oldValue003 = object002->truthValue;
        Object* newValue002 = updateInContext(oldValue003, value002, object002->homeContext, FALSE);

        if (!(((boolean)(oldValue003)) &&
            (oldValue003->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
          object002->truthValue = newValue002;
        }
      }
    }
    else if (newtruthvalue == INCONSISTENT_TRUTH_VALUE) {
      { Proposition* object003 = self;
        TruthValue* value003 = newtruthvalue;
        Object* oldValue004 = object003->truthValue;
        Object* newValue003 = updateInContext(oldValue004, value003, object003->homeContext, FALSE);

        if (!(((boolean)(oldValue004)) &&
            (oldValue004->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
          object003->truthValue = newValue003;
        }
      }
    }
    else {
      { Proposition* object004 = self;
        TruthValue* value004 = newtruthvalue;
        Object* oldValue005 = object004->truthValue;
        Object* newValue004 = updateInContext(oldValue005, value004, object004->homeContext, FALSE);

        if (!(((boolean)(oldValue005)) &&
            (oldValue005->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
          object004->truthValue = newValue004;
        }
      }
      handleOutOfDateInferenceCache(KWD_PROPOSITIONS_ASSERT, self);
      if (newtruthvalue == TRUE_TRUTH_VALUE) {
        updateLinksAndTimestamps(self, KWD_PROPOSITIONS_ASSERT_TRUE);
      }
      else if (newtruthvalue == FALSE_TRUTH_VALUE) {
        updateLinksAndTimestamps(self, KWD_PROPOSITIONS_ASSERT_FALSE);
      }
      else if (newtruthvalue == DEFAULT_TRUE_TRUTH_VALUE) {
        updateLinksAndTimestamps(self, KWD_PROPOSITIONS_PRESUME_TRUE);
      }
      else if (newtruthvalue == DEFAULT_FALSE_TRUTH_VALUE) {
        updateLinksAndTimestamps(self, KWD_PROPOSITIONS_PRESUME_FALSE);
      }
      reactToKbUpdate(oCONTEXTo.get(), self);
    }
  }
}

void deassignTruthValue(Proposition* self, Keyword* updatemode) {
  { Proposition* object000 = self;
    TruthValue* value000 = NULL;
    Object* oldValue000 = object000->truthValue;
    Object* newValue000 = updateInContext(oldValue000, value000, object000->homeContext, FALSE);

    if (!(((boolean)(oldValue000)) &&
        (oldValue000->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
      object000->truthValue = newValue000;
    }
  }
  if (!(updatemode == KWD_PROPOSITIONS_RETRACT_INCONSISTENT)) {
    handleOutOfDateInferenceCache(KWD_PROPOSITIONS_RETRACT, self);
    updateLinksAndTimestamps(self, updatemode);
  }
}

void updatePropositionTruthValue(Proposition* self, Keyword* updatemode) {
  if (updatemode == KWD_PROPOSITIONS_ASSERT_TRUE) {
    assignTruthValue(self, TRUE_WRAPPER);
  }
  else if (updatemode == KWD_PROPOSITIONS_ASSERT_FALSE) {
    assignTruthValue(self, FALSE_WRAPPER);
  }
  else if (updatemode == KWD_PROPOSITIONS_PRESUME_TRUE) {
    assignTruthValue(self, DEFAULT_TRUE_TRUTH_VALUE);
  }
  else if (updatemode == KWD_PROPOSITIONS_PRESUME_FALSE) {
    assignTruthValue(self, DEFAULT_FALSE_TRUTH_VALUE);
  }
  else if (updatemode == KWD_PROPOSITIONS_RETRACT_TRUE) {
    if (trueP(self) &&
        (!(self->kind == KWD_PROPOSITIONS_FUNCTION))) {
      deassignTruthValue(self, updatemode);
    }
  }
  else if (updatemode == KWD_PROPOSITIONS_RETRACT_FALSE) {
    if (falseP(self)) {
      deassignTruthValue(self, updatemode);
    }
  }
  else if (updatemode == KWD_PROPOSITIONS_CONCEIVE) {
  }
  else {
    { OutputStringStream* stream000 = newOutputStringStream();

      *(stream000->nativeStream) << "`" << updatemode << "'" << " is not a valid case option";
      throw *newStellaException(stream000->theStringReader());
    }
  }
}

void assignPropositionWeight(Proposition* self, double weight) {
  { TruthValue* tv = ((TruthValue*)(accessInContext(self->truthValue, self->homeContext, FALSE)));

    if (weight == 1.0) {
      assignTruthValue(self, TRUE_WRAPPER);
    }
    else if (weight == 0.0) {
      assignTruthValue(self, FALSE_WRAPPER);
    }
    else if ((tv == TRUE_TRUTH_VALUE) ||
        ((tv == FALSE_TRUTH_VALUE) ||
         (!((boolean)(tv))))) {
      { TruthValue* self000 = newTruthValue();

        self000->positiveScore = weight;
        { Proposition* object000 = self;
          TruthValue* value000 = self000;
          Object* oldValue001 = object000->truthValue;
          Object* newValue000 = updateInContext(oldValue001, value000, object000->homeContext, FALSE);

          if (!(((boolean)(oldValue001)) &&
              (oldValue001->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
            object000->truthValue = newValue000;
          }
        }
      }
    }
    else {
      ((TruthValue*)(accessInContext(self->truthValue, self->homeContext, FALSE)))->positiveScore = weight;
    }
  }
}

Context* tickleContext() {
  { NamedDescription* description = getDescription(SGT_PROPOSITIONS_PL_KERNEL_KB_HIDDEN_RELATION);

    assertProperty(description, SGT_PROPOSITIONS_PL_KERNEL_KB_HIDDEN_RELATION);
    updateProperty(description, SGT_PROPOSITIONS_PL_KERNEL_KB_HIDDEN_RELATION, KWD_PROPOSITIONS_RETRACT_TRUE);
    return (oCONTEXTo.get());
  }
}

Context* tickleInstances() {
  { Object* temp000 = getInstance(wrapString("HIDDEN-INSTANCE"));

    { Object* instance = (((boolean)(temp000)) ? temp000 : createLogicInstance(SGT_PROPOSITIONS_LOGIC_HIDDEN_INSTANCE, NULL));

      assertProperty(((LogicObject*)(instance)), SGT_PROPOSITIONS_PL_KERNEL_KB_HIDDEN_RELATION);
      updateProperty(((LogicObject*)(instance)), SGT_PROPOSITIONS_PL_KERNEL_KB_HIDDEN_RELATION, KWD_PROPOSITIONS_RETRACT_TRUE);
      return (oCONTEXTo.get());
    }
  }
}

void reactToSkolemValueUpdate(Skolem* skolem, Object* oldvalue, Object* newvalue, boolean toplevelupdateP) {
  { Proposition* proposition = skolem->definingProposition;

    if (toplevelupdateP &&
        nativeSlotPropositionP(proposition)) {
      { Relation* slot = getDescription(((Surrogate*)(proposition->operatoR)))->nativeRelation();
        Object* nativefirstargvalue = valueOf((proposition->arguments->theArray)[0]);

        if (((boolean)(newvalue))) {
          assignNativeSlotValue(((Thing*)(nativefirstargvalue)), ((StorageSlot*)(slot)), newvalue);
        }
        else {
          dropNativeSlotValue(((Thing*)(nativefirstargvalue)), ((StorageSlot*)(slot)), oldvalue);
        }
      }
    }
    if (!eqlP(oldvalue, newvalue)) {
      bumpMemoizationTimestamp(KWD_PROPOSITIONS_KB_UPDATE);
      if ((((boolean)(newvalue)) &&
          isaP(newvalue, SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) ||
          metaPropositionP(proposition)) {
        bumpMemoizationTimestamp(KWD_PROPOSITIONS_META_KB_UPDATE);
      }
      if (((boolean)(oldvalue))) {
        updateLinksAndTimestamps(proposition, KWD_PROPOSITIONS_RETRACT_TRUE);
      }
      if (((boolean)(newvalue))) {
        updateLinksAndTimestamps(proposition, KWD_PROPOSITIONS_ASSERT_TRUE);
      }
    }
  }
}

boolean nativeSlotPropositionP(Proposition* self) {
  { Vector* arguments = self->arguments;

    if ((!(arguments->length()))) {
      return (FALSE);
    }
    { Object* firstarg = (((boolean)((arguments->theArray)[0])) ? valueOf((arguments->theArray)[0]) : ((Object*)(NULL)));

      if (((boolean)(firstarg)) &&
          isaP(firstarg, SGT_PROPOSITIONS_STELLA_THING)) {
        { NamedDescription* description = getDescription(self->operatoR);

          if (((boolean)(description)) &&
              (((boolean)(description->nativeRelation())) &&
               isaP(description->nativeRelation(), SGT_PROPOSITIONS_STELLA_STORAGE_SLOT))) {
            return (TRUE);
          }
        }
      }
      return (FALSE);
    }
  }
}

Cons* updateNativeSlotProposition(Proposition* proposition, Keyword* updatemode) {
  { Relation* slot = surrogateToDescription(((Surrogate*)(proposition->operatoR)))->nativeRelation();
    Object* object = valueOf((proposition->arguments->theArray)[0]);
    Object* value = valueOf((proposition->arguments->theArray)[1]);

    if (updatemode == KWD_PROPOSITIONS_ASSERT_TRUE) {
      if (oCLIPPINGENABLEDpo.get()) {
        clearNativeSlotValue(((Thing*)(object)), ((StorageSlot*)(slot)));
      }
      assignNativeSlotValue(((Thing*)(object)), ((StorageSlot*)(slot)), value);
    }
    else if (updatemode == KWD_PROPOSITIONS_RETRACT_TRUE) {
      dropNativeSlotValue(((Thing*)(object)), ((StorageSlot*)(slot)), value);
    }
    else {
      *(STANDARD_WARNING->nativeStream) << "Warning: " << "Illegal update to native slot value.  Mode:  " << "`" << updatemode << "'" << std::endl << "   Proposition:  " << "`" << proposition << "'" << std::endl << std::endl;
    }
    return (NIL);
  }
}

void assignNativeSlotValue(Thing* self, StorageSlot* slot, Object* value) {
  if (skolemP(value)) {
    return;
  }
  { Object* oldvalue = readSlotValue(self, slot);

    if (eqlP(oldvalue, value)) {
      return;
    }
    if (((boolean)(oldvalue))) {
      signalUnificationClash(oldvalue, value);
    }
    putSlotValue(self, slot, value);
  }
}

void clearNativeSlotValue(Thing* object, StorageSlot* slot) {
  putSlotValue(object, slot, NULL);
}

void dropNativeSlotValue(Thing* self, StorageSlot* slot, Object* value) {
  { Object* oldvalue = readSlotValue(self, slot);

    if (eqlP(value, oldvalue)) {
      clearNativeSlotValue(self, slot);
    }
  }
}

Surrogate* lookupSlotrefOnInstance(Object* term, Symbol* slotname) {
  { Surrogate* termtype = NULL;

    termtype = safeLogicalType(term);
    if (!((boolean)(termtype))) {
      return (NULL);
    }
    if ((!((boolean)(termtype->surrogateValue))) ||
        ((!isaP(termtype->surrogateValue, SGT_PROPOSITIONS_STELLA_CLASS)) ||
         (termtype == SGT_PROPOSITIONS_STELLA_THING))) {
      return (NULL);
    }
    if (((boolean)(lookupFunction(slotname)))) {
      return (NULL);
    }
    { Surrogate* surrogate = lookupSlotref(termtype, slotname);

      if (((boolean)(surrogate))) {
        return (surrogate);
      }
      if (subtypeOfP(safePrimaryType(term), SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
        { Object* term000 = term;
          LogicObject* term = ((LogicObject*)(term000));

          { Proposition* prop = NULL;
            Iterator* iter000 = unfilteredDependentIsaPropositions(term)->allocateIterator();

            while (iter000->nextP()) {
              prop = ((Proposition*)(iter000->value));
              if (prop->kind == KWD_PROPOSITIONS_ISA) {
                surrogate = lookupSlotref(((Surrogate*)(prop->operatoR)), slotname);
                if (((boolean)(surrogate))) {
                  return (surrogate);
                }
              }
            }
          }
        }
      }
      else {
      }
      if (subtypeOfP(safePrimaryType(term), SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
        { Object* term001 = term;
          Skolem* term = ((Skolem*)(term001));

          if (((boolean)(term->definingProposition)) &&
              isaP(term->definingProposition->operatoR, SGT_PROPOSITIONS_STELLA_SURROGATE)) {
            { NamedDescription* description = getDescription(((Surrogate*)(term->definingProposition->operatoR)));

              if (((boolean)(description))) {
                termtype = ((Surrogate*)(description->ioVariableTypes->last()));
                surrogate = lookupSlotref(termtype, slotname);
                if (((boolean)(surrogate))) {
                  return (surrogate);
                }
              }
            }
          }
        }
      }
      else {
      }
      return (NULL);
    }
  }
}

boolean unboundP(Skolem* skolem) {
  return (!((boolean)(((Object*)(accessInContext(skolem->variableValue, skolem->homeContext, FALSE))))));
}

Object* valueOf(Object* self) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
      { Object* self000 = self;
        Skolem* self = ((Skolem*)(self000));

        if (!((boolean)(((Object*)(accessInContext(self->variableValue, self->homeContext, FALSE)))))) {
          return (self);
        }
        else {
          return (valueOf(((Object*)(accessInContext(self->variableValue, self->homeContext, FALSE)))));
        }
      }
    }
    else if (subtypeOfSurrogateP(testValue000)) {
      { Object* self001 = self;
        Surrogate* self = ((Surrogate*)(self001));

        return (surrogateDvalueOf(self));
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_RELATION)) {
      { Object* self002 = self;
        Relation* self = ((Relation*)(self002));

        return (relationDvalueOf(self));
      }
    }
    else {
      return (self);
    }
  }
}

Object* surrogateDvalueOf(Surrogate* self) {
  { Object* value = self->surrogateValue;

    if (((boolean)(value))) {
      return (valueOf(value));
    }
    else {
      {
        if (dottedStringP(self->symbolName)) {
          { Slot* slot = ((Slot*)(stringToSlot(self->symbolName)));

            if (((boolean)(slot))) {
              return (valueOf(slot));
            }
          }
        }
        if (!(oSUPPRESSNONLOGICOBJECTWARNINGpo.get())) {
          signalUndefinedTerm(wrapString(self->symbolName));
        }
        return (NULL);
      }
    }
  }
}

Description* relationDvalueOf(Relation* self) {
  { NamedDescription* description = ((NamedDescription*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTION, NULL)));

    if (((boolean)(description))) {
      return (description);
    }
    if (subtypeOfP(safePrimaryType(self), SGT_PROPOSITIONS_STELLA_RELATION)) {
      { Relation* self000 = self;
        Relation* self = self000;

        description = getDescription(self);
      }
    }
    else {
    }
    if (((boolean)(description))) {
      return (description);
    }
    if (!(oSUPPRESSNONLOGICOBJECTWARNINGpo.get())) {
      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        *(STANDARD_WARNING->nativeStream) << "WARNING: " << "value-of: encountered missing description on class or relation " << self->name() << std::endl;
        helpSignalPropositionError(STANDARD_WARNING, KWD_PROPOSITIONS_WARNING);
      }
    }
    return (NULL);
  }
}

Object* equivalentValueOf(Object* self) {
  if (!((boolean)(self))) {
    return (NULL);
  }
  if (subtypeOfP(safePrimaryType(self), SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
    { Object* self000 = self;
      LogicObject* self = ((LogicObject*)(self000));

      if (((boolean)(((Description*)(accessInContext(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_EQUIVALENT_VALUE, NULL), self->homeContext, FALSE)))))) {
        return (equivalentValueOf(((Description*)(accessInContext(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_EQUIVALENT_VALUE, NULL), self->homeContext, FALSE)))));
      }
      else {
        return (self);
      }
    }
  }
  else {
    return (self);
  }
}

// Used to signal 'ground-value-of' that it can
// call 'bound-to' safely.
DEFINE_STELLA_SPECIAL(oCOMPUTEDQUERYpo, boolean , FALSE);

Object* groundValueOf(Object* self) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE)) {
      { Object* self000 = self;
        PatternVariable* self = ((PatternVariable*)(self000));

        { Object* value = NULL;

          if ((self->boundToOffset != NULL_INTEGER) &&
              oCOMPUTEDQUERYpo.get()) {
            value = (oQUERYITERATORo.get()->currentPatternRecord->variableBindings->theArray)[(self->boundToOffset)];
          }
          if (!((boolean)(value))) {
            value = ((Object*)(accessInContext(self->variableValue, self->homeContext, FALSE)));
          }
          if (((boolean)(value))) {
            return (groundValueOf(value));
          }
          else {
            return (NULL);
          }
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
      { Object* self001 = self;
        Skolem* self = ((Skolem*)(self001));

        { Object* value = ((Object*)(accessInContext(self->variableValue, self->homeContext, FALSE)));

          if (((boolean)(value))) {
            return (groundValueOf(value));
          }
          else {
            return (NULL);
          }
        }
      }
    }
    else if (subtypeOfSurrogateP(testValue000)) {
      { Object* self002 = self;
        Surrogate* self = ((Surrogate*)(self002));

        { Object* value = self->surrogateValue;

          if (((boolean)(value))) {
            return (groundValueOf(value));
          }
          else {
            return (NULL);
          }
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_RELATION)) {
      { Object* self003 = self;
        Relation* self = ((Relation*)(self003));

        return (((NamedDescription*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTION, NULL))));
      }
    }
    else {
      return (self);
    }
  }
}

boolean skolemP(Object* self) {
  return (((boolean)(self)) &&
      isaP(self, SGT_PROPOSITIONS_LOGIC_SKOLEM));
}

boolean variableP(Object* self) {
  return (isaP(self, SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE));
}

boolean nativeValueP(Object* self) {
  if (((boolean)(self))) {
    return (!isaP(self, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT));
  }
  else {
    return (FALSE);
  }
}

Object* nativeValueOf(Object* self) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE)) {
      { Object* self000 = self;
        PatternVariable* self = ((PatternVariable*)(self000));

        { Object* value = groundValueOf(self);

          if (((boolean)(value))) {
            return (nativeValueOf(value));
          }
          else {
            return (NULL);
          }
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
      { Object* self001 = self;
        Skolem* self = ((Skolem*)(self001));

        { Object* value = ((Object*)(accessInContext(self->variableValue, self->homeContext, FALSE)));

          if (((boolean)(value))) {
            return (nativeValueOf(value));
          }
          else {
            return (NULL);
          }
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* self002 = self;
        LogicObject* self = ((LogicObject*)(self002));

        return (NULL);
      }
    }
    else if (subtypeOfSurrogateP(testValue000)) {
      { Object* self003 = self;
        Surrogate* self = ((Surrogate*)(self003));

        { Object* value = self->surrogateValue;

          if (((boolean)(value))) {
            return (nativeValueOf(value));
          }
          else {
            return (NULL);
          }
        }
      }
    }
    else {
      return (self);
    }
  }
}

PatternVariable* innermostVariableOf(PatternVariable* self) {
  for (;;) {
    { PatternVariable* tightestvariable = self;
      Object* value = NULL;

      for (;;) {
        value = ((Object*)(accessInContext(tightestvariable->variableValue, tightestvariable->homeContext, FALSE)));
        if (((boolean)(value)) &&
            variableP(value)) {
          tightestvariable = ((PatternVariable*)(value));
        }
        else {
          return (tightestvariable);
        }
      }
    }
  }
}

Object* innermostOf(Object* self) {
  if (descriptionModeP()) {
    { Object* nextvalue = self;
      PatternVariable* tightestvariable = NULL;
      Object* equivalent = NULL;

      for (;;) {
        { Surrogate* testValue000 = safePrimaryType(nextvalue);

          if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE)) {
            { Object* nextvalue000 = nextvalue;
              PatternVariable* nextvalue = ((PatternVariable*)(nextvalue000));

              equivalent = ((Object*)(accessInContext(nextvalue->variableValue, nextvalue->homeContext, FALSE)));
              if (!((boolean)(equivalent))) {
                return (nextvalue);
              }
              tightestvariable = nextvalue;
            }
          }
          else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) {
            { Object* nextvalue001 = nextvalue;
              Description* nextvalue = ((Description*)(nextvalue001));

              return (nextvalue);
            }
          }
          else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
            { Object* nextvalue002 = nextvalue;
              LogicObject* nextvalue = ((LogicObject*)(nextvalue002));

              return (tightestvariable);
            }
          }
          else if (subtypeOfSurrogateP(testValue000)) {
            { Object* nextvalue003 = nextvalue;
              Surrogate* nextvalue = ((Surrogate*)(nextvalue003));

              return (nextvalue);
            }
          }
          else {
            return (nextvalue);
          }
        }
        nextvalue = equivalent;
      }
    }
  }
  else {
    { Object* tightestvalue = self;
      Object* equivalent = NULL;

      for (;;) {
        { Surrogate* testValue001 = safePrimaryType(tightestvalue);

          if (subtypeOfP(testValue001, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
            { Object* tightestvalue000 = tightestvalue;
              Skolem* tightestvalue = ((Skolem*)(tightestvalue000));

              if (!((boolean)(((Object*)(accessInContext(tightestvalue->variableValue, tightestvalue->homeContext, FALSE)))))) {
                return (tightestvalue);
              }
              equivalent = ((Object*)(accessInContext(tightestvalue->variableValue, tightestvalue->homeContext, FALSE)));
            }
          }
          else if (subtypeOfSurrogateP(testValue001)) {
            { Object* tightestvalue001 = tightestvalue;
              Surrogate* tightestvalue = ((Surrogate*)(tightestvalue001));

              if (!((boolean)(tightestvalue->surrogateValue))) {
                return (tightestvalue);
              }
              equivalent = tightestvalue->surrogateValue;
            }
          }
          else {
            return (tightestvalue);
          }
        }
        tightestvalue = equivalent;
      }
    }
  }
}

int oSKOLEM_ID_COUNTERo = 0;

DEFINE_STELLA_SPECIAL(oVARIABLEIDCOUNTERo, int , 0);

Symbol* yieldAnonymousSkolemName(boolean variableP) {
  { char* suffix = (variableP ? integerToString(oVARIABLEIDCOUNTERo.set(oVARIABLEIDCOUNTERo.get() + 1)) : integerToString(oSKOLEM_ID_COUNTERo = oSKOLEM_ID_COUNTERo + 1));

    return (internSymbol(stringConcatenate((variableP ? (char*)"?V" : (char*)"SK"), ((strlen(suffix) == 1) ? (char*)"0" : (char*)""), 1, suffix)));
  }
}

// Cons-list of top-level existentially-quantified skolems.
DEFINE_STELLA_SPECIAL(oFREESKOLEMSo, Cons* , NULL);

Skolem* helpCreateSkolem(Skolem* self, Surrogate* type, Symbol* name, boolean assertisaP) {
  type = unwrapWrappedType(type);
  if (!(((boolean)(name)))) {
    name = yieldAnonymousSkolemName(variableP(self));
  }
  self->skolemName = name;
  if (assertisaP &&
      ((boolean)(type))) {
    assertIsaProposition(self, type);
  }
  return (self);
}

Skolem* createSkolem(Surrogate* type, Symbol* name) {
  { Skolem* skolem = newSkolem();

    if (((boolean)(oFREESKOLEMSo.get()))) {
      oFREESKOLEMSo.set(cons(skolem, oFREESKOLEMSo.get()));
    }
    helpCreateSkolem(skolem, type, name, TRUE);
    return (skolem);
  }
}

PatternVariable* createVariable(Surrogate* type, Symbol* name, boolean assertisaP) {
  { Keyword* evaluationmode = oEVALUATIONMODEo.get();

    { 
      BIND_STELLA_SPECIAL(oEVALUATIONMODEo, Keyword*, evaluationmode);
      if (assertisaP &&
          (!descriptionModeP())) {
        oEVALUATIONMODEo.set(KWD_PROPOSITIONS_DESCRIPTION);
      }
      return (((PatternVariable*)(helpCreateSkolem(newPatternVariable(), type, name, assertisaP))));
    }
  }
}

Skolem* createVariableOrSkolem(Surrogate* type, Symbol* name) {
  if (descriptionModeP()) {
    return (createVariable(type, name, FALSE));
  }
  else {
    { Skolem* skolem = newSkolem();

      { 
        BIND_STELLA_SPECIAL(oINVISIBLEASSERTIONpo, boolean, TRUE);
        helpCreateSkolem(skolem, type, name, FALSE);
      }
      return (skolem);
    }
  }
}

boolean Skolem::objectEqualP(Object* other) {
  { Skolem* self = this;

    if (self == other) {
      return (TRUE);
    }
    else {
      if (skolemP(other)) {
        { Proposition* propself = self->definingProposition;
          Proposition* propother = ((Skolem*)(other))->definingProposition;

          return (((boolean)(propself)) &&
              (((boolean)(propother)) &&
               equivalentFunctionPropositionsP(propself, propother, NULL)));
        }
      }
    }
    return (FALSE);
  }
}

// Value representing TRUE.
TruthValue* TRUE_TRUTH_VALUE = NULL;

// Value representing FALSE.
TruthValue* FALSE_TRUTH_VALUE = NULL;

// Value representing DEFAULT-TRUE.
TruthValue* DEFAULT_TRUE_TRUTH_VALUE = NULL;

// Value representing DEFAULT-FALSE.
TruthValue* DEFAULT_FALSE_TRUTH_VALUE = NULL;

// Value representing UNKNOWN.  Needed for those cases
// where we need a non-NULL truth value to represents UNKNOWN.
TruthValue* UNKNOWN_TRUTH_VALUE = NULL;

// Value representing a contradiction.
TruthValue* INCONSISTENT_TRUTH_VALUE = NULL;

boolean trueP(Proposition* self) {
  // Return true if `self' is true (or default-true if we are
  // considering default assertions).
  { TruthValue* truthvalue = ((TruthValue*)(accessInContext(self->truthValue, self->homeContext, FALSE)));

    if (truthvalue == TRUE_TRUTH_VALUE) {
      return (TRUE);
    }
    else if (truthvalue == DEFAULT_TRUE_TRUTH_VALUE) {
      if (!oDONTUSEDEFAULTKNOWLEDGEpo.get()) {
        return (TRUE);
      }
    }
    else if (((boolean)(truthvalue)) &&
        ((truthvalue->polarity == KWD_PROPOSITIONS_TRUE) &&
         ((!oDONTUSEDEFAULTKNOWLEDGEpo.get()) ||
          (truthvalue->strength == KWD_PROPOSITIONS_STRICT)))) {
      return (TRUE);
    }
    return (FALSE);
  }
}

boolean falseP(Proposition* self) {
  // Return true if `self' is false (or default-false if we are
  // considering default assertions).
  { TruthValue* truthvalue = ((TruthValue*)(accessInContext(self->truthValue, self->homeContext, FALSE)));

    if (truthvalue == FALSE_TRUTH_VALUE) {
      return (TRUE);
    }
    else if (truthvalue == DEFAULT_FALSE_TRUTH_VALUE) {
      if (!oDONTUSEDEFAULTKNOWLEDGEpo.get()) {
        return (TRUE);
      }
    }
    else if (((boolean)(truthvalue)) &&
        ((truthvalue->polarity == KWD_PROPOSITIONS_FALSE) &&
         ((!oDONTUSEDEFAULTKNOWLEDGEpo.get()) ||
          (truthvalue->strength == KWD_PROPOSITIONS_STRICT)))) {
      return (TRUE);
    }
    return (FALSE);
  }
}

boolean unknownP(Proposition* self) {
  // Return true if the truth of `self' is unknown.
  { TruthValue* tv = ((TruthValue*)(accessInContext(self->truthValue, self->homeContext, FALSE)));

    return (((tv == UNKNOWN_TRUTH_VALUE) ||
        (!((boolean)(tv)))) ||
        (((!(!oDONTUSEDEFAULTKNOWLEDGEpo.get())) &&
        ((tv == DEFAULT_TRUE_TRUTH_VALUE) ||
         (tv == DEFAULT_FALSE_TRUTH_VALUE))) ||
         (tv == INCONSISTENT_TRUTH_VALUE)));
  }
}

boolean inconsistentP(Proposition* self) {
  // Return true if `self' is inconsistent (true and false).
  return (((TruthValue*)(accessInContext(self->truthValue, self->homeContext, FALSE))) == INCONSISTENT_TRUTH_VALUE);
}

boolean defaultTrueP(Proposition* self) {
  // Return true if `self' is default true.
  return (((TruthValue*)(accessInContext(self->truthValue, self->homeContext, FALSE))) == DEFAULT_TRUE_TRUTH_VALUE);
}

boolean defaultFalseP(Proposition* self) {
  // Return true if `self' is default false.
  return (((TruthValue*)(accessInContext(self->truthValue, self->homeContext, FALSE))) == DEFAULT_FALSE_TRUTH_VALUE);
}

TruthValue* defaultifyTruthValue(boolean trueP, boolean bydefaultP) {
  return ((trueP ? ((bydefaultP ? DEFAULT_TRUE_TRUTH_VALUE : TRUE_TRUTH_VALUE)) : ((bydefaultP ? DEFAULT_FALSE_TRUTH_VALUE : FALSE_TRUTH_VALUE))));
}

TruthValue* conjoinTruthValues(TruthValue* tv1, TruthValue* tv2) {
  // Return the logical conjunction of truth values `tv1' and `tv2'.
  if ((tv1 == INCONSISTENT_TRUTH_VALUE) ||
      (tv2 == INCONSISTENT_TRUTH_VALUE)) {
    return (INCONSISTENT_TRUTH_VALUE);
  }
  else if ((tv1 == FALSE_TRUTH_VALUE) ||
      (tv2 == FALSE_TRUTH_VALUE)) {
    return (FALSE_TRUTH_VALUE);
  }
  else if ((tv1 == DEFAULT_FALSE_TRUTH_VALUE) &&
      (tv2 == DEFAULT_TRUE_TRUTH_VALUE)) {
    return (UNKNOWN_TRUTH_VALUE);
  }
  else if ((tv1 == DEFAULT_TRUE_TRUTH_VALUE) &&
      (tv2 == DEFAULT_FALSE_TRUTH_VALUE)) {
    return (UNKNOWN_TRUTH_VALUE);
  }
  else if ((!((boolean)(tv1))) ||
      ((tv1 == UNKNOWN_TRUTH_VALUE) ||
       ((!((boolean)(tv2))) ||
        (tv2 == UNKNOWN_TRUTH_VALUE)))) {
    return (UNKNOWN_TRUTH_VALUE);
  }
  else if ((tv1 == DEFAULT_FALSE_TRUTH_VALUE) ||
      (tv2 == DEFAULT_FALSE_TRUTH_VALUE)) {
    return (DEFAULT_FALSE_TRUTH_VALUE);
  }
  else if ((tv1 == DEFAULT_TRUE_TRUTH_VALUE) ||
      (tv2 == DEFAULT_TRUE_TRUTH_VALUE)) {
    return (DEFAULT_TRUE_TRUTH_VALUE);
  }
  else {
    return (TRUE_TRUTH_VALUE);
  }
}

TruthValue* disjoinTruthValues(TruthValue* tv1, TruthValue* tv2) {
  // Return the logical disjunction of truth values `tv1' and `tv2'.
  if ((tv1 == TRUE_TRUTH_VALUE) ||
      (tv2 == TRUE_TRUTH_VALUE)) {
    return (TRUE_TRUTH_VALUE);
  }
  else if ((tv1 == DEFAULT_TRUE_TRUTH_VALUE) ||
      (tv2 == DEFAULT_TRUE_TRUTH_VALUE)) {
    return (DEFAULT_TRUE_TRUTH_VALUE);
  }
  else if ((!((boolean)(tv1))) ||
      ((tv1 == UNKNOWN_TRUTH_VALUE) ||
       ((!((boolean)(tv2))) ||
        (tv2 == UNKNOWN_TRUTH_VALUE)))) {
    return (UNKNOWN_TRUTH_VALUE);
  }
  else if ((tv1 == DEFAULT_FALSE_TRUTH_VALUE) ||
      (tv2 == DEFAULT_FALSE_TRUTH_VALUE)) {
    return (DEFAULT_FALSE_TRUTH_VALUE);
  }
  else {
    return (FALSE_TRUTH_VALUE);
  }
}

TruthValue* invertTruthValue(TruthValue* self) {
  // Return the logical negation of `self'.
  if (self == TRUE_TRUTH_VALUE) {
    return (FALSE_TRUTH_VALUE);
  }
  else if (self == DEFAULT_TRUE_TRUTH_VALUE) {
    return (DEFAULT_FALSE_TRUTH_VALUE);
  }
  else if (self == FALSE_TRUTH_VALUE) {
    return (TRUE_TRUTH_VALUE);
  }
  else if (self == DEFAULT_FALSE_TRUTH_VALUE) {
    return (DEFAULT_TRUE_TRUTH_VALUE);
  }
  else if (self == INCONSISTENT_TRUTH_VALUE) {
    return (INCONSISTENT_TRUTH_VALUE);
  }
  else {
    return (UNKNOWN_TRUTH_VALUE);
  }
}

TruthValue* weakenTruthValue(TruthValue* tv1, TruthValue* tv2) {
  // If `tv2' has lesser strength than `tv1', adapt the strength of `tv1' (not
  // its value!) and return the result.  Otherwise, return `tv1' unmodified.
  if ((tv2 == UNKNOWN_TRUTH_VALUE) ||
      (!((boolean)(tv2)))) {
    return (UNKNOWN_TRUTH_VALUE);
  }
  else if ((tv2 == DEFAULT_TRUE_TRUTH_VALUE) ||
      (tv2 == DEFAULT_FALSE_TRUTH_VALUE)) {
    if ((tv1 == TRUE_TRUTH_VALUE) ||
        (tv1 == DEFAULT_TRUE_TRUTH_VALUE)) {
      return (DEFAULT_TRUE_TRUTH_VALUE);
    }
    if ((tv1 == FALSE_TRUTH_VALUE) ||
        (tv1 == DEFAULT_FALSE_TRUTH_VALUE)) {
      return (DEFAULT_FALSE_TRUTH_VALUE);
    }
  }
  return (tv1);
}

TruthValue* strengthenTruthValue(TruthValue* tv1, TruthValue* tv2) {
  // If `tv2' has greater strength than `tv1', adapt the strength of `tv1' (not
  // its value!) and return the result.  Otherwise, return `tv1' unmodified.
  if (((tv2 == TRUE_TRUTH_VALUE) ||
      (tv2 == FALSE_TRUTH_VALUE)) &&
      ((tv1 == DEFAULT_TRUE_TRUTH_VALUE) ||
       (tv1 == DEFAULT_FALSE_TRUTH_VALUE))) {
    if ((tv1 == TRUE_TRUTH_VALUE) ||
        (tv1 == DEFAULT_TRUE_TRUTH_VALUE)) {
      return (TRUE_TRUTH_VALUE);
    }
    if ((tv1 == FALSE_TRUTH_VALUE) ||
        (tv1 == DEFAULT_FALSE_TRUTH_VALUE)) {
      return (FALSE_TRUTH_VALUE);
    }
  }
  return (tv1);
}

boolean trueTruthValueP(TruthValue* self) {
  // Return TRUE if `self' represents some form of truth.
  return ((self == TRUE_TRUTH_VALUE) ||
      (self == DEFAULT_TRUE_TRUTH_VALUE));
}

boolean falseTruthValueP(TruthValue* self) {
  // Return TRUE if `self' represents some form of falsehood.
  return ((self == FALSE_TRUTH_VALUE) ||
      (self == DEFAULT_FALSE_TRUTH_VALUE));
}

boolean strictTruthValueP(TruthValue* self) {
  // Return TRUE if `self' is a strict truth value.
  return ((self == TRUE_TRUTH_VALUE) ||
      (self == FALSE_TRUTH_VALUE));
}

boolean defaultTruthValueP(TruthValue* self) {
  // Return TRUE if `self' is a default truth value.
  return ((self == DEFAULT_TRUE_TRUTH_VALUE) ||
      (self == DEFAULT_FALSE_TRUTH_VALUE));
}

boolean unknownTruthValueP(TruthValue* self) {
  // Return TRUE if `self' represents UNKNOWN.
  return ((self == UNKNOWN_TRUTH_VALUE) ||
      (!((boolean)(self))));
}

boolean inconsistentTruthValueP(TruthValue* self) {
  // Return TRUE if `self' represents INCONSISTENT.
  return (self == INCONSISTENT_TRUTH_VALUE);
}

void printTruthValue(TruthValue* self, OutputStream* stream) {
  if (self == TRUE_TRUTH_VALUE) {
    *(stream->nativeStream) << "TRUE";
  }
  else if (self == FALSE_TRUTH_VALUE) {
    *(stream->nativeStream) << "FALSE";
  }
  else if (self == DEFAULT_TRUE_TRUTH_VALUE) {
    *(stream->nativeStream) << "DEFAULT-TRUE";
  }
  else if (self == DEFAULT_FALSE_TRUTH_VALUE) {
    *(stream->nativeStream) << "DEFAULT-FALSE";
  }
  else if (self == UNKNOWN_TRUTH_VALUE) {
    *(stream->nativeStream) << "UNKNOWN";
  }
  else if (self == INCONSISTENT_TRUTH_VALUE) {
    *(stream->nativeStream) << "INCONSISTENT";
  }
  else {
    *(stream->nativeStream) << "|i|" << self->primaryType();
  }
}

void signalUnificationClash(Object* term1, Object* term2) {
  if (((boolean)(term1)) &&
      ((boolean)(term2))) {
    { OutputStringStream* stream000 = newOutputStringStream();

      *(stream000->nativeStream) << "`" << term1 << "'" << " clashes with " << "`" << term2 << "'";
      throw *newClash(stream000->theStringReader());
    }
  }
  else {
    throw *newClash("Unification clash.");
  }
}

void signalVariableValueClash(Skolem* skolem, Object* value1, Object* value2) {
  if (((boolean)(value1)) &&
      ((boolean)(value2))) {
    { OutputStringStream* stream000 = newOutputStringStream();

      *(stream000->nativeStream) << "Skolem " << "`" << skolem << "'" << " is equated with multiple values:" << std::endl << "   " << "`" << value1 << "'" << " and " << "`" << value2 << "'" << ".";
      throw *newClash(stream000->theStringReader());
    }
  }
  else {
    throw *newClash("Variable value clash.");
  }
}

void signalTruthValueClash(Proposition* proposition) {
  assignTruthValue(proposition, INCONSISTENT_TRUTH_VALUE);
  if (oCONTEXTo.get() == ((World*)(dynamicSlotValue(oMODULEo.get()->dynamicSlots, SYM_PROPOSITIONS_LOGIC_META_INFERENCE_CACHE, NULL)))) {
    { OutputStringStream* stream000 = newOutputStringStream();

      *(stream000->nativeStream) << "Derived both TRUE and FALSE for the proposition " << "`" << proposition << "'" << "." << std::endl << "   Clash occurred in the meta inference cache for module " << "`" << oMODULEo.get() << "'" << ".";
      throw *newClash(stream000->theStringReader());
    }
  }
  else if (oCONTEXTo.get() == ((World*)(dynamicSlotValue(oMODULEo.get()->dynamicSlots, SYM_PROPOSITIONS_LOGIC_TRUTH_MAINTAINED_INFERENCE_CACHE, NULL)))) {
    { OutputStringStream* stream001 = newOutputStringStream();

      *(stream001->nativeStream) << "Derived both TRUE and FALSE for the proposition " << "`" << proposition << "'" << "." << std::endl << "   Clash occurred in the default inference cache for module " << "`" << oMODULEo.get() << "'" << ".";
      throw *newClash(stream001->theStringReader());
    }
  }
  else if (oCONTEXTo.get() == oMODULEo.get()) {
    { OutputStringStream* stream002 = newOutputStringStream();

      *(stream002->nativeStream) << "Derived both TRUE and FALSE for the proposition " << "`" << proposition << "'" << "." << std::endl << "   Clash occurred in module " << "`" << oMODULEo.get() << "'" << ".";
      throw *newClash(stream002->theStringReader());
    }
  }
  else {
    { OutputStringStream* stream003 = newOutputStringStream();

      *(stream003->nativeStream) << "Derived both TRUE and FALSE for the proposition " << "`" << proposition << "'" << "." << std::endl << "   Clash occurred in the world " << "`" << oCONTEXTo.get() << "'" << " in module " << "`" << oMODULEo.get() << "'" << ".";
      throw *newClash(stream003->theStringReader());
    }
  }
}

Surrogate* logicalType(Object* self) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) {
      { Object* self000 = self;
        Description* self = ((Description*)(self000));

        if (monadicP(self) &&
            ((BooleanWrapper*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_IOTAp, FALSE_WRAPPER)))->wrapperValue) {
          return (((PatternVariable*)((self->ioVariables->theArray)[0]))->skolemType);
        }
        else if (monadicP(self) &&
            ((self->ioVariables->arraySize > 0) &&
             (((boolean)(((PatternVariable*)((self->ioVariables->theArray)[0]))->skolemType)) &&
              classDescriptionP(surrogateToDescription(((PatternVariable*)((self->ioVariables->theArray)[0]))->skolemType))))) {
          return (SGT_PROPOSITIONS_PL_KERNEL_KB_CLASS);
        }
        else {
          return (SGT_PROPOSITIONS_PL_KERNEL_KB_RELATION);
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE)) {
      { Object* self001 = self;
        PatternVariable* self = ((PatternVariable*)(self001));

        { Surrogate* type = self->skolemType;

          if (((boolean)(type)) &&
              (!(type == SGT_PROPOSITIONS_STELLA_THING))) {
            return (type);
          }
        }
        { Proposition* p = NULL;
          Iterator* iter000 = unfilteredDependentIsaPropositions(self)->allocateIterator();

          while (iter000->nextP()) {
            p = ((Proposition*)(iter000->value));
            if ((p->kind == KWD_PROPOSITIONS_ISA) &&
                coerceWrappedBooleanToBoolean(p->variableTypeP_reader())) {
              return (((Surrogate*)(p->operatoR)));
            }
          }
        }
        return (SGT_PROPOSITIONS_STELLA_THING);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
      { Object* self002 = self;
        Skolem* self = ((Skolem*)(self002));

        if (((boolean)(self->skolemType))) {
          return (self->skolemType);
        }
        else {
          return (SGT_PROPOSITIONS_STELLA_THING);
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* self003 = self;
        LogicObject* self = ((LogicObject*)(self003));

        return (SGT_PROPOSITIONS_STELLA_THING);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_LITERAL_WRAPPER)) {
      { Object* self004 = self;
        LiteralWrapper* self = ((LiteralWrapper*)(self004));

        return (wrapperValueType(self));
      }
    }
    else if (subtypeOfSurrogateP(testValue000)) {
      { Object* self005 = self;
        Surrogate* self = ((Surrogate*)(self005));

        if (((boolean)(self->surrogateValue)) &&
            (!(self->surrogateValue == BOTTOM))) {
          return (logicalType(self->surrogateValue));
        }
        else {
          return (SGT_PROPOSITIONS_STELLA_THING);
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_SET)) {
      { Object* self006 = self;
        Set* self = ((Set*)(self006));

        return (SGT_PROPOSITIONS_PL_KERNEL_KB_SET);
      }
    }
    else {
      return (self->primaryType());
    }
  }
}

// Used by 'safe-logical-type' to ask for a type
// without signalling an error if none exists.
DEFINE_STELLA_SPECIAL(oSUPPRESSUNTYPEDTYPEERRORpo, boolean , FALSE);

Surrogate* safeLogicalType(Object* self) {
  { 
    BIND_STELLA_SPECIAL(oSUPPRESSUNTYPEDTYPEERRORpo, boolean, TRUE);
    return (logicalType(self));
  }
}

Surrogate* classLogicalType(Class* self) {
  return (self->classType);
}

void unifyTypes(Skolem* term1, Object* term2) {
  { Surrogate* type1 = logicalType(term1);
    Surrogate* type2 = logicalType(term2);

    if ((type1 == type2) ||
        (logicalSubtypeOfP(type1, type2) ||
         logicalSubtypeOfP(type2, type1))) {
    }
    else if (type1 == SGT_PROPOSITIONS_STELLA_THING) {
      if (isaP(term1, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
      }
      ((Skolem*)(term1))->skolemType = type2;
    }
    else if (type2 == SGT_PROPOSITIONS_STELLA_THING) {
      if (isaP(term2, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
        ((Skolem*)(term2))->skolemType = type1;
      }
    }
    else {
      if (bottomP(term2)) {
        return;
      }
      signalUnificationClash(term1, term2);
    }
  }
}

void unifySkolemAndValue(Skolem* skolem, Object* value) {
  { Proposition* prop = NULL;
    Iterator* iter000 = unfilteredDependentPropositions(skolem, NULL)->allocateIterator();

    while (iter000->nextP()) {
      prop = ((Proposition*)(iter000->value));
      evaluateProposition(prop);
    }
  }
}

void bindSkolemToValue(Skolem* fromskolem, Object* tovalue, boolean toplevelupdateP) {
  { boolean typeisokP = FALSE;
    Surrogate* type = fromskolem->skolemType;
    Proposition* equatingproposition = NULL;

    if (skolemP(tovalue)) {
      equatingproposition = findEquatingProposition(tovalue, fromskolem);
      if (((boolean)(equatingproposition)) &&
          defaultTrueP(equatingproposition)) {
        { OutputStringStream* stream000 = newOutputStringStream();

          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            *(stream000->nativeStream) << "ERROR: " << "Equating two skolems " << "`" << fromskolem << "'" << " and " << "`" << tovalue << "'" << " by default is not yet supported." << std::endl << "." << std::endl;
            helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
          }
          throw *newPropositionError(stream000->theStringReader());
        }
      }
    }
    typeisokP = checkTypeP(tovalue, type, TRUE, tovalue);
    if (!typeisokP) {
      equatingproposition = findEquatingProposition(tovalue, fromskolem);
      if (!((boolean)(equatingproposition))) {
        { 
          BIND_STELLA_SPECIAL(oTERMUNDERCONSTRUCTIONo, Object*, oTERMUNDERCONSTRUCTIONo.get());
          if (!((boolean)(oTERMUNDERCONSTRUCTIONo.get()))) {
            oTERMUNDERCONSTRUCTIONo.set(fromskolem->definingProposition);
          }
          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            *(STANDARD_WARNING->nativeStream) << "WARNING: " << "Type check violation: Assigned value `" << tovalue << "'" << std::endl << "   to skolem with type `" << type << "'." << std::endl;
            helpSignalPropositionError(STANDARD_WARNING, KWD_PROPOSITIONS_WARNING);
          }
        }
      }
      else {
        verifyArgumentTypesAndCount(equatingproposition);
      }
    }
  }
  { Skolem* object000 = fromskolem;
    Object* value000 = tovalue;
    Object* oldValue000 = object000->variableValue;
    Object* newValue000 = updateInContext(oldValue000, value000, object000->homeContext, FALSE);

    if (!(((boolean)(oldValue000)) &&
        (oldValue000->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
      object000->variableValue = newValue000;
    }
  }
  { Surrogate* testValue000 = safePrimaryType(tovalue);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
      { Object* tovalue000 = tovalue;
        Skolem* tovalue = ((Skolem*)(tovalue000));

        { Object* equivalentskolem = ((Object*)(accessInContext(tovalue->variableValue, tovalue->homeContext, FALSE)));

          for (;;) {
            if (!((boolean)(equivalentskolem))) {
              break;
            }
            if (subtypeOfP(safePrimaryType(equivalentskolem), SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
              { Object* equivalentskolem000 = equivalentskolem;
                Skolem* equivalentskolem = ((Skolem*)(equivalentskolem000));

                if (equivalentskolem == tovalue) {
                  { Skolem* object001 = fromskolem;
                    Object* value001 = NULL;
                    Object* oldValue002 = object001->variableValue;
                    Object* newValue001 = updateInContext(oldValue002, value001, object001->homeContext, FALSE);

                    if (!(((boolean)(oldValue002)) &&
                        (oldValue002->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
                      object001->variableValue = newValue001;
                    }
                  }
                  return;
                }
                equivalentskolem = ((Skolem*)(((Object*)(accessInContext(equivalentskolem->variableValue, equivalentskolem->homeContext, FALSE)))));
              }
            }
            else {
              break;
            }
          }
        }
        { Skolem* object002 = tovalue;
          Cons* value002 = cons(fromskolem, tovalue->variableValueInverse_reader());
          Object* oldValue004 = object002->variableValueInverse;
          Object* newValue002 = updateInContext(oldValue004, value002, object002->homeContext, FALSE);

          if (!(((boolean)(oldValue004)) &&
              (oldValue004->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
            object002->variableValueInverse = newValue002;
          }
        }
        if (!((boolean)(tovalue->skolemType))) {
          tovalue->skolemType = fromskolem->skolemType;
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* tovalue001 = tovalue;
        LogicObject* tovalue = ((LogicObject*)(tovalue001));

        { LogicObject* object003 = tovalue;
          Cons* value003 = cons(fromskolem, tovalue->variableValueInverse_reader());
          Object* oldValue005 = object003->variableValueInverse;
          Object* newValue003 = updateInContext(oldValue005, value003, object003->homeContext, FALSE);

          if (!(((boolean)(oldValue005)) &&
              (oldValue005->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
            object003->variableValueInverse = newValue003;
          }
        }
      }
    }
    else {
      { Proposition* proposition = fromskolem->definingProposition;

        if (((boolean)(proposition)) &&
            rewrapIthArgumentP(tovalue, proposition, proposition->arguments->length())) {
          { Skolem* object004 = fromskolem;
            Object* value004 = rewrapArgument(tovalue);
            Object* oldValue006 = object004->variableValue;
            Object* newValue004 = updateInContext(oldValue006, value004, object004->homeContext, FALSE);

            if (!(((boolean)(oldValue006)) &&
                (oldValue006->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
              object004->variableValue = newValue004;
            }
          }
        }
      }
    }
  }
  handleOutOfDateInferenceCache(KWD_PROPOSITIONS_ASSERT, fromskolem->definingProposition);
  reactToKbUpdate(oCONTEXTo.get(), fromskolem);
  if (!skolemP(tovalue)) {
    if (((boolean)(fromskolem->definingProposition))) {
      reactToSkolemValueUpdate(fromskolem, NULL, tovalue, toplevelupdateP);
    }
  }
}

void unbindSkolemValue(Skolem* skolem, boolean toplevelupdateP) {
  { Object* oldvalue = valueOf(skolem);
    Proposition* definingprop = skolem->definingProposition;
    Proposition* equatingprop = NULL;

    unlinkEquivalentValue(skolem);
    equatingprop = findEquatingProposition(oldvalue, skolem);
    if (((boolean)(equatingprop))) {
      updatePropositionTruthValue(equatingprop, KWD_PROPOSITIONS_RETRACT_TRUE);
      handleOutOfDateInferenceCache(KWD_PROPOSITIONS_RETRACT, equatingprop);
    }
    else {
      handleOutOfDateInferenceCache(KWD_PROPOSITIONS_RETRACT, definingprop);
    }
    if (((boolean)(definingprop))) {
      reactToSkolemValueUpdate(skolem, oldvalue, NULL, toplevelupdateP);
    }
  }
}

void constrainAsSubset(Collection* set1, Collection* set2) {
  if (set1->length() > set2->length()) {
    signalUnificationClash(set1, set2);
  }
  { Cons* copy1 = NIL;
    Cons* copy2 = NIL;

    { Object* i2 = NULL;
      Iterator* iter000 = ((Iterator*)(set2->allocateIterator()));

      while (iter000->nextP()) {
        i2 = iter000->value;
        copy2 = cons(i2, copy2);
      }
    }
    { Object* i1 = NULL;
      Iterator* iter001 = ((Iterator*)(set1->allocateIterator()));

      while (iter001->nextP()) {
        i1 = iter001->value;
        if (copy2->memberP(i1)) {
          copy2 = copy2->remove(i1);
        }
        else {
          copy1 = cons(i1, copy1);
        }
      }
    }
    if (copy1 == NIL) {
      return;
    }
    if (copy2->length() == 1) {
      { Object* item1 = copy1->value;
        Object* item2 = copy2->value;

        equateValues(item1, item2);
        return;
      }
    }
    { int skolemcount1 = 0;
      int skolemcount2 = 0;

      { Object* i1 = NULL;
        Cons* iter002 = copy1;

        while (!(iter002 == NIL)) {
          i1 = iter002->value;
          iter002 = iter002->rest;
          if (skolemP(i1)) {
            skolemcount1 = skolemcount1 + 1;
          }
        }
      }
      { Object* i2 = NULL;
        Cons* iter003 = copy2;

        while (!(iter003 == NIL)) {
          i2 = iter003->value;
          iter003 = iter003->rest;
          if (skolemP(i2)) {
            skolemcount2 = skolemcount2 + 1;
          }
        }
      }
      if ((skolemcount1 + skolemcount2) == 0) {
        signalUnificationClash(set1, set2);
      }
    }
  }
}

void equateCollections(Collection* col1, Collection* col2) {
  if (col1->orderedP() &&
      col2->orderedP()) {
    { Object* i1 = NULL;
      Object* i2 = NULL;
      Iterator* iter000 = ((Iterator*)(col1->allocateIterator()));
      Iterator* iter001 = ((Iterator*)(col2->allocateIterator()));

      while (iter000->nextP() &&
          iter001->nextP()) {
        i1 = iter000->value;
        i2 = iter001->value;
        equateValues(i1, i2);
      }
    }
    return;
  }
  if (col1->noDuplicatesP() &&
      col2->noDuplicatesP()) {
    constrainAsSubset(col1, col2);
    constrainAsSubset(col2, col1);
  }
}

void equateValues(Object* term1, Object* term2) {
  if (eqlP(term1, term2)) {
  }
  else if (skolemP(term1)) {
    bindSkolemToValue(((Skolem*)(term1)), term2, FALSE);
  }
  else if (skolemP(term2)) {
    bindSkolemToValue(((Skolem*)(term2)), term1, FALSE);
  }
  else if (collectionP(term1) &&
      collectionP(term2)) {
    equateCollections(((Collection*)(term1)), ((Collection*)(term2)));
  }
  else if (isaP(term1, SGT_PROPOSITIONS_PL_KERNEL_KB_INTERVAL_CACHE) &&
      isaP(term2, SGT_PROPOSITIONS_PL_KERNEL_KB_INTERVAL_CACHE)) {
    pl_kernel_kb::unifyIntervalCaches(((pl_kernel_kb::IntervalCache*)(term1)), ((pl_kernel_kb::IntervalCache*)(term2)), SGT_PROPOSITIONS_PL_KERNEL_KB_ge);
    pl_kernel_kb::unifyIntervalCaches(((pl_kernel_kb::IntervalCache*)(term2)), ((pl_kernel_kb::IntervalCache*)(term1)), SGT_PROPOSITIONS_PL_KERNEL_KB_ge);
  }
  else {
    signalUnificationClash(term1, term2);
  }
}

Cons* Skolem::conflictingDefaultValues_reader() {
  { Skolem* self = this;

    { Cons* answer = ((Cons*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_CONFLICTING_DEFAULT_VALUES, NULL)));

      if (!((boolean)(answer))) {
        return (NIL);
      }
      else {
        return (answer);
      }
    }
  }
}

boolean unequateConflictingDefaultValuesP(Proposition* newequatingprop) {
  { Vector* arguments = newequatingprop->arguments;
    Object* firstarg = (arguments->theArray)[0];
    boolean firstargisskolemP = skolemP(firstarg);
    Object* firstargvalue = innermostOf(firstarg);
    Object* secondarg = (arguments->theArray)[1];
    boolean secondargisskolemP = skolemP(secondarg);
    Object* secondargvalue = innermostOf(secondarg);
    boolean newisdefaultP = defaultTrueP(newequatingprop);
    Object* newtruthvalue = NULL;
    Proposition* equatingprop = NULL;

    if (eqlP(firstargvalue, secondargvalue)) {
      return (TRUE);
    }
    else if (skolemP(firstargvalue)) {
      if ((((Skolem*)(firstargvalue))->conflictingDefaultValues_reader() == NIL) ||
          (!newisdefaultP)) {
        return (TRUE);
      }
      else {
        return (FALSE);
      }
    }
    else if (skolemP(secondargvalue)) {
      if ((((Skolem*)(secondargvalue))->conflictingDefaultValues_reader() == NIL) ||
          (!newisdefaultP)) {
        return (TRUE);
      }
      else {
        return (FALSE);
      }
    }
    if (firstargisskolemP &&
        (secondargisskolemP &&
         newisdefaultP)) {
      { OutputStringStream* stream000 = newOutputStringStream();

        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          *(stream000->nativeStream) << "ERROR: " << "Equating two skolems " << "`" << firstarg << "'" << " and " << "`" << secondarg << "'" << " by default is not yet supported." << std::endl << "." << std::endl;
          helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
        }
        throw *newPropositionError(stream000->theStringReader());
      }
    }
    newtruthvalue = newequatingprop->truthValue;
    newequatingprop->truthValue = NULL;
    if (firstargisskolemP) {
      { 
        BIND_STELLA_SPECIAL(oDONTUSEDEFAULTKNOWLEDGEpo, boolean, TRUE);
        equatingprop = findEquatingProposition(firstargvalue, ((LogicObject*)(firstarg)));
      }
      if (!((boolean)(equatingprop))) {
        equatingprop = findEquatingProposition(firstargvalue, ((LogicObject*)(firstarg)));
        while (!eqlP(valueOf(firstarg), firstargvalue)) {
          firstarg = valueOf(firstarg);
        }
        { Skolem* skolem = ((Skolem*)(firstarg));

          { Skolem* object000 = skolem;
            Object* value000 = NULL;
            Object* oldValue001 = object000->variableValue;
            Object* newValue000 = updateInContext(oldValue001, value000, object000->homeContext, FALSE);

            if (!(((boolean)(oldValue001)) &&
                (oldValue001->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
              object000->variableValue = newValue000;
            }
          }
          setDynamicSlotValue(skolem->dynamicSlots, SYM_PROPOSITIONS_LOGIC_CONFLICTING_DEFAULT_VALUES, cons(equatingprop, skolem->conflictingDefaultValues_reader()), NULL);
          if (newisdefaultP) {
            setDynamicSlotValue(skolem->dynamicSlots, SYM_PROPOSITIONS_LOGIC_CONFLICTING_DEFAULT_VALUES, cons(newequatingprop, skolem->conflictingDefaultValues_reader()), NULL);
          }
        }
      }
    }
    if (secondargisskolemP) {
      { 
        BIND_STELLA_SPECIAL(oDONTUSEDEFAULTKNOWLEDGEpo, boolean, TRUE);
        equatingprop = findEquatingProposition(secondargvalue, ((LogicObject*)(secondarg)));
      }
      if (!((boolean)(equatingprop))) {
        equatingprop = findEquatingProposition(firstargvalue, ((LogicObject*)(firstarg)));
        while (!eqlP(valueOf(secondarg), secondargvalue)) {
          secondarg = valueOf(secondarg);
        }
        { Skolem* skolem = ((Skolem*)(secondarg));

          { Skolem* object001 = skolem;
            Object* value001 = NULL;
            Object* oldValue002 = object001->variableValue;
            Object* newValue001 = updateInContext(oldValue002, value001, object001->homeContext, FALSE);

            if (!(((boolean)(oldValue002)) &&
                (oldValue002->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
              object001->variableValue = newValue001;
            }
          }
          setDynamicSlotValue(skolem->dynamicSlots, SYM_PROPOSITIONS_LOGIC_CONFLICTING_DEFAULT_VALUES, cons(equatingprop, skolem->conflictingDefaultValues_reader()), NULL);
          if (newisdefaultP) {
            setDynamicSlotValue(skolem->dynamicSlots, SYM_PROPOSITIONS_LOGIC_CONFLICTING_DEFAULT_VALUES, cons(newequatingprop, skolem->conflictingDefaultValues_reader()), NULL);
          }
        }
      }
    }
    newequatingprop->truthValue = ((TruthValue*)(newtruthvalue));
    return (((!firstargisskolemP) &&
        (!secondargisskolemP)) ||
        (!newisdefaultP));
  }
}

boolean unequateValuesP(Object* term1, Object* term2) {
  if (subtypeOfP(safePrimaryType(term1), SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
    { Object* term1000 = term1;
      Skolem* term1 = ((Skolem*)(term1000));

      if (eqlP(((Object*)(accessInContext(term1->variableValue, term1->homeContext, FALSE))), term2)) {
        unbindSkolemValue(term1, FALSE);
        return (TRUE);
      }
    }
  }
  else {
  }
  return (FALSE);
}

void bindToEquivalentValue(LogicObject* from, Description* to) {
  if (from == to) {
    return;
  }
  if (subtypeOfP(safePrimaryType(from), SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
    { LogicObject* from000 = from;
      Skolem* from = ((Skolem*)(from000));

      { Skolem* object000 = from;
        Object* value000 = to;
        Object* oldValue000 = object000->variableValue;
        Object* newValue000 = updateInContext(oldValue000, value000, object000->homeContext, FALSE);

        if (!(((boolean)(oldValue000)) &&
            (oldValue000->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
          object000->variableValue = newValue000;
        }
      }
    }
  }
  else {
    if (((Description*)(accessInContext(dynamicSlotValue(from->dynamicSlots, SYM_PROPOSITIONS_LOGIC_EQUIVALENT_VALUE, NULL), from->homeContext, FALSE))) == to) {
      return;
    }
    { LogicObject* object001 = from;
      Description* value001 = to;
      Object* oldValue002 = dynamicSlotValue(object001->dynamicSlots, SYM_PROPOSITIONS_LOGIC_EQUIVALENT_VALUE, NULL);
      Object* newValue001 = updateInContext(oldValue002, value001, object001->homeContext, FALSE);

      if (!(((boolean)(oldValue002)) &&
          (oldValue002->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
        setDynamicSlotValue(object001->dynamicSlots, SYM_PROPOSITIONS_LOGIC_EQUIVALENT_VALUE, newValue001, NULL);
      }
    }
  }
  { Description* object002 = to;
    Cons* value002 = cons(from, from->variableValueInverse_reader());
    Object* oldValue003 = object002->variableValueInverse;
    Object* newValue002 = updateInContext(oldValue003, value002, object002->homeContext, FALSE);

    if (!(((boolean)(oldValue003)) &&
        (oldValue003->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
      object002->variableValueInverse = newValue002;
    }
  }
}

void unbindEquivalentValue(LogicObject* self) {
  if (subtypeOfP(safePrimaryType(self), SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
    { LogicObject* self000 = self;
      Skolem* self = ((Skolem*)(self000));

      unbindSkolemValue(self, FALSE);
    }
  }
  else {
    unlinkEquivalentValue(self);
    handleOutOfDateInferenceCache(KWD_PROPOSITIONS_RETRACT, NULL);
  }
}

void unlinkEquivalentValue(LogicObject* self) {
  { Object* oldvalue = NULL;

    if (subtypeOfP(safePrimaryType(self), SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
      { LogicObject* self000 = self;
        Skolem* self = ((Skolem*)(self000));

        oldvalue = valueOf(self);
        { Skolem* object000 = self;
          Object* value000 = NULL;
          Object* oldValue000 = object000->variableValue;
          Object* newValue000 = updateInContext(oldValue000, value000, object000->homeContext, FALSE);

          if (!(((boolean)(oldValue000)) &&
              (oldValue000->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
            object000->variableValue = newValue000;
          }
        }
      }
    }
    else {
      oldvalue = equivalentValueOf(self);
      { LogicObject* object001 = self;
        Description* value001 = NULL;
        Object* oldValue001 = dynamicSlotValue(object001->dynamicSlots, SYM_PROPOSITIONS_LOGIC_EQUIVALENT_VALUE, NULL);
        Object* newValue001 = updateInContext(oldValue001, value001, object001->homeContext, FALSE);

        if (!(((boolean)(oldValue001)) &&
            (oldValue001->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
          setDynamicSlotValue(object001->dynamicSlots, SYM_PROPOSITIONS_LOGIC_EQUIVALENT_VALUE, newValue001, NULL);
        }
      }
    }
    if (subtypeOfP(safePrimaryType(oldvalue), SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* oldvalue000 = oldvalue;
        LogicObject* oldvalue = ((LogicObject*)(oldvalue000));

        if (((LogicObject*)(oldvalue->variableValueInverse_reader()->value)) == self) {
          { LogicObject* object002 = oldvalue;
            Cons* value002 = oldvalue->variableValueInverse_reader()->rest;
            Object* oldValue002 = object002->variableValueInverse;
            Object* newValue002 = updateInContext(oldValue002, value002, object002->homeContext, FALSE);

            if (!(((boolean)(oldValue002)) &&
                (oldValue002->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
              object002->variableValueInverse = newValue002;
            }
          }
        }
        else {
          { LogicObject* object003 = oldvalue;
            Cons* value003 = copyConsList(oldvalue->variableValueInverse_reader())->remove(self);
            Object* oldValue003 = object003->variableValueInverse;
            Object* newValue003 = updateInContext(oldValue003, value003, object003->homeContext, FALSE);

            if (!(((boolean)(oldValue003)) &&
                (oldValue003->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
              object003->variableValueInverse = newValue003;
            }
          }
        }
      }
    }
    else {
    }
  }
}

int equivalentCollectionPriority(LogicObject* self) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
      { LogicObject* self000 = self;
        NamedDescription* self = ((NamedDescription*)(self000));

        return (1);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) {
      { LogicObject* self001 = self;
        Description* self = ((Description*)(self001));

        return (2);
      }
    }
    else {
      return (0);
    }
  }
}

// Experiment with equality reasoning on collections.
boolean oEQUIVALENCE_COLLECTIONSpo = FALSE;

void equivalenceCollections(LogicObject* collection1, LogicObject* collection2) {
  if (!oEQUIVALENCE_COLLECTIONSpo) {
    return;
  }
  { Object* value1 = equivalentValueOf(valueOf(collection1));
    Object* value2 = equivalentValueOf(valueOf(collection2));

    if (equivalentCollectionPriority(((LogicObject*)(value1))) > equivalentCollectionPriority(((LogicObject*)(value2)))) {
      bindToEquivalentValue(((LogicObject*)(value2)), ((Description*)(value1)));
    }
    else {
      bindToEquivalentValue(((LogicObject*)(value1)), ((Description*)(value2)));
    }
  }
}

void reviseEquivalences(Proposition* proposition, boolean goestrueP) {
  { LogicObject* arg1 = ((LogicObject*)((proposition->arguments->theArray)[0]));
    LogicObject* arg2 = ((LogicObject*)((proposition->arguments->theArray)[1]));

    if (!oEQUIVALENCE_COLLECTIONSpo) {
      return;
    }
    if (goestrueP) {
      if (collectionImpliesCollectionP(arg1, arg2) &&
          collectionImpliesCollectionP(arg2, arg1)) {
        equivalenceCollections(arg1, arg2);
      }
    }
    else if (((boolean)(((Description*)(accessInContext(dynamicSlotValue(arg1->dynamicSlots, SYM_PROPOSITIONS_LOGIC_EQUIVALENT_VALUE, NULL), arg1->homeContext, FALSE))))) ||
        ((boolean)(((Description*)(accessInContext(dynamicSlotValue(arg2->dynamicSlots, SYM_PROPOSITIONS_LOGIC_EQUIVALENT_VALUE, NULL), arg2->homeContext, FALSE)))))) {
      std::cout << "DIDN'T FINISH 'revise-equivalences' YET. DON'T KNOW HOW" << std::endl << "   TO REMOVE AN 'equivalent-value' LINK" << std::endl;
    }
    else if ((isaP(arg1, SGT_PROPOSITIONS_LOGIC_SKOLEM) &&
        ((boolean)(((Object*)(accessInContext(((Skolem*)(arg1))->variableValue, ((Skolem*)(arg1))->homeContext, FALSE)))))) ||
        (isaP(arg2, SGT_PROPOSITIONS_LOGIC_SKOLEM) &&
         ((boolean)(((Object*)(accessInContext(((Skolem*)(arg2))->variableValue, ((Skolem*)(arg2))->homeContext, FALSE))))))) {
      std::cout << "DIDN'T FINISH 'revise-equivalences' YET. DON'T KNOW HOW" << std::endl << "   TO REMOVE AN 'equivalent-value' LINK" << std::endl;
    }
  }
}

Proposition* findEquatingProposition(Object* directobject, LogicObject* indirectobject) {
  { Object* indirectobjectvalue = valueOf(indirectobject);
    Object* backlinkedindirectobject = indirectobjectvalue;

    if (!isaP(backlinkedindirectobject, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      backlinkedindirectobject = indirectobject;
    }
    { Proposition* value000 = NULL;

      { Proposition* prop = NULL;
        Iterator* iter000 = ((Iterator*)((descriptionModeP() ? unfilteredDependentPropositions(backlinkedindirectobject, SGT_PROPOSITIONS_PL_KERNEL_KB_EQUIVALENT)->allocateIterator() : allTrueDependentPropositions(backlinkedindirectobject, SGT_PROPOSITIONS_PL_KERNEL_KB_EQUIVALENT, FALSE))));

        while (iter000->nextP()) {
          prop = ((Proposition*)(iter000->value));
          if ((prop->operatoR == SGT_PROPOSITIONS_PL_KERNEL_KB_EQUIVALENT) &&
              ((eqlP((prop->arguments->theArray)[0], directobject) &&
              eqlP(valueOf((prop->arguments->theArray)[1]), indirectobjectvalue)) ||
               (eqlP((prop->arguments->theArray)[1], directobject) &&
                eqlP(valueOf((prop->arguments->theArray)[0]), indirectobjectvalue)))) {
            value000 = prop;
            break;
          }
        }
      }
      { Proposition* value001 = value000;

        return (value001);
      }
    }
  }
}

Proposition* TRUE_PROPOSITION = NULL;

Proposition* FALSE_PROPOSITION = NULL;

Object* evaluateTerm(Object* self) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_LITERAL_WRAPPER)) {
      { Object* self000 = self;
        LiteralWrapper* self = ((LiteralWrapper*)(self000));

        return (evaluateLiteralWrapperTerm(self));
      }
    }
    else if (subtypeOfSurrogateP(testValue000)) {
      { Object* self001 = self;
        Surrogate* self = ((Surrogate*)(self001));

        return (evaluateSurrogateTerm(self));
      }
    }
    else if (subtypeOfSymbolP(testValue000)) {
      { Object* self002 = self;
        Symbol* self = ((Symbol*)(self002));

        return (evaluateSymbolTerm(self));
      }
    }
    else if (testValue000 == SGT_PROPOSITIONS_STELLA_CONS) {
      { Object* self003 = self;
        Cons* self = ((Cons*)(self003));

        { GeneralizedSymbol* testValue001 = ((GeneralizedSymbol*)(self->value));

          if ((testValue001 == SYM_PROPOSITIONS_LOGIC_SETOFALL) ||
              ((testValue001 == SYM_PROPOSITIONS_LOGIC_THE_ONLY) ||
               (testValue001 == SYM_PROPOSITIONS_LOGIC_KAPPA))) {
            return (evaluateLogicObjectTerm(evaluateDescriptionTerm(self, TRUE)));
          }
          else if (testValue001 == SYM_PROPOSITIONS_LOGIC_CREATE) {
            return (evaluateCreateTerm(self));
          }
          else {
            return (evaluateFunctionTerm(self));
          }
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* self004 = self;
        LogicObject* self = ((LogicObject*)(self004));

        return (evaluateLogicObjectTerm(self));
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_THING)) {
      { Object* self005 = self;
        Thing* self = ((Thing*)(self005));

        return (self);
      }
    }
    else if (subtypeOfKeywordP(testValue000)) {
      { Object* self006 = self;
        Keyword* self = ((Keyword*)(self006));

        return (self);
      }
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "Can't evaluate the term " << "`" << self << "'";
        throw *newStellaException(stream000->theStringReader());
      }
    }
  }
}

Object* evaluateLiteralWrapperTerm(LiteralWrapper* self) {
  return (((LiteralWrapper*)(self->permanentify())));
}

Object* evaluateSurrogateTerm(Surrogate* self) {
  if (oEVALUATIONMODEo.get() == KWD_PROPOSITIONS_EXTENSIONAL_ASSERTION) {
    { Object* value = self->surrogateValue;
      ObjectStore* store = NULL;

      if (((boolean)(value))) {
        if (subtypeOfP(safePrimaryType(value), SGT_PROPOSITIONS_STELLA_RELATION)) {
          { Object* value000 = value;
            Relation* value = ((Relation*)(value000));

            return (getDescription(self));
          }
        }
        else {
          return (value);
        }
      }
      else {
        self = coerceToBoundOrLocalSurrogate(self);
        store = homeObjectStore(self);
        if (((boolean)(store))) {
          store->fetchInstance(self);
        }
        if (!((boolean)(self->surrogateValue))) {
          createLogicInstance(self, NULL);
        }
        return (self->surrogateValue);
      }
    }
  }
  else if (oEVALUATIONMODEo.get() == KWD_PROPOSITIONS_DESCRIPTION) {
    { Object* surrogatevalue = self->surrogateValue;
      ObjectStore* store = NULL;

      if (!((boolean)(surrogatevalue))) {
        self = coerceToBoundOrLocalSurrogate(self);
        store = homeObjectStore(self);
        if (((boolean)(store))) {
          store->fetchInstance(self);
        }
        if (!((boolean)(self->surrogateValue))) {
          createLogicInstance(self, NULL);
        }
        return (self);
      }
      else if (isaP(surrogatevalue, SGT_PROPOSITIONS_STELLA_LITERAL_WRAPPER)) {
        return (surrogatevalue);
      }
      else {
        return (self);
      }
    }
  }
  else if (oEVALUATIONMODEo.get() == KWD_PROPOSITIONS_INTENSIONAL_ASSERTION) {
    { Object* directvalue = self->surrogateValue;
      Skolem* intensionalskolem = NULL;

      if (skolemP(directvalue)) {
        return (directvalue);
      }
      intensionalskolem = createSkolem((((boolean)(directvalue)) ? logicalType(valueOf(directvalue)) : NULL), NULL);
      self->surrogateValue = intensionalskolem;
      bindSkolemToValue(intensionalskolem, directvalue, FALSE);
      return (intensionalskolem);
    }
  }
  else {
    { OutputStringStream* stream000 = newOutputStringStream();

      *(stream000->nativeStream) << "`" << oEVALUATIONMODEo.get() << "'" << " is not a valid case option";
      throw *newStellaException(stream000->theStringReader());
    }
  }
}

Object* evaluateLogicObjectTerm(LogicObject* self) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE)) {
      { LogicObject* self000 = self;
        PatternVariable* self = ((PatternVariable*)(self000));

      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
      { LogicObject* self001 = self;
        Skolem* self = ((Skolem*)(self001));

        if (!descriptionModeP()) {
          return (self);
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) {
      { LogicObject* self002 = self;
        Description* self = ((Description*)(self002));

        if (descriptionModeP()) {
          { Surrogate* name = self->surrogateValueInverse;

            if (((boolean)(name))) {
              return (name);
            }
          }
        }
        return (self);
      }
    }
    else {
      if (descriptionModeP()) {
        { Surrogate* name = self->surrogateValueInverse;

          if (((boolean)(name))) {
            return (name);
          }
        }
      }
      else {
        return (self);
      }
    }
  }
  { OutputStringStream* stream000 = newOutputStringStream();

    { 
      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
      *(stream000->nativeStream) << "ERROR: " << "Illegal LOGIC-OBJECT term: " << "`" << self << "'" << "." << std::endl;
      helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
    }
    throw *newPropositionError(stream000->theStringReader());
  }
}

// Variables with name 'ANONYMOUS-VARIABLE-NAME' are considered
// anonymous, and are not assumed to be identical to any other variable also named
// 'ANONYMOUS-VARIABLE-NAME'.
Symbol* ANONYMOUS_VARIABLE_NAME = NULL;

boolean anonymousVariableP(Skolem* self) {
  return ((self->skolemName == ANONYMOUS_VARIABLE_NAME) &&
      variableP(self));
}

Object* evaluateSymbolTerm(Symbol* self) {
  { Object* value = lookupLogicVariableBinding(self);

    if (((boolean)(value))) {
      return (value);
    }
    else if (self == ANONYMOUS_VARIABLE_NAME) {
      return (createVariable(NULL, ANONYMOUS_VARIABLE_NAME, FALSE));
    }
    else if (questionMarkSymbolP(self)) {
      { OutputStringStream* stream000 = newOutputStringStream();

        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          *(stream000->nativeStream) << "ERROR: " << "Reference to undeclared variable: " << "`" << self << "'" << "." << std::endl;
          helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
        }
        throw *newPropositionError(stream000->theStringReader());
      }
    }
    else if (automaticInstanceSymbolP(self)) {
      return (evaluateAutomaticInstance(self));
    }
    else {
      value = symbolToSurrogate(self)->surrogateValue;
      if (((boolean)(value))) {
        return (value);
      }
      else {
        { OutputStringStream* stream001 = newOutputStringStream();

          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            *(stream001->nativeStream) << "ERROR: " << "Reference to undeclared symbol: " << "`" << self << "'" << "." << std::endl;
            helpSignalPropositionError(stream001, KWD_PROPOSITIONS_ERROR);
          }
          throw *newPropositionError(stream001->theStringReader());
        }
      }
    }
  }
}

// Represents a key-value list that should never be used.
KeyValueList* oMISSING_KEY_VALUE_LISTo = NULL;

// Used by 'evaluate-automatic-instance' to
// record current bindings of automatic instance symbols.
DEFINE_STELLA_SPECIAL(oAUTOMATICINSTANCETABLEo, KeyValueList* , NULL);

Object* evaluateAutomaticInstance(Symbol* name) {
  if (oAUTOMATICINSTANCETABLEo.get() == oMISSING_KEY_VALUE_LISTo) {
    { OutputStringStream* stream000 = newOutputStringStream();

      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        *(stream000->nativeStream) << "ERROR: " << "Automatic instance " << "`" << name << "'" << " found outside" << std::endl << "   the scope of an assertion." << "." << std::endl;
        helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
      }
      throw *newPropositionError(stream000->theStringReader());
    }
  }
  { LogicObject* instance = (((boolean)(oAUTOMATICINSTANCETABLEo.get())) ? ((LogicObject*)(oAUTOMATICINSTANCETABLEo.get()->lookup(name))) : ((LogicObject*)(NULL)));

    if (!((boolean)(instance))) {
      { Surrogate* surrogate = stringToSurrogate(gensym(stringSubsequence(name->symbolName, 1, NULL_INTEGER))->symbolName);

        instance = ((LogicObject*)(createLogicInstance(surrogate, NULL)));
        if (!((boolean)(oAUTOMATICINSTANCETABLEo.get()))) {
          oAUTOMATICINSTANCETABLEo.set(newKeyValueList());
        }
        oAUTOMATICINSTANCETABLEo.get()->insertAt(name, instance);
      }
    }
    return (instance);
  }
}

void signalUndefinedTerm(Object* term) {
  { OutputStringStream* stream000 = newOutputStringStream();

    { 
      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
      *(stream000->nativeStream) << "ERROR: " << "Reference to undefined term: " << "`" << term << "'" << "." << std::endl;
      helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
    }
    throw *newPropositionError(stream000->theStringReader());
  }
}

void signalUntypedTerm(Object* term) {
  if (!oSUPPRESSUNTYPEDTYPEERRORpo.get()) {
    { OutputStringStream* stream000 = newOutputStringStream();

      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        *(stream000->nativeStream) << "ERROR: " << "Need a type for the untyped term: " << "`" << term << "'" << "." << std::endl;
        helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
      }
      throw *newPropositionError(stream000->theStringReader());
    }
  }
}

Object* evaluateCreateTerm(Cons* term) {
  { Symbol* namesymbol = ((Symbol*)(term->rest->value));
    Surrogate* name = (((boolean)(namesymbol)) ? symbolToSurrogate(namesymbol) : ((Surrogate*)(NULL)));
    Symbol* typesymbol = ((Symbol*)(term->rest->rest->value));
    Surrogate* type = (((boolean)(typesymbol)) ? symbolToSurrogate(typesymbol) : ((Surrogate*)(NULL)));
    Object* term = NULL;

    try {
      term = createLogicInstance(name, type);
    }
    catch (LogicException& _e) {
      LogicException* e = &_e;

      *(STANDARD_ERROR->nativeStream) << exceptionMessage(e);
    }
    return (term);
  }
}

Object* create(GeneralizedSymbol* name, Cons* type) {
  // Create a logic object with name 'name' and return it.
  // If `type' is also supplied, assert that the object belongs to that type.
  { Cons* createtree = listO(3, SYM_PROPOSITIONS_LOGIC_CREATE, name, type->concatenate(NIL, 0));
    Object* result = evaluateCreateTerm(createtree);

    return (result);
  }
}

Object* createEvaluatorWrapper(Cons* arguments) {
  return (create(((GeneralizedSymbol*)(arguments->value)), arguments->rest));
}

void updateSkolemType(Skolem* self, Surrogate* type) {
  { Surrogate* oldtype = self->skolemType;

    if ((!((boolean)(oldtype))) ||
        logicalSubtypeOfP(type, oldtype)) {
      self->skolemType = type;
    }
  }
}

void updateSkolemTypeFromIsaAssertions(Proposition* self) {
  { Keyword* testValue000 = self->kind;

    if (testValue000 == KWD_PROPOSITIONS_ISA) {
      { Object* object = (self->arguments->theArray)[0];

        if (isaP(object, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
          updateSkolemType(((Skolem*)(object)), ((Surrogate*)(self->operatoR)));
        }
      }
    }
    else if (testValue000 == KWD_PROPOSITIONS_AND) {
      { Object* arg = NULL;
        Vector* vector000 = self->arguments;
        int index000 = 0;
        int length000 = vector000->length();

        while (index000 < length000) {
          arg = (vector000->theArray)[index000];
          index000 = index000 + 1;
          updateSkolemTypeFromIsaAssertions(((Proposition*)(arg)));
        }
      }
    }
    else {
    }
  }
}

boolean descriptionExtensionL(Description* desc1, Description* desc2) {
  return (getDescriptionExtension(((NamedDescription*)(desc1)), TRUE)->estimatedLength() < getDescriptionExtension(((NamedDescription*)(desc2)), TRUE)->estimatedLength());
}

void printExtensionSizes(Module* module, int sizecutoff) {
  // Print the extension sizes of concepts visible in `module'.
  // If `module' is NULL the current module is used.  Do not report extensions
  // with size less than `sizeCutoff' (default is 10).
  if (sizecutoff == NULL_INTEGER) {
    sizecutoff = 10;
  }
  if (!((boolean)(module))) {
    module = oMODULEo.get();
  }
  { List* descriptions = newList();
    char* size = NULL;

    { NamedDescription* d = NULL;
      Iterator* iter000 = allNamedDescriptions(module, FALSE);
      Cons* collect000 = NULL;

      while (iter000->nextP()) {
        d = ((NamedDescription*)(iter000->value));
        if (((boolean)(d->extension)) &&
            (d->extension->estimatedLength() >= sizecutoff)) {
          if (!((boolean)(collect000))) {
            {
              collect000 = cons(d, NIL);
              if (descriptions->theConsList == NIL) {
                descriptions->theConsList = collect000;
              }
              else {
                addConsToEndOfConsList(descriptions->theConsList, collect000);
              }
            }
          }
          else {
            {
              collect000->rest = cons(d, NIL);
              collect000 = collect000->rest;
            }
          }
        }
      }
    }
    descriptions->sort(((cpp_function_code)(&descriptionExtensionL)));
    { NamedDescription* desc = NULL;
      Cons* iter001 = descriptions->reverse()->theConsList;

      while (!(iter001 == NIL)) {
        desc = ((NamedDescription*)(iter001->value));
        iter001 = iter001->rest;
        size = integerToString(desc->extension->estimatedLength());
        size = stringConcatenate(makeString(8 - strlen(size), ' '), size, 0);
        std::cout << size << " : " << desc->descriptionName() << std::endl;
      }
    }
  }
}

void transferLogicInformationFromRelationHook(List* relationslist) {
  { Relation* oldrelation = ((Relation*)(relationslist->first()));
    Relation* newrelation = ((Relation*)(relationslist->second()));
    NamedDescription* olddescription = ((NamedDescription*)(dynamicSlotValue(oldrelation->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTION, NULL)));
    NamedDescription* newdescription = ((NamedDescription*)(dynamicSlotValue(newrelation->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTION, NULL)));

    if (((boolean)(olddescription))) {
      retractRelationAxioms(olddescription);
      if (((boolean)(newdescription))) {
        {
          transferPropositionsAndBacklinks(((NamedDescription*)(dynamicSlotValue(oldrelation->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTION, NULL))), ((NamedDescription*)(dynamicSlotValue(newrelation->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTION, NULL))));
          destroyInstance(olddescription);
        }
      }
      else {
        {
          setDynamicSlotValue(newrelation->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTION, olddescription, NULL);
          olddescription->nativeRelationSetter(newrelation);
          finalizeSuperrelationLinks(newrelation);
        }
      }
      setDynamicSlotValue(oldrelation->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTION, NULL, NULL);
    }
    if (subtypeOfClassP(safePrimaryType(oldrelation))) {
      { Relation* oldrelation000 = oldrelation;
        Class* oldrelation = ((Class*)(oldrelation000));

        setDynamicSlotValue(((Class*)(newrelation))->dynamicSlots, SYM_PROPOSITIONS_STELLA_CLASS_EXTENSION, ((ClassExtension*)(dynamicSlotValue(oldrelation->dynamicSlots, SYM_PROPOSITIONS_STELLA_CLASS_EXTENSION, NULL))), NULL);
        setDynamicSlotValue(oldrelation->dynamicSlots, SYM_PROPOSITIONS_STELLA_CLASS_EXTENSION, NULL, NULL);
      }
    }
    else {
    }
  }
}

void retractRelationAxioms(NamedDescription* oldrelation) {
  { List* originatedpropositions = originatedPropositions(oldrelation);

    if (((boolean)(originatedpropositions))) {
      { Proposition* proposition = NULL;
        Cons* iter000 = originatedpropositions->theConsList;

        while (!(iter000 == NIL)) {
          proposition = ((Proposition*)(iter000->value));
          iter000 = iter000->rest;
          unassertProposition(proposition);
        }
      }
    }
  }
}

Proposition* assertDescriptionImpliesDescription(Description* sub, Description* super, boolean dontupdateP) {
  { Proposition* impliesprop = createImpliesProposition(sub, super);

    impliesprop = fastenDownOneProposition(impliesprop, FALSE);
    if (!(dontupdateP)) {
      updatePropositionTruthValue(impliesprop, KWD_PROPOSITIONS_ASSERT_TRUE);
    }
    return (impliesprop);
  }
}

Proposition* assertImpliesLinkBetweenRelations(NamedDescription* sub, NamedDescription* super) {
  { Proposition* impliesprop = assertDescriptionImpliesDescription(sub, super, TRUE);

    setDynamicSlotValue(impliesprop->dynamicSlots, SYM_PROPOSITIONS_LOGIC_SUBRELATION_LINKp, (TRUE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
    updatePropositionTruthValue(impliesprop, KWD_PROPOSITIONS_ASSERT_TRUE);
    linkOriginatedProposition(sub, impliesprop);
    return (impliesprop);
  }
}

void finalizeSuperrelationLinks(Relation* self) {
  { NamedDescription* subdescription = getDescription(self);

    { 
      BIND_STELLA_SPECIAL(oEVALUATIONMODEo, Keyword*, KWD_PROPOSITIONS_EXTENSIONAL_ASSERTION);
      BIND_STELLA_SPECIAL(oFILLINGCONSTRAINTPROPAGATIONQUEUESpo, boolean, FALSE);
      { 
        BIND_STELLA_SPECIAL(oMODULEo, Module*, ((Module*)(subdescription->homeContext)));
        BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
        { Relation* super = NULL;
          Iterator* iter000 = directSuperrelations(self);

          while (iter000->nextP()) {
            super = ((Relation*)(iter000->value));
            if (((boolean)(((NamedDescription*)(dynamicSlotValue(super->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTION, NULL)))))) {
              assertImpliesLinkBetweenRelations(subdescription, getDescription(super));
            }
          }
        }
      }
    }
  }
}

Proposition* updateIsaProposition(Object* self, Surrogate* type, Keyword* updatemode) {
  { Proposition* value000 = NULL;

    { Proposition* p = NULL;
      Iterator* iter000 = unfilteredDependentIsaPropositions(self)->allocateIterator();

      while (iter000->nextP()) {
        p = ((Proposition*)(iter000->value));
        if (((Surrogate*)(p->operatoR)) == type) {
          value000 = p;
          break;
        }
      }
    }
    { Proposition* isaproposition = value000;

      if (!((boolean)(isaproposition))) {
        isaproposition = createProposition(SYM_PROPOSITIONS_STELLA_ISA, 1);
        isaproposition->operatoR = type;
        (isaproposition->arguments->theArray)[0] = self;
        fastenDownOneProposition(isaproposition, TRUE);
      }
      updatePropositionTruthValue(isaproposition, updatemode);
      return (isaproposition);
    }
  }
}

Proposition* assertIsaProposition(Object* self, Surrogate* type) {
  if (((boolean)(type))) {
    return (updateIsaProposition(self, type, KWD_PROPOSITIONS_ASSERT_TRUE));
  }
  else {
    return (NULL);
  }
}

Proposition* retractIsaProposition(Object* self, Surrogate* type) {
  if (((boolean)(type))) {
    return (updateIsaProposition(self, type, KWD_PROPOSITIONS_RETRACT_TRUE));
  }
  else {
    return (NULL);
  }
}

Proposition* updateProperty(LogicObject* self, Surrogate* surrogate, Keyword* updatemode) {
  { Proposition* value000 = NULL;

    { Proposition* p = NULL;
      Iterator* iter000 = unfilteredDependentPropositions(self, surrogate)->allocateIterator();

      while (iter000->nextP()) {
        p = ((Proposition*)(iter000->value));
        if ((((Surrogate*)(p->operatoR)) == surrogate) &&
            (p->arguments->length() == 1)) {
          value000 = p;
          break;
        }
      }
    }
    { Proposition* propertyproposition = value000;

      if ((!((boolean)(surrogate))) ||
          (!((boolean)(surrogate->surrogateValue)))) {
        std::cout << "Error: Property " << surrogate << " is not defined." << std::endl;
        return (NULL);
      }
      if (!((boolean)(propertyproposition))) {
        propertyproposition = createProposition(SYM_PROPOSITIONS_STELLA_PREDICATE, 1);
        (propertyproposition->arguments->theArray)[0] = self;
        propertyproposition->operatoR = surrogate;
        fastenDownOneProposition(propertyproposition, TRUE);
      }
      updatePropositionTruthValue(propertyproposition, updatemode);
      return (propertyproposition);
    }
  }
}

Proposition* assertProperty(LogicObject* instance, Surrogate* property) {
  return (updateProperty(instance, property, KWD_PROPOSITIONS_ASSERT_TRUE));
}

Proposition* retractProperty(LogicObject* instance, Surrogate* property) {
  return (updateProperty(instance, property, KWD_PROPOSITIONS_RETRACT_TRUE));
}

Proposition* updateTuple(Surrogate* relation, Cons* arguments, Keyword* updatemode) {
  { NamedDescription* description = getDescription(relation);
    Proposition* proposition = NULL;

    if (functionDescriptionP(description)) {
      { Proposition* definingproposition = findOrCreateFunctionProposition(relation, arguments->butLast()->consify());

        proposition = updateEquivalenceProposition(createEquivalenceProposition((definingproposition->arguments->theArray)[(definingproposition->arguments->length() - 1)], arguments->last()), updatemode);
        if (!((boolean)(proposition))) {
          return (definingproposition);
        }
      }
    }
    else if (classDescriptionP(description)) {
      return (updateIsaProposition(arguments->value, relation, updatemode));
    }
    else {
      proposition = createProposition(SYM_PROPOSITIONS_STELLA_PREDICATE, arguments->length());
      proposition->operatoR = relation;
      { Object* arg = NULL;
        int i = NULL_INTEGER;
        Cons* iter000 = arguments;
        int iter001 = 0;

        while (!(iter000 == NIL)) {
          arg = iter000->value;
          iter000 = iter000->rest;
          i = iter001;
          iter001 = iter001 + 1;
          (proposition->arguments->theArray)[i] = arg;
        }
      }
    }
    proposition = fastenDownOneProposition(proposition, FALSE);
    updatePropositionTruthValue(proposition, updatemode);
    return (proposition);
  }
}

Proposition* assertTuple(Surrogate* relation, Cons* arguments) {
  return (updateTuple(relation, arguments, KWD_PROPOSITIONS_ASSERT_TRUE));
}

Proposition* updateBinaryValue(Surrogate* relation, Object* self, Object* value, Keyword* updatemode) {
  { 
    BIND_STELLA_SPECIAL(oCLIPPINGENABLEDpo, boolean, TRUE);
    if (functionDescriptionP(getDescription(relation))) {
      { Proposition* value000 = NULL;

        { Proposition* p = NULL;
          Iterator* iter000 = unfilteredDependentPropositions(self, relation)->allocateIterator();

          while (iter000->nextP()) {
            p = ((Proposition*)(iter000->value));
            if (((Surrogate*)(p->operatoR)) == relation) {
              value000 = p;
              break;
            }
          }
        }
        { Proposition* proposition = value000;

          if ((!((boolean)(proposition))) ||
              skolemP(value)) {
            return (updateTuple(relation, consList(2, self, value), updatemode));
          }
          { Object* skolemoutput = (proposition->arguments->theArray)[(proposition->arguments->length() - 1)];

            if (updatemode == KWD_PROPOSITIONS_ASSERT_TRUE) {
              if (clipValueP(skolemoutput, value)) {
              }
              else {
                equateValues(valueOf(skolemoutput), valueOf(value));
              }
            }
            else if (updatemode == KWD_PROPOSITIONS_RETRACT_TRUE) {
              unequateValuesP(skolemoutput, value);
            }
            else {
              { OutputStringStream* stream000 = newOutputStringStream();

                *(stream000->nativeStream) << "`" << updatemode << "'" << " is not a valid case option";
                throw *newStellaException(stream000->theStringReader());
              }
            }
            return (proposition);
          }
        }
      }
    }
    else {
      return (updateTuple(relation, consList(2, self, value), updatemode));
    }
  }
}

Proposition* assertBinaryValue(Surrogate* relation, Object* self, Object* value) {
  return (updateBinaryValue(relation, self, value, KWD_PROPOSITIONS_ASSERT_TRUE));
}

Proposition* retractBinaryValue(Surrogate* relation, Object* self, Object* value) {
  return (updateBinaryValue(relation, self, value, KWD_PROPOSITIONS_RETRACT_TRUE));
}

Proposition* assertMemberOfProposition(Object* self, Object* collection) {
  if (!((boolean)(collection))) {
    return (NULL);
  }
  if (subtypeOfP(safePrimaryType(collection), SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
    { Object* collection000 = collection;
      NamedDescription* collection = ((NamedDescription*)(collection000));

      if (collection->surrogateValueInverse == SGT_PROPOSITIONS_STELLA_THING) {
        return (NULL);
      }
      return (assertIsaProposition(self, collection->surrogateValueInverse));
    }
  }
  else {
    return (assertTuple(SGT_PROPOSITIONS_PL_KERNEL_KB_MEMBER_OF, consList(2, self, collection)));
  }
}

DEFINE_STELLA_SPECIAL(oDEFAULTCREATIONTYPEo, Surrogate* , NULL);

Object* createLogicInstance(Surrogate* name, Surrogate* type) {
  enforceCodeOnly();
  if (!((boolean)(type))) {
    type = oDEFAULTCREATIONTYPEo.get();
  }
  { NamedDescription* description = (((boolean)(type)) ? getDescription(type) : ((NamedDescription*)(NULL)));
    Class* nativeclass = NULL;
    Object* newobject = NULL;

    if (((boolean)(type)) &&
        (!((boolean)(description)))) {
      std::cout << "Can't create an instance of the undefined class " << "`" << type->symbolName << "'" << std::endl << std::endl;
      return (NULL);
    }
    if (!((boolean)(description))) {
      newobject = newLogicObject();
    }
    else {
      {
        if (!classDescriptionP(description)) {
          std::cout << "Can't create an instance of a relation: " << "`" << type->symbolName << "'" << std::endl;
          return (NULL);
        }
        nativeclass = ((Class*)(description->nativeRelation()));
        if ((!((boolean)(nativeclass))) ||
            nativeclass->abstractP) {
          newobject = newLogicObject();
        }
        else if (subtypeOfP(nativeclass->classType, SGT_PROPOSITIONS_STELLA_THING)) {
          newobject = createObject(type, 0);
        }
        else {
          std::cout << "Can't create an instance of the class " << "`" << type->symbolName << "'" << std::endl << "   because it doesn't inherit the class 'THING'." << std::endl;
          return (NULL);
        }
      }
    }
    if (((boolean)(name))) {
      if (((boolean)(name->surrogateValue))) {
        std::cout << "Redefining the surrogate " << name->symbolName << " will cause the existing" << std::endl << "   object named " << name->symbolName << " to be destroyed. ";
        if (!yOrNP("Do it? ")) {
          std::cout << "Returning existing object:" << std::endl;
          return (name->surrogateValue);
        }
        destroyInstance(name->surrogateValue);
      }
      name->surrogateValue = newobject;
      { Surrogate* testValue000 = safePrimaryType(newobject);

        if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
          { Object* newobject000 = newobject;
            LogicObject* newobject = ((LogicObject*)(newobject000));

            newobject->surrogateValueInverse = name;
          }
        }
        else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_THING)) {
          { Object* newobject001 = newobject;
            Thing* newobject = ((Thing*)(newobject001));

            newobject->surrogateValueInverse = name;
          }
        }
        else {
          { OutputStringStream* stream000 = newOutputStringStream();

            *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
            throw *newStellaException(stream000->theStringReader());
          }
        }
      }
    }
    return (newobject);
  }
}

// Contains a table of propositions, indexed by their
// stringifications.  Tricky: Propositions in two
// different modules could have the same stringification but be different.
StringHashTable* oSTRINGIFIED_PROPOSITION_INDEXo = NULL;

Proposition* findExactDuplicateProposition(Proposition* self) {
  { char* stringification = stringify(self);
    List* bucket = ((List*)(oSTRINGIFIED_PROPOSITION_INDEXo->lookup(stringification)));

    if (!((boolean)(bucket))) {
      oSTRINGIFIED_PROPOSITION_INDEXo->insertAt(stringification, list(1, self));
      return (NULL);
    }
    bucket->removeIf(((cpp_function_code)(&deletedPropositionP)));
    { Proposition* p = NULL;
      Cons* iter000 = bucket->theConsList;

      while (!(iter000 == NIL)) {
        p = ((Proposition*)(iter000->value));
        iter000 = iter000->rest;
        if (equivalentPropositionsP(self, p, NULL) &&
            ((self->homeContext == p->homeContext) ||
             self->homeContext->allSuperContexts->membP(p->homeContext))) {
          return (p);
        }
      }
    }
    bucket->push(self);
    return (NULL);
  }
}

Proposition* findDuplicateProposition(Proposition* self) {
  { Keyword* testValue000 = self->kind;

    if ((testValue000 == KWD_PROPOSITIONS_FORALL) ||
        (testValue000 == KWD_PROPOSITIONS_EXISTS)) {
      if (!descriptionModeP()) {
        return (findExactDuplicateProposition(self));
      }
    }
    else {
    }
  }
  { Proposition* value000 = NULL;

    { Proposition* p = NULL;
      Iterator* iter000 = findSimilarPropositions(self)->allocateIterator();

      while (iter000->nextP()) {
        p = ((Proposition*)(iter000->value));
        if (equivalentPropositionsP(self, p, NULL) &&
            ((self->homeContext == p->homeContext) ||
             self->homeContext->allSuperContexts->membP(p->homeContext))) {
          value000 = p;
          break;
        }
      }
    }
    { Proposition* value001 = value000;

      return (value001);
    }
  }
}

// If TRUE never check for the existence of duplicate
// propositions when building a new proposition.
DEFINE_STELLA_SPECIAL(oDONT_CHECK_FOR_DUPLICATE_PROPOSITIONSpo, boolean , FALSE);

Proposition* fastenDownOneProposition(Proposition* self, boolean dontcheckforduplicatesP) {
  if (nativeSlotPropositionP(self)) {
    return (self);
  }
  if (!(oLOADINGREGENERABLEOBJECTSpo.get())) {
    surrogatifyDiscouragedArguments(self);
  }
  { Proposition* duplicate = NULL;

    if (!(dontcheckforduplicatesP ||
        (oDONT_CHECK_FOR_DUPLICATE_PROPOSITIONSpo.get() ||
         oLOADINGREGENERABLEOBJECTSpo.get()))) {
      duplicate = findDuplicateProposition(self);
    }
    if (((boolean)(duplicate))) {
      {
        if (!equalP(((PropertyList*)(dynamicSlotValue(duplicate->dynamicSlots, SYM_PROPOSITIONS_LOGIC_ANNOTATIONS, NULL))), ((PropertyList*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_ANNOTATIONS, NULL))))) {
          clearPropositionAnnotations(duplicate);
          setDynamicSlotValue(duplicate->dynamicSlots, SYM_PROPOSITIONS_LOGIC_ANNOTATIONS, ((PropertyList*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_ANNOTATIONS, NULL))), NULL);
          finalizePropositionAnnotations(duplicate);
          evaluateNewProposition(duplicate);
        }
        return (duplicate);
      }
    }
    else {
      {
        if (!(((BooleanWrapper*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTIVEp, FALSE_WRAPPER)))->wrapperValue ||
            descriptionModeP())) {
          if (!oLOADINGREGENERABLEOBJECTSpo.get()) {
            logNewlyConceivedProposition(oMODULEo.get(), self);
          }
        }
        if (!oLOADINGREGENERABLEOBJECTSpo.get()) {
          rewrapPropositionArguments(self);
        }
        if (!oLOADINGREGENERABLEOBJECTSpo.get()) {
          { Object* arg = NULL;
            Vector* vector000 = self->arguments;
            int index000 = 0;
            int length000 = vector000->length();

            while (index000 < length000) {
              arg = (vector000->theArray)[index000];
              index000 = index000 + 1;
              addDependentPropositionLink(arg, self);
            }
          }
        }
        updateDescriptionExtension(self);
        setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_UNFASTENEDp, NULL, NULL);
        finalizePropositionAnnotations(self);
        evaluateNewProposition(self);
        return (self);
      }
    }
  }
}

// Used by 'recursively-fasten-down-propositions'.
DEFINE_STELLA_SPECIAL(oVISITEDUNFASTENEDDEFININGPROPOSITIONSo, List* , NULL);

Proposition* helpFastenDownPropositions(Proposition* self, boolean dontcheckforduplicatesP) {
  { Object* arg = NULL;
    int i = NULL_INTEGER;
    Vector* vector000 = self->arguments;
    int index000 = 0;
    int length000 = vector000->length();
    int iter000 = 0;

    while (index000 < length000) {
      arg = (vector000->theArray)[index000];
      index000 = index000 + 1;
      i = iter000;
      iter000 = iter000 + 1;
      { Surrogate* testValue000 = safePrimaryType(arg);

        if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
          { Object* arg000 = arg;
            Proposition* arg = ((Proposition*)(arg000));

            if (coerceWrappedBooleanToBoolean(arg->unfastenedP_reader())) {
              (self->arguments->theArray)[i] = (helpFastenDownPropositions(arg, dontcheckforduplicatesP));
            }
          }
        }
        else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
          { Object* arg001 = arg;
            Skolem* arg = ((Skolem*)(arg001));

            { Proposition* definingprop = arg->definingProposition;

              if (((boolean)(definingprop)) &&
                  (coerceWrappedBooleanToBoolean(definingprop->unfastenedP_reader()) &&
                   (!oVISITEDUNFASTENEDDEFININGPROPOSITIONSo.get()->memberP(definingprop)))) {
                oVISITEDUNFASTENEDDEFININGPROPOSITIONSo.get()->insert(definingprop);
                helpFastenDownPropositions(definingprop, dontcheckforduplicatesP);
              }
            }
          }
        }
        else {
        }
      }
    }
  }
  return (fastenDownOneProposition(self, dontcheckforduplicatesP));
}

Proposition* recursivelyFastenDownPropositions(Proposition* self, boolean dontcheckforduplicatesP) {
  if (!coerceWrappedBooleanToBoolean(self->unfastenedP_reader())) {
    return (self);
  }
  { 
    BIND_STELLA_SPECIAL(oVISITEDUNFASTENEDDEFININGPROPOSITIONSo, List*, list(0));
    return (helpFastenDownPropositions(self, dontcheckforduplicatesP));
  }
}

void unfastenProposition(Proposition* proposition) {
  { Object* arg = NULL;
    Vector* vector000 = proposition->arguments;
    int index000 = 0;
    int length000 = vector000->length();

    while (index000 < length000) {
      arg = (vector000->theArray)[index000];
      index000 = index000 + 1;
      removeDependentPropositionLink(arg, proposition);
    }
  }
  setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_UNFASTENEDp, TRUE_WRAPPER, NULL);
}

void instantiateUndefinedSurrogates(Proposition* self) {
  { Keyword* testValue000 = self->kind;

    if ((testValue000 == KWD_PROPOSITIONS_AND) ||
        ((testValue000 == KWD_PROPOSITIONS_OR) ||
         ((testValue000 == KWD_PROPOSITIONS_NOT) ||
          (testValue000 == KWD_PROPOSITIONS_EXISTS)))) {
      { Object* arg = NULL;
        Vector* vector000 = self->arguments;
        int index000 = 0;
        int length000 = vector000->length();

        while (index000 < length000) {
          arg = (vector000->theArray)[index000];
          index000 = index000 + 1;
          instantiateUndefinedSurrogates(((Proposition*)(arg)));
        }
      }
    }
    else if ((testValue000 == KWD_PROPOSITIONS_PREDICATE) ||
        ((testValue000 == KWD_PROPOSITIONS_FUNCTION) ||
         (testValue000 == KWD_PROPOSITIONS_ISA))) {
      { Object* arg = NULL;
        Vector* vector001 = self->arguments;
        int index001 = 0;
        int length001 = vector001->length();

        while (index001 < length001) {
          arg = (vector001->theArray)[index001];
          index001 = index001 + 1;
          if (subtypeOfSurrogateP(safePrimaryType(arg))) {
            { Object* arg000 = arg;
              Surrogate* arg = ((Surrogate*)(arg000));

              if (!((boolean)(arg->surrogateValue))) {
                createLogicInstance(arg, NULL);
              }
            }
          }
          else {
          }
        }
      }
    }
    else {
    }
  }
}

Proposition* buildTopLevelProposition(Object* tree, boolean trueassertionP) {
  if (isaP(tree, SGT_PROPOSITIONS_STELLA_STRING_WRAPPER)) {
    return (buildTopLevelProposition(readSExpressionFromString(((StringWrapper*)(tree))->wrapperValue), trueassertionP));
  }
  { Proposition* proposition = NULL;
    Cons* logicvariabletable = oLOGICVARIABLETABLEo.get();

    { 
      BIND_STELLA_SPECIAL(oLOGICVARIABLETABLEo, Cons*, ((Cons*)((((boolean)(logicvariabletable)) ? logicvariabletable : NIL))));
      BIND_STELLA_SPECIAL(oVARIABLEIDCOUNTERo, int, oVARIABLEIDCOUNTERo.get());
      BIND_STELLA_SPECIAL(oTERMUNDERCONSTRUCTIONo, Object*, tree);
      proposition = ((Proposition*)(buildProposition(tree)));
      if (((boolean)(proposition))) {
        normalizeProposition(proposition);
        verifyForallPropositions(proposition);
        if (!(collectUnresolvedSlotReferences(proposition) == NIL)) {
          if (trueassertionP) {
            updateSkolemTypeFromIsaAssertions(proposition);
          }
          resolveUnresolvedSlotReferences(proposition);
        }
        if (trueassertionP &&
            ((proposition->kind == KWD_PROPOSITIONS_EXISTS) &&
             (!descriptionModeP()))) {
          instantiateUndefinedSurrogates(proposition);
        }
        proposition = recursivelyFastenDownPropositions(proposition, FALSE);
      }
    }
    return (proposition);
  }
}

Object* buildProposition(Object* tree) {
  { Surrogate* testValue000 = safePrimaryType(tree);

    if (testValue000 == SGT_PROPOSITIONS_STELLA_CONS) {
      { Object* tree000 = tree;
        Cons* tree = ((Cons*)(tree000));

        return (consDbuildProposition(tree));
      }
    }
    else if (subtypeOfSurrogateP(testValue000)) {
      { Object* tree001 = tree;
        Surrogate* tree = ((Surrogate*)(tree001));

        return (surrogateDbuildProposition(tree));
      }
    }
    else if (subtypeOfSymbolP(testValue000)) {
      { Object* tree002 = tree;
        Symbol* tree = ((Symbol*)(tree002));

        return (symbolDbuildProposition(tree));
      }
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          *(stream000->nativeStream) << "ERROR: " << "Illegal logical sentence: " << "`" << tree << "'" << "." << std::endl;
          helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
        }
        throw *newPropositionError(stream000->theStringReader());
      }
    }
  }
}

Proposition* surrogateDbuildProposition(Surrogate* self) {
  { Object* value = self->surrogateValue;
    Proposition* proposition = NULL;

    if (((boolean)(value))) {
      if (subtypeOfP(safePrimaryType(value), SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
        { Object* value000 = value;
          Proposition* value = ((Proposition*)(value000));

          proposition = value;
        }
      }
      else {
        { OutputStringStream* stream000 = newOutputStringStream();

          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            *(stream000->nativeStream) << "ERROR: " << "Expected " << "`" << self->symbolName << "'" << " to have type proposition" << "." << std::endl;
            helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
          }
          throw *newPropositionError(stream000->theStringReader());
        }
      }
    }
    else {
      {
        proposition = createProposition(SYM_PROPOSITIONS_LOGIC_CONSTANT, 0);
        proposition->operatoR = self;
        self->surrogateValue = proposition;
        setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_STELLA_SURROGATE_VALUE_INVERSE, self, NULL);
      }
    }
    return (proposition);
  }
}

Object* symbolDbuildProposition(Symbol* symbol) {
  { Proposition* proposition = NULL;

    if (symbol == SYM_PROPOSITIONS_STELLA_TRUE) {
      proposition = TRUE_PROPOSITION;
    }
    else if (symbol == SYM_PROPOSITIONS_STELLA_FALSE) {
      proposition = FALSE_PROPOSITION;
    }
    else {
      return (evaluateSymbolTerm(symbol));
    }
    return (proposition);
  }
}

Proposition* consDbuildProposition(Cons* tree) {
  { Proposition* proposition = NULL;

    if (!isaP(tree->value, SGT_PROPOSITIONS_STELLA_SYMBOL)) {
      std::cerr << "Safety violation: " << "INTERNAL ERROR: Illegal operator in proposition tree.";
    }
    { GeneralizedSymbol* testValue000 = ((GeneralizedSymbol*)(tree->value));

      if (testValue000 == SYM_PROPOSITIONS_STELLA_e) {
        proposition = buildEquivalenceProposition(tree);
      }
      else if ((testValue000 == SYM_PROPOSITIONS_STELLA_AND) ||
          ((testValue000 == SYM_PROPOSITIONS_STELLA_OR) ||
           (testValue000 == SYM_PROPOSITIONS_STELLA_NOT))) {
        proposition = buildAndOrNotProposition(tree);
      }
      else if (testValue000 == SYM_PROPOSITIONS_STELLA_ISA) {
        proposition = buildIsaProposition(tree);
      }
      else if (testValue000 == SYM_PROPOSITIONS_PL_KERNEL_KB_MEMBER_OF) {
        proposition = buildMemberOfProposition(tree);
      }
      else if ((testValue000 == SYM_PROPOSITIONS_LOGIC_IMPLIES) ||
          (testValue000 == SYM_PROPOSITIONS_PL_KERNEL_KB_SUBSET_OF)) {
        proposition = buildImpliesProposition(tree);
      }
      else if (testValue000 == SYM_PROPOSITIONS_STELLA_EXISTS) {
        proposition = buildExistsProposition(tree);
      }
      else if (testValue000 == SYM_PROPOSITIONS_STELLA_FORALL) {
        proposition = buildForallProposition(tree);
      }
      else if (testValue000 == SYM_PROPOSITIONS_LOGIC_FAIL) {
        proposition = buildFailProposition(tree);
      }
      else if (testValue000 == SYM_PROPOSITIONS_STELLA_COLLECT_INTO) {
        proposition = buildCollectIntoProposition(tree);
      }
      else if (testValue000 == SYM_PROPOSITIONS_LOGIC_ABOUT) {
        proposition = buildAnnotatedProposition(tree);
      }
      else {
        proposition = buildPredicateProposition(tree);
      }
    }
    if (!((boolean)(proposition))) {
      return (NULL);
    }
    if (containsNestedArgumentP(proposition)) {
      proposition = flattenNestedFunctionArguments(proposition);
    }
    return (proposition);
  }
}

void verifyNumberOfPropositionArguments(Cons* tree, int correctnumber) {
  if (!(tree->rest->length() == correctnumber)) {
    { OutputStringStream* stream000 = newOutputStringStream();

      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        *(stream000->nativeStream) << "ERROR: " << "Wrong number of arguments in proposition: " << "`" << tree << "'" << std::endl << "   Expected " << "`" << correctnumber << "'" << " arguments." << "." << std::endl;
        helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
      }
      throw *newPropositionError(stream000->theStringReader());
    }
  }
}

Surrogate* Description::descriptionSurrogate() {
  { Description* self = this;

    { Relation* nativerelation = self->nativeRelation();

      if (((boolean)(nativerelation))) {
        { Surrogate* testValue000 = safePrimaryType(nativerelation);

          if (subtypeOfClassP(testValue000)) {
            { Relation* nativerelation000 = nativerelation;
              Class* nativerelation = ((Class*)(nativerelation000));

              return (classLogicalType(nativerelation));
            }
          }
          else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_SLOT)) {
            { Relation* nativerelation001 = nativerelation;
              Slot* nativerelation = ((Slot*)(nativerelation001));

              return (nativerelation->slotSlotref);
            }
          }
          else {
          }
        }
      }
      return (NULL);
    }
  }
}

Proposition* buildIsaProposition(Cons* tree) {
  verifyNumberOfPropositionArguments(tree, 2);
  { Proposition* proposition = createProposition(SYM_PROPOSITIONS_STELLA_ISA, 1);
    Surrogate* type = NULL;
    Object* symbol = tree->rest->rest->value;

    { Surrogate* testValue000 = safePrimaryType(symbol);

      if (subtypeOfSurrogateP(testValue000)) {
        { Object* symbol000 = symbol;
          Surrogate* symbol = ((Surrogate*)(symbol000));

          type = symbol;
        }
      }
      else if (subtypeOfSymbolP(testValue000)) {
        { Object* symbol001 = symbol;
          Symbol* symbol = ((Symbol*)(symbol001));

          type = symbolToSurrogate(symbol);
        }
      }
      else {
        { OutputStringStream* stream000 = newOutputStringStream();

          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            *(stream000->nativeStream) << "ERROR: " << "Illegal token where symbol expected: " << "`" << symbol << "'" << "." << std::endl;
            helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
          }
          throw *newPropositionError(stream000->theStringReader());
        }
      }
    }
    { NamedDescription* clasS = getDescription(type);

      if (!((boolean)(clasS))) {
        { OutputStringStream* stream001 = newOutputStringStream();

          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            *(stream001->nativeStream) << "ERROR: " << "No class is associated with the term " << "`" << type->symbolName << "'" << "." << std::endl;
            helpSignalPropositionError(stream001, KWD_PROPOSITIONS_ERROR);
          }
          throw *newPropositionError(stream001->theStringReader());
        }
      }
      if (!classDescriptionP(clasS)) {
        { OutputStringStream* stream002 = newOutputStringStream();

          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            *(stream002->nativeStream) << "ERROR: " << "Term " << "`" << type->symbolName << "'" << " found where class expected." << "." << std::endl;
            helpSignalPropositionError(stream002, KWD_PROPOSITIONS_ERROR);
          }
          throw *newPropositionError(stream002->theStringReader());
        }
      }
    }
    proposition->operatoR = type;
    (proposition->arguments->theArray)[0] = (evaluateTypedArgument(tree->rest->value, type));
    return (proposition);
  }
}

Proposition* buildMemberOfProposition(Cons* tree) {
  verifyNumberOfPropositionArguments(tree, 2);
  { Object* collectionref = tree->rest->rest->value;

    { Surrogate* testValue000 = safePrimaryType(collectionref);

      if (subtypeOfSurrogateP(testValue000)) {
        { Object* collectionref000 = collectionref;
          Surrogate* collectionref = ((Surrogate*)(collectionref000));

          if (((boolean)(collectionref->getStellaClass(FALSE)))) {
            return (buildIsaProposition(tree));
          }
        }
      }
      else if (subtypeOfSymbolP(testValue000)) {
        { Object* collectionref001 = collectionref;
          Symbol* collectionref = ((Symbol*)(collectionref001));

          if (((boolean)(collectionref->getStellaClass(FALSE)))) {
            return (buildIsaProposition(tree));
          }
        }
      }
      else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) {
        { Object* collectionref002 = collectionref;
          Description* collectionref = ((Description*)(collectionref002));

          if (namedDescriptionP(collectionref) &&
              classDescriptionP(((NamedDescription*)(collectionref)))) {
            return (buildIsaProposition(listO(3, SYM_PROPOSITIONS_STELLA_ISA, tree->rest->value, cons(collectionref->descriptionSurrogate(), NIL))));
          }
        }
      }
      else {
      }
    }
  }
  return (buildPredicateProposition(tree));
}

boolean memberOfPropositionP(Proposition* proposition) {
  return (proposition->operatoR == SGT_PROPOSITIONS_PL_KERNEL_KB_MEMBER_OF);
}

boolean subsetOfPropositionP(Proposition* proposition) {
  return (proposition->operatoR == SGT_PROPOSITIONS_PL_KERNEL_KB_SUBSET_OF);
}

boolean simpleSubrelationPropositionP(Proposition* proposition) {
  return ((proposition->operatoR == SGT_PROPOSITIONS_PL_KERNEL_KB_SUBSET_OF) &&
      (isaP((proposition->arguments->theArray)[0], SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION) &&
       isaP((proposition->arguments->theArray)[1], SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)));
}

Object* extractAnnotation(Object* tree, Keyword* option, Object*& _Return1) {
  { Object* value = NULL;

    if (safePrimaryType(tree) == SGT_PROPOSITIONS_STELLA_CONS) {
      { Object* tree000 = tree;
        Cons* tree = ((Cons*)(tree000));

        if (tree->value == SYM_PROPOSITIONS_LOGIC_ABOUT) {
          { PropertyList* self000 = newPropertyList();

            self000->thePlist = tree->rest->rest;
            { PropertyList* options = self000;

              value = options->lookup(option);
              if (((boolean)(value))) {
                options->removeAt(option);
                tree->rest->rest = options->thePlist;
              }
            }
          }
        }
      }
    }
    else {
    }
    _Return1 = value;
    return (tree);
  }
}

Object* lookupAnnotation(Proposition* proposition, Keyword* key) {
  { PropertyList* annotations = ((PropertyList*)(dynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_ANNOTATIONS, NULL)));

    if (((boolean)(annotations))) {
      return (annotations->lookup(key));
    }
    else {
      return (NULL);
    }
  }
}

void insertAnnotation(Proposition* proposition, Keyword* key, Object* value) {
  { PropertyList* annotations = ((PropertyList*)(dynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_ANNOTATIONS, NULL)));

    if (!((boolean)(annotations))) {
      annotations = newPropertyList();
      setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_ANNOTATIONS, annotations, NULL);
    }
    annotations->insertAt(key, value);
  }
}

void removeAnnotation(Proposition* proposition, Keyword* key) {
  { PropertyList* annotations = ((PropertyList*)(dynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_ANNOTATIONS, NULL)));

    if (!((boolean)(annotations))) {
      return;
    }
    annotations->removeAt(key);
    if (annotations->thePlist == NIL) {
      setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_ANNOTATIONS, NULL, NULL);
    }
  }
}

Proposition* buildAnnotatedProposition(Cons* tree) {
  { Proposition* proposition = ((Proposition*)(buildProposition(tree->rest->value)));
    Cons* options = tree->rest->rest;

    while (!(options == NIL)) {
      if (isaP(options->value, SGT_PROPOSITIONS_STELLA_GENERALIZED_SYMBOL)) {
        options->firstSetter(((GeneralizedSymbol*)(options->value))->keywordify());
      }
      else {
        { OutputStringStream* stream000 = newOutputStringStream();

          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            *(stream000->nativeStream) << "PARSING ERROR: " << "Illegal annotation option: " << "`" << options->value << "'" << "." << std::endl;
            helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
          }
          throw *newParsingError(stream000->theStringReader());
        }
      }
      if (!(((boolean)(options->rest->value)))) {
        { OutputStringStream* stream001 = newOutputStringStream();

          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            *(stream001->nativeStream) << "PARSING ERROR: " << "Missing annotation value for option: " << "`" << options->value << "'" << "." << std::endl;
            helpSignalPropositionError(stream001, KWD_PROPOSITIONS_ERROR);
          }
          throw *newParsingError(stream001->theStringReader());
        }
      }
      options = options->rest->rest;
    }
    { PropertyList* self004 = newPropertyList();

      self004->thePlist = tree->rest->rest;
      setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_ANNOTATIONS, self004, NULL);
    }
    finalizePropositionAnnotations(proposition);
    return (proposition);
  }
}

void finalizePropositionAnnotations(Proposition* proposition) {
  { PropertyList* annotations = ((PropertyList*)(dynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_ANNOTATIONS, NULL)));

    if (((boolean)(annotations))) {
      { Keyword* option = NULL;
        Object* value = NULL;
        Cons* iter000 = annotations->thePlist;

        while (!(iter000 == NIL)) {
          option = ((Keyword*)(iter000->value));
          value = iter000->rest->value;
          iter000 = iter000->rest->rest;
          if (option == KWD_PROPOSITIONS_DIRECTION) {
            if (subtypeOfP(safePrimaryType(value), SGT_PROPOSITIONS_STELLA_GENERALIZED_SYMBOL)) {
              { Object* value000 = value;
                GeneralizedSymbol* value = ((GeneralizedSymbol*)(value000));

                { char* testValue000 = stringUpcase(value->symbolName);

                  if (stringEqlP(testValue000, "FORWARD")) {
                    setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_FORWARD_ONLYp, (TRUE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
                  }
                  else if (stringEqlP(testValue000, "BACKWARD")) {
                    setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_BACKWARD_ONLYp, (TRUE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
                  }
                  else {
                    annotations->removeAt(option);
                    { OutputStringStream* stream000 = newOutputStringStream();

                      { 
                        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                        *(stream000->nativeStream) << "PARSING ERROR: " << "Illegal value for :direction option: " << "`" << value << "'" << "." << std::endl;
                        helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
                      }
                      throw *newParsingError(stream000->theStringReader());
                    }
                  }
                }
              }
            }
            else {
              annotations->removeAt(option);
              { OutputStringStream* stream001 = newOutputStringStream();

                { 
                  BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                  *(stream001->nativeStream) << "PARSING ERROR: " << "Illegal value for :direction option: " << "`" << value << "'" << "." << std::endl;
                  helpSignalPropositionError(stream001, KWD_PROPOSITIONS_ERROR);
                }
                throw *newParsingError(stream001->theStringReader());
              }
            }
          }
          else if (option == KWD_PROPOSITIONS_WEIGHT) {
            { Surrogate* testValue001 = safePrimaryType(value);

              if (subtypeOfIntegerP(testValue001)) {
                { Object* value001 = value;
                  IntegerWrapper* value = ((IntegerWrapper*)(value001));

                  setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_WEIGHT, wrapFloat(value->numberWrapperToFloat()), NULL_FLOAT_WRAPPER);
                }
              }
              else if (subtypeOfFloatP(testValue001)) {
                { Object* value002 = value;
                  FloatWrapper* value = ((FloatWrapper*)(value002));

                  setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_WEIGHT, wrapFloat(value->wrapperValue), NULL_FLOAT_WRAPPER);
                }
              }
              else {
                annotations->removeAt(option);
                { OutputStringStream* stream002 = newOutputStringStream();

                  { 
                    BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                    *(stream002->nativeStream) << "PARSING ERROR: " << "Illegal value for :weight option: " << "`" << value << "'" << "." << std::endl;
                    helpSignalPropositionError(stream002, KWD_PROPOSITIONS_ERROR);
                  }
                  throw *newParsingError(stream002->theStringReader());
                }
              }
            }
          }
          else if (option == KWD_PROPOSITIONS_CONFIDENCE_LEVEL) {
          }
          else {
            if (!(runOptionHandlerP(proposition, option, value))) {
              annotations->removeAt(option);
              { OutputStringStream* stream003 = newOutputStringStream();

                { 
                  BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
                  *(stream003->nativeStream) << "PARSING ERROR: " << "Unrecognized annotation option: " << "`" << option << "'" << "." << std::endl;
                  helpSignalPropositionError(stream003, KWD_PROPOSITIONS_ERROR);
                }
                throw *newParsingError(stream003->theStringReader());
              }
            }
          }
        }
      }
    }
  }
}

void clearPropositionAnnotations(Proposition* proposition) {
  { PropertyList* annotations = ((PropertyList*)(dynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_ANNOTATIONS, NULL)));

    if (((boolean)(annotations))) {
      { Keyword* option = NULL;
        Object* value = NULL;
        Cons* iter000 = annotations->thePlist;

        while (!(iter000 == NIL)) {
          option = ((Keyword*)(iter000->value));
          value = iter000->rest->value;
          iter000 = iter000->rest->rest;
          value = value;
          if (option == KWD_PROPOSITIONS_DIRECTION) {
            setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_FORWARD_ONLYp, (FALSE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
            setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_BACKWARD_ONLYp, (FALSE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
          }
          else if (option == KWD_PROPOSITIONS_WEIGHT) {
            setDynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_WEIGHT, wrapFloat(NULL_FLOAT), NULL_FLOAT_WRAPPER);
          }
          else {
            runOptionHandlerP(proposition, option, NULL);
          }
        }
      }
    }
  }
}

boolean functionInducedExistsP(Proposition* proposition) {
  if (proposition->kind == KWD_PROPOSITIONS_EXISTS) {
    { Object* firstargument = (proposition->arguments->theArray)[0];

      if (isaP(firstargument, SGT_PROPOSITIONS_LOGIC_PROPOSITION) &&
          (((Proposition*)(firstargument))->kind == KWD_PROPOSITIONS_AND)) {
        { Vector* existsvariables = ((Vector*)(dynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_IO_VARIABLES, NULL)));
          Proposition* andprop = ((Proposition*)(firstargument));
          Vector* andarguments = andprop->arguments;
          int nonfunctionscount = 0;

          { boolean alwaysP000 = TRUE;

            { PatternVariable* variable = NULL;
              Vector* vector000 = existsvariables;
              int index000 = 0;
              int length000 = vector000->length();

              while (index000 < length000) {
                variable = ((PatternVariable*)((vector000->theArray)[index000]));
                index000 = index000 + 1;
                { boolean testValue000 = FALSE;

                  { boolean foundP000 = FALSE;

                    { Proposition* arg = NULL;
                      Vector* vector001 = andarguments;
                      int index001 = 0;
                      int length001 = vector001->length();

                      while (index001 < length001) {
                        arg = ((Proposition*)((vector001->theArray)[index001]));
                        index001 = index001 + 1;
                        if ((arg->kind == KWD_PROPOSITIONS_FUNCTION) &&
                            (arg->arguments->last() == variable)) {
                          foundP000 = TRUE;
                          break;
                        }
                      }
                    }
                    testValue000 = foundP000;
                  }
                  testValue000 = !testValue000;
                  if (testValue000) {
                    alwaysP000 = FALSE;
                    break;
                  }
                }
              }
            }
            if (alwaysP000) {
              { Proposition* arg = NULL;
                Vector* vector002 = andarguments;
                int index002 = 0;
                int length002 = vector002->length();

                while (index002 < length002) {
                  arg = ((Proposition*)((vector002->theArray)[index002]));
                  index002 = index002 + 1;
                  if ((!(arg->kind == KWD_PROPOSITIONS_FUNCTION)) &&
                      (!coerceWrappedBooleanToBoolean(arg->variableTypeP_reader()))) {
                    nonfunctionscount = nonfunctionscount + 1;
                  }
                }
              }
              return (nonfunctionscount == 1);
            }
          }
        }
      }
    }
  }
  return (FALSE);
}

Proposition* predicateOfFunctionInducedExists(Proposition* existsproposition) {
  { Proposition* value000 = NULL;

    { Proposition* arg = NULL;
      Vector* vector000 = ((Vector*)(((Proposition*)((existsproposition->arguments->theArray)[0]))->arguments));
      int index000 = 0;
      int length000 = vector000->length();

      while (index000 < length000) {
        arg = ((Proposition*)((vector000->theArray)[index000]));
        index000 = index000 + 1;
        if ((!(arg->kind == KWD_PROPOSITIONS_FUNCTION)) &&
            (!coerceWrappedBooleanToBoolean(arg->variableTypeP_reader()))) {
          value000 = arg;
          break;
        }
      }
    }
    { Proposition* value001 = value000;

      return (value001);
    }
  }
}

Proposition* embedNegationWithinFunctionInducedExists(Proposition* existsproposition) {
  invertProposition(predicateOfFunctionInducedExists(existsproposition));
  return (existsproposition);
}

Proposition* buildAndOrNotProposition(Cons* tree) {
  { Object* term = NULL;
    Cons* arguments = NIL;
    Proposition* proposition = NULL;
    Vector* argumentsvector = NULL;

    { GeneralizedSymbol* testValue000 = ((GeneralizedSymbol*)(tree->value));

      if ((testValue000 == SYM_PROPOSITIONS_STELLA_AND) ||
          ((testValue000 == SYM_PROPOSITIONS_STELLA_OR) ||
           (testValue000 == SYM_PROPOSITIONS_STELLA_NOT))) {
        { Object* arg = NULL;
          Cons* iter000 = tree->rest;
          Cons* collect000 = NULL;

          while (!(iter000 == NIL)) {
            arg = iter000->value;
            iter000 = iter000->rest;
            { 
              term = buildProposition(arg);
              if (((boolean)(term))) {
                if (!((boolean)(collect000))) {
                  {
                    collect000 = cons(term, NIL);
                    if (arguments == NIL) {
                      arguments = collect000;
                    }
                    else {
                      addConsToEndOfConsList(arguments, collect000);
                    }
                  }
                }
                else {
                  {
                    collect000->rest = cons(term, NIL);
                    collect000 = collect000->rest;
                  }
                }
              }
            }
          }
        }
      }
      else {
        { OutputStringStream* stream000 = newOutputStringStream();

          *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
          throw *newStellaException(stream000->theStringReader());
        }
      }
    }
    if (arguments == NIL) {
      return (NULL);
    }
    if ((tree->value == SYM_PROPOSITIONS_STELLA_NOT) &&
        (isaP(arguments->value, SGT_PROPOSITIONS_LOGIC_PROPOSITION) &&
         functionInducedExistsP(((Proposition*)(arguments->value))))) {
      return (embedNegationWithinFunctionInducedExists(((Proposition*)(arguments->value))));
    }
    proposition = createProposition(((Symbol*)(tree->value)), arguments->length());
    argumentsvector = proposition->arguments;
    { Object* arg = NULL;
      int i = NULL_INTEGER;
      Cons* iter001 = arguments;
      int iter002 = 0;

      while (!(iter001 == NIL)) {
        arg = iter001->value;
        iter001 = iter001->rest;
        i = iter002;
        iter002 = iter002 + 1;
        (argumentsvector->theArray)[i] = arg;
      }
    }
    return (proposition);
  }
}

boolean functionalTermP(Object* term) {
  if (safePrimaryType(term) == SGT_PROPOSITIONS_STELLA_CONS) {
    { Object* term000 = term;
      Cons* term = ((Cons*)(term000));

      { GeneralizedSymbol* testValue000 = ((GeneralizedSymbol*)(term->value));

        if ((testValue000 == SYM_PROPOSITIONS_LOGIC_SETOFALL) ||
            ((testValue000 == SYM_PROPOSITIONS_LOGIC_THE_ONLY) ||
             ((testValue000 == SYM_PROPOSITIONS_STELLA_SETOF) ||
              ((testValue000 == SYM_PROPOSITIONS_LOGIC_LISTOF) ||
               (testValue000 == SYM_PROPOSITIONS_LOGIC_KAPPA))))) {
          return (FALSE);
        }
        else {
          return (TRUE);
        }
      }
    }
  }
  else {
    return (FALSE);
  }
}

boolean clipValueP(Object* term1, Object* term2) {
  if (subtypeOfP(safePrimaryType(term1), SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
    { Object* term1000 = term1;
      Skolem* term1 = ((Skolem*)(term1000));

      { Object* value1 = valueOf(term1);

        if (eqlP(value1, term2)) {
          return (TRUE);
        }
        if ((!skolemP(value1)) &&
            (((boolean)(term1->definingProposition)) &&
             oCLIPPINGENABLEDpo.get())) {
          unbindSkolemValue(term1, TRUE);
          bindSkolemToValue(term1, valueOf(term2), TRUE);
          return (TRUE);
        }
      }
    }
  }
  else {
  }
  return (FALSE);
}

Proposition* createEquivalenceProposition(Object* term1, Object* term2) {
  { Proposition* proposition = createProposition(SYM_PROPOSITIONS_LOGIC_EQUIVALENT, 2);
    Vector* argumentsvector = proposition->arguments;

    (argumentsvector->theArray)[0] = term1;
    (argumentsvector->theArray)[1] = term2;
    return (proposition);
  }
}

Proposition* buildEquivalenceProposition(Cons* tree) {
  switch (tree->length()) {
    case 3: 
      return (createEquivalenceProposition(evaluateTerm(tree->rest->value), evaluateTerm(tree->rest->rest->value)));
    case 1: 
    case 2: 
      { OutputStringStream* stream000 = newOutputStringStream();

        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          *(stream000->nativeStream) << "ERROR: " << "Missing argument/s in `=' proposition." << "." << std::endl;
          helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
        }
        throw *newPropositionError(stream000->theStringReader());
      }
    break;
    default:
      return (buildConjunctionOfEqualityPropositions(tree->rest, NIL));
  }
}

Proposition* buildConjunctionOfEqualityPropositions(Cons* remainingterms, Cons* equivalences) {
  if (remainingterms->rest == NIL) {
    return (conjoinPropositions(equivalences));
  }
  equivalences = cons(buildEquivalenceProposition(listO(3, SYM_PROPOSITIONS_STELLA_e, remainingterms->value, cons(remainingterms->rest->value, NIL))), equivalences);
  return (buildConjunctionOfEqualityPropositions(remainingterms->rest, equivalences));
}

Proposition* buildFailProposition(Cons* tree) {
  { Proposition* proposition = createProposition(SYM_PROPOSITIONS_LOGIC_FAIL, 1);
    Object* argument = buildProposition(tree->rest->value);

    if (!((boolean)(argument))) {
      return (NULL);
    }
    (proposition->arguments->theArray)[0] = argument;
    return (proposition);
  }
}

Proposition* buildCollectIntoProposition(Cons* tree) {
  { Proposition* proposition = createProposition(SYM_PROPOSITIONS_STELLA_COLLECT_INTO, 2);
    Vector* argumentsvector = proposition->arguments;

    (argumentsvector->theArray)[0] = (evaluateTerm(tree->rest->value));
    (argumentsvector->theArray)[1] = (evaluateTerm(tree->rest->rest->value));
    return (proposition);
  }
}

boolean holdsOperatorP(Object* operatoR) {
  { Surrogate* testValue000 = safePrimaryType(operatoR);

    if (subtypeOfSymbolP(testValue000)) {
      { Object* operator000 = operatoR;
        Symbol* operatoR = ((Symbol*)(operator000));

        return ((operatoR == SYM_PROPOSITIONS_PL_KERNEL_KB_HOLDS) ||
            ((boolean)(lookupLogicVariableBinding(operatoR))));
      }
    }
    else if (testValue000 == SGT_PROPOSITIONS_STELLA_CONS) {
      { Object* operator001 = operatoR;
        Cons* operatoR = ((Cons*)(operator001));

        return (TRUE);
      }
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
        throw *newStellaException(stream000->theStringReader());
      }
    }
  }
}

void verifyEvaluableOperator(Object* operatoR, Keyword* kind) {
  if (subtypeOfP(safePrimaryType(operatoR), SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE)) {
    { Object* operator000 = operatoR;
      PatternVariable* operatoR = ((PatternVariable*)(operator000));

      if (kind == KWD_PROPOSITIONS_PREDICATE) {
        operatoR->skolemType = SGT_PROPOSITIONS_PL_KERNEL_KB_RELATION;
      }
      else if (kind == KWD_PROPOSITIONS_FUNCTION) {
        operatoR->skolemType = SGT_PROPOSITIONS_PL_KERNEL_KB_FUNCTION;
      }
      else {
        { OutputStringStream* stream000 = newOutputStringStream();

          *(stream000->nativeStream) << "`" << kind << "'" << " is not a valid case option";
          throw *newStellaException(stream000->theStringReader());
        }
      }
    }
  }
  else {
  }
}

Proposition* buildHoldsProposition(Cons* tree) {
  { Object* operatorexpression = tree->value;
    boolean explicitholdsP = operatorexpression == SYM_PROPOSITIONS_PL_KERNEL_KB_HOLDS;
    int argcount = tree->rest->length();
    Proposition* proposition = createProposition(SYM_PROPOSITIONS_STELLA_PREDICATE, (explicitholdsP ? argcount : (argcount + 1)));

    proposition->operatoR = SGT_PROPOSITIONS_PL_KERNEL_KB_HOLDS;
    if (explicitholdsP) {
      { Object* arg = NULL;
        int i = NULL_INTEGER;
        Cons* iter000 = tree->rest;
        int iter001 = 0;

        while (!(iter000 == NIL)) {
          arg = iter000->value;
          iter000 = iter000->rest;
          i = iter001;
          iter001 = iter001 + 1;
          (proposition->arguments->theArray)[i] = (((i == 0) ? evaluateFirstArgument(arg, SYM_PROPOSITIONS_PL_KERNEL_KB_HOLDS) : evaluateTerm(arg)));
        }
      }
      return (proposition);
    }
    { Object* operatoR = NULL;

      { Surrogate* testValue000 = safePrimaryType(operatorexpression);

        if (subtypeOfSymbolP(testValue000)) {
          { Object* operatorexpression000 = operatorexpression;
            Symbol* operatorexpression = ((Symbol*)(operatorexpression000));

            operatoR = lookupLogicVariableBinding(operatorexpression);
          }
        }
        else if (testValue000 == SGT_PROPOSITIONS_STELLA_CONS) {
          { Object* operatorexpression001 = operatorexpression;
            Cons* operatorexpression = ((Cons*)(operatorexpression001));

            operatoR = evaluateFirstArgument(operatorexpression, SYM_PROPOSITIONS_PL_KERNEL_KB_HOLDS);
          }
        }
        else {
          { OutputStringStream* stream000 = newOutputStringStream();

            *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
            throw *newStellaException(stream000->theStringReader());
          }
        }
      }
      verifyEvaluableOperator(operatoR, KWD_PROPOSITIONS_PREDICATE);
      (proposition->arguments->theArray)[0] = operatoR;
      { Object* arg = NULL;
        int i = NULL_INTEGER;
        Cons* iter002 = tree->rest;
        int iter003 = 1;

        while (!(iter002 == NIL)) {
          arg = iter002->value;
          iter002 = iter002->rest;
          i = iter003;
          iter003 = iter003 + 1;
          (proposition->arguments->theArray)[i] = (evaluateTerm(arg));
        }
      }
      return (proposition);
    }
  }
}

Surrogate* mostGeneralEquivalentSlotref(Surrogate* surrogate) {
  { Slot* slot = ((Slot*)(surrogate->surrogateValue));

    if (!((boolean)(slot))) {
      return (NULL);
    }
    if (((boolean)(slot->slotDirectEquivalent))) {
      while (((boolean)(slot->slotDirectEquivalent))) {
        slot = slot->slotDirectEquivalent;
      }
      if (!((boolean)(slot->slotSlotref))) {
        lookupSlotref(slot->slotOwner, slot->slotName);
      }
    }
    return (slot->slotSlotref);
  }
}

boolean polymorphicRelationP(Relation* self) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfStorageSlotP(testValue000)) {
      { Relation* self000 = self;
        StorageSlot* self = ((StorageSlot*)(self000));

        return (TRUE);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_TABLE)) {
      { Relation* self001 = self;
        Table* self = ((Table*)(self001));

        return (TRUE);
      }
    }
    else {
      return (FALSE);
    }
  }
}

boolean nonPolymorphicPredicateP(LogicObject* self) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
      { LogicObject* self000 = self;
        NamedDescription* self = ((NamedDescription*)(self000));

        { Relation* relation = self->nativeRelation();

          return (!(((boolean)(relation)) &&
              polymorphicRelationP(relation)));
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { LogicObject* self001 = self;
        LogicObject* self = self001;

        return (!testIsaP(self, SGT_PROPOSITIONS_PL_KERNEL_KB_POLYMORPHIC));
      }
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
        throw *newStellaException(stream000->theStringReader());
      }
    }
  }
}

LogicObject* evaluatePredicate(GeneralizedSymbol* name, Object* firstargument) {
  { Surrogate* surrogate = NULL;
    LogicObject* predicatevalue = NULL;

    { Surrogate* testValue000 = safePrimaryType(name);

      if (subtypeOfSymbolP(testValue000)) {
        { GeneralizedSymbol* name000 = name;
          Symbol* name = ((Symbol*)(name000));

          surrogate = symbolToSurrogate(name);
        }
      }
      else if (subtypeOfSurrogateP(testValue000)) {
        { GeneralizedSymbol* name001 = name;
          Surrogate* name = ((Surrogate*)(name001));

          surrogate = name;
        }
      }
      else {
        { OutputStringStream* stream000 = newOutputStringStream();

          *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
          throw *newStellaException(stream000->theStringReader());
        }
      }
    }
    predicatevalue = getDescription(surrogate);
    if (((boolean)(predicatevalue))) {
      return (predicatevalue);
    }
    else {
      {
        predicatevalue = ((LogicObject*)(surrogate->surrogateValue));
        if (((boolean)(predicatevalue)) &&
            (testIsaP(predicatevalue, SGT_PROPOSITIONS_PL_KERNEL_KB_RELATION) &&
             nonPolymorphicPredicateP(predicatevalue))) {
          return (predicatevalue);
        }
      }
    }
    if (((boolean)(firstargument))) {
      { Object* firstargvalue = valueOf(firstargument);
        Symbol* symbol = surrogateToSymbol(surrogate);

        if (!((boolean)(firstargvalue))) {
          signalUndefinedTerm(firstargument);
        }
        surrogate = lookupSlotrefOnInstance(firstargvalue, symbol);
        if (((boolean)(surrogate))) {
          return (getDescription(mostGeneralEquivalentSlotref(surrogate)));
        }
      }
    }
    return (NULL);
  }
}

// If TRUE, automatically coerce propositional arguments of a
// proposition, even if the corresponding argument type of the hosting relation
// doesn't indicate that.
boolean oAUTO_COERCE_PROPOSITIONAL_ARGUMENTSpo = FALSE;

boolean propositionalArgumentP(Object* argument) {
  if (safePrimaryType(argument) == SGT_PROPOSITIONS_STELLA_CONS) {
    { Object* argument000 = argument;
      Cons* argument = ((Cons*)(argument000));

      { Object* operatoR = argument->value;
        Object* firstarg = argument->rest->value;

        if ((!(operatoR == SYM_PROPOSITIONS_PL_KERNEL_KB_VALUE)) &&
            holdsOperatorP(operatoR)) {
          return (TRUE);
        }
        if (symbolP(operatoR)) {
          { GeneralizedSymbol* testValue000 = ((GeneralizedSymbol*)(operatoR));

            if ((testValue000 == SYM_PROPOSITIONS_STELLA_AND) ||
                ((testValue000 == SYM_PROPOSITIONS_STELLA_OR) ||
                 ((testValue000 == SYM_PROPOSITIONS_STELLA_NOT) ||
                  ((testValue000 == SYM_PROPOSITIONS_STELLA_EXISTS) ||
                   ((testValue000 == SYM_PROPOSITIONS_STELLA_FORALL) ||
                    (testValue000 == SYM_PROPOSITIONS_STELLA_e)))))) {
              return (TRUE);
            }
            else {
              { LogicObject* predicate = evaluatePredicate(((GeneralizedSymbol*)(operatoR)), firstarg);
                int arity = 0;

                if (((boolean)(predicate))) {
                  if (subtypeOfP(safePrimaryType(predicate), SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
                    { LogicObject* predicate000 = predicate;
                      NamedDescription* predicate = ((NamedDescription*)(predicate000));

                      if (functionP(predicate)) {
                        arity = predicate->arity();
                        if ((arity == -1) ||
                            (arity == argument->length())) {
                          return (FALSE);
                        }
                        else {
                          return (TRUE);
                        }
                      }
                      else {
                        return (TRUE);
                      }
                    }
                  }
                  else {
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  else {
  }
  return (FALSE);
}

Object* evaluateTypedArgument(Object* argument, Surrogate* type) {
  if ((type == SGT_PROPOSITIONS_LOGIC_PROPOSITION) ||
      (oAUTO_COERCE_PROPOSITIONAL_ARGUMENTSpo &&
       propositionalArgumentP(argument))) {
    return (buildProposition(argument));
  }
  else if (((boolean)(type))) {
    { 
      BIND_STELLA_SPECIAL(oDEFAULTCREATIONTYPEo, Surrogate*, type);
      return (evaluateTerm(argument));
    }
  }
  else {
    return (evaluateTerm(argument));
  }
}

Object* evaluateFirstArgument(Object* argument, Symbol* relationname) {
  { LogicObject* predicatevalue = evaluatePredicate(relationname, NULL);
    Surrogate* type = NULL;
    Object* firstargument = NULL;

    if (((boolean)(predicatevalue)) &&
        isaP(predicatevalue, SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
      type = ((Surrogate*)(((NamedDescription*)(predicatevalue))->ioVariableTypes->first()));
    }
    firstargument = evaluateTypedArgument(argument, type);
    if ((relationname == SYM_PROPOSITIONS_PL_KERNEL_KB_VALUE) ||
        (relationname == SYM_PROPOSITIONS_PL_KERNEL_KB_HOLDS)) {
      if (descriptionModeP() &&
          variableP(firstargument)) {
        { Proposition* fnprop = ((PatternVariable*)(firstargument))->definingProposition;

          { boolean testValue000 = FALSE;

            testValue000 = ((boolean)(fnprop));
            if (testValue000) {
              { boolean foundP000 = FALSE;

                { int i = NULL_INTEGER;
                  int iter000 = 0;
                  int upperBound000 = fnprop->arguments->length() - 2;

                  while (iter000 <= upperBound000) {
                    i = iter000;
                    iter000 = iter000 + 1;
                    if (variableP((fnprop->arguments->theArray)[i])) {
                      foundP000 = TRUE;
                      break;
                    }
                  }
                }
                testValue000 = foundP000;
              }
              testValue000 = !testValue000;
            }
            if (testValue000) {
              { 
                BIND_STELLA_SPECIAL(oEVALUATIONMODEo, Keyword*, KWD_PROPOSITIONS_EXTENSIONAL_ASSERTION);
                firstargument = evaluateTypedArgument(argument, type);
              }
            }
          }
        }
      }
    }
    else {
    }
    if (!((boolean)(firstargument))) {
      signalUndefinedTerm(argument);
    }
    { 
      BIND_STELLA_SPECIAL(oSUPPRESSNONLOGICOBJECTWARNINGpo, boolean, TRUE);
      if (valueOf(firstargument) == BOTTOM) {
        signalUndefinedTerm(firstargument);
      }
    }
    return (firstargument);
  }
}

Cons* evaluateRemainingArguments(LogicObject* predicatevalue, Cons* arguments) {
  { Cons* evaluatedarguments = NIL;

    if (((boolean)(predicatevalue)) &&
        isaP(predicatevalue, SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
      { List* variabletypes = ((NamedDescription*)(predicatevalue))->ioVariableTypes;

        { Surrogate* type = NULL;
          Cons* iter000 = variabletypes->rest();

          while (!(iter000 == NIL)) {
            type = ((Surrogate*)(iter000->value));
            iter000 = iter000->rest;
            { Object* head000 = arguments->value;

              arguments = arguments->rest;
              { Object* arg = head000;

                if (((boolean)(arg))) {
                  evaluatedarguments = cons(evaluateTypedArgument(arg, type), evaluatedarguments);
                }
              }
            }
          }
        }
        { Object* arg = NULL;
          Cons* iter001 = arguments;

          while (!(iter001 == NIL)) {
            arg = iter001->value;
            iter001 = iter001->rest;
            evaluatedarguments = cons(evaluateTypedArgument(arg, ((Surrogate*)(variabletypes->last()))), evaluatedarguments);
          }
        }
        return (evaluatedarguments->reverse());
      }
    }
    else {
      {
        { Object* arg = NULL;
          Cons* iter002 = arguments;
          Cons* collect000 = NULL;

          while (!(iter002 == NIL)) {
            arg = iter002->value;
            iter002 = iter002->rest;
            if (!((boolean)(collect000))) {
              {
                collect000 = cons(evaluateTerm(arg), NIL);
                if (evaluatedarguments == NIL) {
                  evaluatedarguments = collect000;
                }
                else {
                  addConsToEndOfConsList(evaluatedarguments, collect000);
                }
              }
            }
            else {
              {
                collect000->rest = cons(evaluateTerm(arg), NIL);
                collect000 = collect000->rest;
              }
            }
          }
        }
        return (evaluatedarguments);
      }
    }
  }
}

boolean classOperatorP(GeneralizedSymbol* operatoR) {
  { NamedDescription* description = getDescription(operatoR);

    return (((boolean)(description)) &&
        classDescriptionP(description));
  }
}

Proposition* buildPredicateProposition(Cons* tree) {
  if ((tree->length() == 2) &&
      classOperatorP(((GeneralizedSymbol*)(tree->value)))) {
    return (buildIsaProposition(listO(3, SYM_PROPOSITIONS_STELLA_ISA, tree->rest->value, cons(tree->value, NIL))));
  }
  if (holdsOperatorP(tree->value)) {
    return (buildHoldsProposition(tree));
  }
  { Object* predicatename = tree->value;
    Proposition* proposition = createProposition(SYM_PROPOSITIONS_STELLA_PREDICATE, tree->rest->length());
    Object* firstargument = NULL;
    LogicObject* predicatevalue = NULL;

    if (tree->rest == NIL) {
      if (predicatename == SYM_PROPOSITIONS_PL_KERNEL_KB_CUT) {
      }
      else {
        std::cout << "Predicate " << "`" << predicatename << "'" << " applied to zero arguments." << std::endl;
      }
    }
    else {
      firstargument = evaluateFirstArgument(tree->rest->value, ((Symbol*)(predicatename)));
      (proposition->arguments->theArray)[0] = firstargument;
    }
    predicatevalue = evaluatePredicate(((GeneralizedSymbol*)(predicatename)), firstargument);
    if (((boolean)(predicatevalue)) &&
        functionP(predicatevalue)) {
      if (descriptionModeP()) {
        proposition->kind = KWD_PROPOSITIONS_FUNCTION;
      }
      else {
        { int lastargindex = tree->length() - 1;
          Object* resultargument = tree->nth(lastargindex);

          tree->nthSetter(NULL, lastargindex);
          tree = tree->remove(NULL);
          return (((Proposition*)(buildProposition(listO(3, SYM_PROPOSITIONS_STELLA_e, tree, cons(resultargument, NIL))))));
        }
      }
    }
    { Object* arg = NULL;
      int i = NULL_INTEGER;
      Cons* iter000 = evaluateRemainingArguments(predicatevalue, tree->rest->rest);
      int iter001 = 1;

      while (!(iter000 == NIL)) {
        arg = iter000->value;
        iter000 = iter000->rest;
        i = iter001;
        iter001 = iter001 + 1;
        (proposition->arguments->theArray)[i] = arg;
      }
    }
    proposition->operatoR = (((boolean)(predicatevalue)) ? predicatevalue->surrogateValueInverse : ((GeneralizedSymbol*)(predicatename)));
    if (proposition->operatoR == SGT_PROPOSITIONS_PL_KERNEL_KB_VALUE) {
      proposition = normalizeValueFunction(proposition);
    }
    verifyArgumentTypesAndCount(proposition);
    return (proposition);
  }
}

boolean checkTypeP(Object* self, Surrogate* type, boolean shallowP, Object*& _Return1) {
  shallowP = shallowP;
  if ((!((boolean)(type))) ||
      ((!((boolean)(self))) ||
       (type == SGT_PROPOSITIONS_STELLA_THING))) {
    _Return1 = self;
    return (TRUE);
  }
  { boolean typeisokP = TRUE;

    { Surrogate* testValue000 = safePrimaryType(self);

      if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE)) {
        { Object* self000 = self;
          PatternVariable* self = ((PatternVariable*)(self000));

        }
      }
      else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_LITERAL_WRAPPER)) {
        { Object* self001 = self;
          LiteralWrapper* self = ((LiteralWrapper*)(self001));

          if (isaP(type->surrogateValue, SGT_PROPOSITIONS_STELLA_CLASS)) {
            typeisokP = subtypeOfP(logicalType(self), type);
          }
          else {
            typeisokP = FALSE;
          }
        }
      }
      else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
        { Object* self002 = self;
          Proposition* self = ((Proposition*)(self002));

          if (isaP(type->surrogateValue, SGT_PROPOSITIONS_STELLA_CLASS)) {
            typeisokP = subtypeOfP(logicalType(self), type);
          }
          else {
            typeisokP = FALSE;
          }
        }
      }
      else {
        typeisokP = testIsaP(self, type) ||
            logicalSubtypeOfP(logicalType(self), type);
      }
    }
    if (typeisokP) {
      _Return1 = self;
      return (TRUE);
    }
    { Object* substitutedvalue = coerceIncompatibleValue(self, type);

      if (((boolean)(substitutedvalue))) {
        _Return1 = substitutedvalue;
        return (TRUE);
      }
    }
    if (((boolean)(oQUERYITERATORo.get())) &&
        ((boolean)(oQUERYITERATORo.get()->partialMatchStrategy))) {
      _Return1 = self;
      return (TRUE);
    }
    _Return1 = self;
    return (FALSE);
  }
}

Object* coerceIncompatibleValue(Object* value, Surrogate* type) {
  { Object* substituteValue = NULL;

    if (relationrefSpecializesRelationrefP(type, SGT_PROPOSITIONS_PL_KERNEL_KB_QUANTITY)) {
      return (coerceQuantity(value, type));
    }
    else if (relationrefSpecializesRelationrefP(type, SGT_PROPOSITIONS_PL_KERNEL_KB_DATE)) {
      return (coerceDate(value, type));
    }
    { Surrogate* testValue000 = safePrimaryType(value);

      if (subtypeOfIntegerP(testValue000)) {
        { Object* value000 = value;
          IntegerWrapper* value = ((IntegerWrapper*)(value000));

          if (type == SGT_PROPOSITIONS_STELLA_FLOAT) {
            { int intvalue = value->wrapperValue;
              double floatvalue = ((double)(intvalue));

              substituteValue = wrapFloat(floatvalue);
            }
          }
        }
      }
      else if (subtypeOfFloatP(testValue000)) {
        { Object* value001 = value;
          FloatWrapper* value = ((FloatWrapper*)(value001));

          if (type == SGT_PROPOSITIONS_STELLA_INTEGER) {
            { double floatvalue = value->wrapperValue;
              int intvalue = stella::floor(floatvalue);

              if (((double)(intvalue)) == floatvalue) {
                substituteValue = wrapInteger(intvalue);
              }
            }
          }
        }
      }
      else {
      }
    }
    return (substituteValue);
  }
}

Object* coerceQuantity(Object* value, Surrogate* quantitytyperef) {
  return (value);
}

Object* coerceDate(Object* value, Surrogate* datetyperef) {
  return (value);
}

// Three policies are implemented:
//    :AUTOMATICALLY-FIX-TYPE-VIOLATIONS asserts missing types to fix type
//        violations (default),
//    :REPORT-TYPE-VIOLATIONS complains about missing or incorrect types, and
//    :IGNORE-TYPE-VIOLATIONS which disables all type checking.
Keyword* oTYPE_CHECK_POLICYo = NULL;

// Controls the behavior of the type-checking
// routines in the event that a type-check fails.  Options are:
//   :POST-TYPE-VIOLATIONS              push offending proposition onto queue,
//   :REPORT-TYPE-VIOLATIONS            print occasions of failed type checks,
//   :AUTOMATICALLY-FIX-TYPE-VIOLATIONS assert missing types on propositions, and
//   :IGNORE-TYPE-VIOLATIONS            don't perform any type checking at all.
DEFINE_STELLA_SPECIAL(oTYPECHECKMODEo, Keyword* , NULL);

void verifyOneArgumentType(Object* arg, Surrogate* type, Proposition* proposition, NamedDescription* description) {
  if ((oTYPECHECKMODEo.get() == KWD_PROPOSITIONS_IGNORE_TYPE_VIOLATIONS) ||
      (!((boolean)(type->surrogateValue)))) {
    return;
  }
  { Object* argvalue = valueOf(arg);
    boolean typeisokP = FALSE;
    Object* substitutevalue = NULL;

    typeisokP = checkTypeP(argvalue, type, FALSE, substitutevalue);
    if (!eqlP(argvalue, substitutevalue)) {
      { Object* argtoo = NULL;
        Surrogate* t = NULL;
        int i = NULL_INTEGER;
        Vector* vector000 = proposition->arguments;
        int index000 = 0;
        int length000 = vector000->length();
        Cons* iter000 = description->ioVariableTypes->theConsList;
        int iter001 = 0;

        while ((index000 < length000) &&
            (!(iter000 == NIL))) {
          argtoo = (vector000->theArray)[index000];
          index000 = index000 + 1;
          t = ((Surrogate*)(iter000->value));
          iter000 = iter000->rest;
          i = iter001;
          iter001 = iter001 + 1;
          if (eqlP(argtoo, arg) &&
              (t == type)) {
            (proposition->arguments->theArray)[i] = substitutevalue;
          }
        }
      }
    }
    else if (!typeisokP) {
      handleArgumentTypeViolation(proposition, arg, type);
    }
  }
}

void verifyArgumentTypesAndCount(Proposition* proposition) {
  if ((oTYPECHECKMODEo.get() == KWD_PROPOSITIONS_IGNORE_TYPE_VIOLATIONS) ||
      ((!worldStateP(oCONTEXTo.get())) ||
       (!isaP(proposition->operatoR, SGT_PROPOSITIONS_STELLA_SURROGATE)))) {
    return;
  }
  { NamedDescription* description = getDescription(((Surrogate*)(proposition->operatoR)));
    int argumentcount = proposition->arguments->length();

    { Keyword* testValue000 = proposition->kind;

      if ((testValue000 == KWD_PROPOSITIONS_PREDICATE) ||
          ((testValue000 == KWD_PROPOSITIONS_FUNCTION) ||
           (testValue000 == KWD_PROPOSITIONS_ISA))) {
        if (!((boolean)(description))) {
          return;
        }
        { int arity = description->arity();
          boolean variablearityP = !(arity > 0);
          boolean functionP = proposition->kind == KWD_PROPOSITIONS_FUNCTION;

          if ((!variablearityP) &&
              (!(argumentcount == arity))) {
            handleArityViolation(proposition, arity);
          }
          { Object* arg = NULL;
            Object* type = NULL;
            Vector* vector000 = proposition->arguments;
            int index000 = 0;
            int length000 = vector000->length();
            IntegerIntervalIterator* i = interval(1, ((variablearityP &&
                functionP) ? (argumentcount - 1) : argumentcount));
            Iterator* iter000 = allDomainTypes(description);

            while ((index000 < length000) &&
                (i->nextP() &&
                 iter000->nextP())) {
              arg = (vector000->theArray)[index000];
              index000 = index000 + 1;
              type = iter000->value;
              if (((boolean)(type)) &&
                  ((!(type == SGT_PROPOSITIONS_STELLA_THING)) &&
                   ((boolean)(arg)))) {
                verifyOneArgumentType(arg, ((Surrogate*)(type)), proposition, description);
              }
            }
          }
          if (functionP) {
            verifyOneArgumentType(proposition->arguments->last(), ((Surrogate*)(description->ioVariableTypes->last())), proposition, description);
          }
        }
      }
      else if (testValue000 == KWD_PROPOSITIONS_EQUIVALENT) {
        { Object* arg1 = (proposition->arguments->theArray)[0];
          Object* arg2 = (proposition->arguments->theArray)[1];
          Surrogate* requiredtype = NULL;
          Object* constrainingarg = NULL;
          Object* offendingarg = NULL;
          boolean typeisokP = FALSE;
          Object* substitutevalue = NULL;

          requiredtype = safeLogicalType(arg2);
          typeisokP = checkTypeP(arg1, requiredtype, FALSE, substitutevalue);
          if (!eqlP(arg1, substitutevalue)) {
            (proposition->arguments->theArray)[0] = substitutevalue;
            return;
          }
          else if (!typeisokP) {
            constrainingarg = arg2;
            offendingarg = arg1;
          }
          if (typeisokP) {
            requiredtype = safeLogicalType(arg1);
            typeisokP = checkTypeP(arg2, requiredtype, FALSE, substitutevalue);
            if (!eqlP(arg2, substitutevalue)) {
              (proposition->arguments->theArray)[1] = substitutevalue;
              return;
            }
            else if (!typeisokP) {
              constrainingarg = arg1;
              offendingarg = arg2;
            }
          }
          if (!(typeisokP)) {
            { 
              BIND_STELLA_SPECIAL(oTERMUNDERCONSTRUCTIONo, Object*, oTERMUNDERCONSTRUCTIONo.get());
              if ((!((boolean)(oTERMUNDERCONSTRUCTIONo.get()))) &&
                  skolemP(constrainingarg)) {
                oTERMUNDERCONSTRUCTIONo.set(((Skolem*)(constrainingarg))->definingProposition);
              }
              handleArgumentTypeViolation(proposition, offendingarg, requiredtype);
            }
          }
        }
      }
      else {
      }
    }
  }
}

void handleArgumentTypeViolation(Proposition* proposition, Object* arg, Surrogate* requiredtype) {
  if (oTYPECHECKMODEo.get() == KWD_PROPOSITIONS_POST_TYPE_VIOLATIONS) {
    postToCheckTypesAgenda(proposition);
  }
  else if (oTYPECHECKMODEo.get() == KWD_PROPOSITIONS_REPORT_TYPE_VIOLATIONS) {
    { 
      BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
      *(STANDARD_WARNING->nativeStream) << "WARNING: " << "Type check violation on argument `" << arg << "' in proposition" << std::endl << "   " << proposition << "." << std::endl << "   Argument must have type `" << requiredtype->symbolName << "'." << std::endl;
      helpSignalPropositionError(STANDARD_WARNING, KWD_PROPOSITIONS_WARNING);
    }
  }
  else if (oTYPECHECKMODEo.get() == KWD_PROPOSITIONS_AUTOMATICALLY_FIX_TYPE_VIOLATIONS) {
    { Proposition* isaproposition = assertIsaProposition(arg, requiredtype);

      setDynamicSlotValue(isaproposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_ASSERTED_BY_TYPE_CHECKERp, (TRUE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
    }
  }
  else {
    { OutputStringStream* stream000 = newOutputStringStream();

      *(stream000->nativeStream) << "`" << oTYPECHECKMODEo.get() << "'" << " is not a valid case option";
      throw *newStellaException(stream000->theStringReader());
    }
  }
}

void handleArityViolation(Proposition* proposition, int requiredarity) {
  if (oTYPECHECKMODEo.get() == KWD_PROPOSITIONS_POST_TYPE_VIOLATIONS) {
    postToCheckTypesAgenda(proposition);
  }
  else if ((oTYPECHECKMODEo.get() == KWD_PROPOSITIONS_REPORT_TYPE_VIOLATIONS) ||
      (oTYPECHECKMODEo.get() == KWD_PROPOSITIONS_AUTOMATICALLY_FIX_TYPE_VIOLATIONS)) {
    { OutputStringStream* stream000 = newOutputStringStream();

      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        *(stream000->nativeStream) << "ERROR: " << "Too " << "`" << (((proposition->arguments->length() > requiredarity) ? (char*)"many" : (char*)"few")) << "'" << " arguments in the proposition" << std::endl << "   " << "`" << generateTerm(proposition) << "'" << std::endl << "   Expected " << "`" << requiredarity << "'" << " argument(s)." << "." << std::endl;
        helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
      }
      throw *newPropositionError(stream000->theStringReader());
    }
  }
  else {
    { OutputStringStream* stream001 = newOutputStringStream();

      *(stream001->nativeStream) << "`" << oTYPECHECKMODEo.get() << "'" << " is not a valid case option";
      throw *newStellaException(stream001->theStringReader());
    }
  }
}

CheckTypesRecord* newCheckTypesRecord() {
  { CheckTypesRecord* self = NULL;

    self = new CheckTypesRecord();
    self->module = NULL;
    self->parentDescription = NULL;
    self->parentProposition = NULL;
    return (self);
  }
}

Surrogate* CheckTypesRecord::primaryType() {
  { CheckTypesRecord* self = this;

    return (SGT_PROPOSITIONS_LOGIC_CHECK_TYPES_RECORD);
  }
}

Object* accessCheckTypesRecordSlotValue(CheckTypesRecord* self, Symbol* slotname, Object* value, boolean setvalueP) {
  if (slotname == SYM_PROPOSITIONS_LOGIC_PARENT_PROPOSITION) {
    if (setvalueP) {
      self->parentProposition = value;
    }
    else {
      value = self->parentProposition;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_LOGIC_PARENT_DESCRIPTION) {
    if (setvalueP) {
      self->parentDescription = ((Description*)(value));
    }
    else {
      value = self->parentDescription;
    }
  }
  else if (slotname == SYM_PROPOSITIONS_STELLA_MODULE) {
    if (setvalueP) {
      self->module = ((Module*)(value));
    }
    else {
      value = self->module;
    }
  }
  else {
    { OutputStringStream* stream000 = newOutputStringStream();

      *(stream000->nativeStream) << "`" << slotname << "'" << " is not a valid case option";
      throw *newStellaException(stream000->theStringReader());
    }
  }
  return (value);
}

// List of propositions that have failed a type check,
// but might pass once finalization is complete.
KeyValueList* oCHECK_TYPES_AGENDAo = NULL;

void postToCheckTypesAgenda(Proposition* self) {
  { CheckTypesRecord* self000 = newCheckTypesRecord();

    self000->parentProposition = oTERMUNDERCONSTRUCTIONo.get();
    self000->parentDescription = ((Description*)(oDESCRIPTIONUNDERCONSTRUCTIONo.get()));
    self000->module = oMODULEo.get();
    oCHECK_TYPES_AGENDAo->insertAt(self, self000);
  }
}

void processCheckTypesAgenda() {
  if (oCHECK_TYPES_AGENDAo->emptyP() ||
      oINHIBITOBJECTFINALIZATIONpo.get()) {
    return;
  }
  std::cout << "Processing check-types agenda..." << std::endl;
  { KeyValueList* agenda = oCHECK_TYPES_AGENDAo->reverse();

    oCHECK_TYPES_AGENDAo = newKeyValueList();
    { Proposition* proposition = NULL;
      CheckTypesRecord* record = NULL;
      KvCons* iter000 = agenda->theKvList;

      while (((boolean)(iter000))) {
        proposition = ((Proposition*)(iter000->key));
        record = ((CheckTypesRecord*)(iter000->value));
        iter000 = iter000->rest;
        if (!proposition->deletedP()) {
          { 
            BIND_STELLA_SPECIAL(oTYPECHECKMODEo, Keyword*, ((((BooleanWrapper*)(dynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTIVEp, FALSE_WRAPPER)))->wrapperValue &&
                (!(oTYPE_CHECK_POLICYo == KWD_PROPOSITIONS_IGNORE_TYPE_VIOLATIONS))) ? KWD_PROPOSITIONS_REPORT_TYPE_VIOLATIONS : oTYPE_CHECK_POLICYo));
            BIND_STELLA_SPECIAL(oTERMUNDERCONSTRUCTIONo, Object*, record->parentProposition);
            BIND_STELLA_SPECIAL(oDESCRIPTIONUNDERCONSTRUCTIONo, Object*, record->parentDescription);
            if (proposition->deletedP() ||
                ((((boolean)(oTERMUNDERCONSTRUCTIONo.get())) &&
                oTERMUNDERCONSTRUCTIONo.get()->deletedP()) ||
                 (((boolean)(oDESCRIPTIONUNDERCONSTRUCTIONo.get())) &&
                  oDESCRIPTIONUNDERCONSTRUCTIONo.get()->deletedP()))) {
              continue;
            }
            try {
              { 
                BIND_STELLA_SPECIAL(oMODULEo, Module*, record->module);
                BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
                verifyArgumentTypesAndCount(proposition);
              }
            }
            catch (LogicException& _e) {
              LogicException* e = &_e;

              *(STANDARD_ERROR->nativeStream) << exceptionMessage(e);
            }
          }
        }
      }
    }
  }
}

void processDefinitions() {
  // Finish processing all definitions and assertions that have
  // been evaluated/loaded since that last call to 'process-definitions'.
  // PowerLoom defers complete processing of definitions to make it easier
  // to support cyclic definitions.  Following finalization of definitions,
  // this call performs semantic validation of any assertions evaluated since
  // the last call to 'process-definitions'.  PowerLoom calls this function
  // internally before each query; the primary reason to call it explicitly
  // is to force the production of any diagnostic information that results from
  // the processing and validation.
  finalizeObjects();
  processCheckTypesAgenda();
}

Proposition* buildExistsProposition(Cons* tree) {
  { Cons* variables = NULL;
    Proposition* proposition = NULL;
    Proposition* unused = NULL;

    variables = buildQuantifiedProposition(tree, TRUE, proposition, unused);
    unused = unused;
    { Proposition* existsprop = createProposition(SYM_PROPOSITIONS_STELLA_EXISTS, 1);

      existsprop->kind = KWD_PROPOSITIONS_EXISTS;
      (existsprop->arguments->theArray)[0] = proposition;
      setDynamicSlotValue(existsprop->dynamicSlots, SYM_PROPOSITIONS_LOGIC_IO_VARIABLES, copyConsListToVariablesVector(variables), NULL);
      return (existsprop);
    }
  }
}

boolean topLevelExistsPropositionP(Proposition* proposition) {
  return ((!unknownP(proposition)) &&
      helpCollectSkolemizedExistsVariablesP(proposition, NULL));
}

Vector* collectSkolemizedExistsVariables(Proposition* proposition) {
  { List* skolems = newList();
    Vector* result = NULL;

    helpCollectSkolemizedExistsVariablesP(proposition, skolems);
    if (skolems->emptyP()) {
      result = ZERO_VARIABLES_VECTOR;
    }
    else {
      result = newVector(skolems->length());
    }
    { Skolem* skolem = NULL;
      int i = NULL_INTEGER;
      Cons* iter000 = skolems->theConsList;
      int iter001 = 0;

      while (!(iter000 == NIL)) {
        skolem = ((Skolem*)(iter000->value));
        iter000 = iter000->rest;
        i = iter001;
        iter001 = iter001 + 1;
        (result->theArray)[i] = skolem;
      }
    }
    skolems->free();
    return (result);
  }
}

boolean helpCollectSkolemizedExistsVariablesP(Proposition* proposition, List* result) {
  { Vector* arguments = proposition->arguments;
    int argumentcount = arguments->length();

    if (proposition->kind == KWD_PROPOSITIONS_FUNCTION) {
      argumentcount = argumentcount - 1;
    }
    { Object* arg = NULL;
      int i = NULL_INTEGER;
      Vector* vector000 = arguments;
      int index000 = 0;
      int length000 = vector000->length();
      int iter000 = 1;
      int upperBound000 = argumentcount;
      boolean unboundedP000 = upperBound000 == NULL_INTEGER;

      while ((index000 < length000) &&
          (unboundedP000 ||
           (iter000 <= upperBound000))) {
        arg = (vector000->theArray)[index000];
        index000 = index000 + 1;
        i = iter000;
        iter000 = iter000 + 1;
        i = i;
        { Surrogate* testValue000 = safePrimaryType(arg);

          if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
            { Object* arg000 = arg;
              Proposition* arg = ((Proposition*)(arg000));

              if (helpCollectSkolemizedExistsVariablesP(arg, result) &&
                  (!((boolean)(result)))) {
                return (TRUE);
              }
            }
          }
          else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE)) {
            { Object* arg001 = arg;
              PatternVariable* arg = ((PatternVariable*)(arg001));

            }
          }
          else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
            { Object* arg002 = arg;
              Skolem* arg = ((Skolem*)(arg002));

              { boolean foundP000 = FALSE;

                { Proposition* prop = NULL;
                  Iterator* iter001 = unfilteredDependentIsaPropositions(arg)->allocateIterator();

                  while (iter001->nextP()) {
                    prop = ((Proposition*)(iter001->value));
                    if ((prop->kind == KWD_PROPOSITIONS_ISA) &&
                        (trueP(prop) &&
                         coerceWrappedBooleanToBoolean(prop->variableTypeP_reader()))) {
                      foundP000 = TRUE;
                      break;
                    }
                  }
                }
                if (foundP000) {
                  if (((boolean)(result))) {
                    result->insertNew(arg);
                  }
                  else {
                    return (TRUE);
                  }
                }
              }
            }
          }
          else {
          }
        }
      }
    }
    if (((boolean)(result))) {
      return (result->nonEmptyP());
    }
    else {
      return (FALSE);
    }
  }
}

Proposition* buildForallProposition(Cons* tree) {
  { Proposition* forallprop = createProposition(SYM_PROPOSITIONS_STELLA_FORALL, 2);

    { Cons* variables = NULL;
      Proposition* antecedent = NULL;
      Proposition* consequent = NULL;

      variables = buildQuantifiedProposition(tree, TRUE, antecedent, consequent);
      forallprop->kind = KWD_PROPOSITIONS_FORALL;
      setDynamicSlotValue(forallprop->dynamicSlots, SYM_PROPOSITIONS_LOGIC_IO_VARIABLES, copyConsListToVariablesVector(variables), NULL);
      (forallprop->arguments->theArray)[0] = ((((boolean)(antecedent)) ? antecedent : TRUE_PROPOSITION));
      (forallprop->arguments->theArray)[1] = consequent;
    }
    return (forallprop);
  }
}

void verifyForallPropositions(Proposition* proposition) {
  { Keyword* testValue000 = proposition->kind;

    if (testValue000 == KWD_PROPOSITIONS_FORALL) {
      { Proposition* antecedent = ((Proposition*)((proposition->arguments->theArray)[0]));
        Proposition* consequent = ((Proposition*)((proposition->arguments->theArray)[1]));
        List* antvars = NULL;
        List* cqvars = NULL;

        if (antecedent->kind == KWD_PROPOSITIONS_CONSTANT) {
          return;
        }
        antvars = list(0);
        cqvars = list(0);
        collectFreeVariables(antecedent, antvars, list(0), list(0));
        collectFreeVariables(consequent, cqvars, list(0), list(0));
        if (!(cqvars->nonEmptyP() &&
            cqvars->subsetP(antvars))) {
          { OutputStringStream* stream000 = newOutputStringStream();

            { 
              BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
              *(stream000->nativeStream) << "ERROR: " << "Illegal implication rule, consequent has variables" << std::endl << "   that are not restricted by the antecedent: " << std::endl << "   " << "`" << proposition << "'" << "." << std::endl;
              helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
            }
            throw *newPropositionError(stream000->theStringReader());
          }
        }
      }
    }
    else if (testValue000 == KWD_PROPOSITIONS_AND) {
      { Object* arg = NULL;
        Vector* vector000 = proposition->arguments;
        int index000 = 0;
        int length000 = vector000->length();

        while (index000 < length000) {
          arg = (vector000->theArray)[index000];
          index000 = index000 + 1;
          verifyForallPropositions(((Proposition*)(arg)));
        }
      }
    }
    else {
    }
  }
}

Object* maybeSubstituteSurrogateArgument(Object* arg1, Object* arg2) {
  if (skolemP(arg2) ||
      descriptionModeP()) {
    if (subtypeOfP(safePrimaryType(arg1), SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) {
      { Object* arg1000 = arg1;
        Description* arg1 = ((Description*)(arg1000));

        if (((boolean)(arg1->descriptionSurrogate()))) {
          return (arg1->descriptionSurrogate());
        }
      }
    }
    else {
    }
  }
  return (arg1);
}

Proposition* createImpliesProposition(Object* tail, Object* head) {
  { Proposition* impliesprop = createProposition(SYM_PROPOSITIONS_LOGIC_IMPLIES, 2);

    (impliesprop->arguments->theArray)[0] = (maybeSubstituteSurrogateArgument(tail, head));
    (impliesprop->arguments->theArray)[1] = (maybeSubstituteSurrogateArgument(head, tail));
    verifyArgumentTypesAndCount(impliesprop);
    return (impliesprop);
  }
}

Proposition* buildImpliesProposition(Cons* tree) {
  verifyNumberOfPropositionArguments(tree, 2);
  { Object* tailarg = evaluateTerm(tree->rest->value);
    Object* headarg = evaluateTerm(tree->rest->rest->value);

    return (createImpliesProposition(tailarg, headarg));
  }
}

Proposition* findDuplicateFunctionProposition(Proposition* proposition) {
  if (oLOADINGREGENERABLEOBJECTSpo.get()) {
    return (NULL);
  }
  { GeneralizedSymbol* surrogate = proposition->operatoR;
    Vector* arguments = proposition->arguments;

    if (!isaP(surrogate, SGT_PROPOSITIONS_STELLA_SURROGATE)) {
      return (NULL);
    }
    { Proposition* value000 = NULL;

      { Proposition* p = NULL;
        Iterator* iter000 = findSimilarPropositions(proposition)->allocateIterator();

        while (iter000->nextP()) {
          p = ((Proposition*)(iter000->value));
          { boolean testValue000 = FALSE;

            testValue000 = p->operatoR == surrogate;
            if (testValue000) {
              { boolean alwaysP000 = TRUE;

                { Object* arg1 = NULL;
                  Object* arg2 = NULL;
                  Iterator* iter001 = arguments->butLast();
                  Vector* vector000 = p->arguments;
                  int index000 = 0;
                  int length000 = vector000->length();

                  while (iter001->nextP() &&
                      (index000 < length000)) {
                    arg1 = iter001->value;
                    arg2 = (vector000->theArray)[index000];
                    index000 = index000 + 1;
                    if (!eqlP(valueOf(arg1), valueOf(arg2))) {
                      alwaysP000 = FALSE;
                      break;
                    }
                  }
                }
                testValue000 = alwaysP000;
              }
              if (testValue000) {
                testValue000 = (arguments->length() == p->arguments->length()) &&
                    ((proposition->homeContext == p->homeContext) ||
                     proposition->homeContext->allSuperContexts->membP(p->homeContext));
              }
            }
            if (testValue000) {
              value000 = p;
              break;
            }
          }
        }
      }
      { Proposition* value001 = value000;

        return (value001);
      }
    }
  }
}

Surrogate* unwrapWrappedType(Surrogate* type) {
  if (((boolean)(type)) &&
      (((boolean)(type->surrogateValue)) &&
       (isaP(type->surrogateValue, SGT_PROPOSITIONS_STELLA_CLASS) &&
        subtypeOfP(type, SGT_PROPOSITIONS_STELLA_LITERAL_WRAPPER)))) {
    type = wrappedTypeToType(type);
  }
  return (type);
}

Proposition* beginCreatingFunctionProposition(GeneralizedSymbol* surrogate, Cons* inputarguments) {
  { int argcount = inputarguments->length() + 1;
    Proposition* proposition = createProposition(SYM_PROPOSITIONS_STELLA_FUNCTION, argcount);

    proposition->operatoR = surrogate;
    { Object* arg = NULL;
      int i = NULL_INTEGER;
      Cons* iter000 = inputarguments;
      int iter001 = 0;

      while (!(iter000 == NIL)) {
        arg = iter000->value;
        iter000 = iter000->rest;
        i = iter001;
        iter001 = iter001 + 1;
        (proposition->arguments->theArray)[i] = arg;
      }
    }
    return (proposition);
  }
}

Proposition* finishCreatingFunctionProposition(Proposition* proposition) {
  { int argcount = proposition->arguments->length();
    NamedDescription* description = getDescription(proposition->operatoR);
    Surrogate* type = NULL;
    Skolem* skolem = NULL;

    if (((boolean)(description))) {
      type = unwrapWrappedType(((Surrogate*)(description->ioVariableTypes->last())));
    }
    { 
      BIND_STELLA_SPECIAL(oCONTEXTo, Context*, proposition->homeContext);
      BIND_STELLA_SPECIAL(oMODULEo, Module*, oCONTEXTo.get()->baseModule);
      skolem = createVariableOrSkolem(type, NULL);
    }
    (proposition->arguments->theArray)[(argcount - 1)] = skolem;
    skolem->definingProposition = proposition;
    if (((boolean)(type))) {
      updateSkolemType(skolem, type);
    }
    if (proposition->operatoR == SGT_PROPOSITIONS_PL_KERNEL_KB_VALUE) {
      proposition = normalizeValueFunction(proposition);
    }
    verifyArgumentTypesAndCount(proposition);
    if (isaP(proposition->operatoR, SGT_PROPOSITIONS_STELLA_SURROGATE)) {
      fastenDownOneProposition(proposition, TRUE);
    }
    return (proposition);
  }
}

Proposition* findOrCreateFunctionProposition(GeneralizedSymbol* predicate, Cons* inputarguments) {
  { Proposition* proposition = beginCreatingFunctionProposition(predicate, inputarguments);
    Proposition* duplicate = ((!descriptionModeP()) ? findDuplicateFunctionProposition(proposition) : ((Proposition*)(NULL)));

    if (((boolean)(duplicate))) {
      return (duplicate);
    }
    proposition = finishCreatingFunctionProposition(proposition);
    { boolean testValue000 = FALSE;

      if (!((BooleanWrapper*)(dynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTIVEp, FALSE_WRAPPER)))->wrapperValue) {
        testValue000 = TRUE;
      }
      else {
        { boolean alwaysP000 = TRUE;

          { int i = NULL_INTEGER;
            int iter000 = 0;
            int upperBound000 = proposition->arguments->length() - 2;

            while (iter000 <= upperBound000) {
              i = iter000;
              iter000 = iter000 + 1;
              if (!(!variableP((proposition->arguments->theArray)[i]))) {
                alwaysP000 = FALSE;
                break;
              }
            }
          }
          testValue000 = alwaysP000;
        }
      }
      if (testValue000) {
        runGoesTrueDemons(proposition);
      }
    }
    return (proposition);
  }
}

Proposition* createFunctionProposition(GeneralizedSymbol* predicate, Cons* inputarguments) {
  { Proposition* proposition = beginCreatingFunctionProposition(predicate, inputarguments);

    return (finishCreatingFunctionProposition(proposition));
  }
}

Object* evaluateFunctionTerm(Cons* tree) {
  if (!(isaP(tree->value, SGT_PROPOSITIONS_STELLA_GENERALIZED_SYMBOL))) {
    { OutputStringStream* stream000 = newOutputStringStream();

      { 
        BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
        *(stream000->nativeStream) << "ERROR: " << "Illegal function name in function term: " << "`" << tree << "'" << "." << std::endl;
        helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
      }
      throw *newPropositionError(stream000->theStringReader());
    }
  }
  { GeneralizedSymbol* name = ((GeneralizedSymbol*)(tree->value));
    Object* firstargument = evaluateFirstArgument(tree->rest->value, ((Symbol*)(name)));
    LogicObject* predicatevalue = evaluatePredicate(name, firstargument);
    Cons* inputarguments = evaluateRemainingArguments(predicatevalue, tree->rest->rest);

    if (((boolean)(predicatevalue)) &&
        (isaP(predicatevalue, SGT_PROPOSITIONS_LOGIC_DESCRIPTION) &&
         (!functionDescriptionP(((NamedDescription*)(predicatevalue)))))) {
      { OutputStringStream* stream001 = newOutputStringStream();

        { 
          BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
          *(stream001->nativeStream) << "ERROR: " << "Relation " << "`" << ((NamedDescription*)(predicatevalue))->descriptionName() << "'" << " found in function position." << "." << std::endl;
          helpSignalPropositionError(stream001, KWD_PROPOSITIONS_ERROR);
        }
        throw *newPropositionError(stream001->theStringReader());
      }
    }
    inputarguments = cons(firstargument, inputarguments);
    { Proposition* functionproposition = findOrCreateFunctionProposition((((boolean)(predicatevalue)) ? predicatevalue->surrogateValueInverse : name), inputarguments);

      if (symbolP(functionproposition->operatoR) &&
          (!descriptionModeP())) {
        complainAboutUndeclaredReference(functionproposition);
        createDummyRelation(functionproposition);
      }
      return ((functionproposition->arguments->theArray)[(functionproposition->arguments->length() - 1)]);
    }
  }
}

boolean extensionalIndividualP(Object* individual) {
  { Surrogate* testValue000 = safePrimaryType(individual);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE)) {
      { Object* individual000 = individual;
        PatternVariable* individual = ((PatternVariable*)(individual000));

        return (FALSE);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
      { Object* individual001 = individual;
        Skolem* individual = ((Skolem*)(individual001));

        { Proposition* definingproposition = individual->definingProposition;

          { boolean foundP000 = FALSE;

            { Proposition* prop = NULL;
              Iterator* iter000 = unfilteredDependentPropositions(individual, NULL)->allocateIterator();

              while (iter000->nextP()) {
                prop = ((Proposition*)(iter000->value));
                if ((!(prop == definingproposition)) &&
                    ((!prop->deletedP()) &&
                     ((oREVERSEPOLARITYpo.get() ? falseP(prop) : (trueP(prop) ||
                    functionWithDefinedValueP(prop)))))) {
                  foundP000 = TRUE;
                  break;
                }
              }
            }
            if (foundP000) {
              return (TRUE);
            }
          }
        }
        { boolean foundP001 = FALSE;

          { LogicObject* sk = NULL;
            Cons* iter001 = individual->variableValueInverse_reader();

            while (!(iter001 == NIL)) {
              sk = ((LogicObject*)(iter001->value));
              iter001 = iter001->rest;
              if (extensionalIndividualP(sk)) {
                foundP001 = TRUE;
                break;
              }
            }
          }
          { boolean value000 = foundP001;

            return (value000);
          }
        }
      }
    }
    else {
      return (TRUE);
    }
  }
}

boolean functionWithDefinedValueP(Proposition* proposition) {
  if (!(proposition->kind == KWD_PROPOSITIONS_FUNCTION)) {
    return (FALSE);
  }
  { Object* lastargument = proposition->arguments->last();
    boolean resultP = testPropertyP(getDescription(((Surrogate*)(proposition->operatoR))), SGT_PROPOSITIONS_PL_KERNEL_KB_TOTAL) ||
        extensionalIndividualP(valueOf(lastargument));

    if (resultP &&
        (!worldStateP(oCONTEXTo.get()))) {
      assignTruthValue(proposition, TRUE_WRAPPER);
      if (subtypeOfP(safePrimaryType(lastargument), SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
        { Object* lastargument000 = lastargument;
          Skolem* lastargument = ((Skolem*)(lastargument000));

          if (((boolean)(lastargument->skolemType))) {
            { 
              BIND_STELLA_SPECIAL(oEVALUATIONMODEo, Keyword*, KWD_PROPOSITIONS_EXTENSIONAL_ASSERTION);
              BIND_STELLA_SPECIAL(oINVISIBLEASSERTIONpo, boolean, TRUE);
              assertIsaProposition(lastargument, lastargument->skolemType);
            }
          }
        }
      }
      else {
      }
    }
    return (resultP);
  }
}

Skolem* createEnumeratedSet(List* set) {
  { Proposition* setprop = beginCreatingFunctionProposition(SGT_PROPOSITIONS_PL_KERNEL_KB_SETOF, set->theConsList);
    Skolem* skolem = createSkolem(SGT_PROPOSITIONS_PL_KERNEL_KB_SET, NULL);

    (setprop->arguments->theArray)[(set->length())] = skolem;
    skolem->definingProposition = setprop;
    addDependentPropositionLink(skolem, setprop);
    setDynamicSlotValue(setprop->dynamicSlots, SYM_PROPOSITIONS_LOGIC_UNFASTENEDp, NULL, NULL);
    return (skolem);
  }
}

Skolem* createLogicalList(List* list) {
  { Proposition* listprop = beginCreatingFunctionProposition(SGT_PROPOSITIONS_PL_KERNEL_KB_LISTOF, list->theConsList);
    Skolem* skolem = createSkolem(SGT_PROPOSITIONS_PL_KERNEL_KB_LIST, NULL);

    (listprop->arguments->theArray)[(list->length())] = skolem;
    skolem->definingProposition = listprop;
    addDependentPropositionLink(skolem, listprop);
    setDynamicSlotValue(listprop->dynamicSlots, SYM_PROPOSITIONS_LOGIC_UNFASTENEDp, NULL, NULL);
    return (skolem);
  }
}

boolean logicalCollectionP(Object* self) {
  if (subtypeOfP(safePrimaryType(self), SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
    { Object* self000 = self;
      Skolem* self = ((Skolem*)(self000));

      { Proposition* definingproposition = self->definingProposition;

        return (((boolean)(definingproposition)) &&
            collectionofPropositionP(definingproposition));
      }
    }
  }
  else {
    return (FALSE);
  }
}

boolean skolemDefinedByOperatorP(Object* self, Surrogate* operatoR) {
  if (subtypeOfP(safePrimaryType(self), SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
    { Object* self000 = self;
      Skolem* self = ((Skolem*)(self000));

      { Proposition* definingproposition = self->definingProposition;

        return (((boolean)(definingproposition)) &&
            (definingproposition->operatoR == operatoR));
      }
    }
  }
  else {
    return (FALSE);
  }
}

boolean enumeratedSetP(Object* self) {
  return (skolemDefinedByOperatorP(self, SGT_PROPOSITIONS_PL_KERNEL_KB_SETOF));
}

boolean enumeratedListP(Object* self) {
  return (skolemDefinedByOperatorP(self, SGT_PROPOSITIONS_PL_KERNEL_KB_LISTOF));
}

Object* removeTrueAndFalse(Cons* tree, boolean& _Return1) {
  { Symbol* t = SYM_PROPOSITIONS_STELLA_TRUE;
    Symbol* f = SYM_PROPOSITIONS_STELLA_FALSE;

    { boolean testValue000 = FALSE;

      { boolean foundP000 = FALSE;

        { Object* arg = NULL;
          Cons* iter000 = tree->rest;

          while (!(iter000 == NIL)) {
            arg = iter000->value;
            iter000 = iter000->rest;
            if ((arg == t) ||
                (arg == f)) {
              foundP000 = TRUE;
              break;
            }
          }
        }
        testValue000 = foundP000;
      }
      testValue000 = !testValue000;
      if (testValue000) {
        _Return1 = FALSE;
        return (tree);
      }
    }
    { Symbol* testValue001 = ((Symbol*)(tree->value));

      if (testValue001 == SYM_PROPOSITIONS_STELLA_AND) {
        tree->rest = tree->rest->remove(t);
        { boolean foundP001 = FALSE;

          { Object* arg = NULL;
            Cons* iter001 = tree->rest;

            while (!(iter001 == NIL)) {
              arg = iter001->value;
              iter001 = iter001->rest;
              if (arg == f) {
                foundP001 = TRUE;
                break;
              }
            }
          }
          if (foundP001) {
            _Return1 = TRUE;
            return (SYM_PROPOSITIONS_STELLA_FALSE);
          }
        }
      }
      else if (testValue001 == SYM_PROPOSITIONS_STELLA_OR) {
        tree->rest = tree->rest->remove(f);
        { boolean foundP002 = FALSE;

          { Object* arg = NULL;
            Cons* iter002 = tree->rest;

            while (!(iter002 == NIL)) {
              arg = iter002->value;
              iter002 = iter002->rest;
              if (arg == t) {
                foundP002 = TRUE;
                break;
              }
            }
          }
          if (foundP002) {
            _Return1 = TRUE;
            return (SYM_PROPOSITIONS_STELLA_TRUE);
          }
        }
      }
      else if (testValue001 == SYM_PROPOSITIONS_STELLA_NOT) {
        { Symbol* testValue002 = ((Symbol*)(tree->rest->value));

          if (testValue002 == SYM_PROPOSITIONS_STELLA_TRUE) {
            _Return1 = TRUE;
            return (SYM_PROPOSITIONS_STELLA_FALSE);
          }
          else if (testValue002 == SYM_PROPOSITIONS_STELLA_FALSE) {
            _Return1 = FALSE;
            return (SYM_PROPOSITIONS_STELLA_TRUE);
          }
          else {
            { OutputStringStream* stream000 = newOutputStringStream();

              *(stream000->nativeStream) << "`" << testValue002 << "'" << " is not a valid case option";
              throw *newStellaException(stream000->theStringReader());
            }
          }
        }
      }
      else {
        { OutputStringStream* stream001 = newOutputStringStream();

          *(stream001->nativeStream) << "`" << testValue001 << "'" << " is not a valid case option";
          throw *newStellaException(stream001->theStringReader());
        }
      }
    }
    _Return1 = TRUE;
    return (canonicalizePropositionTree(tree));
  }
}

Object* flattenNestedOperators(Cons* tree, boolean& _Return1) {
  { boolean testValue000 = FALSE;

    { boolean foundP000 = FALSE;

      { Object* arg = NULL;
        Cons* iter000 = tree->rest;

        while (!(iter000 == NIL)) {
          arg = iter000->value;
          iter000 = iter000->rest;
          if (consP(arg) &&
              eqlP(((Cons*)(arg))->value, tree->value)) {
            foundP000 = TRUE;
            break;
          }
        }
      }
      testValue000 = foundP000;
    }
    testValue000 = !testValue000;
    if (testValue000) {
      _Return1 = FALSE;
      return (tree);
    }
  }
  { Symbol* targetoperator = ((Symbol*)(tree->value));
    Cons* tail = tree->rest;

    for (;;) {
      if (consP(tail->value) &&
          (((Cons*)(tail->value))->value == targetoperator)) {
        { Cons* nestedargs = ((Cons*)(tail->value))->rest;

          if (nestedargs == NIL) {
            if (targetoperator == SYM_PROPOSITIONS_STELLA_AND) {
              tail->value = SYM_PROPOSITIONS_STELLA_TRUE;
            }
            else if (targetoperator == SYM_PROPOSITIONS_STELLA_OR) {
              tail->value = SYM_PROPOSITIONS_LOGIC_FALS;
            }
            else {
              { OutputStringStream* stream000 = newOutputStringStream();

                *(stream000->nativeStream) << "`" << targetoperator << "'" << " is not a valid case option";
                throw *newStellaException(stream000->theStringReader());
              }
            }
          }
          else if (nestedargs->rest == NIL) {
            tail->value = nestedargs->value;
          }
          else {
            tail->value = nestedargs->value;
            lastCons(nestedargs)->rest = tail->rest;
            tail->rest = nestedargs->rest;
          }
        }
      }
      tail = tail->rest;
      if (tail == NIL) {
        break;
      }
    }
    _Return1 = TRUE;
    return (canonicalizePropositionTree(tree));
  }
}

Object* canonicalizeAndOrTree(Cons* tree) {
  { Symbol* operatoR = ((Symbol*)(tree->value));
    int argcount = tree->rest->length();
    boolean canonicalP = FALSE;
    Object* resulttree = NULL;
    Object* arg = NULL;

    resulttree = removeTrueAndFalse(tree, canonicalP);
    if (canonicalP) {
      return (resulttree);
    }
    else {
      tree = ((Cons*)(resulttree));
    }
    if (argcount > 1) {
      resulttree = flattenNestedOperators(tree, canonicalP);
      return (resulttree);
    }
    else if (argcount == 1) {
      arg = tree->rest->value;
      return (canonicalizePropositionTree(arg));
    }
    else {
      if (operatoR == SYM_PROPOSITIONS_STELLA_AND) {
        return (SYM_PROPOSITIONS_STELLA_TRUE);
      }
      else if (operatoR == SYM_PROPOSITIONS_STELLA_OR) {
        return (SYM_PROPOSITIONS_STELLA_FALSE);
      }
      else {
        { OutputStringStream* stream000 = newOutputStringStream();

          *(stream000->nativeStream) << "`" << operatoR << "'" << " is not a valid case option";
          throw *newStellaException(stream000->theStringReader());
        }
      }
    }
  }
}

Object* canonicalizeNegationTree(Cons* tree) {
  { Cons* argument = ((Cons*)(tree->rest->value));
    Symbol* nestedoperator = ((Symbol*)(argument->value));

    if (nestedoperator == SYM_PROPOSITIONS_STELLA_NOT) {
      return (canonicalizePropositionTree(argument->rest->value));
    }
    else if ((nestedoperator == SYM_PROPOSITIONS_STELLA_AND) ||
        (nestedoperator == SYM_PROPOSITIONS_STELLA_OR)) {
      { ConsIterator* it = argument->rest->allocateIterator();

        while (it->nextP()) {
          it->valueSetter(listO(3, SYM_PROPOSITIONS_STELLA_NOT, it->value, NIL));
        }
      }
      if (nestedoperator == SYM_PROPOSITIONS_STELLA_AND) {
        argument->firstSetter(SYM_PROPOSITIONS_STELLA_OR);
      }
      else if (nestedoperator == SYM_PROPOSITIONS_STELLA_OR) {
        argument->firstSetter(SYM_PROPOSITIONS_STELLA_AND);
      }
      else {
        { OutputStringStream* stream000 = newOutputStringStream();

          *(stream000->nativeStream) << "`" << nestedoperator << "'" << " is not a valid case option";
          throw *newStellaException(stream000->theStringReader());
        }
      }
    }
    else if (nestedoperator == SYM_PROPOSITIONS_STELLA_EXISTS) {
      { Object* whereclause = argument->rest->rest->value;

        return (canonicalizePropositionTree(listO(3, SYM_PROPOSITIONS_STELLA_FORALL, argument->rest->value, listO(3, SYM_PROPOSITIONS_STELLA_TRUE, listO(3, SYM_PROPOSITIONS_STELLA_NOT, whereclause, NIL), NIL))));
      }
    }
    else if (nestedoperator == SYM_PROPOSITIONS_STELLA_FORALL) {
      { Object* whereclause = argument->rest->rest->value;
        Object* alwaysclause = argument->fourth();

        return (canonicalizePropositionTree(listO(3, SYM_PROPOSITIONS_STELLA_EXISTS, argument->rest->value, cons(listO(3, SYM_PROPOSITIONS_STELLA_AND, whereclause, cons(listO(3, SYM_PROPOSITIONS_STELLA_NOT, alwaysclause, NIL), NIL)), NIL))));
      }
    }
    else {
      return (tree);
    }
    return (canonicalizePropositionTree(argument));
  }
}

boolean quantifierSymbolP(Symbol* self) {
  return ((self == SYM_PROPOSITIONS_STELLA_EXISTS) ||
      ((self == SYM_PROPOSITIONS_STELLA_FORALL) ||
       ((self == SYM_PROPOSITIONS_LOGIC_SETOFALL) ||
        ((self == SYM_PROPOSITIONS_LOGIC_THE_ONLY) ||
         ((self == SYM_PROPOSITIONS_LOGIC_KAPPA) ||
          (self == SYM_PROPOSITIONS_STELLA_SOME))))));
}

boolean allVariablesGoalP(Object* tree) {
  if (!consP(tree)) {
    return (FALSE);
  }
  { Cons* goal = ((Cons*)(tree));
    Object* operatoR = goal->value;

    { boolean testValue000 = FALSE;

      testValue000 = symbolP(operatoR);
      if (testValue000) {
        testValue000 = !oKIF_SENTENCE_OPERATORSo->memberP(operatoR);
        if (testValue000) {
          { boolean alwaysP000 = TRUE;

            { Object* arg = NULL;
              Cons* iter000 = goal->rest;

              while (!(iter000 == NIL)) {
                arg = iter000->value;
                iter000 = iter000->rest;
                if (!questionMarkSymbolP(arg)) {
                  alwaysP000 = FALSE;
                  break;
                }
              }
            }
            testValue000 = alwaysP000;
          }
        }
      }
      { boolean value000 = testValue000;

        return (value000);
      }
    }
  }
}

boolean equalArgumentsP(Cons* goal1, Cons* goal2) {
  { boolean testValue000 = FALSE;

    { boolean alwaysP000 = TRUE;

      { Object* arg1 = NULL;
        Object* arg2 = NULL;
        Cons* iter000 = goal1->rest;
        Cons* iter001 = goal2->rest;

        while ((!(iter000 == NIL)) &&
            (!(iter001 == NIL))) {
          arg1 = iter000->value;
          iter000 = iter000->rest;
          arg2 = iter001->value;
          iter001 = iter001->rest;
          if (!eqlP(arg1, arg2)) {
            alwaysP000 = FALSE;
            break;
          }
        }
      }
      testValue000 = alwaysP000;
    }
    if (testValue000) {
      testValue000 = goal1->length() == goal2->length();
    }
    { boolean value000 = testValue000;

      return (value000);
    }
  }
}

Cons* explodeForallTree(Cons* tree) {
  { Object* antecedent = tree->rest->rest->value;
    Object* consequent = tree->fourth();

    if ((!allVariablesGoalP(antecedent)) ||
        ((!consP(consequent)) ||
         (!(((Cons*)(consequent))->value == SYM_PROPOSITIONS_STELLA_AND)))) {
      return (tree);
    }
    { Cons* antecedentgoal = ((Cons*)(antecedent));
      Cons* conjuncts = ((Cons*)(consequent))->rest;

      { Cons* value000 = NIL;

        { Object* goal = NULL;
          Cons* iter000 = conjuncts;
          Cons* collect000 = NULL;

          while (!(iter000 == NIL)) {
            goal = iter000->value;
            iter000 = iter000->rest;
            if (allVariablesGoalP(goal) &&
                equalArgumentsP(((Cons*)(goal)), antecedentgoal)) {
              if (!((boolean)(collect000))) {
                {
                  collect000 = cons(goal, NIL);
                  if (value000 == NIL) {
                    value000 = collect000;
                  }
                  else {
                    addConsToEndOfConsList(value000, collect000);
                  }
                }
              }
              else {
                {
                  collect000->rest = cons(goal, NIL);
                  collect000 = collect000->rest;
                }
              }
            }
          }
        }
        { Cons* matchinggoals = value000;
          Cons* othergoals = conjuncts->difference(matchinggoals);
          Cons* rules = NIL;

          if (matchinggoals == NIL) {
            return (tree);
          }
          { Object* g = NULL;
            Cons* iter001 = matchinggoals;

            while (!(iter001 == NIL)) {
              g = iter001->value;
              iter001 = iter001->rest;
              rules = cons(((Cons*)(copyConsTree(listO(3, SYM_PROPOSITIONS_STELLA_FORALL, tree->rest->value, cons(antecedentgoal, cons(g, NIL)))))), rules);
            }
          }
          if (!(othergoals == NIL)) {
            rules = cons(listO(3, SYM_PROPOSITIONS_STELLA_FORALL, tree->rest->value, cons(antecedentgoal, cons(cons(SYM_PROPOSITIONS_STELLA_AND, othergoals->concatenate(NIL, 0)), NIL))), rules);
          }
          return (((Cons*)(canonicalizePropositionTree(cons(SYM_PROPOSITIONS_STELLA_AND, rules->concatenate(NIL, 0))))));
        }
      }
    }
  }
}

Cons* canonicalizeQuantificationTree(Cons* tree) {
  { Object* whereclause = tree->rest->rest->value;
    Object* alwaysclause = tree->fourth();

    if (((boolean)(whereclause))) {
      tree->thirdSetter(canonicalizePropositionTree(whereclause));
    }
    else {
      tree->concatenate(cons(SYM_PROPOSITIONS_STELLA_TRUE, NIL), 0);
    }
    if (((boolean)(alwaysclause))) {
      tree->fourthSetter(canonicalizePropositionTree(alwaysclause));
    }
    if (((GeneralizedSymbol*)(tree->value)) == SYM_PROPOSITIONS_STELLA_FORALL) {
      return (explodeForallTree(tree));
    }
    else {
      return (tree);
    }
  }
}

Object* consDcanonicalizePropositionTree(Cons* tree) {
  { Symbol* operatoR = ((Symbol*)(tree->value));

    if (quantifierSymbolP(operatoR)) {
      return (canonicalizeQuantificationTree(tree));
    }
    { ConsIterator* it = tree->rest->allocateIterator();

      while (it->nextP()) {
        it->valueSetter(canonicalizePropositionTree(it->value));
      }
    }
    if ((operatoR == SYM_PROPOSITIONS_STELLA_AND) ||
        (operatoR == SYM_PROPOSITIONS_STELLA_OR)) {
      return (canonicalizeAndOrTree(tree));
    }
    else if (operatoR == SYM_PROPOSITIONS_STELLA_NOT) {
      { Object* arg = tree->rest->value;

        if (consP(arg)) {
          return (canonicalizeNegationTree(tree));
        }
        if (arg == SGT_PROPOSITIONS_STELLA_TRUE) {
          return (SGT_PROPOSITIONS_STELLA_FALSE);
        }
        else if (arg == SGT_PROPOSITIONS_STELLA_FALSE) {
          return (SGT_PROPOSITIONS_STELLA_TRUE);
        }
        else {
          return (tree);
        }
      }
    }
    else {
    }
    return (tree);
  }
}

Object* canonicalizePropositionTree(Object* tree) {
  if (safePrimaryType(tree) == SGT_PROPOSITIONS_STELLA_CONS) {
    { Object* tree000 = tree;
      Cons* tree = ((Cons*)(tree000));

      return (consDcanonicalizePropositionTree(tree));
    }
  }
  else {
    return (tree);
  }
}

void eraseProposition(Proposition* proposition) {
  { Object* arg = NULL;
    Vector* vector000 = proposition->arguments;
    int index000 = 0;
    int length000 = vector000->length();

    while (index000 < length000) {
      arg = (vector000->theArray)[index000];
      index000 = index000 + 1;
      if (((boolean)(arg))) {
        removeDependentPropositionLink(arg, proposition);
      }
    }
  }
  proposition->arguments->free();
  proposition->arguments = NULL;
  proposition->deletedPSetter(TRUE);
}

Proposition* updateEquivalenceProposition(Proposition* proposition, Keyword* updatemode) {
  { Object* term1 = (proposition->arguments->theArray)[0];
    Object* term2 = (proposition->arguments->theArray)[1];

    if (skolemP(term1) &&
        skolemP(term2)) {
      if (updatemode == KWD_PROPOSITIONS_PRESUME_TRUE) {
        { OutputStringStream* stream000 = newOutputStringStream();

          { 
            BIND_STELLA_SPECIAL(oPRINTREADABLYpo, boolean, TRUE);
            *(stream000->nativeStream) << "ERROR: " << "Equating two skolems by default is not yet supported." << std::endl << "." << std::endl;
            helpSignalPropositionError(stream000, KWD_PROPOSITIONS_ERROR);
          }
          throw *newPropositionError(stream000->theStringReader());
        }
      }
      return (proposition);
    }
    if ((updatemode == KWD_PROPOSITIONS_ASSERT_TRUE) ||
        (updatemode == KWD_PROPOSITIONS_PRESUME_TRUE)) {
      updatePropositionTruthValue(proposition, updatemode);
      if (clipValueP(term1, term2)) {
      }
      else if (clipValueP(term2, term1)) {
      }
      else {
        equateValues(valueOf(term1), valueOf(term2));
      }
    }
    else if (updatemode == KWD_PROPOSITIONS_RETRACT_TRUE) {
      if (unequateValuesP(term1, term2)) {
      }
      else if (unequateValuesP(term2, term1)) {
      }
    }
    else if (updatemode == KWD_PROPOSITIONS_CONCEIVE) {
      return (proposition);
    }
    else {
      { OutputStringStream* stream001 = newOutputStringStream();

        *(stream001->nativeStream) << "`" << updatemode << "'" << " is not a valid case option";
        throw *newStellaException(stream001->theStringReader());
      }
    }
    if (functionOutputSkolemP(term1) &&
        (!functionOutputSkolemP(term2))) {
      return (((Skolem*)(term1))->definingProposition);
    }
    if (functionOutputSkolemP(term2) &&
        (!functionOutputSkolemP(term1))) {
      return (((Skolem*)(term2))->definingProposition);
    }
    return (NULL);
  }
}

Proposition* skolemizeExistsProposition(Proposition* existsproposition) {
  { KeyValueList* mapping = newKeyValueList();
    Proposition* skolemizedproposition = NULL;

    { PatternVariable* var = NULL;
      Vector* vector000 = ((Vector*)(dynamicSlotValue(existsproposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_IO_VARIABLES, NULL)));
      int index000 = 0;
      int length000 = vector000->length();

      while (index000 < length000) {
        var = ((PatternVariable*)((vector000->theArray)[index000]));
        index000 = index000 + 1;
        createSkolemForUnmappedVariable(var, mapping);
      }
    }
    skolemizedproposition = recursivelyFastenDownPropositions(inheritProposition(((Proposition*)((existsproposition->arguments->theArray)[0])), mapping), FALSE);
    eraseProposition(existsproposition);
    return (skolemizedproposition);
  }
}

Cons* helpUpdateTopLevelProposition(Proposition* proposition, Keyword* updatemode) {
  if (!((boolean)(proposition))) {
    return (NULL);
  }
  if (lookupAnnotation(proposition, KWD_PROPOSITIONS_CONFIDENCE_LEVEL) == KWD_PROPOSITIONS_DEFAULT) {
    updatemode = defaultifyUpdateMode(updatemode);
    if ((updatemode == KWD_PROPOSITIONS_PRESUME_TRUE) ||
        (updatemode == KWD_PROPOSITIONS_PRESUME_FALSE)) {
      removeAnnotation(proposition, KWD_PROPOSITIONS_CONFIDENCE_LEVEL);
    }
  }
  if (nativeSlotPropositionP(proposition)) {
    return (updateNativeSlotProposition(proposition, updatemode));
  }
  { 
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, getTopLevelUpdateContext());
    { Keyword* testValue000 = proposition->kind;

      if (testValue000 == KWD_PROPOSITIONS_AND) {
        { Cons* propositions = NIL;

          { Object* arg = NULL;
            Vector* vector000 = proposition->arguments;
            int index000 = 0;
            int length000 = vector000->length();

            while (index000 < length000) {
              arg = (vector000->theArray)[index000];
              index000 = index000 + 1;
              { Proposition* p = NULL;
                Cons* iter000 = helpUpdateTopLevelProposition(((Proposition*)(arg)), updatemode);
                Cons* collect000 = NULL;

                while (!(iter000 == NIL)) {
                  p = ((Proposition*)(iter000->value));
                  iter000 = iter000->rest;
                  if (!((boolean)(collect000))) {
                    {
                      collect000 = cons(p, NIL);
                      if (propositions == NIL) {
                        propositions = collect000;
                      }
                      else {
                        addConsToEndOfConsList(propositions, collect000);
                      }
                    }
                  }
                  else {
                    {
                      collect000->rest = cons(p, NIL);
                      collect000 = collect000->rest;
                    }
                  }
                }
              }
            }
          }
          eraseProposition(proposition);
          return (propositions);
        }
      }
      else if (testValue000 == KWD_PROPOSITIONS_EQUIVALENT) {
        proposition = updateEquivalenceProposition(proposition, updatemode);
        if (!((boolean)(proposition))) {
          return (NULL);
        }
      }
      else if (testValue000 == KWD_PROPOSITIONS_NOT) {
        { Object* argument = (proposition->arguments->theArray)[0];

          if ((updatemode == KWD_PROPOSITIONS_ASSERT_TRUE) ||
              ((updatemode == KWD_PROPOSITIONS_ASSERT_FALSE) ||
               ((updatemode == KWD_PROPOSITIONS_PRESUME_TRUE) ||
                ((updatemode == KWD_PROPOSITIONS_PRESUME_FALSE) ||
                 ((updatemode == KWD_PROPOSITIONS_RETRACT_TRUE) ||
                  (updatemode == KWD_PROPOSITIONS_RETRACT_FALSE)))))) {
            updatePropositionTruthValue(((Proposition*)(argument)), invertUpdateMode(updatemode));
            eraseProposition(proposition);
            return (consList(1, argument));
          }
          else {
          }
        }
      }
      else if (testValue000 == KWD_PROPOSITIONS_EXISTS) {
        return (helpUpdateTopLevelProposition(skolemizeExistsProposition(proposition), updatemode));
      }
      else {
      }
    }
    if (!(proposition->kind == KWD_PROPOSITIONS_FUNCTION)) {
      updatePropositionTruthValue(proposition, updatemode);
    }
    return (consList(1, proposition));
  }
}

Keyword* invertUpdateMode(Keyword* updatemode) {
  if (updatemode == KWD_PROPOSITIONS_ASSERT_TRUE) {
    return (KWD_PROPOSITIONS_ASSERT_FALSE);
  }
  else if (updatemode == KWD_PROPOSITIONS_PRESUME_TRUE) {
    return (KWD_PROPOSITIONS_PRESUME_FALSE);
  }
  else if (updatemode == KWD_PROPOSITIONS_RETRACT_TRUE) {
    return (KWD_PROPOSITIONS_RETRACT_FALSE);
  }
  else if (updatemode == KWD_PROPOSITIONS_ASSERT_FALSE) {
    return (KWD_PROPOSITIONS_ASSERT_TRUE);
  }
  else if (updatemode == KWD_PROPOSITIONS_PRESUME_FALSE) {
    return (KWD_PROPOSITIONS_PRESUME_TRUE);
  }
  else if (updatemode == KWD_PROPOSITIONS_RETRACT_FALSE) {
    return (KWD_PROPOSITIONS_RETRACT_TRUE);
  }
  else {
    { OutputStringStream* stream000 = newOutputStringStream();

      *(stream000->nativeStream) << "`" << updatemode << "'" << " is not a valid case option";
      throw *newStellaException(stream000->theStringReader());
    }
  }
}

Keyword* defaultifyUpdateMode(Keyword* updatemode) {
  if (updatemode == KWD_PROPOSITIONS_ASSERT_TRUE) {
    return (KWD_PROPOSITIONS_PRESUME_TRUE);
  }
  else if (updatemode == KWD_PROPOSITIONS_ASSERT_FALSE) {
    return (KWD_PROPOSITIONS_PRESUME_FALSE);
  }
  else {
    return (updatemode);
  }
}

Cons* updateTopLevelProposition(Object* tree, Keyword* updatemode) {
  if ((updatemode == KWD_PROPOSITIONS_ASSERT_FALSE) ||
      ((updatemode == KWD_PROPOSITIONS_PRESUME_FALSE) ||
       (updatemode == KWD_PROPOSITIONS_RETRACT_FALSE))) {
    tree = listO(3, SYM_PROPOSITIONS_STELLA_NOT, tree, NIL);
    updatemode = invertUpdateMode(updatemode);
  }
  else {
  }
  if (updatemode == KWD_PROPOSITIONS_ASSERT_TRUE) {
    { 
      BIND_STELLA_SPECIAL(oAUTOMATICINSTANCETABLEo, KeyValueList*, NULL);
      return (helpUpdateTopLevelProposition(buildTopLevelProposition(tree, TRUE), updatemode));
    }
  }
  else {
    return (helpUpdateTopLevelProposition(buildTopLevelProposition(tree, FALSE), updatemode));
  }
}

Object* standardizePropositionTree(Object* tree) {
  if (stringP(tree)) {
    tree = readSExpressionFromString(((StringWrapper*)(tree))->wrapperValue);
  }
  tree = standardizeLogicalParseTree(tree);
  if (consP(tree)) {
    tree = canonicalizePropositionTree(tree);
  }
  return (tree);
}

Cons* updateProposition(Object* tree, Keyword* updatemode) {
  try {
    { 
      BIND_STELLA_SPECIAL(oINHIBITOBJECTFINALIZATIONpo, boolean, TRUE);
      return (updateTopLevelProposition(standardizePropositionTree(tree), updatemode));
    }
  }
  catch (LogicException& _e) {
    LogicException* e = &_e;

    *(STANDARD_ERROR->nativeStream) << exceptionMessage(e);
  }
  catch (ReadException& _e) {
    ReadException* e = &_e;

    *(STANDARD_ERROR->nativeStream) << exceptionMessage(e);
  }
  return (NULL);
}

Object* buildTopLevelTerm(Object* tree) {
  { Object* term = NULL;

    { 
      BIND_STELLA_SPECIAL(oLOGICVARIABLETABLEo, Cons*, NIL);
      BIND_STELLA_SPECIAL(oVARIABLEIDCOUNTERo, int, 0);
      BIND_STELLA_SPECIAL(oTERMUNDERCONSTRUCTIONo, Object*, tree);
      if (safePrimaryType(tree) == SGT_PROPOSITIONS_STELLA_CONS) {
        { Object* tree000 = tree;
          Cons* tree = ((Cons*)(tree000));

          if (getQuotedTree("((FORALL EXISTS IMPLIES AND OR NOT) \"/LOGIC\")", "/LOGIC")->memberP(tree->value)) {
            *(STANDARD_WARNING->nativeStream) << "Warning: " << "Sentence where term expected: " << "`" << tree << "'" << std::endl << std::endl;
            return (NULL);
          }
        }
      }
      else {
      }
      term = evaluateTerm(tree);
    }
    return (term);
  }
}

Object* conceiveTerm(Object* tree) {
  // `tree' is a term expression (a string or an s-expression),
  // or is a class reference (a symbol or surrogate).  Return a (possibly 
  // newly-conceived) term representing the internalized representation of that term.
  try {
    { Object* standardizedtree = standardizePropositionTree(tree);
      Object* operatoR = NULL;
      Object* term = NULL;

      if (consP(standardizedtree)) {
        operatoR = ((Cons*)(standardizedtree))->value;
      }
      if (((boolean)(operatoR)) &&
          (((boolean)(getRelation(operatoR))) &&
           (!functionP(getRelation(operatoR))))) {
        term = conceiveSentence(tree);
      }
      else {
        term = buildTopLevelTerm(standardizedtree);
      }
      return (term);
    }
  }
  catch (LogicException& _e) {
    LogicException* e = &_e;

    *(STANDARD_ERROR->nativeStream) << exceptionMessage(e);
  }
  catch (ReadException& _e) {
    ReadException* e = &_e;

    *(STANDARD_ERROR->nativeStream) << exceptionMessage(e);
  }
  return (NULL);
}

Object* conceiveSentence(Object* tree) {
  if (subtypeOfStringP(safePrimaryType(tree))) {
    { Object* tree000 = tree;
      StringWrapper* tree = ((StringWrapper*)(tree000));

      return (conceiveSentence(readSExpressionFromString(tree->wrapperValue)));
    }
  }
  else {
    return (smartUpdateProposition(tree, KWD_PROPOSITIONS_CONCEIVE));
  }
}

Object* conceiveFormula(Object* tree) {
  { Surrogate* testValue000 = safePrimaryType(tree);

    if (subtypeOfStringP(testValue000)) {
      { Object* tree000 = tree;
        StringWrapper* tree = ((StringWrapper*)(tree000));

        return (conceiveFormula(readSExpressionFromString(tree->wrapperValue)));
      }
    }
    else if (testValue000 == SGT_PROPOSITIONS_STELLA_CONS) {
      { Object* tree001 = tree;
        Cons* tree = ((Cons*)(tree001));

        if (symbolP(tree->value)) {
          { GeneralizedSymbol* testValue001 = ((GeneralizedSymbol*)(tree->value));

            if ((testValue001 == SYM_PROPOSITIONS_LOGIC_KAPPA) ||
                ((testValue001 == SYM_PROPOSITIONS_LOGIC_LAMBDA) ||
                 ((testValue001 == SYM_PROPOSITIONS_LOGIC_SETOFALL) ||
                  ((testValue001 == SYM_PROPOSITIONS_PL_KERNEL_KB_SETOF) ||
                   ((testValue001 == SYM_PROPOSITIONS_PL_KERNEL_KB_LISTOF) ||
                    (testValue001 == SYM_PROPOSITIONS_LOGIC_BAGOF)))))) {
              return (conceiveTerm(tree));
            }
            else {
            }
          }
        }
        return (conceiveSentence(tree));
      }
    }
    else if (subtypeOfSymbolP(testValue000)) {
      { Object* tree002 = tree;
        Symbol* tree = ((Symbol*)(tree002));

        return (conceiveTerm(tree));
      }
    }
    else if (subtypeOfSurrogateP(testValue000)) {
      { Object* tree003 = tree;
        Surrogate* tree = ((Surrogate*)(tree003));

        return (conceiveTerm(surrogateToSymbol(tree)));
      }
    }
    else {
      return (conceiveSentence(tree));
    }
  }
}

Object* smartUpdateProposition(Object* tree, Keyword* updatemode) {
  { Cons* propositions = updateProposition(tree, updatemode);

    if (((boolean)(propositions)) &&
        (propositions->rest == NIL)) {
      return (((Proposition*)(propositions->value)));
    }
    else {
      return (propositions);
    }
  }
}

Object* asserT(Object* proposition) {
  // Assert the truth of `proposition'.  Return the asserted proposition object.
  // KIF example:  "(assert (happy Fred))" asserts that Fred is indeed happy.
  // Note that for this assertion to succeed, the relation `happy' must already
  // be defined.  If the constant `Fred' has not yet been created, it is
  // automatically created as a side-effect of calling `assert'.
  return (smartUpdateProposition(proposition, KWD_PROPOSITIONS_ASSERT_TRUE));
}

Object* presume(Object* proposition) {
  // Presume the default truth of `proposition'.  Return the presumed
  // proposition object.  KIF example:  "(presume (happy Fred))" states that Fred
  // is most probably happy.  Note, that for this to succeed, the relation
  // `happy' must already be defined (see `assert').
  return (smartUpdateProposition(proposition, KWD_PROPOSITIONS_PRESUME_TRUE));
}

Object* retract(Object* proposition) {
  // Retract the truth of `proposition'.  Return the retracted proposition
  // object.  KIF example:  "(retract (happy Fred))" retracts that Fred is
  // happy.  Note that for this assertion to succeed, the relation `happy' must already
  // be defined.  If the constant `Fred' has not yet been created, it is
  // automatically created as a side-effect of calling `retract'.
  return (smartUpdateProposition(proposition, KWD_PROPOSITIONS_RETRACT_TRUE));
}

Object* deny(Object* proposition) {
  // Assert the falsity of `proposition'.  Return the asserted proposition
  // object.  KIF example:  "(deny (happy Fred))" asserts that Fred is not happy,
  // which could have been done equivalently by "(assert (not (happy Fred)))".
  // Note, that for this to succeed, the relation `happy' must already be defined
  // (see `assert').
  return (smartUpdateProposition(proposition, KWD_PROPOSITIONS_ASSERT_FALSE));
}

Object* conceive(Object* formula) {
  // Guess whether 'formula' represents a term or a sentence/proposition.
  // If we are not sure, assume its a proposition.
  // If its, a term, return its internal representation.  If a proposition,
  // construct a proposition for 'formula' without asserting its truth value.
  // Return the conceived proposition object.  KIF example: "(conceive (happy Fred))"
  // builds the proposition expressing that Fred is happy without explictly asserting
  // or denying it.  Note, that for this to succeed, the relation `happy' must already
  // be defined (see `assert').  If the logic constant `Fred' has not yet been
  // created, it is automatically created as a side-effect of calling `conceive'.
  return (conceiveFormula(formula));
}

Object* unassert(Object* proposition) {
  // Retract the truth, falsity or inconsistency of `proposition'.  This is a
  // more general version of `retract' that also handles falsity.  For example, if
  // we assert the proposition "(not (sad Fred))", and then execute the statement
  // "(unassert (sad Fred))", the truth value of the proposition "(sad Fred)"
  // will be set to UNKNOWN.  If we had called `retract' in place of `unassert',
  // the proposition "(sad Fred)" would remain set to FALSE.   Note that for this
  // unassertion to succeed, the logic constant `Fred' and the relation `sad' must
  // already be defined.
  { Cons* parsedprops = updateProposition(proposition, KWD_PROPOSITIONS_CONCEIVE);
    Cons* result = NIL;

    if (((boolean)(parsedprops))) {
      { Proposition* prop = NULL;
        Cons* iter000 = parsedprops;
        Cons* collect000 = NULL;

        while (!(iter000 == NIL)) {
          prop = ((Proposition*)(iter000->value));
          iter000 = iter000->rest;
          if (prop->kind == KWD_PROPOSITIONS_NOT) {
            prop = ((Proposition*)((prop->arguments->theArray)[0]));
          }
          unassertProposition(prop);
          if (!((boolean)(collect000))) {
            {
              collect000 = cons(prop, NIL);
              if (result == NIL) {
                result = collect000;
              }
              else {
                addConsToEndOfConsList(result, collect000);
              }
            }
          }
          else {
            {
              collect000->rest = cons(prop, NIL);
              collect000 = collect000->rest;
            }
          }
        }
      }
    }
    return (((result->length() > 1) ? result : result->value));
  }
}

void unassertProposition(Proposition* self) {
  // Retract the truth, falsity or inconsistency of the proposition `self'.
  if (((boolean)(self))) {
    { boolean equivalenceP = self->kind == KWD_PROPOSITIONS_EQUIVALENT;

      if (trueP(self)) {
        if (equivalenceP) {
          updateEquivalenceProposition(self, KWD_PROPOSITIONS_RETRACT_TRUE);
        }
        else {
          updatePropositionTruthValue(self, KWD_PROPOSITIONS_RETRACT_TRUE);
        }
      }
      else if (falseP(self)) {
        if (equivalenceP) {
          updateEquivalenceProposition(self, KWD_PROPOSITIONS_RETRACT_FALSE);
        }
        else {
          updatePropositionTruthValue(self, KWD_PROPOSITIONS_RETRACT_FALSE);
        }
      }
      else if (inconsistentP(self)) {
        if (equivalenceP) {
          deassignTruthValue(self, KWD_PROPOSITIONS_RETRACT_INCONSISTENT);
          updateEquivalenceProposition(self, KWD_PROPOSITIONS_RETRACT_TRUE);
          updateEquivalenceProposition(self, KWD_PROPOSITIONS_RETRACT_FALSE);
        }
        else {
          deassignTruthValue(self, KWD_PROPOSITIONS_RETRACT_INCONSISTENT);
          updatePropositionTruthValue(self, KWD_PROPOSITIONS_RETRACT_TRUE);
          updatePropositionTruthValue(self, KWD_PROPOSITIONS_RETRACT_FALSE);
        }
      }
    }
  }
}

boolean Proposition::deletedP() {
  { Proposition* self = this;

    return (deletedPropositionP(self));
  }
}

boolean Proposition::deletedPSetter(boolean value) {
  { Proposition* self = this;

    if (value) {
      self->kind = KWD_PROPOSITIONS_DELETED;
    }
    return (value);
  }
}

boolean deletedPropositionP(Proposition* self) {
  if (self->kind == KWD_PROPOSITIONS_DELETED) {
    return (TRUE);
  }
  if (((boolean)(self->homeContext)) &&
      self->homeContext->deletedP()) {
    self->homeContext = NULL;
    self->kind = KWD_PROPOSITIONS_DELETED;
    return (TRUE);
  }
  return (FALSE);
}

Proposition* destroyProposition(Proposition* proposition) {
  // Retract and destroy the proposition 'proposition'.
  // Recursively destroy all propositions that reference 'proposition'.
  // Also, destroy all satellite propositions of 'proposition'.
  if (proposition->deletedP()) {
    return (proposition);
  }
  proposition->deletedPSetter(TRUE);
  if (((boolean)(((Surrogate*)(dynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_STELLA_SURROGATE_VALUE_INVERSE, NULL)))))) {
    ((Surrogate*)(dynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_STELLA_SURROGATE_VALUE_INVERSE, NULL)))->surrogateValue = NULL;
  }
  { Object* arg = NULL;
    Vector* vector000 = proposition->arguments;
    int index000 = 0;
    int length000 = vector000->length();

    while (index000 < length000) {
      arg = (vector000->theArray)[index000];
      index000 = index000 + 1;
      if (((boolean)(arg)) &&
          (!arg->deletedP())) {
        if (skolemP(arg) &&
            (((Skolem*)(arg))->definingProposition == proposition)) {
          destroyTerm(((LogicObject*)(arg)));
        }
        removeDependentPropositionLink(arg, proposition);
      }
    }
  }
  { Object* dep = NULL;
    Iterator* iter000 = proposition->dependentPropositions->copy()->allocateIterator();

    while (iter000->nextP()) {
      dep = iter000->value;
      destroyProposition(((Proposition*)(dep)));
    }
  }
  { Proposition* satellite = NULL;
    Cons* iter001 = proposition->satellitePropositions_reader()->copy()->theConsList;

    while (!(iter001 == NIL)) {
      satellite = ((Proposition*)(iter001->value));
      iter001 = iter001->rest;
      destroyProposition(satellite);
    }
  }
  if (((boolean)(((Proposition*)(dynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_MASTER_PROPOSITION, NULL)))))) {
    ((Proposition*)(dynamicSlotValue(proposition->dynamicSlots, SYM_PROPOSITIONS_LOGIC_MASTER_PROPOSITION, NULL)))->satellitePropositions_reader()->remove(proposition);
  }
  if (!(!((boolean)(proposition->homeContext)))) {
    handleOutOfDateInferenceCache(KWD_PROPOSITIONS_RETRACT, proposition);
  }
  proposition->dependentPropositions = NULL;
  proposition->operatoR = NULL;
  proposition->arguments = NULL;
  proposition->homeContext = NULL;
  proposition->truthValue = NULL;
  proposition->dynamicSlots = NULL;
  return (proposition);
}

void destroyLogicInstance(Object* self) {
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* self000 = self;
        LogicObject* self = ((LogicObject*)(self000));

        if ((!((boolean)(self))) ||
            self->deletedP()) {
          return;
        }
        self->deletedPSetter(TRUE);
        { Surrogate* surrogate = self->surrogateValueInverse;

          if (((boolean)(surrogate)) &&
              (surrogate->surrogateValue == self)) {
            surrogate->surrogateValue = NULL;
          }
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_THING)) {
      { Object* self001 = self;
        Thing* self = ((Thing*)(self001));

        if ((!((boolean)(self))) ||
            self->deletedP()) {
          return;
        }
        self->deletedPSetter(TRUE);
        { Surrogate* surrogate = self->surrogateValueInverse;

          if (((boolean)(surrogate)) &&
              (surrogate->surrogateValue == self)) {
            surrogate->surrogateValue = NULL;
          }
        }
      }
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
        throw *newStellaException(stream000->theStringReader());
      }
    }
  }
  { Proposition* p = NULL;
    Iterator* iter000 = unfilteredDependentPropositions(self, NULL)->allocateIterator();

    while (iter000->nextP()) {
      p = ((Proposition*)(iter000->value));
      destroyProposition(p);
    }
  }
  if (subtypeOfP(safePrimaryType(self), SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
    { Object* self004 = self;
      LogicObject* self = ((LogicObject*)(self004));

      unlinkEquivalentValue(self);
      { LogicObject* equivalent = NULL;
        Cons* iter001 = self->variableValueInverse_reader();

        while (!(iter001 == NIL)) {
          equivalent = ((LogicObject*)(iter001->value));
          iter001 = iter001->rest;
          if (!equivalent->deletedP()) {
            { Surrogate* testValue001 = safePrimaryType(equivalent);

              if (subtypeOfP(testValue001, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
                { LogicObject* equivalent000 = equivalent;
                  Skolem* equivalent = ((Skolem*)(equivalent000));

                  { Skolem* object000 = equivalent;
                    Object* value000 = NULL;
                    Object* oldValue000 = object000->variableValue;
                    Object* newValue000 = updateInContext(oldValue000, value000, object000->homeContext, FALSE);

                    if (!(((boolean)(oldValue000)) &&
                        (oldValue000->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
                      object000->variableValue = newValue000;
                    }
                  }
                }
              }
              else if (subtypeOfP(testValue001, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
                { LogicObject* equivalent001 = equivalent;
                  LogicObject* equivalent = equivalent001;

                  { LogicObject* object001 = equivalent;
                    Description* value001 = NULL;
                    Object* oldValue001 = dynamicSlotValue(object001->dynamicSlots, SYM_PROPOSITIONS_LOGIC_EQUIVALENT_VALUE, NULL);
                    Object* newValue001 = updateInContext(oldValue001, value001, object001->homeContext, FALSE);

                    if (!(((boolean)(oldValue001)) &&
                        (oldValue001->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
                      setDynamicSlotValue(object001->dynamicSlots, SYM_PROPOSITIONS_LOGIC_EQUIVALENT_VALUE, newValue001, NULL);
                    }
                  }
                }
              }
              else {
                { OutputStringStream* stream001 = newOutputStringStream();

                  *(stream001->nativeStream) << "`" << testValue001 << "'" << " is not a valid case option";
                  throw *newStellaException(stream001->theStringReader());
                }
              }
            }
          }
        }
      }
      { LogicObject* object002 = self;
        Cons* value002 = NULL;
        Object* oldValue002 = object002->variableValueInverse;
        Object* newValue002 = updateInContext(oldValue002, value002, object002->homeContext, FALSE);

        if (!(((boolean)(oldValue002)) &&
            (oldValue002->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
          object002->variableValueInverse = newValue002;
        }
      }
    }
  }
  else {
  }
}

void destroyTerm(LogicObject* self) {
  // Destroy all propositions that reference 'self',
  // and mark it as 'deleted?', thereby making it invisible within class
  // extensions.  Unlink descriptions from native relations.
  if ((!((boolean)(self))) ||
      self->deletedP()) {
    return;
  }
  destroyLogicInstance(self);
  if (subtypeOfP(safePrimaryType(self), SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) {
    { LogicObject* self000 = self;
      Description* self = ((Description*)(self000));

      if (subtypeOfP(safePrimaryType(self), SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
        { Description* self001 = self;
          NamedDescription* self = ((NamedDescription*)(self001));

          { Relation* nativerelation = self->nativeRelation();

            self->nativeRelationSetter(NULL);
            if (((boolean)(nativerelation))) {
              setDynamicSlotValue(nativerelation->dynamicSlots, SYM_PROPOSITIONS_LOGIC_DESCRIPTION, NULL, NULL);
            }
            if (((boolean)(self->extension))) {
              self->extension = NULL;
            }
          }
        }
      }
      else {
      }
      { Description* complement = ((Description*)(dynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_COMPLEMENT_DESCRIPTION, NULL)));

        if (((boolean)(complement))) {
          setDynamicSlotValue(complement->dynamicSlots, SYM_PROPOSITIONS_LOGIC_COMPLEMENT_DESCRIPTION, NULL, NULL);
          setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_COMPLEMENT_DESCRIPTION, NULL, NULL);
        }
      }
      removeAllDependentPropositionLinks(self);
      self->ioVariables = NULL;
      self->internalVariables = NULL;
      self->proposition = NULL;
      self->homeContext = NULL;
    }
  }
  else {
  }
  self->dynamicSlots->clear();
  self->deletedPSetter(TRUE);
}

void destroyInstance(Object* self) {
  // Destroy all propositions that reference 'self',
  // and mark it as 'deleted?', thereby making it invisible within class
  // extensions.
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* self000 = self;
        LogicObject* self = ((LogicObject*)(self000));

        destroyTerm(self);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_THING)) {
      { Object* self001 = self;
        Thing* self = ((Thing*)(self001));

        destroyLogicInstance(self);
      }
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
        throw *newStellaException(stream000->theStringReader());
      }
    }
  }
}

void destroyObject(Object* self) {
  // Destroy `self' which can be a term or a proposition.  Destroy all
  // propositions that reference 'self' and mark it as 'deleted?' (thereby
  // making it invisible within class extensions).
  { Surrogate* testValue000 = safePrimaryType(self);

    if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* self000 = self;
        LogicObject* self = ((LogicObject*)(self000));

        destroyTerm(self);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_THING)) {
      { Object* self001 = self;
        Thing* self = ((Thing*)(self001));

        destroyLogicInstance(self);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
      { Object* self002 = self;
        Proposition* self = ((Proposition*)(self002));

        destroyProposition(self);
      }
    }
    else {
      { OutputStringStream* stream000 = newOutputStringStream();

        *(stream000->nativeStream) << "`" << testValue000 << "'" << " is not a valid case option";
        throw *newStellaException(stream000->theStringReader());
      }
    }
  }
}

Object* destroy(Object* objectspec) {
  // Find an object or proposition as specified by `objectSpec', and destroy all
  // propositions and indices that reference it.  `objectSpec' must be a name or
  // a parse tree that evaluates to a proposition.  Return the deleted object, or
  // NULL if no matching object was found.
  { Object* object = NULL;

    if (safePrimaryType(objectspec) == SGT_PROPOSITIONS_STELLA_CONS) {
      { Object* objectspec000 = objectspec;
        Cons* objectspec = ((Cons*)(objectspec000));

        object = conceiveFormula(objectspec);
      }
    }
    else {
      object = coerceToInstanceInModule(objectspec, NULL, FALSE, NULL);
    }
    if (((boolean)(object))) {
      destroyObject(object);
    }
    return (object);
  }
}

boolean relationInModuleP(NamedDescription* self) {
  return (self->homeContext == oMODULEo.get());
}

void introduceModule(Module* module) {
  module->surrogateValueInverse = NULL;
  { 
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oPL_KERNEL_MODULEo);
    BIND_STELLA_SPECIAL(oMODULEo, Module*, oCONTEXTo.get()->baseModule);
    if (!((boolean)(module->surrogateValueInverse))) {
      { boolean successP = FALSE;
        Object* unused = NULL;
        Surrogate* surrogate = NULL;

        successP = bindToSurrogateP(module, module->name(), TRUE, TRUE, unused, surrogate);
        unused = unused;
        if (successP) {
          module->surrogateValueInverse = surrogate;
        }
      }
    }
    assertIsaProposition(module, SGT_PROPOSITIONS_STELLA_MODULE);
  }
}

void clearLogicModuleHook(Module* module) {
  { 
    BIND_STELLA_SPECIAL(oMODULEo, Module*, module);
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
    destroyInferenceCaches(module);
    flushInlineQueryCaches();
    { Object* term = NULL;
      Cons* iter000 = allTerms(module, TRUE)->listify()->theConsList;

      while (!(iter000 == NIL)) {
        term = iter000->value;
        iter000 = iter000->rest;
        destroyInstance(term);
      }
    }
    { Proposition* proposition = NULL;
      Iterator* iter001 = allPropositions(module, TRUE);

      while (iter001->nextP()) {
        proposition = ((Proposition*)(iter001->value));
        destroyProposition(proposition);
      }
    }
    locallyConceivedPropositions(module)->removeDeletedMembers();
    cleanupAllDescriptionExtensions();
    cleanupUnfinalizedObjects();
    clearMemoizationTables(KWD_PROPOSITIONS_KB_UPDATE);
    clearMemoizationTables(KWD_PROPOSITIONS_META_KB_UPDATE);
    oSTRINGIFIED_PROPOSITION_INDEXo = newStringHashTable();
    oSTRINGIFIED_DESCRIPTION_INDEXo = newStringHashTable();
  }
}

void clearInstances(Cons* name) {
  // Destroy all instances belonging to module `name' or any of its children.
  // Leave meta-objects, e.g., concepts and relations, alone.
  // If no `name' is supplied, the current module will be cleared after
  // confirming with the user.
  { Cons* optionalname = name;
    Module* module = coerceToModule(optionalname->value, TRUE);

    if (((boolean)(module))) {
      if ((!(optionalname == NIL)) ||
          yOrNP(stringConcatenate("Really clear instances of module ", stringConcatenate(module->moduleFullName, "? ", 0), 0))) {
        doClearInstances(module);
      }
    }
  }
}

void clearInstancesEvaluatorWrapper(Cons* arguments) {
  clearInstances(arguments);
}

void doClearInstances(Module* module) {
  // Function version of `clear-instances' that evaluates its argument.
  { Context* child = NULL;
    Cons* iter000 = module->childContexts->theConsList;

    while (!(iter000 == NIL)) {
      child = ((Context*)(iter000->value));
      iter000 = iter000->rest;
      if (isaP(child, SGT_PROPOSITIONS_STELLA_MODULE)) {
        doClearInstances(((Module*)(child)));
      }
    }
  }
  { 
    BIND_STELLA_SPECIAL(oMODULEo, Module*, module);
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
    destroyInferenceCaches(module);
    { LogicObject* instance = NULL;
      Iterator* iter001 = allInstances(module, TRUE);

      while (iter001->nextP()) {
        instance = ((LogicObject*)(iter001->value));
        destroyInstance(instance);
      }
    }
    locallyConceivedPropositions(module)->removeDeletedMembers();
  }
}

void destroyLogicContextHook(Context* self) {
  setDynamicSlotValue(self->dynamicSlots, SYM_PROPOSITIONS_LOGIC_MONOTONICp, (FALSE ? TRUE_WRAPPER : FALSE_WRAPPER), FALSE_WRAPPER);
  { 
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, self);
    BIND_STELLA_SPECIAL(oMODULEo, Module*, oCONTEXTo.get()->baseModule);
    { LogicObject* i = NULL;
      Iterator* iter000 = locallyConceivedInstances(self)->allocateIterator();

      while (iter000->nextP()) {
        i = ((LogicObject*)(iter000->value));
        destroyTerm(i);
      }
    }
  }
  if (isaP(self, SGT_PROPOSITIONS_STELLA_MODULE)) {
    clearLogicModuleHook(((Module*)(self)));
    self->surrogateValueInverse->surrogateValue = NULL;
  }
  if (isaP(self, SGT_PROPOSITIONS_STELLA_WORLD)) {
    unlinkInferenceCache(((World*)(self)));
  }
}

boolean consLessThanP(Cons* o1, Cons* o2) {
  { Object* i1 = NULL;
    Object* i2 = NULL;
    Cons* iter000 = o1;
    Cons* iter001 = o2;

    while ((!(iter000 == NIL)) &&
        (!(iter001 == NIL))) {
      i1 = iter000->value;
      iter000 = iter000->rest;
      i2 = iter001->value;
      iter001 = iter001->rest;
      if (logicFormLessP(i1, i2)) {
        return (TRUE);
      }
      if (logicFormLessP(i2, i1)) {
        return (FALSE);
      }
    }
  }
  return (o1->length() < o2->length());
}

boolean safeStringLessP(char* s1, char* s2) {
  if ((s1 != NULL) &&
      (s2 != NULL)) {
    return (stringLessP(s1, s2));
  }
  else if (s2 != NULL) {
    return (FALSE);
  }
  else {
    return (TRUE);
  }
}

boolean objectNameLessThanP(LogicObject* o1, LogicObject* o2) {
  return (safeStringLessP(objectName(o1)->symbolName, objectName(o2)->symbolName));
}

boolean moduleNameLessThanP(Context* m1, Context* m2) {
  return (safeStringLessP(m1->contextName(), m2->contextName()));
}

boolean propositionLessThanP(Proposition* p1, Proposition* p2) {
  if (stringLessP(p1->operatoR->symbolName, p2->operatoR->symbolName)) {
    return (TRUE);
  }
  if (stringLessP(p2->operatoR->symbolName, p1->operatoR->symbolName)) {
    return (FALSE);
  }
  { Object* a1 = NULL;
    Object* a2 = NULL;
    Vector* vector000 = p1->arguments;
    int index000 = 0;
    int length000 = vector000->length();
    Vector* vector001 = p2->arguments;
    int index001 = 0;
    int length001 = vector001->length();

    while ((index000 < length000) &&
        (index001 < length001)) {
      a1 = (vector000->theArray)[index000];
      index000 = index000 + 1;
      a2 = (vector001->theArray)[index001];
      index001 = index001 + 1;
      if (logicFormLessP(a1, a2)) {
        return (TRUE);
      }
      else {
        if (logicFormLessP(a2, a1)) {
          return (FALSE);
        }
      }
    }
  }
  return (p1->arguments->length() < p2->arguments->length());
}

char* propositionHeadSortString(Proposition* p) {
  if (p->kind == KWD_PROPOSITIONS_IMPLIES) {
    if (isaP((p->arguments->theArray)[0], SGT_PROPOSITIONS_LOGIC_DESCRIPTION) &&
        isaP((p->arguments->theArray)[1], SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) {
      return ("FORALL");
    }
    else {
      return ("SUBSET-OF");
    }
  }
  else {
    return (p->operatoR->symbolName);
  }
}

char* objectSortString(Object* o) {
  if (!((boolean)(o))) {
    return ("NULL");
  }
  { Surrogate* testValue000 = safePrimaryType(o);

    if (subtypeOfIntegerP(testValue000)) {
      { Object* o000 = o;
        IntegerWrapper* o = ((IntegerWrapper*)(o000));

        return (stringify(o));
      }
    }
    else if (subtypeOfFloatP(testValue000)) {
      { Object* o001 = o;
        FloatWrapper* o = ((FloatWrapper*)(o001));

        return (stringify(o));
      }
    }
    else if (subtypeOfBooleanP(testValue000)) {
      { Object* o002 = o;
        BooleanWrapper* o = ((BooleanWrapper*)(o002));

        if (o == FALSE_WRAPPER) {
          return ("FALSE");
        }
        else if (o == TRUE_WRAPPER) {
          return ("TRUE");
        }
        else {
          return ("UNKNOWN");
        }
      }
    }
    else if (subtypeOfStringP(testValue000)) {
      { Object* o003 = o;
        StringWrapper* o = ((StringWrapper*)(o003));

        return (o->wrapperValue);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_MUTABLE_STRING_WRAPPER)) {
      { Object* o004 = o;
        MutableStringWrapper* o = ((MutableStringWrapper*)(o004));

        return ((o->wrapperValue));
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_GENERALIZED_SYMBOL)) {
      { Object* o005 = o;
        GeneralizedSymbol* o = ((GeneralizedSymbol*)(o005));

        return (o->symbolName);
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
      { Object* o006 = o;
        Skolem* o = ((Skolem*)(o006));

        std::cout << "Skolem " << o << " value = " << ((Object*)(accessInContext(o->variableValue, o->homeContext, FALSE)));
        return (objectSortString(((Object*)(accessInContext(o->variableValue, o->homeContext, FALSE)))));
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT)) {
      { Object* o007 = o;
        LogicObject* o = ((LogicObject*)(o007));

        return (objectStringName(o));
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_STELLA_CONTEXT)) {
      { Object* o008 = o;
        Context* o = ((Context*)(o008));

        return (o->contextName());
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_COMPUTED_PROCEDURE)) {
      { Object* o009 = o;
        ComputedProcedure* o = ((ComputedProcedure*)(o009));

        return (o->surrogateValueInverse->symbolName);
      }
    }
    else {
      *(STANDARD_WARNING->nativeStream) << "Warning: " << "OBJECT-SORT-STRING: Doesn't specifically handle object " << "`" << o << "'" << " yet!" << std::endl;
      return (stringify(o));
    }
  }
}

boolean logicFormLessP(Object* o1, Object* o2) {
  // A sorting predicate for objects `o1' and `o2' that can appear in logical
  // forms.  Performs a combined numeric and lexocographic sort that accounts
  // for lists, collections and propositions.  Numbers precede all other values,
  // `null' follows all other values.
  if (isaP(o1, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
    o1 = ((Object*)(accessInContext(((Skolem*)(o1))->variableValue, ((Skolem*)(o1))->homeContext, FALSE)));
  }
  if (isaP(o2, SGT_PROPOSITIONS_LOGIC_SKOLEM)) {
    o2 = ((Object*)(accessInContext(((Skolem*)(o2))->variableValue, ((Skolem*)(o2))->homeContext, FALSE)));
  }
  if (!((boolean)(o1))) {
    return (FALSE);
  }
  if (!((boolean)(o2))) {
    return (TRUE);
  }
  { Surrogate* testValue000 = safePrimaryType(o1);

    if (subtypeOfIntegerP(testValue000)) {
      { Object* o1000 = o1;
        IntegerWrapper* o1 = ((IntegerWrapper*)(o1000));

        { Surrogate* testValue001 = safePrimaryType(o2);

          if (subtypeOfIntegerP(testValue001)) {
            { Object* o2000 = o2;
              IntegerWrapper* o2 = ((IntegerWrapper*)(o2000));

              return (o1->wrapperValue < o2->wrapperValue);
            }
          }
          else if (subtypeOfFloatP(testValue001)) {
            { Object* o2001 = o2;
              FloatWrapper* o2 = ((FloatWrapper*)(o2001));

              return (o1->wrapperValue < o2->wrapperValue);
            }
          }
          else {
            return (TRUE);
          }
        }
      }
    }
    else if (subtypeOfFloatP(testValue000)) {
      { Object* o1001 = o1;
        FloatWrapper* o1 = ((FloatWrapper*)(o1001));

        { Surrogate* testValue002 = safePrimaryType(o2);

          if (subtypeOfIntegerP(testValue002)) {
            { Object* o2002 = o2;
              IntegerWrapper* o2 = ((IntegerWrapper*)(o2002));

              return (o1->wrapperValue < o2->wrapperValue);
            }
          }
          else if (subtypeOfFloatP(testValue002)) {
            { Object* o2003 = o2;
              FloatWrapper* o2 = ((FloatWrapper*)(o2003));

              return (o1->wrapperValue < o2->wrapperValue);
            }
          }
          else {
            return (TRUE);
          }
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
      { Object* o1002 = o1;
        Proposition* o1 = ((Proposition*)(o1002));

        { Surrogate* testValue003 = safePrimaryType(o2);

          if (subtypeOfP(testValue003, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
            { Object* o2004 = o2;
              Proposition* o2 = ((Proposition*)(o2004));

              return (propositionLessThanP(o1, o2));
            }
          }
          else if (subtypeOfP(testValue003, SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
            { Object* o2005 = o2;
              NamedDescription* o2 = ((NamedDescription*)(o2005));

              return (stringLessP(propositionHeadSortString(o1), objectSortString(o2)));
            }
          }
          else if (subtypeOfP(testValue003, SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) {
            { Object* o2006 = o2;
              Description* o2 = ((Description*)(o2006));

              return (propositionLessThanP(o1, o2->proposition));
            }
          }
          else {
            return (stringLessP(propositionHeadSortString(o1), objectSortString(o2)));
          }
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
      { Object* o1003 = o1;
        NamedDescription* o1 = ((NamedDescription*)(o1003));

        { Surrogate* testValue004 = safePrimaryType(o2);

          if (subtypeOfP(testValue004, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
            { Object* o2007 = o2;
              Proposition* o2 = ((Proposition*)(o2007));

              return (stringLessP(objectSortString(o1), propositionHeadSortString(o2)));
            }
          }
          else if (subtypeOfP(testValue004, SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
            { Object* o2008 = o2;
              NamedDescription* o2 = ((NamedDescription*)(o2008));

              return (stringLessP(objectSortString(o1), objectSortString(o2)));
            }
          }
          else if (subtypeOfP(testValue004, SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) {
            { Object* o2009 = o2;
              Description* o2 = ((Description*)(o2009));

              return (stringLessP(objectSortString(o1), propositionHeadSortString(o2->proposition)));
            }
          }
          else {
            return (stringLessP(objectSortString(o1), objectSortString(o2)));
          }
        }
      }
    }
    else if (subtypeOfP(testValue000, SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) {
      { Object* o1004 = o1;
        Description* o1 = ((Description*)(o1004));

        return (logicFormLessP(o1->proposition, o2));
      }
    }
    else {
      { Surrogate* testValue005 = safePrimaryType(o2);

        if (subtypeOfP(testValue005, SGT_PROPOSITIONS_LOGIC_PROPOSITION)) {
          { Object* o2010 = o2;
            Proposition* o2 = ((Proposition*)(o2010));

            return (!stringLessP(objectSortString(o1), propositionHeadSortString(o2)));
          }
        }
        else if (subtypeOfP(testValue005, SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION)) {
          { Object* o2011 = o2;
            NamedDescription* o2 = ((NamedDescription*)(o2011));

            return (stringLessP(objectSortString(o1), objectSortString(o2)));
          }
        }
        else if (subtypeOfP(testValue005, SGT_PROPOSITIONS_LOGIC_DESCRIPTION)) {
          { Object* o2012 = o2;
            Description* o2 = ((Description*)(o2012));

            return (stringLessP(objectSortString(o1), propositionHeadSortString(o2->proposition)));
          }
        }
        else {
          return (stringLessP(objectSortString(o1), objectSortString(o2)));
        }
      }
    }
  }
}

void helpStartupPropositions1() {
  {
    SYM_PROPOSITIONS_LOGIC_PROPOSITIONS_INDEX = ((Symbol*)(internRigidSymbolWrtModule("PROPOSITIONS-INDEX", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_INSTANCES_INDEX = ((Symbol*)(internRigidSymbolWrtModule("INSTANCES-INDEX", NULL, 0)));
    SGT_PROPOSITIONS_LOGIC_BACKLINKS_INDEX = ((Surrogate*)(internRigidSymbolWrtModule("BACKLINKS-INDEX", NULL, 1)));
    SYM_PROPOSITIONS_LOGIC_DEPENDENT_PROPOSITIONS_LIST = ((Symbol*)(internRigidSymbolWrtModule("DEPENDENT-PROPOSITIONS-LIST", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_DEPENDENT_ISA_PROPOSITIONS_LIST = ((Symbol*)(internRigidSymbolWrtModule("DEPENDENT-ISA-PROPOSITIONS-LIST", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_PREDICATE_PROPOSITIONS_TABLE = ((Symbol*)(internRigidSymbolWrtModule("PREDICATE-PROPOSITIONS-TABLE", NULL, 0)));
    SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT = ((Surrogate*)(internRigidSymbolWrtModule("LOGIC-OBJECT", NULL, 1)));
    SYM_PROPOSITIONS_LOGIC_DEPENDENT_DESCRIPTIONS = ((Symbol*)(internRigidSymbolWrtModule("DEPENDENT-DESCRIPTIONS", NULL, 0)));
    SYM_PROPOSITIONS_STELLA_BADp = ((Symbol*)(internRigidSymbolWrtModule("BAD?", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_LOGIC_VARIABLE_VALUE_INVERSE = ((Symbol*)(internRigidSymbolWrtModule("VARIABLE-VALUE-INVERSE", NULL, 0)));
    SYM_PROPOSITIONS_STELLA_SURROGATE_VALUE_INVERSE = ((Symbol*)(internRigidSymbolWrtModule("SURROGATE-VALUE-INVERSE", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_LOGIC_OBJECT_STRINGIFIED_SOURCE = ((Symbol*)(internRigidSymbolWrtModule("OBJECT-STRINGIFIED-SOURCE", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_OBJECT_ORIGINATED_PROPOSITIONS = ((Symbol*)(internRigidSymbolWrtModule("OBJECT-ORIGINATED-PROPOSITIONS", NULL, 0)));
    SGT_PROPOSITIONS_STELLA_CS_VALUE = ((Surrogate*)(internRigidSymbolWrtModule("CS-VALUE", getStellaModule("/STELLA", TRUE), 1)));
    SYM_PROPOSITIONS_STELLA_DELETED_OBJECTp = ((Symbol*)(internRigidSymbolWrtModule("DELETED-OBJECT?", getStellaModule("/STELLA", TRUE), 0)));
    SGT_PROPOSITIONS_LOGIC_LOGIC_THING = ((Surrogate*)(internRigidSymbolWrtModule("LOGIC-THING", NULL, 1)));
    SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE = ((Surrogate*)(internRigidSymbolWrtModule("PATTERN-VARIABLE", NULL, 1)));
    SGT_PROPOSITIONS_LOGIC_SKOLEM = ((Surrogate*)(internRigidSymbolWrtModule("SKOLEM", NULL, 1)));
    SYM_PROPOSITIONS_LOGIC_SKOLEM_TYPE = ((Symbol*)(internRigidSymbolWrtModule("SKOLEM-TYPE", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_SKOLEM_NAME = ((Symbol*)(internRigidSymbolWrtModule("SKOLEM-NAME", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_VARIABLE_VALUE = ((Symbol*)(internRigidSymbolWrtModule("VARIABLE-VALUE", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_DEFINING_PROPOSITION = ((Symbol*)(internRigidSymbolWrtModule("DEFINING-PROPOSITION", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_BOUND_TO_OFFSET = ((Symbol*)(internRigidSymbolWrtModule("BOUND-TO-OFFSET", NULL, 0)));
    SYM_PROPOSITIONS_STELLA_VARIABLE_NAME = ((Symbol*)(internRigidSymbolWrtModule("VARIABLE-NAME", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_LOGIC_OBJECT_VECTOR = ((Symbol*)(internRigidSymbolWrtModule("OBJECT-VECTOR", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_ARGUMENTS_VECTOR = ((Symbol*)(internRigidSymbolWrtModule("ARGUMENTS-VECTOR", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_VARIABLES_VECTOR = ((Symbol*)(internRigidSymbolWrtModule("VARIABLES-VECTOR", NULL, 0)));
    SGT_PROPOSITIONS_LOGIC_FORWARD_CHAINING_INDEX = ((Surrogate*)(internRigidSymbolWrtModule("FORWARD-CHAINING-INDEX", NULL, 1)));
    SYM_PROPOSITIONS_LOGIC_FORWARD_GOAL = ((Symbol*)(internRigidSymbolWrtModule("FORWARD-GOAL", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_MASTER_RULE = ((Symbol*)(internRigidSymbolWrtModule("MASTER-RULE", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_IO_VARIABLES = ((Symbol*)(internRigidSymbolWrtModule("IO-VARIABLES", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_QUERY_BODY = ((Symbol*)(internRigidSymbolWrtModule("QUERY-BODY", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_INPUT_BINDINGS = ((Symbol*)(internRigidSymbolWrtModule("INPUT-BINDINGS", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_CACHE_ID = ((Symbol*)(internRigidSymbolWrtModule("CACHE-ID", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_CONSEQUENT = ((Symbol*)(internRigidSymbolWrtModule("CONSEQUENT", NULL, 0)));
    SGT_PROPOSITIONS_LOGIC_DESCRIPTION = ((Surrogate*)(internRigidSymbolWrtModule("DESCRIPTION", NULL, 1)));
    SYM_PROPOSITIONS_LOGIC_FORWARD_CHAINING_INDICES = ((Symbol*)(internRigidSymbolWrtModule("FORWARD-CHAINING-INDICES", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_DEFERRED_CONTRAPOSITIVESp = ((Symbol*)(internRigidSymbolWrtModule("DEFERRED-CONTRAPOSITIVES?", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_PROPOSITION = ((Symbol*)(internRigidSymbolWrtModule("PROPOSITION", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_INTERNAL_VARIABLES = ((Symbol*)(internRigidSymbolWrtModule("INTERNAL-VARIABLES", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_QUERY_PATTERNS = ((Symbol*)(internRigidSymbolWrtModule("QUERY-PATTERNS", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_RELATIVE_COMPLEMENT = ((Symbol*)(internRigidSymbolWrtModule("RELATIVE-COMPLEMENT", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_COMPLEMENT_DESCRIPTION = ((Symbol*)(internRigidSymbolWrtModule("COMPLEMENT-DESCRIPTION", NULL, 0)));
    SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION = ((Surrogate*)(internRigidSymbolWrtModule("NAMED-DESCRIPTION", NULL, 1)));
    SYM_PROPOSITIONS_LOGIC_NATIVE_RELATION_BACK_POINTER = ((Symbol*)(internRigidSymbolWrtModule("NATIVE-RELATION-BACK-POINTER", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_IO_VARIABLE_NAMES = ((Symbol*)(internRigidSymbolWrtModule("IO-VARIABLE-NAMES", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_IO_VARIABLE_TYPES = ((Symbol*)(internRigidSymbolWrtModule("IO-VARIABLE-TYPES", NULL, 0)));
    SYM_PROPOSITIONS_STELLA_EXTENSION = ((Symbol*)(internRigidSymbolWrtModule("EXTENSION", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_LOGIC_PARTIAL_MATCH_SCORE = ((Symbol*)(internRigidSymbolWrtModule("PARTIAL-MATCH-SCORE", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_SET_OF_SUPPORT = ((Symbol*)(internRigidSymbolWrtModule("SET-OF-SUPPORT", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_TIMESTAMP = ((Symbol*)(internRigidSymbolWrtModule("TIMESTAMP", NULL, 0)));
    KWD_PROPOSITIONS_UNKNOWN = ((Keyword*)(internRigidSymbolWrtModule("UNKNOWN", NULL, 2)));
    SGT_PROPOSITIONS_LOGIC_TRUTH_VALUE = ((Surrogate*)(internRigidSymbolWrtModule("TRUTH-VALUE", NULL, 1)));
    SYM_PROPOSITIONS_LOGIC_POLARITY = ((Symbol*)(internRigidSymbolWrtModule("POLARITY", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_STRENGTH = ((Symbol*)(internRigidSymbolWrtModule("STRENGTH", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_POSITIVE_SCORE = ((Symbol*)(internRigidSymbolWrtModule("POSITIVE-SCORE", NULL, 0)));
    SGT_PROPOSITIONS_LOGIC_PARTIAL_SUPPORT = ((Surrogate*)(internRigidSymbolWrtModule("PARTIAL-SUPPORT", NULL, 1)));
    SYM_PROPOSITIONS_LOGIC_FACT = ((Symbol*)(internRigidSymbolWrtModule("FACT", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_AXIOM = ((Symbol*)(internRigidSymbolWrtModule("AXIOM", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_SCORE = ((Symbol*)(internRigidSymbolWrtModule("SCORE", NULL, 0)));
  }
}

void helpStartupPropositions2() {
  {
    SYM_PROPOSITIONS_LOGIC_ID = ((Symbol*)(internRigidSymbolWrtModule("ID", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_ARGUMENT_SCORES = ((Symbol*)(internRigidSymbolWrtModule("ARGUMENT-SCORES", NULL, 0)));
    SGT_PROPOSITIONS_LOGIC_PROPOSITION = ((Surrogate*)(internRigidSymbolWrtModule("PROPOSITION", NULL, 1)));
    SYM_PROPOSITIONS_LOGIC_VARIABLE_TYPEp = ((Symbol*)(internRigidSymbolWrtModule("VARIABLE-TYPE?", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_SATELLITE_PROPOSITIONS = ((Symbol*)(internRigidSymbolWrtModule("SATELLITE-PROPOSITIONS", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_UNFASTENEDp = ((Symbol*)(internRigidSymbolWrtModule("UNFASTENED?", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_WEIGHT_DELTA = ((Symbol*)(internRigidSymbolWrtModule("WEIGHT-DELTA", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_SUPPORT = ((Symbol*)(internRigidSymbolWrtModule("SUPPORT", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_KIND = ((Symbol*)(internRigidSymbolWrtModule("KIND", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_TRUTH_VALUE = ((Symbol*)(internRigidSymbolWrtModule("TRUTH-VALUE", NULL, 0)));
    SYM_PROPOSITIONS_STELLA_ARGUMENTS = ((Symbol*)(internRigidSymbolWrtModule("ARGUMENTS", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_STELLA_OPERATOR = ((Symbol*)(internRigidSymbolWrtModule("OPERATOR", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_LOGIC_RELATIONREF = ((Symbol*)(internRigidSymbolWrtModule("RELATIONREF", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_DEPENDENT_PROPOSITIONS = ((Symbol*)(internRigidSymbolWrtModule("DEPENDENT-PROPOSITIONS", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_PROPOSITION_ORIGINATED_PROPOSITIONS = ((Symbol*)(internRigidSymbolWrtModule("PROPOSITION-ORIGINATED-PROPOSITIONS", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_PROPOSITION_STRINGIFIED_SOURCE = ((Symbol*)(internRigidSymbolWrtModule("PROPOSITION-STRINGIFIED-SOURCE", NULL, 0)));
    SGT_PROPOSITIONS_LOGIC_INTEGER_LOGIC_WRAPPER = ((Surrogate*)(internRigidSymbolWrtModule("INTEGER-LOGIC-WRAPPER", NULL, 1)));
    SGT_PROPOSITIONS_LOGIC_FLOAT_LOGIC_WRAPPER = ((Surrogate*)(internRigidSymbolWrtModule("FLOAT-LOGIC-WRAPPER", NULL, 1)));
    SGT_PROPOSITIONS_LOGIC_STRING_LOGIC_WRAPPER = ((Surrogate*)(internRigidSymbolWrtModule("STRING-LOGIC-WRAPPER", NULL, 1)));
    KWD_PROPOSITIONS_AND = ((Keyword*)(internRigidSymbolWrtModule("AND", NULL, 2)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_AND = ((Surrogate*)(internRigidSymbolWrtModule("AND", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    KWD_PROPOSITIONS_OR = ((Keyword*)(internRigidSymbolWrtModule("OR", NULL, 2)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_OR = ((Surrogate*)(internRigidSymbolWrtModule("OR", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    KWD_PROPOSITIONS_NOT = ((Keyword*)(internRigidSymbolWrtModule("NOT", NULL, 2)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_NOT = ((Surrogate*)(internRigidSymbolWrtModule("NOT", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    KWD_PROPOSITIONS_FORALL = ((Keyword*)(internRigidSymbolWrtModule("FORALL", NULL, 2)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_FORALL = ((Surrogate*)(internRigidSymbolWrtModule("FORALL", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    KWD_PROPOSITIONS_EXISTS = ((Keyword*)(internRigidSymbolWrtModule("EXISTS", NULL, 2)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_EXISTS = ((Surrogate*)(internRigidSymbolWrtModule("EXISTS", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    KWD_PROPOSITIONS_EQUIVALENT = ((Keyword*)(internRigidSymbolWrtModule("EQUIVALENT", NULL, 2)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_EQUIVALENT = ((Surrogate*)(internRigidSymbolWrtModule("EQUIVALENT", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    KWD_PROPOSITIONS_FAIL = ((Keyword*)(internRigidSymbolWrtModule("FAIL", NULL, 2)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_FAIL = ((Surrogate*)(internRigidSymbolWrtModule("FAIL", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    KWD_PROPOSITIONS_COLLECT_INTO = ((Keyword*)(internRigidSymbolWrtModule("COLLECT-INTO", NULL, 2)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_COLLECT_INTO = ((Surrogate*)(internRigidSymbolWrtModule("COLLECT-INTO", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    KWD_PROPOSITIONS_IMPLIES = ((Keyword*)(internRigidSymbolWrtModule("IMPLIES", NULL, 2)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_SUBSET_OF = ((Surrogate*)(internRigidSymbolWrtModule("SUBSET-OF", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SYM_PROPOSITIONS_STELLA_FUNCTION = ((Symbol*)(internRigidSymbolWrtModule("FUNCTION", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_STELLA_PREDICATE = ((Symbol*)(internRigidSymbolWrtModule("PREDICATE", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_STELLA_ISA = ((Symbol*)(internRigidSymbolWrtModule("ISA", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_LOGIC_DESCRIPTIVEp = ((Symbol*)(internRigidSymbolWrtModule("DESCRIPTIVE?", NULL, 0)));
    SYM_PROPOSITIONS_STELLA_CODE_ONLYp = ((Symbol*)(internRigidSymbolWrtModule("CODE-ONLY?", getStellaModule("/STELLA", TRUE), 0)));
    KWD_PROPOSITIONS_ERROR = ((Keyword*)(internRigidSymbolWrtModule("ERROR", NULL, 2)));
    SYM_PROPOSITIONS_LOGIC_LOCALLY_CONCEIVED_PROPOSITIONS_INTERNAL = ((Symbol*)(internRigidSymbolWrtModule("LOCALLY-CONCEIVED-PROPOSITIONS-INTERNAL", NULL, 0)));
    KWD_PROPOSITIONS_PAGING = ((Keyword*)(internRigidSymbolWrtModule("PAGING", NULL, 2)));
    KWD_PROPOSITIONS_CONTEXT_PROPOSITIONS = ((Keyword*)(internRigidSymbolWrtModule("CONTEXT-PROPOSITIONS", NULL, 2)));
    SYM_PROPOSITIONS_LOGIC_LOCALLY_CONCEIVED_INSTANCES_INTERNAL = ((Symbol*)(internRigidSymbolWrtModule("LOCALLY-CONCEIVED-INSTANCES-INTERNAL", NULL, 0)));
    KWD_PROPOSITIONS_CONTEXT_INSTANCES = ((Keyword*)(internRigidSymbolWrtModule("CONTEXT-INSTANCES", NULL, 2)));
    KWD_PROPOSITIONS_UPDATE_PROPOSITION = ((Keyword*)(internRigidSymbolWrtModule("UPDATE-PROPOSITION", NULL, 2)));
    KWD_PROPOSITIONS_EXECUTE_QUERY = ((Keyword*)(internRigidSymbolWrtModule("EXECUTE-QUERY", NULL, 2)));
    SYM_PROPOSITIONS_LOGIC_BOTTOM = ((Symbol*)(internRigidSymbolWrtModule("BOTTOM", NULL, 0)));
    SGT_PROPOSITIONS_STELLA_LITERAL = ((Surrogate*)(internRigidSymbolWrtModule("LITERAL", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_LOGIC_F_LOGICAL_SUBTYPE_OF_LITERALp_MEMO_TABLE_000 = ((Surrogate*)(internRigidSymbolWrtModule("F-LOGICAL-SUBTYPE-OF-LITERAL?-MEMO-TABLE-000", NULL, 1)));
    SGT_PROPOSITIONS_STELLA_CLASS = ((Surrogate*)(internRigidSymbolWrtModule("CLASS", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_LITERAL_WRAPPER = ((Surrogate*)(internRigidSymbolWrtModule("LITERAL-WRAPPER", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_BOOLEAN = ((Surrogate*)(internRigidSymbolWrtModule("BOOLEAN", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_BOOLEAN_WRAPPER = ((Surrogate*)(internRigidSymbolWrtModule("BOOLEAN-WRAPPER", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_VARIABLE_ARITY = ((Surrogate*)(internRigidSymbolWrtModule("VARIABLE-ARITY", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_COLLECTION = ((Surrogate*)(internRigidSymbolWrtModule("COLLECTION", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_RELATION_COMPUTATION = ((Surrogate*)(internRigidSymbolWrtModule("RELATION-COMPUTATION", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
  }
}

void helpStartupPropositions3() {
  {
    SGT_PROPOSITIONS_STELLA_THING = ((Surrogate*)(internRigidSymbolWrtModule("THING", getStellaModule("/STELLA", TRUE), 1)));
    KWD_PROPOSITIONS_EXTENSIONAL_ASSERTION = ((Keyword*)(internRigidSymbolWrtModule("EXTENSIONAL-ASSERTION", NULL, 2)));
    KWD_PROPOSITIONS_DESCRIPTION = ((Keyword*)(internRigidSymbolWrtModule("DESCRIPTION", NULL, 2)));
    KWD_PROPOSITIONS_RETRACT_TRUE = ((Keyword*)(internRigidSymbolWrtModule("RETRACT-TRUE", NULL, 2)));
    KWD_PROPOSITIONS_FUNCTION = ((Keyword*)(internRigidSymbolWrtModule("FUNCTION", NULL, 2)));
    KWD_PROPOSITIONS_ISA = ((Keyword*)(internRigidSymbolWrtModule("ISA", NULL, 2)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_GOES_TRUE_DEMON = ((Surrogate*)(internRigidSymbolWrtModule("GOES-TRUE-DEMON", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    KWD_PROPOSITIONS_PREDICATE = ((Keyword*)(internRigidSymbolWrtModule("PREDICATE", NULL, 2)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_GOES_FALSE_DEMON = ((Surrogate*)(internRigidSymbolWrtModule("GOES-FALSE-DEMON", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_GOES_UNKNOWN_DEMON = ((Surrogate*)(internRigidSymbolWrtModule("GOES-UNKNOWN-DEMON", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SYM_PROPOSITIONS_LOGIC_MASTER_PROPOSITION = ((Symbol*)(internRigidSymbolWrtModule("MASTER-PROPOSITION", NULL, 0)));
    KWD_PROPOSITIONS_ASSERT_TRUE = ((Keyword*)(internRigidSymbolWrtModule("ASSERT-TRUE", NULL, 2)));
    KWD_PROPOSITIONS_PRESUME_TRUE = ((Keyword*)(internRigidSymbolWrtModule("PRESUME-TRUE", NULL, 2)));
    KWD_PROPOSITIONS_ASSERT_FALSE = ((Keyword*)(internRigidSymbolWrtModule("ASSERT-FALSE", NULL, 2)));
    KWD_PROPOSITIONS_PRESUME_FALSE = ((Keyword*)(internRigidSymbolWrtModule("PRESUME-FALSE", NULL, 2)));
    KWD_PROPOSITIONS_RETRACT_FALSE = ((Keyword*)(internRigidSymbolWrtModule("RETRACT-FALSE", NULL, 2)));
    KWD_PROPOSITIONS_ASSERT = ((Keyword*)(internRigidSymbolWrtModule("ASSERT", NULL, 2)));
    KWD_PROPOSITIONS_RETRACT_INCONSISTENT = ((Keyword*)(internRigidSymbolWrtModule("RETRACT-INCONSISTENT", NULL, 2)));
    KWD_PROPOSITIONS_RETRACT = ((Keyword*)(internRigidSymbolWrtModule("RETRACT", NULL, 2)));
    KWD_PROPOSITIONS_CONCEIVE = ((Keyword*)(internRigidSymbolWrtModule("CONCEIVE", NULL, 2)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_HIDDEN_RELATION = ((Surrogate*)(internRigidSymbolWrtModule("HIDDEN-RELATION", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_LOGIC_HIDDEN_INSTANCE = ((Surrogate*)(internRigidSymbolWrtModule("HIDDEN-INSTANCE", NULL, 1)));
    KWD_PROPOSITIONS_KB_UPDATE = ((Keyword*)(internRigidSymbolWrtModule("KB-UPDATE", NULL, 2)));
    KWD_PROPOSITIONS_META_KB_UPDATE = ((Keyword*)(internRigidSymbolWrtModule("META-KB-UPDATE", NULL, 2)));
    SGT_PROPOSITIONS_STELLA_STORAGE_SLOT = ((Surrogate*)(internRigidSymbolWrtModule("STORAGE-SLOT", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_SURROGATE = ((Surrogate*)(internRigidSymbolWrtModule("SURROGATE", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_RELATION = ((Surrogate*)(internRigidSymbolWrtModule("RELATION", getStellaModule("/STELLA", TRUE), 1)));
    SYM_PROPOSITIONS_LOGIC_DESCRIPTION = ((Symbol*)(internRigidSymbolWrtModule("DESCRIPTION", NULL, 0)));
    KWD_PROPOSITIONS_WARNING = ((Keyword*)(internRigidSymbolWrtModule("WARNING", NULL, 2)));
    SYM_PROPOSITIONS_LOGIC_EQUIVALENT_VALUE = ((Symbol*)(internRigidSymbolWrtModule("EQUIVALENT-VALUE", NULL, 0)));
    KWD_PROPOSITIONS_TRUE = ((Keyword*)(internRigidSymbolWrtModule("TRUE", NULL, 2)));
    KWD_PROPOSITIONS_STRICT = ((Keyword*)(internRigidSymbolWrtModule("STRICT", NULL, 2)));
    KWD_PROPOSITIONS_FALSE = ((Keyword*)(internRigidSymbolWrtModule("FALSE", NULL, 2)));
    KWD_PROPOSITIONS_DEFAULT = ((Keyword*)(internRigidSymbolWrtModule("DEFAULT", NULL, 2)));
    KWD_PROPOSITIONS_INCONSISTENT = ((Keyword*)(internRigidSymbolWrtModule("INCONSISTENT", NULL, 2)));
    SYM_PROPOSITIONS_LOGIC_META_INFERENCE_CACHE = ((Symbol*)(internRigidSymbolWrtModule("META-INFERENCE-CACHE", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_TRUTH_MAINTAINED_INFERENCE_CACHE = ((Symbol*)(internRigidSymbolWrtModule("TRUTH-MAINTAINED-INFERENCE-CACHE", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_IOTAp = ((Symbol*)(internRigidSymbolWrtModule("IOTA?", NULL, 0)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_CLASS = ((Surrogate*)(internRigidSymbolWrtModule("CLASS", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_RELATION = ((Surrogate*)(internRigidSymbolWrtModule("RELATION", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_SET = ((Surrogate*)(internRigidSymbolWrtModule("SET", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_SET = ((Surrogate*)(internRigidSymbolWrtModule("SET", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_INTERVAL_CACHE = ((Surrogate*)(internRigidSymbolWrtModule("INTERVAL-CACHE", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_ge = ((Surrogate*)(internRigidSymbolWrtModule(">=", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SYM_PROPOSITIONS_LOGIC_CONFLICTING_DEFAULT_VALUES = ((Symbol*)(internRigidSymbolWrtModule("CONFLICTING-DEFAULT-VALUES", NULL, 0)));
    KWD_PROPOSITIONS_CONSTANT = ((Keyword*)(internRigidSymbolWrtModule("CONSTANT", NULL, 2)));
    SGT_PROPOSITIONS_STELLA_TRUE = ((Surrogate*)(internRigidSymbolWrtModule("TRUE", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_FALSE = ((Surrogate*)(internRigidSymbolWrtModule("FALSE", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_CONS = ((Surrogate*)(internRigidSymbolWrtModule("CONS", getStellaModule("/STELLA", TRUE), 1)));
    SYM_PROPOSITIONS_LOGIC_SETOFALL = ((Symbol*)(internRigidSymbolWrtModule("SETOFALL", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_THE_ONLY = ((Symbol*)(internRigidSymbolWrtModule("THE-ONLY", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_KAPPA = ((Symbol*)(internRigidSymbolWrtModule("KAPPA", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_CREATE = ((Symbol*)(internRigidSymbolWrtModule("CREATE", NULL, 0)));
    KWD_PROPOSITIONS_INTENSIONAL_ASSERTION = ((Keyword*)(internRigidSymbolWrtModule("INTENSIONAL-ASSERTION", NULL, 2)));
    SYM_PROPOSITIONS_LOGIC_p = ((Symbol*)(internRigidSymbolWrtModule("?", NULL, 0)));
    SYM_PROPOSITIONS_STELLA_CLASS_EXTENSION = ((Symbol*)(internRigidSymbolWrtModule("CLASS-EXTENSION", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_LOGIC_TRANSFER_LOGIC_INFORMATION_FROM_RELATION_HOOK = ((Symbol*)(internRigidSymbolWrtModule("TRANSFER-LOGIC-INFORMATION-FROM-RELATION-HOOK", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_SUBRELATION_LINKp = ((Symbol*)(internRigidSymbolWrtModule("SUBRELATION-LINK?", NULL, 0)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_MEMBER_OF = ((Surrogate*)(internRigidSymbolWrtModule("MEMBER-OF", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SYM_PROPOSITIONS_LOGIC_ANNOTATIONS = ((Symbol*)(internRigidSymbolWrtModule("ANNOTATIONS", NULL, 0)));
  }
}

void helpStartupPropositions4() {
  {
    SGT_PROPOSITIONS_STELLA_STRING_WRAPPER = ((Surrogate*)(internRigidSymbolWrtModule("STRING-WRAPPER", getStellaModule("/STELLA", TRUE), 1)));
    SYM_PROPOSITIONS_LOGIC_CONSTANT = ((Symbol*)(internRigidSymbolWrtModule("CONSTANT", NULL, 0)));
    SYM_PROPOSITIONS_STELLA_TRUE = ((Symbol*)(internRigidSymbolWrtModule("TRUE", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_STELLA_FALSE = ((Symbol*)(internRigidSymbolWrtModule("FALSE", getStellaModule("/STELLA", TRUE), 0)));
    SGT_PROPOSITIONS_STELLA_SYMBOL = ((Surrogate*)(internRigidSymbolWrtModule("SYMBOL", getStellaModule("/STELLA", TRUE), 1)));
    SYM_PROPOSITIONS_STELLA_e = ((Symbol*)(internRigidSymbolWrtModule("=", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_STELLA_AND = ((Symbol*)(internRigidSymbolWrtModule("AND", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_STELLA_OR = ((Symbol*)(internRigidSymbolWrtModule("OR", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_STELLA_NOT = ((Symbol*)(internRigidSymbolWrtModule("NOT", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_PL_KERNEL_KB_MEMBER_OF = ((Symbol*)(internRigidSymbolWrtModule("MEMBER-OF", getStellaModule("/PL-KERNEL-KB", TRUE), 0)));
    SYM_PROPOSITIONS_LOGIC_IMPLIES = ((Symbol*)(internRigidSymbolWrtModule("IMPLIES", NULL, 0)));
    SYM_PROPOSITIONS_PL_KERNEL_KB_SUBSET_OF = ((Symbol*)(internRigidSymbolWrtModule("SUBSET-OF", getStellaModule("/PL-KERNEL-KB", TRUE), 0)));
    SYM_PROPOSITIONS_STELLA_EXISTS = ((Symbol*)(internRigidSymbolWrtModule("EXISTS", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_STELLA_FORALL = ((Symbol*)(internRigidSymbolWrtModule("FORALL", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_LOGIC_FAIL = ((Symbol*)(internRigidSymbolWrtModule("FAIL", NULL, 0)));
    SYM_PROPOSITIONS_STELLA_COLLECT_INTO = ((Symbol*)(internRigidSymbolWrtModule("COLLECT-INTO", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_LOGIC_ABOUT = ((Symbol*)(internRigidSymbolWrtModule("ABOUT", NULL, 0)));
    SGT_PROPOSITIONS_STELLA_SLOT = ((Surrogate*)(internRigidSymbolWrtModule("SLOT", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_GENERALIZED_SYMBOL = ((Surrogate*)(internRigidSymbolWrtModule("GENERALIZED-SYMBOL", getStellaModule("/STELLA", TRUE), 1)));
    KWD_PROPOSITIONS_DIRECTION = ((Keyword*)(internRigidSymbolWrtModule("DIRECTION", NULL, 2)));
    SYM_PROPOSITIONS_LOGIC_FORWARD_ONLYp = ((Symbol*)(internRigidSymbolWrtModule("FORWARD-ONLY?", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_BACKWARD_ONLYp = ((Symbol*)(internRigidSymbolWrtModule("BACKWARD-ONLY?", NULL, 0)));
    KWD_PROPOSITIONS_WEIGHT = ((Keyword*)(internRigidSymbolWrtModule("WEIGHT", NULL, 2)));
    SYM_PROPOSITIONS_LOGIC_WEIGHT = ((Symbol*)(internRigidSymbolWrtModule("WEIGHT", NULL, 0)));
    KWD_PROPOSITIONS_CONFIDENCE_LEVEL = ((Keyword*)(internRigidSymbolWrtModule("CONFIDENCE-LEVEL", NULL, 2)));
    SYM_PROPOSITIONS_STELLA_SETOF = ((Symbol*)(internRigidSymbolWrtModule("SETOF", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_LOGIC_LISTOF = ((Symbol*)(internRigidSymbolWrtModule("LISTOF", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_EQUIVALENT = ((Symbol*)(internRigidSymbolWrtModule("EQUIVALENT", NULL, 0)));
    SYM_PROPOSITIONS_PL_KERNEL_KB_HOLDS = ((Symbol*)(internRigidSymbolWrtModule("HOLDS", getStellaModule("/PL-KERNEL-KB", TRUE), 0)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_FUNCTION = ((Surrogate*)(internRigidSymbolWrtModule("FUNCTION", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_HOLDS = ((Surrogate*)(internRigidSymbolWrtModule("HOLDS", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_TABLE = ((Surrogate*)(internRigidSymbolWrtModule("TABLE", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_POLYMORPHIC = ((Surrogate*)(internRigidSymbolWrtModule("POLYMORPHIC", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SYM_PROPOSITIONS_PL_KERNEL_KB_VALUE = ((Symbol*)(internRigidSymbolWrtModule("VALUE", getStellaModule("/PL-KERNEL-KB", TRUE), 0)));
    SYM_PROPOSITIONS_PL_KERNEL_KB_CUT = ((Symbol*)(internRigidSymbolWrtModule("CUT", getStellaModule("/PL-KERNEL-KB", TRUE), 0)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_VALUE = ((Surrogate*)(internRigidSymbolWrtModule("VALUE", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_QUANTITY = ((Surrogate*)(internRigidSymbolWrtModule("QUANTITY", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_DATE = ((Surrogate*)(internRigidSymbolWrtModule("DATE", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_FLOAT = ((Surrogate*)(internRigidSymbolWrtModule("FLOAT", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_INTEGER = ((Surrogate*)(internRigidSymbolWrtModule("INTEGER", getStellaModule("/STELLA", TRUE), 1)));
    KWD_PROPOSITIONS_AUTOMATICALLY_FIX_TYPE_VIOLATIONS = ((Keyword*)(internRigidSymbolWrtModule("AUTOMATICALLY-FIX-TYPE-VIOLATIONS", NULL, 2)));
    KWD_PROPOSITIONS_POST_TYPE_VIOLATIONS = ((Keyword*)(internRigidSymbolWrtModule("POST-TYPE-VIOLATIONS", NULL, 2)));
    KWD_PROPOSITIONS_IGNORE_TYPE_VIOLATIONS = ((Keyword*)(internRigidSymbolWrtModule("IGNORE-TYPE-VIOLATIONS", NULL, 2)));
    KWD_PROPOSITIONS_REPORT_TYPE_VIOLATIONS = ((Keyword*)(internRigidSymbolWrtModule("REPORT-TYPE-VIOLATIONS", NULL, 2)));
    SYM_PROPOSITIONS_LOGIC_ASSERTED_BY_TYPE_CHECKERp = ((Symbol*)(internRigidSymbolWrtModule("ASSERTED-BY-TYPE-CHECKER?", NULL, 0)));
    SGT_PROPOSITIONS_LOGIC_CHECK_TYPES_RECORD = ((Surrogate*)(internRigidSymbolWrtModule("CHECK-TYPES-RECORD", NULL, 1)));
    SYM_PROPOSITIONS_LOGIC_PARENT_PROPOSITION = ((Symbol*)(internRigidSymbolWrtModule("PARENT-PROPOSITION", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_PARENT_DESCRIPTION = ((Symbol*)(internRigidSymbolWrtModule("PARENT-DESCRIPTION", NULL, 0)));
    SYM_PROPOSITIONS_STELLA_MODULE = ((Symbol*)(internRigidSymbolWrtModule("MODULE", getStellaModule("/STELLA", TRUE), 0)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_TOTAL = ((Surrogate*)(internRigidSymbolWrtModule("TOTAL", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_SETOF = ((Surrogate*)(internRigidSymbolWrtModule("SETOF", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_LISTOF = ((Surrogate*)(internRigidSymbolWrtModule("LISTOF", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SGT_PROPOSITIONS_PL_KERNEL_KB_LIST = ((Surrogate*)(internRigidSymbolWrtModule("LIST", getStellaModule("/PL-KERNEL-KB", TRUE), 1)));
    SYM_PROPOSITIONS_LOGIC_FALS = ((Symbol*)(internRigidSymbolWrtModule("FALS", NULL, 0)));
    SYM_PROPOSITIONS_STELLA_SOME = ((Symbol*)(internRigidSymbolWrtModule("SOME", getStellaModule("/STELLA", TRUE), 0)));
    SYM_PROPOSITIONS_LOGIC_LAMBDA = ((Symbol*)(internRigidSymbolWrtModule("LAMBDA", NULL, 0)));
    SYM_PROPOSITIONS_PL_KERNEL_KB_SETOF = ((Symbol*)(internRigidSymbolWrtModule("SETOF", getStellaModule("/PL-KERNEL-KB", TRUE), 0)));
    SYM_PROPOSITIONS_PL_KERNEL_KB_LISTOF = ((Symbol*)(internRigidSymbolWrtModule("LISTOF", getStellaModule("/PL-KERNEL-KB", TRUE), 0)));
    SYM_PROPOSITIONS_LOGIC_BAGOF = ((Symbol*)(internRigidSymbolWrtModule("BAGOF", NULL, 0)));
    SYM_PROPOSITIONS_STELLA_ASSERT = ((Symbol*)(internRigidSymbolWrtModule("ASSERT", getStellaModule("/STELLA", TRUE), 0)));
  }
}

void helpStartupPropositions5() {
  {
    KWD_PROPOSITIONS_CPP = ((Keyword*)(internRigidSymbolWrtModule("CPP", NULL, 2)));
    KWD_PROPOSITIONS_DELETED = ((Keyword*)(internRigidSymbolWrtModule("DELETED", NULL, 2)));
    SGT_PROPOSITIONS_STELLA_MODULE = ((Surrogate*)(internRigidSymbolWrtModule("MODULE", getStellaModule("/STELLA", TRUE), 1)));
    SYM_PROPOSITIONS_LOGIC_INTRODUCE_MODULE = ((Symbol*)(internRigidSymbolWrtModule("INTRODUCE-MODULE", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_CLEAR_LOGIC_MODULE_HOOK = ((Symbol*)(internRigidSymbolWrtModule("CLEAR-LOGIC-MODULE-HOOK", NULL, 0)));
    SYM_PROPOSITIONS_LOGIC_MONOTONICp = ((Symbol*)(internRigidSymbolWrtModule("MONOTONIC?", NULL, 0)));
    SGT_PROPOSITIONS_STELLA_WORLD = ((Surrogate*)(internRigidSymbolWrtModule("WORLD", getStellaModule("/STELLA", TRUE), 1)));
    SYM_PROPOSITIONS_LOGIC_DESTROY_LOGIC_CONTEXT_HOOK = ((Symbol*)(internRigidSymbolWrtModule("DESTROY-LOGIC-CONTEXT-HOOK", NULL, 0)));
    SGT_PROPOSITIONS_STELLA_MUTABLE_STRING_WRAPPER = ((Surrogate*)(internRigidSymbolWrtModule("MUTABLE-STRING-WRAPPER", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_STELLA_CONTEXT = ((Surrogate*)(internRigidSymbolWrtModule("CONTEXT", getStellaModule("/STELLA", TRUE), 1)));
    SGT_PROPOSITIONS_LOGIC_COMPUTED_PROCEDURE = ((Surrogate*)(internRigidSymbolWrtModule("COMPUTED-PROCEDURE", NULL, 1)));
    SYM_PROPOSITIONS_LOGIC_STARTUP_PROPOSITIONS = ((Symbol*)(internRigidSymbolWrtModule("STARTUP-PROPOSITIONS", NULL, 0)));
    SYM_PROPOSITIONS_STELLA_METHOD_STARTUP_CLASSNAME = ((Symbol*)(internRigidSymbolWrtModule("METHOD-STARTUP-CLASSNAME", getStellaModule("/STELLA", TRUE), 0)));
  }
}

void helpStartupPropositions6() {
  {
    ZERO_VARIABLES_VECTOR = newVector(0);
    { PropertyList* self003 = newPropertyList();

      self003->thePlist = listO(19, KWD_PROPOSITIONS_AND, SGT_PROPOSITIONS_PL_KERNEL_KB_AND, KWD_PROPOSITIONS_OR, SGT_PROPOSITIONS_PL_KERNEL_KB_OR, KWD_PROPOSITIONS_NOT, SGT_PROPOSITIONS_PL_KERNEL_KB_NOT, KWD_PROPOSITIONS_FORALL, SGT_PROPOSITIONS_PL_KERNEL_KB_FORALL, KWD_PROPOSITIONS_EXISTS, SGT_PROPOSITIONS_PL_KERNEL_KB_EXISTS, KWD_PROPOSITIONS_EQUIVALENT, SGT_PROPOSITIONS_PL_KERNEL_KB_EQUIVALENT, KWD_PROPOSITIONS_FAIL, SGT_PROPOSITIONS_PL_KERNEL_KB_FAIL, KWD_PROPOSITIONS_COLLECT_INTO, SGT_PROPOSITIONS_PL_KERNEL_KB_COLLECT_INTO, KWD_PROPOSITIONS_IMPLIES, SGT_PROPOSITIONS_PL_KERNEL_KB_SUBSET_OF, NIL);
      oOPERATOR_NAME_TO_SURROGATE_TABLEo = self003;
    }
    oLOGIC_MODULEo = getStellaModule("LOGIC", TRUE);
    oPL_KERNEL_MODULEo = getStellaModule("PL-KERNEL", TRUE);
    oLAST_KB_ACTIONo = KWD_PROPOSITIONS_UPDATE_PROPOSITION;
    oEVALUATIONMODEo.set(KWD_PROPOSITIONS_EXTENSIONAL_ASSERTION);
    oCLIPPINGENABLEDpo.set(TRUE);
    oNATURALDEDUCTIONMODEpo.set(TRUE);
    oCONTEXT_DEPENDENT_SEARCH_MODEpo = FALSE;
    oCOMPUTEDQUERYpo.set(FALSE);
    { TruthValue* self004 = newTruthValue();

      self004->polarity = KWD_PROPOSITIONS_TRUE;
      self004->strength = KWD_PROPOSITIONS_STRICT;
      self004->positiveScore = 1.0;
      TRUE_TRUTH_VALUE = self004;
    }
    { TruthValue* self005 = newTruthValue();

      self005->polarity = KWD_PROPOSITIONS_FALSE;
      self005->strength = KWD_PROPOSITIONS_STRICT;
      self005->positiveScore = -1.0;
      FALSE_TRUTH_VALUE = self005;
    }
    { TruthValue* self006 = newTruthValue();

      self006->polarity = KWD_PROPOSITIONS_TRUE;
      self006->strength = KWD_PROPOSITIONS_DEFAULT;
      self006->positiveScore = 0.8;
      DEFAULT_TRUE_TRUTH_VALUE = self006;
    }
    { TruthValue* self007 = newTruthValue();

      self007->polarity = KWD_PROPOSITIONS_FALSE;
      self007->strength = KWD_PROPOSITIONS_DEFAULT;
      self007->positiveScore = -0.8;
      DEFAULT_FALSE_TRUTH_VALUE = self007;
    }
    { TruthValue* self008 = newTruthValue();

      self008->polarity = KWD_PROPOSITIONS_UNKNOWN;
      UNKNOWN_TRUTH_VALUE = self008;
    }
    { TruthValue* self009 = newTruthValue();

      self009->polarity = KWD_PROPOSITIONS_INCONSISTENT;
      self009->strength = KWD_PROPOSITIONS_STRICT;
      INCONSISTENT_TRUTH_VALUE = self009;
    }
    oSUPPRESSUNTYPEDTYPEERRORpo.set(FALSE);
    oEQUIVALENCE_COLLECTIONSpo = TRUE;
    { Proposition* self010 = newProposition();

      self010->kind = KWD_PROPOSITIONS_CONSTANT;
      self010->operatoR = SGT_PROPOSITIONS_STELLA_TRUE;
      self010->arguments = newVector(0);
      { Proposition* object012 = self010;
        TruthValue* value013 = TRUE_TRUTH_VALUE;
        Object* oldValue014 = object012->truthValue;
        Object* newValue015 = updateInContext(oldValue014, value013, object012->homeContext, FALSE);

        if (!(((boolean)(oldValue014)) &&
            (oldValue014->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
          object012->truthValue = newValue015;
        }
      }
      TRUE_PROPOSITION = self010;
    }
    { Proposition* self016 = newProposition();

      self016->kind = KWD_PROPOSITIONS_CONSTANT;
      self016->operatoR = SGT_PROPOSITIONS_STELLA_FALSE;
      self016->arguments = newVector(0);
      { Proposition* object018 = self016;
        TruthValue* value019 = FALSE_TRUTH_VALUE;
        Object* oldValue020 = object018->truthValue;
        Object* newValue021 = updateInContext(oldValue020, value019, object018->homeContext, FALSE);

        if (!(((boolean)(oldValue020)) &&
            (oldValue020->primaryType() == SGT_PROPOSITIONS_STELLA_CS_VALUE))) {
          object018->truthValue = newValue021;
        }
      }
      FALSE_PROPOSITION = self016;
    }
    ANONYMOUS_VARIABLE_NAME = SYM_PROPOSITIONS_LOGIC_p;
    oMISSING_KEY_VALUE_LISTo = newKeyValueList();
    oAUTOMATICINSTANCETABLEo.set(oMISSING_KEY_VALUE_LISTo);
    oSTRINGIFIED_PROPOSITION_INDEXo = newStringHashTable();
    oDONT_CHECK_FOR_DUPLICATE_PROPOSITIONSpo.set(FALSE);
    oAUTO_COERCE_PROPOSITIONAL_ARGUMENTSpo = FALSE;
    oTYPE_CHECK_POLICYo = KWD_PROPOSITIONS_AUTOMATICALLY_FIX_TYPE_VIOLATIONS;
    oTYPECHECKMODEo.set(KWD_PROPOSITIONS_POST_TYPE_VIOLATIONS);
    oCHECK_TYPES_AGENDAo = newKeyValueList();
  }
}

void helpStartupPropositions7() {
  {
    defineStellaTypeFromStringifiedSource("(DEFTYPE PROPOSITIONS-INDEX (SEQUENCE-INDEX OF PROPOSITION))");
    defineStellaTypeFromStringifiedSource("(DEFTYPE INSTANCES-INDEX (SEQUENCE-INDEX OF LOGIC-OBJECT))");
    { Class* clasS = defineClassFromStringifiedSource("BACKLINKS-INDEX", "(DEFCLASS BACKLINKS-INDEX (STANDARD-OBJECT) :DOCUMENTATION \"Contains several types of indices pointing at dependent propositions.\" :SLOTS ((DEPENDENT-PROPOSITIONS-LIST :TYPE PROPOSITIONS-INDEX) (DEPENDENT-ISA-PROPOSITIONS-LIST :TYPE PROPOSITIONS-INDEX) (PREDICATE-PROPOSITIONS-TABLE :TYPE (HASH-TABLE OF SURROGATE PROPOSITIONS-INDEX))))");

      clasS->classConstructorCode = ((cpp_function_code)(&newBacklinksIndex));
      clasS->classSlotAccessorCode = ((cpp_function_code)(&accessBacklinksIndexSlotValue));
    }
    defineClassFromStringifiedSource("BACKLINKS-MIXIN", "(DEFCLASS BACKLINKS-MIXIN () :DOCUMENTATION \"Inherited by objects that point back at propositions that reference\nthem.\" :SLOTS ((DEPENDENT-PROPOSITIONS-INDEX :TYPE BACKLINKS-INDEX)) :MIXIN? TRUE)");
    { Class* clasS = defineClassFromStringifiedSource("LOGIC-OBJECT", "(DEFCLASS LOGIC-OBJECT (CONTEXT-SENSITIVE-OBJECT DYNAMIC-SLOTS-MIXIN BACKLINKS-MIXIN) :DOCUMENTATION \"Object with propositions, and possibly a variable value.\" :PUBLIC-SLOTS ((DEPENDENT-DESCRIPTIONS :TYPE (LIST OF DESCRIPTION) :ALLOCATION :DYNAMIC) (VARIABLE-VALUE-INVERSE :TYPE (CONS OF LOGIC-OBJECT) :DEFAULT NIL :CONTEXT-SENSITIVE? TRUE) (EQUIVALENT-VALUE :TYPE DESCRIPTION :CONTEXT-SENSITIVE? TRUE :ALLOCATION :DYNAMIC) (SURROGATE-VALUE-INVERSE :TYPE SURROGATE) (UNDECLARED? :TYPE BOOLEAN :ALLOCATION :DYNAMIC) (HYPOTHESIZED-INSTANCE? :TYPE BOOLEAN :ALLOCATION :DYNAMIC) (OBJECT-STRINGIFIED-SOURCE :TYPE STRING :ALLOCATION :DYNAMIC :READER STRINGIFIED-SOURCE :WRITER STRINGIFIED-SOURCE-SETTER) (OBJECT-ORIGINATED-PROPOSITIONS :TYPE (LIST OF PROPOSITION) :DEFAULT NIL-LIST :ALLOCATION :DYNAMIC :READER ORIGINATED-PROPOSITIONS :WRITER ORIGINATED-PROPOSITIONS-SETTER :DOCUMENTATION \"List of propositions generated by the definition of\nthis object.\")) :PRINT-FORM (PRINT-TOP-LEVEL-OBJECT SELF STREAM) :INITIALIZER LOG-LOGIC-OBJECT)");

      clasS->classConstructorCode = ((cpp_function_code)(&newLogicObject));
      clasS->classSlotAccessorCode = ((cpp_function_code)(&accessLogicObjectSlotValue));
    }
    { Class* clasS = defineClassFromStringifiedSource("LOGIC-THING", "(DEFCLASS LOGIC-THING (THING) :DOCUMENTATION \"A subclass of THING so we can add methods to it in\nthe PowerLoom LOGIC module.\")");

      clasS->classConstructorCode = ((cpp_function_code)(&newLogicThing));
    }
    { Class* clasS = defineClassFromStringifiedSource("SKOLEM", "(DEFCLASS SKOLEM (LOGIC-OBJECT) :DOCUMENTATION \"Object whose identity and type are not fixed at creation time.\nPotentially unifiable with any other object.\" :PUBLIC-SLOTS ((SKOLEM-TYPE :TYPE TYPE) (SKOLEM-NAME :TYPE SYMBOL) (VARIABLE-VALUE :TYPE OBJECT :CONTEXT-SENSITIVE? TRUE) (DEFINING-PROPOSITION :TYPE PROPOSITION)))");

      clasS->classConstructorCode = ((cpp_function_code)(&newSkolem));
      clasS->classSlotAccessorCode = ((cpp_function_code)(&accessSkolemSlotValue));
    }
    { Class* clasS = defineClassFromStringifiedSource("PATTERN-VARIABLE", "(DEFCLASS PATTERN-VARIABLE (SKOLEM) :DOCUMENTATION \"A variable in a logical expression.\" :SLOTS ((BOUND-TO-OFFSET :TYPE INTEGER) (ORIGINAL-SKOLEM-NAME :TYPE SYMBOL :ALLOCATION :DYNAMIC) (VARIABLE-NAME :RENAMES SKOLEM-NAME)))");

      clasS->classConstructorCode = ((cpp_function_code)(&newPatternVariable));
      clasS->classSlotAccessorCode = ((cpp_function_code)(&accessPatternVariableSlotValue));
    }
    defineStellaTypeFromStringifiedSource("(DEFTYPE OBJECT-VECTOR (VECTOR OF OBJECT))");
    defineStellaTypeFromStringifiedSource("(DEFTYPE ARGUMENTS-VECTOR OBJECT-VECTOR)");
    defineStellaTypeFromStringifiedSource("(DEFTYPE VARIABLES-VECTOR (VECTOR OF PATTERN-VARIABLE))");
    { Class* clasS = defineClassFromStringifiedSource("FORWARD-CHAINING-INDEX", "(DEFCLASS FORWARD-CHAINING-INDEX (STANDARD-OBJECT) :SLOTS ((FORWARD-GOAL :TYPE PROPOSITION) (MASTER-RULE :TYPE PROPOSITION) (IO-VARIABLES :TYPE CONS) (QUERY-BODY :TYPE CONS) (INPUT-BINDINGS :TYPE CONS) (CACHE-ID :TYPE SYMBOL) (CONSEQUENT :TYPE DESCRIPTION)))");

      clasS->classConstructorCode = ((cpp_function_code)(&newForwardChainingIndex));
      clasS->classSlotAccessorCode = ((cpp_function_code)(&accessForwardChainingIndexSlotValue));
    }
    { Class* clasS = defineClassFromStringifiedSource("DESCRIPTION", "(DEFCLASS DESCRIPTION (LOGIC-OBJECT) :PUBLIC-SLOTS ((PROPOSITION :TYPE PROPOSITION) (IO-VARIABLES :TYPE VARIABLES-VECTOR) (INTERNAL-VARIABLES :TYPE VARIABLES-VECTOR :INITIALLY ZERO-VARIABLES-VECTOR :DOCUMENTATION \"The set of internally quantified variables besides the\ndescriptions's IO variables.  This set includes any top-level implicit\nexistential variables plus any variables introduced by some nested quantifier.\nIMPORTANT: therefore, a description's internal variables are not necessarily\nequivalent to its top-level implicit existential variables.\") (EXTERNAL-VARIABLES :TYPE VARIABLES-VECTOR :ALLOCATION :DYNAMIC) (QUERY-PATTERNS :TYPE (KEY-VALUE-LIST OF BOOLEAN-VECTOR DESCRIPTION)) (COMPLEMENT-DESCRIPTION :TYPE DESCRIPTION :ALLOCATION :DYNAMIC) (RELATIVE-COMPLEMENT :RENAMES COMPLEMENT-DESCRIPTION) (TAXONOMY-NODE :TYPE TAXONOMY-NODE :ALLOCATION :DYNAMIC) (IOTA? :TYPE BOOLEAN :ALLOCATION :DYNAMIC) (DONT-OPTIMIZE? :TYPE BOOLEAN :ALLOCATION :DYNAMIC) (FORWARD-CHAINING-INDICES :TYPE (LIST OF FORWARD-CHAINING-INDEX) :ALLOCATION :DYNAMIC) (RULES-WITH-DEFERRED-SATELLITES :TYPE PROPOSITIONS-INDEX :ALLOCATION :DYNAMIC) (DEFERRED-CONTRAPOSITIVES? :TYPE BOOLEAN-WRAPPER :ALLOCATION :DYNAMIC :DEFAULT FALSE)) :METHODS ((NATIVE-RELATION ((SELF DESCRIPTION)) :TYPE RELATION (RETURN NULL))))");

      clasS->classConstructorCode = ((cpp_function_code)(&newDescription));
      clasS->classSlotAccessorCode = ((cpp_function_code)(&accessDescriptionSlotValue));
    }
    { Class* clasS = defineClassFromStringifiedSource("NAMED-DESCRIPTION", "(DEFCLASS NAMED-DESCRIPTION (DESCRIPTION) :PUBLIC-SLOTS ((IO-VARIABLE-NAMES :TYPE (LIST OF SYMBOL)) (IO-VARIABLE-TYPES :TYPE (LIST OF TYPE)) (EXTENSION :TYPE PROPOSITIONS-INDEX) (OBJECT-STRINGIFIED-SOURCE :ALLOCATION :INSTANCE) (OBJECT-ORIGINATED-PROPOSITIONS :TYPE (LIST OF PROPOSITION) :DEFAULT NIL-LIST :ALLOCATION :INSTANCE :DOCUMENTATION \"List of propositions generated by the definition of\nthis description.\") (NATIVE-RELATION-BACK-POINTER :TYPE RELATION :ALLOCATION :DYNAMIC) (CHECK-FOR-CACHED-GOALS? :TYPE BOOLEAN :ALLOCATION :DYNAMIC) (INVERSE-DESCRIPTION :TYPE NAMED-DESCRIPTION :ALLOCATION :DYNAMIC)) :METHODS ((NATIVE-RELATION ((SELF NAMED-DESCRIPTION)) :TYPE RELATION (RETURN (NATIVE-RELATION-BACK-POINTER SELF))) (NATIVE-RELATION-SETTER ((SELF NAMED-DESCRIPTION) (VALUE RELATION)) (SETF (NATIVE-RELATION-BACK-POINTER SELF) VALUE))))");

      clasS->classConstructorCode = ((cpp_function_code)(&newNamedDescription));
      clasS->classSlotAccessorCode = ((cpp_function_code)(&accessNamedDescriptionSlotValue));
    }
    defineStellaTypeFromStringifiedSource("(DEFTYPE PARTIAL-MATCH-SCORE FLOAT)");
    defineStellaTypeFromStringifiedSource("(DEFTYPE SET-OF-SUPPORT LIST)");
    defineStellaTypeFromStringifiedSource("(DEFTYPE TIMESTAMP INTEGER)");
    { Class* clasS = defineClassFromStringifiedSource("TRUTH-VALUE", "(DEFCLASS TRUTH-VALUE (LOGIC-OBJECT) :PUBLIC-SLOTS ((POLARITY :TYPE KEYWORD :INITIALLY :UNKNOWN :DOCUMENTATION \"One of :TRUE, :FALSE, :UNKNOWN or :INCONSISTENT.\") (STRENGTH :TYPE KEYWORD :INITIALLY NULL :DOCUMENTATION \"Either NULL, :STRICT or :DEFAULT.\") (POSITIVE-SCORE :TYPE PARTIAL-MATCH-SCORE :INITIALLY 0.0)))");

      clasS->classConstructorCode = ((cpp_function_code)(&newTruthValue));
      clasS->classSlotAccessorCode = ((cpp_function_code)(&accessTruthValueSlotValue));
    }
    { Class* clasS = defineClassFromStringifiedSource("PARTIAL-SUPPORT", "(DEFCLASS PARTIAL-SUPPORT (STANDARD-OBJECT) :PUBLIC-SLOTS ((FACT :TYPE PROPOSITION) (AXIOM :TYPE PROPOSITION) (SCORE :TYPE FLOAT) (ID :TYPE INTEGER) (ARGUMENT-SCORES :TYPE (CONS OF FLOAT-WRAPPER))))");

      clasS->classConstructorCode = ((cpp_function_code)(&newPartialSupport));
      clasS->classSlotAccessorCode = ((cpp_function_code)(&accessPartialSupportSlotValue));
    }
    { Class* clasS = defineClassFromStringifiedSource("PROPOSITION", "(DEFCLASS PROPOSITION (CONTEXT-SENSITIVE-OBJECT DYNAMIC-SLOTS-MIXIN BACKLINKS-MIXIN) :DOCUMENTATION \"\" :PUBLIC-SLOTS ((KIND :TYPE KEYWORD) (TRUTH-VALUE :TYPE TRUTH-VALUE :CONTEXT-SENSITIVE? TRUE) (ARGUMENTS :TYPE VECTOR) (OPERATOR :TYPE GENERALIZED-SYMBOL) (RELATIONREF :TYPE SURROGATE :RENAMES OPERATOR) (DEPENDENT-PROPOSITIONS :TYPE (NON-PAGING-INDEX OF PROPOSITION) :INITIALLY NIL-NON-PAGING-INDEX) (SURROGATE-VALUE-INVERSE :TYPE SURROGATE :ALLOCATION :DYNAMIC) (IO-VARIABLES :TYPE VARIABLES-VECTOR :ALLOCATION :DYNAMIC) (VARIABLE-TYPE? :TYPE BOOLEAN-WRAPPER :ALLOCATION :DYNAMIC :DEFAULT FALSE) (ASSERTED-BY-TYPE-CHECKER? :TYPE BOOLEAN :ALLOCATION :DYNAMIC) (DESCRIPTIVE? :TYPE BOOLEAN :ALLOCATION :DYNAMIC) (SUBRELATION-LINK? :TYPE BOOLEAN :ALLOCATION :DYNAMIC) (SUBSUMPTION-LINK? :TYPE BOOLEAN :ALLOCATION :DYNAMIC) (SATELLITE-PROPOSITIONS :TYPE (LIST OF PROPOSITION) :ALLOCATION :DYNAMIC) (MASTER-PROPOSITION :TYPE PROPOSITION :ALLOCATION :DYNAMIC) (PROPOSITION-ORIGINATED-PROPOSITIONS :TYPE (LIST OF PROPOSITION) :DEFAULT NIL-LIST :ALLOCATION :DYNAMIC :READER ORIGINATED-PROPOSITIONS :WRITER ORIGINATED-PROPOSITIONS-SETTER :DOCUMENTATION \"List of propositions generated by the definition of\nthis proposition.\") (DONT-OPTIMIZE? :TYPE BOOLEAN :ALLOCATION :DYNAMIC :OPTION-KEYWORD :DONT-OPTIMIZE?) (FORWARD-ONLY? :TYPE BOOLEAN :ALLOCATION :DYNAMIC :OPTION-KEYWORD :FORWARD-ONLY?) (BACKWARD-ONLY? :TYPE BOOLEAN :ALLOCATION :DYNAMIC :OPTION-KEYWORD :BACKWARD-ONLY?) (PROPOSITION-STRINGIFIED-SOURCE :TYPE STRING :ALLOCATION :DYNAMIC :READER STRINGIFIED-SOURCE :WRITER STRINGIFIED-SOURCE-SETTER) (UNFASTENED? :TYPE BOOLEAN-WRAPPER :ALLOCATION :DYNAMIC :DEFAULT FALSE) (ANNOTATIONS :TYPE (PROPERTY-LIST OF KEYWORD OBJECT) :ALLOCATION :DYNAMIC) (WEIGHT :TYPE FLOAT :ALLOCATION :DYNAMIC) (NEURAL-NETWORK :TYPE PROPOSITION-NEURAL-NETWORK :ALLOCATION :DYNAMIC) (WEIGHT-DELTA :TYPE FLOAT :DEFAULT 0.0 :ALLOCATION :DYNAMIC) (MATCH-SCORE :TYPE FLOAT :ALLOCATION :DYNAMIC) (SUPPORT :TYPE (CONS OF PARTIAL-SUPPORT) :ALLOCATION :DYNAMIC)) :PRINT-FORM (PRINT-TOP-LEVEL-OBJECT SELF STREAM))");

      clasS->classConstructorCode = ((cpp_function_code)(&newProposition));
      clasS->classSlotAccessorCode = ((cpp_function_code)(&accessPropositionSlotValue));
    }
    { Class* clasS = defineClassFromStringifiedSource("INTEGER-LOGIC-WRAPPER", "(DEFCLASS INTEGER-LOGIC-WRAPPER (INTEGER-WRAPPER BACKLINKS-MIXIN) :PRINT-FORM (IF *PRINTREADABLY?* (PRINT-NATIVE-STREAM STREAM (WRAPPER-VALUE SELF)) (PRINT-NATIVE-STREAM STREAM \"|W|\" (WRAPPER-VALUE SELF))))");

      clasS->classConstructorCode = ((cpp_function_code)(&newIntegerLogicWrapper));
    }
    { Class* clasS = defineClassFromStringifiedSource("FLOAT-LOGIC-WRAPPER", "(DEFCLASS FLOAT-LOGIC-WRAPPER (FLOAT-WRAPPER BACKLINKS-MIXIN) :PRINT-FORM (IF *PRINTREADABLY?* (PRINT-NATIVE-STREAM STREAM (WRAPPER-VALUE SELF)) (PRINT-NATIVE-STREAM STREAM \"|W|\" (WRAPPER-VALUE SELF))))");

      clasS->classConstructorCode = ((cpp_function_code)(&newFloatLogicWrapper));
    }
    { Class* clasS = defineClassFromStringifiedSource("STRING-LOGIC-WRAPPER", "(DEFCLASS STRING-LOGIC-WRAPPER (STRING-WRAPPER BACKLINKS-MIXIN) :PRINT-FORM (IF *PRINTREADABLY?* (PRINT-NATIVE-STREAM STREAM (WRAPPER-VALUE SELF)) (PRINT-NATIVE-STREAM STREAM \"|W|\" (WRAPPER-VALUE SELF))))");

      clasS->classConstructorCode = ((cpp_function_code)(&newStringLogicWrapper));
    }
    { Class* clasS = defineClassFromStringifiedSource("CHECK-TYPES-RECORD", "(DEFCLASS CHECK-TYPES-RECORD (STANDARD-OBJECT) :SLOTS ((PARENT-PROPOSITION :TYPE OBJECT) (PARENT-DESCRIPTION :TYPE DESCRIPTION) (MODULE :TYPE MODULE)))");

      clasS->classConstructorCode = ((cpp_function_code)(&newCheckTypesRecord));
      clasS->classSlotAccessorCode = ((cpp_function_code)(&accessCheckTypesRecordSlotValue));
    }
  }
}

void helpStartupPropositions8() {
  {
    defineFunctionObject("LOG-LOGIC-OBJECT", "(DEFUN LOG-LOGIC-OBJECT ((SELF LOGIC-OBJECT)))", ((cpp_function_code)(&logLogicObject)), NULL);
    defineFunctionObject("STRINGIFIED-SOURCE", "(DEFUN (STRINGIFIED-SOURCE STRING) ((SELF OBJECT)))", ((cpp_function_code)(&stringifiedSource)), NULL);
    defineFunctionObject("STRINGIFIED-SOURCE-SETTER", "(DEFUN (STRINGIFIED-SOURCE-SETTER STRING) ((SELF OBJECT) (VALUE STRING)))", ((cpp_function_code)(&stringifiedSourceSetter)), NULL);
    defineFunctionObject("ORIGINATED-PROPOSITIONS", "(DEFUN (ORIGINATED-PROPOSITIONS (LIST OF PROPOSITION)) ((SELF OBJECT)))", ((cpp_function_code)(&originatedPropositions)), NULL);
    defineFunctionObject("ORIGINATED-PROPOSITIONS-SETTER", "(DEFUN (ORIGINATED-PROPOSITIONS-SETTER (LIST OF PROPOSITION)) ((SELF OBJECT) (VALUE (LIST OF PROPOSITION))))", ((cpp_function_code)(&originatedPropositionsSetter)), NULL);
    defineFunctionObject("CREATE-PROPOSITION", "(DEFUN (CREATE-PROPOSITION PROPOSITION) ((KIND SYMBOL) (ARGUMENTCOUNT INTEGER)))", ((cpp_function_code)(&createProposition)), NULL);
    defineFunctionObject("ENFORCE-CODE-ONLY", "(DEFUN ENFORCE-CODE-ONLY ())", ((cpp_function_code)(&enforceCodeOnly)), NULL);
    defineExternalSlotFromStringifiedSource("(DEFSLOT CONTEXT MONOTONIC? :TYPE BOOLEAN :PUBLIC? TRUE :ALLOCATION :DYNAMIC)");
    defineExternalSlotFromStringifiedSource("(DEFSLOT WORLD TRUTH-MAINTAINED? :TYPE BOOLEAN :PUBLIC? TRUE :ALLOCATION :DYNAMIC)");
    defineExternalSlotFromStringifiedSource("(DEFSLOT WORLD UP-TO-DATE-INFERENCES? :TYPE BOOLEAN :PUBLIC? TRUE :ALLOCATION :DYNAMIC)");
    defineFunctionObject("LOGIC-MODULE?", "(DEFUN (LOGIC-MODULE? BOOLEAN) ((SELF MODULE)) :DOCUMENTATION \"Return TRUE if 'self' is a logic module, implying\nthat relations defined within it define a knowledge base.  A module\nis a logic module iff it inherits the module 'PL-KERNEL'.\")", ((cpp_function_code)(&logicModuleP)), NULL);
    defineExternalSlotFromStringifiedSource("(DEFSLOT MODULE LOCALLY-CONCEIVED-PROPOSITIONS-INTERNAL :TYPE PROPOSITIONS-INDEX :ALLOCATION :DYNAMIC)");
    defineFunctionObject("LOCALLY-CONCEIVED-PROPOSITIONS", "(DEFUN (LOCALLY-CONCEIVED-PROPOSITIONS PROPOSITIONS-INDEX) ((SELF MODULE)))", ((cpp_function_code)(&locallyConceivedPropositions)), NULL);
    defineFunctionObject("LOCALLY-CONCEIVED-PROPOSITIONS-SETTER", "(DEFUN LOCALLY-CONCEIVED-PROPOSITIONS-SETTER ((SELF MODULE) (VALUE PROPOSITIONS-INDEX)))", ((cpp_function_code)(&locallyConceivedPropositionsSetter)), NULL);
    defineExternalSlotFromStringifiedSource("(DEFSLOT CONTEXT LOCALLY-CONCEIVED-INSTANCES-INTERNAL :TYPE INSTANCES-INDEX :ALLOCATION :DYNAMIC)");
    defineFunctionObject("LOCALLY-CONCEIVED-INSTANCES", "(DEFUN (LOCALLY-CONCEIVED-INSTANCES INSTANCES-INDEX) ((SELF CONTEXT)))", ((cpp_function_code)(&locallyConceivedInstances)), NULL);
    defineFunctionObject("LOCALLY-CONCEIVED-INSTANCES-SETTER", "(DEFUN LOCALLY-CONCEIVED-INSTANCES-SETTER ((SELF CONTEXT) (VALUE INSTANCES-INDEX)))", ((cpp_function_code)(&locallyConceivedInstancesSetter)), NULL);
    defineFunctionObject("INCREMENT-NOW-TIMESTAMP", "(DEFUN INCREMENT-NOW-TIMESTAMP ())", ((cpp_function_code)(&incrementNowTimestamp)), NULL);
    defineFunctionObject("GET-NOW-TIMESTAMP", "(DEFUN (GET-NOW-TIMESTAMP TIMESTAMP) ())", ((cpp_function_code)(&getNowTimestamp)), NULL);
    defineFunctionObject("UPDATE-NOW-TIMESTAMP", "(DEFUN UPDATE-NOW-TIMESTAMP ((KBACTION KEYWORD)))", ((cpp_function_code)(&updateNowTimestamp)), NULL);
    defineFunctionObject("BOTTOM?", "(DEFUN (BOTTOM? BOOLEAN) ((SELF OBJECT)) :PUBLIC? TRUE :DOCUMENTATION \"Return TRUE if 'self' is the undefined individual BOTTOM.\")", ((cpp_function_code)(&bottomP)), NULL);
    defineFunctionObject("SUBRELATION-OF?", "(DEFUN (SUBRELATION-OF? BOOLEAN) ((DESC1 NAMED-DESCRIPTION) (DESC2 NAMED-DESCRIPTION)) :PUBLIC? TRUE)", ((cpp_function_code)(&subrelationOfP)), NULL);
    defineFunctionObject("LOGICAL-SUBTYPE-OF?", "(DEFUN (LOGICAL-SUBTYPE-OF? BOOLEAN) ((TYPE1 SURROGATE) (TYPE2 SURROGATE)) :PUBLIC? TRUE)", ((cpp_function_code)(&logicalSubtypeOfP)), NULL);
    defineFunctionObject("LOGICAL-SUBTYPE-OF-LITERAL?", "(DEFUN (LOGICAL-SUBTYPE-OF-LITERAL? BOOLEAN) ((TYPE SURROGATE)) :PUBLIC? TRUE)", ((cpp_function_code)(&logicalSubtypeOfLiteralP)), NULL);
    defineFunctionObject("LITERAL-TYPE?", "(DEFUN (LITERAL-TYPE? BOOLEAN) ((TYPE SURROGATE)) :PUBLIC? TRUE)", ((cpp_function_code)(&literalTypeP)), NULL);
    defineFunctionObject("BOOLEAN-TYPE?", "(DEFUN (BOOLEAN-TYPE? BOOLEAN) ((SELF TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&booleanTypeP)), NULL);
    defineFunctionObject("CLASS-DESCRIPTION?", "(DEFUN (CLASS-DESCRIPTION? BOOLEAN) ((SELF NAMED-DESCRIPTION)) :PUBLIC? TRUE)", ((cpp_function_code)(&classDescriptionP)), NULL);
    defineFunctionObject("FUNCTION-DESCRIPTION?", "(DEFUN (FUNCTION-DESCRIPTION? BOOLEAN) ((SELF NAMED-DESCRIPTION)) :PUBLIC? TRUE)", ((cpp_function_code)(&functionDescriptionP)), NULL);
    defineFunctionObject("VARIABLE-ARITY?", "(DEFUN (VARIABLE-ARITY? BOOLEAN) ((SELF DESCRIPTION)) :PUBLIC? TRUE)", ((cpp_function_code)(&variableArityP)), NULL);
    defineMethodObject("(DEFMETHOD (ARITY INTEGER) ((SELF NAMED-DESCRIPTION)) :PUBLIC? TRUE)", ((cpp_method_code)(&NamedDescription::arity)), ((cpp_method_code)(NULL)));
    defineMethodObject("(DEFMETHOD (ARITY INTEGER) ((SELF DESCRIPTION)) :PUBLIC? TRUE)", ((cpp_method_code)(&Description::arity)), ((cpp_method_code)(NULL)));
    defineFunctionObject("MONADIC?", "(DEFUN (MONADIC? BOOLEAN) ((SELF OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&monadicP)), NULL);
    defineFunctionObject("COMPUTED-RELATION?", "(DEFUN (COMPUTED-RELATION? BOOLEAN) ((DESCRIPTION NAMED-DESCRIPTION)) :PUBLIC? TRUE)", ((cpp_function_code)(&computedRelationP)), NULL);
    defineFunctionObject("LAST-ARGUMENT", "(DEFUN (LAST-ARGUMENT OBJECT) ((PROPOSITION PROPOSITION)) :PUBLIC? TRUE :GLOBALLY-INLINE? TRUE (RETURN (NTH (ARGUMENTS PROPOSITION) (1- (LENGTH (ARGUMENTS PROPOSITION))))))", ((cpp_function_code)(&lastArgument)), NULL);
    defineFunctionObject("OBJECT-SURROGATE", "(DEFUN (OBJECT-SURROGATE SURROGATE) ((SELF OBJECT)) :DOCUMENTATION \"Return the surrogate naming the logic object `self'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&objectSurrogate)), NULL);
    defineFunctionObject("OBJECT-SURROGATE-SETTER", "(DEFUN (OBJECT-SURROGATE-SETTER SURROGATE) ((SELF OBJECT) (NAME SURROGATE)) :DOCUMENTATION \"Return the name of the logic object `self' to `name'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&objectSurrogateSetter)), NULL);
    defineFunctionObject("OBJECT-NAME", "(DEFUN (OBJECT-NAME SYMBOL) ((SELF OBJECT)) :DOCUMENTATION \"Return the name symbol for the logic object `self'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&objectName)), NULL);
    defineFunctionObject("OBJECT-NAME-STRING", "(DEFUN (OBJECT-NAME-STRING STRING) ((SELF OBJECT)) :DOCUMENTATION \"Return the name string for the logic object `self'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&objectNameString)), NULL);
    defineFunctionObject("OBJECT-STRING-NAME", "(DEFUN (OBJECT-STRING-NAME STRING) ((SELF OBJECT)))", ((cpp_function_code)(&objectStringName)), NULL);
    defineFunctionObject("DESCRIPTION-MODE?", "(DEFUN (DESCRIPTION-MODE? BOOLEAN) ())", ((cpp_function_code)(&descriptionModeP)), NULL);
    defineFunctionObject("NATURAL-DEDUCTION-MODE?", "(DEFUN (NATURAL-DEDUCTION-MODE? BOOLEAN) () :DOCUMENTATION \"True if normalization is governed by natural\ndeduction semantics.\")", ((cpp_function_code)(&naturalDeductionModeP)), NULL);
    defineFunctionObject("CONTEXT-DEPENDENT-SEARCH-MODE?", "(DEFUN (CONTEXT-DEPENDENT-SEARCH-MODE? BOOLEAN) ())", ((cpp_function_code)(&contextDependentSearchModeP)), NULL);
    defineFunctionObject("LOG-NEWLY-CONCEIVED-PROPOSITION", "(DEFUN LOG-NEWLY-CONCEIVED-PROPOSITION ((SELF MODULE) (PROPOSITION PROPOSITION)))", ((cpp_function_code)(&logNewlyConceivedProposition)), NULL);
    defineFunctionObject("CLIP-OR-CLASH-WITH-OLD-PREDICATE-VALUE", "(DEFUN CLIP-OR-CLASH-WITH-OLD-PREDICATE-VALUE ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&clipOrClashWithOldPredicateValue)), NULL);
    defineFunctionObject("RUN-TRUTH-CHANGE-DEMON", "(DEFUN RUN-TRUTH-CHANGE-DEMON ((PROPOSITION PROPOSITION) (TRUTHCHANGERELATION SURROGATE)))", ((cpp_function_code)(&runTruthChangeDemon)), NULL);
    defineFunctionObject("RUN-GOES-TRUE-DEMONS", "(DEFUN RUN-GOES-TRUE-DEMONS ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&runGoesTrueDemons)), NULL);
    defineFunctionObject("RUN-GOES-FALSE-DEMONS", "(DEFUN RUN-GOES-FALSE-DEMONS ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&runGoesFalseDemons)), NULL);
    defineFunctionObject("RUN-GOES-UNKNOWN-DEMONS", "(DEFUN RUN-GOES-UNKNOWN-DEMONS ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&runGoesUnknownDemons)), NULL);
    defineFunctionObject("UPDATE-LINKS-AND-TIMESTAMPS", "(DEFUN UPDATE-LINKS-AND-TIMESTAMPS ((PROPOSITION PROPOSITION) (UPDATEMODE KEYWORD)))", ((cpp_function_code)(&updateLinksAndTimestamps)), NULL);
    defineFunctionObject("ASSIGN-TRUTH-VALUE", "(DEFUN ASSIGN-TRUTH-VALUE ((SELF PROPOSITION) (TRUTHVALUE OBJECT)))", ((cpp_function_code)(&assignTruthValue)), NULL);
    defineFunctionObject("DEASSIGN-TRUTH-VALUE", "(DEFUN DEASSIGN-TRUTH-VALUE ((SELF PROPOSITION) (UPDATEMODE KEYWORD)))", ((cpp_function_code)(&deassignTruthValue)), NULL);
    defineFunctionObject("UPDATE-PROPOSITION-TRUTH-VALUE", "(DEFUN UPDATE-PROPOSITION-TRUTH-VALUE ((SELF PROPOSITION) (UPDATEMODE KEYWORD)))", ((cpp_function_code)(&updatePropositionTruthValue)), NULL);
    defineFunctionObject("ASSIGN-PROPOSITION-WEIGHT", "(DEFUN ASSIGN-PROPOSITION-WEIGHT ((SELF PROPOSITION) (WEIGHT PARTIAL-MATCH-SCORE)))", ((cpp_function_code)(&assignPropositionWeight)), NULL);
    defineFunctionObject("TICKLE-CONTEXT", "(DEFUN (TICKLE-CONTEXT CONTEXT) ())", ((cpp_function_code)(&tickleContext)), NULL);
    defineFunctionObject("TICKLE-INSTANCES", "(DEFUN (TICKLE-INSTANCES CONTEXT) ())", ((cpp_function_code)(&tickleInstances)), NULL);
    defineFunctionObject("REACT-TO-SKOLEM-VALUE-UPDATE", "(DEFUN REACT-TO-SKOLEM-VALUE-UPDATE ((SKOLEM SKOLEM) (OLDVALUE OBJECT) (NEWVALUE OBJECT) (TOPLEVELUPDATE? BOOLEAN)))", ((cpp_function_code)(&reactToSkolemValueUpdate)), NULL);
    defineFunctionObject("NATIVE-SLOT-PROPOSITION?", "(DEFUN (NATIVE-SLOT-PROPOSITION? BOOLEAN) ((SELF PROPOSITION)))", ((cpp_function_code)(&nativeSlotPropositionP)), NULL);
    defineFunctionObject("UPDATE-NATIVE-SLOT-PROPOSITION", "(DEFUN (UPDATE-NATIVE-SLOT-PROPOSITION (CONS OF PROPOSITION)) ((PROPOSITION PROPOSITION) (UPDATEMODE KEYWORD)))", ((cpp_function_code)(&updateNativeSlotProposition)), NULL);
    defineFunctionObject("ASSIGN-NATIVE-SLOT-VALUE", "(DEFUN ASSIGN-NATIVE-SLOT-VALUE ((SELF THING) (SLOT STORAGE-SLOT) (VALUE OBJECT)))", ((cpp_function_code)(&assignNativeSlotValue)), NULL);
    defineFunctionObject("CLEAR-NATIVE-SLOT-VALUE", "(DEFUN CLEAR-NATIVE-SLOT-VALUE ((OBJECT THING) (SLOT STORAGE-SLOT)))", ((cpp_function_code)(&clearNativeSlotValue)), NULL);
  }
}

void helpStartupPropositions9() {
  {
    defineFunctionObject("DROP-NATIVE-SLOT-VALUE", "(DEFUN DROP-NATIVE-SLOT-VALUE ((SELF THING) (SLOT STORAGE-SLOT) (VALUE OBJECT)))", ((cpp_function_code)(&dropNativeSlotValue)), NULL);
    defineFunctionObject("LOOKUP-SLOTREF-ON-INSTANCE", "(DEFUN (LOOKUP-SLOTREF-ON-INSTANCE SURROGATE) ((TERM OBJECT) (SLOTNAME SYMBOL)))", ((cpp_function_code)(&lookupSlotrefOnInstance)), NULL);
    defineFunctionObject("UNBOUND?", "(DEFUN (UNBOUND? BOOLEAN) ((SKOLEM SKOLEM)) :GLOBALLY-INLINE? TRUE :PUBLIC? TRUE (RETURN (NULL? (VARIABLE-VALUE SKOLEM))))", ((cpp_function_code)(&unboundP)), NULL);
    defineFunctionObject("VALUE-OF", "(DEFUN (VALUE-OF OBJECT) ((SELF OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&valueOf)), NULL);
    defineFunctionObject("SURROGATE.VALUE-OF", "(DEFUN (SURROGATE.VALUE-OF OBJECT) ((SELF SURROGATE)))", ((cpp_function_code)(&surrogateDvalueOf)), NULL);
    defineFunctionObject("RELATION.VALUE-OF", "(DEFUN (RELATION.VALUE-OF DESCRIPTION) ((SELF RELATION)))", ((cpp_function_code)(&relationDvalueOf)), NULL);
    defineFunctionObject("EQUIVALENT-VALUE-OF", "(DEFUN (EQUIVALENT-VALUE-OF OBJECT) ((SELF OBJECT)))", ((cpp_function_code)(&equivalentValueOf)), NULL);
    defineFunctionObject("GROUND-VALUE-OF", "(DEFUN (GROUND-VALUE-OF OBJECT) ((SELF OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&groundValueOf)), NULL);
    defineFunctionObject("SKOLEM?", "(DEFUN (SKOLEM? BOOLEAN) ((SELF OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&skolemP)), NULL);
    defineFunctionObject("VARIABLE?", "(DEFUN (VARIABLE? BOOLEAN) ((SELF OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&variableP)), NULL);
    defineFunctionObject("NATIVE-VALUE?", "(DEFUN (NATIVE-VALUE? BOOLEAN) ((SELF OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&nativeValueP)), NULL);
    defineFunctionObject("NATIVE-VALUE-OF", "(DEFUN (NATIVE-VALUE-OF OBJECT) ((SELF OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&nativeValueOf)), NULL);
    defineFunctionObject("INNERMOST-VARIABLE-OF", "(DEFUN (INNERMOST-VARIABLE-OF PATTERN-VARIABLE) ((SELF PATTERN-VARIABLE)))", ((cpp_function_code)(&innermostVariableOf)), NULL);
    defineFunctionObject("INNERMOST-OF", "(DEFUN (INNERMOST-OF OBJECT) ((SELF OBJECT)))", ((cpp_function_code)(&innermostOf)), NULL);
    defineFunctionObject("YIELD-ANONYMOUS-SKOLEM-NAME", "(DEFUN (YIELD-ANONYMOUS-SKOLEM-NAME SYMBOL) ((VARIABLE? BOOLEAN)))", ((cpp_function_code)(&yieldAnonymousSkolemName)), NULL);
    defineFunctionObject("HELP-CREATE-SKOLEM", "(DEFUN (HELP-CREATE-SKOLEM SKOLEM) ((SELF SKOLEM) (TYPE TYPE) (NAME SYMBOL) (ASSERTISA? BOOLEAN)))", ((cpp_function_code)(&helpCreateSkolem)), NULL);
    defineFunctionObject("CREATE-SKOLEM", "(DEFUN (CREATE-SKOLEM SKOLEM) ((TYPE TYPE) (NAME SYMBOL)) :PUBLIC? TRUE)", ((cpp_function_code)(&createSkolem)), NULL);
    defineFunctionObject("CREATE-VARIABLE", "(DEFUN (CREATE-VARIABLE PATTERN-VARIABLE) ((TYPE TYPE) (NAME SYMBOL) (ASSERTISA? BOOLEAN)) :PUBLIC? TRUE)", ((cpp_function_code)(&createVariable)), NULL);
    defineFunctionObject("CREATE-VARIABLE-OR-SKOLEM", "(DEFUN (CREATE-VARIABLE-OR-SKOLEM SKOLEM) ((TYPE TYPE) (NAME SYMBOL)) :PUBLIC? TRUE)", ((cpp_function_code)(&createVariableOrSkolem)), NULL);
    defineMethodObject("(DEFMETHOD (OBJECT-EQUAL? BOOLEAN) ((SELF SKOLEM) (OTHER OBJECT)))", ((cpp_method_code)(&Skolem::objectEqualP)), ((cpp_method_code)(NULL)));
    defineFunctionObject("TRUE?", "(DEFUN (TRUE? BOOLEAN) ((SELF PROPOSITION)) :DOCUMENTATION \"Return true if `self' is true (or default-true if we are\nconsidering default assertions).\" :PUBLIC? TRUE)", ((cpp_function_code)(&trueP)), NULL);
    defineFunctionObject("FALSE?", "(DEFUN (FALSE? BOOLEAN) ((SELF PROPOSITION)) :DOCUMENTATION \"Return true if `self' is false (or default-false if we are\nconsidering default assertions).\" :PUBLIC? TRUE)", ((cpp_function_code)(&falseP)), NULL);
    defineFunctionObject("UNKNOWN?", "(DEFUN (UNKNOWN? BOOLEAN) ((SELF PROPOSITION)) :DOCUMENTATION \"Return true if the truth of `self' is unknown.\" :PUBLIC? TRUE)", ((cpp_function_code)(&unknownP)), NULL);
    defineFunctionObject("INCONSISTENT?", "(DEFUN (INCONSISTENT? BOOLEAN) ((SELF PROPOSITION)) :DOCUMENTATION \"Return true if `self' is inconsistent (true and false).\" :PUBLIC? TRUE)", ((cpp_function_code)(&inconsistentP)), NULL);
    defineFunctionObject("DEFAULT-TRUE?", "(DEFUN (DEFAULT-TRUE? BOOLEAN) ((SELF PROPOSITION)) :DOCUMENTATION \"Return true if `self' is default true.\" :PUBLIC? TRUE)", ((cpp_function_code)(&defaultTrueP)), NULL);
    defineFunctionObject("DEFAULT-FALSE?", "(DEFUN (DEFAULT-FALSE? BOOLEAN) ((SELF PROPOSITION)) :DOCUMENTATION \"Return true if `self' is default false.\" :PUBLIC? TRUE)", ((cpp_function_code)(&defaultFalseP)), NULL);
    defineFunctionObject("DEFAULTIFY-TRUTH-VALUE", "(DEFUN (DEFAULTIFY-TRUTH-VALUE TRUTH-VALUE) ((TRUE? BOOLEAN) (BYDEFAULT? BOOLEAN)))", ((cpp_function_code)(&defaultifyTruthValue)), NULL);
    defineFunctionObject("CONJOIN-TRUTH-VALUES", "(DEFUN (CONJOIN-TRUTH-VALUES TRUTH-VALUE) ((TV1 TRUTH-VALUE) (TV2 TRUTH-VALUE)) :DOCUMENTATION \"Return the logical conjunction of truth values `tv1' and `tv2'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&conjoinTruthValues)), NULL);
    defineFunctionObject("DISJOIN-TRUTH-VALUES", "(DEFUN (DISJOIN-TRUTH-VALUES TRUTH-VALUE) ((TV1 TRUTH-VALUE) (TV2 TRUTH-VALUE)) :DOCUMENTATION \"Return the logical disjunction of truth values `tv1' and `tv2'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&disjoinTruthValues)), NULL);
    defineFunctionObject("INVERT-TRUTH-VALUE", "(DEFUN (INVERT-TRUTH-VALUE TRUTH-VALUE) ((SELF TRUTH-VALUE)) :DOCUMENTATION \"Return the logical negation of `self'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&invertTruthValue)), NULL);
    defineFunctionObject("WEAKEN-TRUTH-VALUE", "(DEFUN (WEAKEN-TRUTH-VALUE TRUTH-VALUE) ((TV1 TRUTH-VALUE) (TV2 TRUTH-VALUE)) :DOCUMENTATION \"If `tv2' has lesser strength than `tv1', adapt the strength of `tv1' (not\nits value!) and return the result.  Otherwise, return `tv1' unmodified.\" :PUBLIC? TRUE)", ((cpp_function_code)(&weakenTruthValue)), NULL);
    defineFunctionObject("STRENGTHEN-TRUTH-VALUE", "(DEFUN (STRENGTHEN-TRUTH-VALUE TRUTH-VALUE) ((TV1 TRUTH-VALUE) (TV2 TRUTH-VALUE)) :DOCUMENTATION \"If `tv2' has greater strength than `tv1', adapt the strength of `tv1' (not\nits value!) and return the result.  Otherwise, return `tv1' unmodified.\" :PUBLIC? TRUE)", ((cpp_function_code)(&strengthenTruthValue)), NULL);
    defineFunctionObject("TRUE-TRUTH-VALUE?", "(DEFUN (TRUE-TRUTH-VALUE? BOOLEAN) ((SELF TRUTH-VALUE)) :DOCUMENTATION \"Return TRUE if `self' represents some form of truth.\" :PUBLIC? TRUE :GLOBALLY-INLINE? TRUE (RETURN (OR (EQL? SELF TRUE-TRUTH-VALUE) (EQL? SELF DEFAULT-TRUE-TRUTH-VALUE))))", ((cpp_function_code)(&trueTruthValueP)), NULL);
    defineFunctionObject("FALSE-TRUTH-VALUE?", "(DEFUN (FALSE-TRUTH-VALUE? BOOLEAN) ((SELF TRUTH-VALUE)) :DOCUMENTATION \"Return TRUE if `self' represents some form of falsehood.\" :PUBLIC? TRUE :GLOBALLY-INLINE? TRUE (RETURN (OR (EQL? SELF FALSE-TRUTH-VALUE) (EQL? SELF DEFAULT-FALSE-TRUTH-VALUE))))", ((cpp_function_code)(&falseTruthValueP)), NULL);
    defineFunctionObject("STRICT-TRUTH-VALUE?", "(DEFUN (STRICT-TRUTH-VALUE? BOOLEAN) ((SELF TRUTH-VALUE)) :DOCUMENTATION \"Return TRUE if `self' is a strict truth value.\" :PUBLIC? TRUE :GLOBALLY-INLINE? TRUE (RETURN (OR (EQL? SELF TRUE-TRUTH-VALUE) (EQL? SELF FALSE-TRUTH-VALUE))))", ((cpp_function_code)(&strictTruthValueP)), NULL);
    defineFunctionObject("DEFAULT-TRUTH-VALUE?", "(DEFUN (DEFAULT-TRUTH-VALUE? BOOLEAN) ((SELF TRUTH-VALUE)) :DOCUMENTATION \"Return TRUE if `self' is a default truth value.\" :PUBLIC? TRUE :GLOBALLY-INLINE? TRUE (RETURN (OR (EQL? SELF DEFAULT-TRUE-TRUTH-VALUE) (EQL? SELF DEFAULT-FALSE-TRUTH-VALUE))))", ((cpp_function_code)(&defaultTruthValueP)), NULL);
    defineFunctionObject("UNKNOWN-TRUTH-VALUE?", "(DEFUN (UNKNOWN-TRUTH-VALUE? BOOLEAN) ((SELF TRUTH-VALUE)) :DOCUMENTATION \"Return TRUE if `self' represents UNKNOWN.\" :PUBLIC? TRUE :GLOBALLY-INLINE? TRUE (RETURN (OR (EQL? SELF UNKNOWN-TRUTH-VALUE) (NULL? SELF))))", ((cpp_function_code)(&unknownTruthValueP)), NULL);
    defineFunctionObject("INCONSISTENT-TRUTH-VALUE?", "(DEFUN (INCONSISTENT-TRUTH-VALUE? BOOLEAN) ((SELF TRUTH-VALUE)) :DOCUMENTATION \"Return TRUE if `self' represents INCONSISTENT.\" :PUBLIC? TRUE :GLOBALLY-INLINE? TRUE (RETURN (EQL? SELF INCONSISTENT-TRUTH-VALUE)))", ((cpp_function_code)(&inconsistentTruthValueP)), NULL);
    defineFunctionObject("PRINT-TRUTH-VALUE", "(DEFUN PRINT-TRUTH-VALUE ((SELF TRUTH-VALUE) (STREAM OUTPUT-STREAM)))", ((cpp_function_code)(&printTruthValue)), NULL);
    defineFunctionObject("SIGNAL-UNIFICATION-CLASH", "(DEFUN SIGNAL-UNIFICATION-CLASH ((TERM1 OBJECT) (TERM2 OBJECT)))", ((cpp_function_code)(&signalUnificationClash)), NULL);
    defineFunctionObject("SIGNAL-VARIABLE-VALUE-CLASH", "(DEFUN SIGNAL-VARIABLE-VALUE-CLASH ((SKOLEM SKOLEM) (VALUE1 OBJECT) (VALUE2 OBJECT)))", ((cpp_function_code)(&signalVariableValueClash)), NULL);
    defineFunctionObject("SIGNAL-TRUTH-VALUE-CLASH", "(DEFUN SIGNAL-TRUTH-VALUE-CLASH ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&signalTruthValueClash)), NULL);
    defineFunctionObject("LOGICAL-TYPE", "(DEFUN (LOGICAL-TYPE TYPE) ((SELF OBJECT)))", ((cpp_function_code)(&logicalType)), NULL);
    defineFunctionObject("SAFE-LOGICAL-TYPE", "(DEFUN (SAFE-LOGICAL-TYPE TYPE) ((SELF OBJECT)))", ((cpp_function_code)(&safeLogicalType)), NULL);
    defineFunctionObject("CLASS-LOGICAL-TYPE", "(DEFUN (CLASS-LOGICAL-TYPE TYPE) ((SELF CLASS)))", ((cpp_function_code)(&classLogicalType)), NULL);
    defineFunctionObject("UNIFY-TYPES", "(DEFUN UNIFY-TYPES ((TERM1 SKOLEM) (TERM2 OBJECT)))", ((cpp_function_code)(&unifyTypes)), NULL);
    defineFunctionObject("UNIFY-SKOLEM-AND-VALUE", "(DEFUN UNIFY-SKOLEM-AND-VALUE ((SKOLEM SKOLEM) (VALUE OBJECT)))", ((cpp_function_code)(&unifySkolemAndValue)), NULL);
    defineFunctionObject("BIND-SKOLEM-TO-VALUE", "(DEFUN BIND-SKOLEM-TO-VALUE ((FROMSKOLEM SKOLEM) (TOVALUE OBJECT) (TOPLEVELUPDATE? BOOLEAN)))", ((cpp_function_code)(&bindSkolemToValue)), NULL);
    defineFunctionObject("UNBIND-SKOLEM-VALUE", "(DEFUN UNBIND-SKOLEM-VALUE ((SKOLEM SKOLEM) (TOPLEVELUPDATE? BOOLEAN)))", ((cpp_function_code)(&unbindSkolemValue)), NULL);
    defineFunctionObject("CONSTRAIN-AS-SUBSET", "(DEFUN CONSTRAIN-AS-SUBSET ((SET1 COLLECTION) (SET2 COLLECTION)))", ((cpp_function_code)(&constrainAsSubset)), NULL);
    defineFunctionObject("EQUATE-COLLECTIONS", "(DEFUN EQUATE-COLLECTIONS ((COL1 COLLECTION) (COL2 COLLECTION)))", ((cpp_function_code)(&equateCollections)), NULL);
    defineFunctionObject("EQUATE-VALUES", "(DEFUN EQUATE-VALUES ((TERM1 OBJECT) (TERM2 OBJECT)))", ((cpp_function_code)(&equateValues)), NULL);
    defineExternalSlotFromStringifiedSource("(DEFSLOT SKOLEM CONFLICTING-DEFAULT-VALUES :TYPE (CONS OF PROPOSITION) :DEFAULT NIL :ALLOCATION :DYNAMIC)");
    defineFunctionObject("UNEQUATE-CONFLICTING-DEFAULT-VALUES?", "(DEFUN (UNEQUATE-CONFLICTING-DEFAULT-VALUES? BOOLEAN) ((NEWEQUATINGPROP PROPOSITION)))", ((cpp_function_code)(&unequateConflictingDefaultValuesP)), NULL);
    defineFunctionObject("UNEQUATE-VALUES?", "(DEFUN (UNEQUATE-VALUES? BOOLEAN) ((TERM1 OBJECT) (TERM2 OBJECT)))", ((cpp_function_code)(&unequateValuesP)), NULL);
    defineFunctionObject("BIND-TO-EQUIVALENT-VALUE", "(DEFUN BIND-TO-EQUIVALENT-VALUE ((FROM LOGIC-OBJECT) (TO DESCRIPTION)))", ((cpp_function_code)(&bindToEquivalentValue)), NULL);
    defineFunctionObject("UNBIND-EQUIVALENT-VALUE", "(DEFUN UNBIND-EQUIVALENT-VALUE ((SELF LOGIC-OBJECT)))", ((cpp_function_code)(&unbindEquivalentValue)), NULL);
    defineFunctionObject("UNLINK-EQUIVALENT-VALUE", "(DEFUN UNLINK-EQUIVALENT-VALUE ((SELF LOGIC-OBJECT)))", ((cpp_function_code)(&unlinkEquivalentValue)), NULL);
    defineFunctionObject("EQUIVALENT-COLLECTION-PRIORITY", "(DEFUN (EQUIVALENT-COLLECTION-PRIORITY INTEGER) ((SELF LOGIC-OBJECT)))", ((cpp_function_code)(&equivalentCollectionPriority)), NULL);
    defineFunctionObject("EQUIVALENCE-COLLECTIONS", "(DEFUN EQUIVALENCE-COLLECTIONS ((COLLECTION1 LOGIC-OBJECT) (COLLECTION2 LOGIC-OBJECT)))", ((cpp_function_code)(&equivalenceCollections)), NULL);
  }
}

void helpStartupPropositions10() {
  {
    defineFunctionObject("REVISE-EQUIVALENCES", "(DEFUN REVISE-EQUIVALENCES ((PROPOSITION PROPOSITION) (GOESTRUE? BOOLEAN)))", ((cpp_function_code)(&reviseEquivalences)), NULL);
    defineFunctionObject("FIND-EQUATING-PROPOSITION", "(DEFUN (FIND-EQUATING-PROPOSITION PROPOSITION) ((DIRECTOBJECT OBJECT) (INDIRECTOBJECT LOGIC-OBJECT)))", ((cpp_function_code)(&findEquatingProposition)), NULL);
    defineFunctionObject("EVALUATE-TERM", "(DEFUN (EVALUATE-TERM OBJECT) ((SELF OBJECT)))", ((cpp_function_code)(&evaluateTerm)), NULL);
    defineFunctionObject("EVALUATE-LITERAL-WRAPPER-TERM", "(DEFUN (EVALUATE-LITERAL-WRAPPER-TERM OBJECT) ((SELF LITERAL-WRAPPER)))", ((cpp_function_code)(&evaluateLiteralWrapperTerm)), NULL);
    defineFunctionObject("EVALUATE-SURROGATE-TERM", "(DEFUN (EVALUATE-SURROGATE-TERM OBJECT) ((SELF SURROGATE)))", ((cpp_function_code)(&evaluateSurrogateTerm)), NULL);
    defineFunctionObject("EVALUATE-LOGIC-OBJECT-TERM", "(DEFUN (EVALUATE-LOGIC-OBJECT-TERM OBJECT) ((SELF LOGIC-OBJECT)))", ((cpp_function_code)(&evaluateLogicObjectTerm)), NULL);
    defineFunctionObject("ANONYMOUS-VARIABLE?", "(DEFUN (ANONYMOUS-VARIABLE? BOOLEAN) ((SELF SKOLEM)))", ((cpp_function_code)(&anonymousVariableP)), NULL);
    defineFunctionObject("EVALUATE-SYMBOL-TERM", "(DEFUN (EVALUATE-SYMBOL-TERM OBJECT) ((SELF SYMBOL)))", ((cpp_function_code)(&evaluateSymbolTerm)), NULL);
    defineFunctionObject("EVALUATE-AUTOMATIC-INSTANCE", "(DEFUN (EVALUATE-AUTOMATIC-INSTANCE OBJECT) ((NAME SYMBOL)))", ((cpp_function_code)(&evaluateAutomaticInstance)), NULL);
    defineFunctionObject("SIGNAL-UNDEFINED-TERM", "(DEFUN SIGNAL-UNDEFINED-TERM ((TERM OBJECT)))", ((cpp_function_code)(&signalUndefinedTerm)), NULL);
    defineFunctionObject("SIGNAL-UNTYPED-TERM", "(DEFUN SIGNAL-UNTYPED-TERM ((TERM OBJECT)))", ((cpp_function_code)(&signalUntypedTerm)), NULL);
    defineFunctionObject("EVALUATE-CREATE-TERM", "(DEFUN (EVALUATE-CREATE-TERM OBJECT) ((TERM CONS)))", ((cpp_function_code)(&evaluateCreateTerm)), NULL);
    defineFunctionObject("CREATE", "(DEFUN (CREATE OBJECT) ((NAME GENERALIZED-SYMBOL) |&REST| (TYPE GENERALIZED-SYMBOL)) :DOCUMENTATION \"Create a logic object with name 'name' and return it.\nIf `type' is also supplied, assert that the object belongs to that type.\" :PUBLIC? TRUE :COMMAND? TRUE :EVALUATE-ARGUMENTS? FALSE)", ((cpp_function_code)(&create)), ((cpp_function_code)(&createEvaluatorWrapper)));
    defineFunctionObject("UPDATE-SKOLEM-TYPE", "(DEFUN UPDATE-SKOLEM-TYPE ((SELF SKOLEM) (TYPE TYPE)))", ((cpp_function_code)(&updateSkolemType)), NULL);
    defineFunctionObject("UPDATE-SKOLEM-TYPE-FROM-ISA-ASSERTIONS", "(DEFUN UPDATE-SKOLEM-TYPE-FROM-ISA-ASSERTIONS ((SELF PROPOSITION)))", ((cpp_function_code)(&updateSkolemTypeFromIsaAssertions)), NULL);
    defineFunctionObject("DESCRIPTION-EXTENSION<", "(DEFUN (DESCRIPTION-EXTENSION< BOOLEAN) ((DESC1 DESCRIPTION) (DESC2 DESCRIPTION)))", ((cpp_function_code)(&descriptionExtensionL)), NULL);
    defineFunctionObject("PRINT-EXTENSION-SIZES", "(DEFUN PRINT-EXTENSION-SIZES ((MODULE MODULE) (SIZECUTOFF INTEGER)) :DOCUMENTATION \"Print the extension sizes of concepts visible in `module'.\nIf `module' is NULL the current module is used.  Do not report extensions\nwith size less than `sizeCutoff' (default is 10).\" :PUBLIC? TRUE)", ((cpp_function_code)(&printExtensionSizes)), NULL);
    defineFunctionObject("TRANSFER-LOGIC-INFORMATION-FROM-RELATION-HOOK", "(DEFUN TRANSFER-LOGIC-INFORMATION-FROM-RELATION-HOOK ((RELATIONSLIST (LIST OF RELATION))))", ((cpp_function_code)(&transferLogicInformationFromRelationHook)), NULL);
    defineFunctionObject("RETRACT-RELATION-AXIOMS", "(DEFUN RETRACT-RELATION-AXIOMS ((OLDRELATION NAMED-DESCRIPTION)))", ((cpp_function_code)(&retractRelationAxioms)), NULL);
    defineFunctionObject("ASSERT-DESCRIPTION-IMPLIES-DESCRIPTION", "(DEFUN (ASSERT-DESCRIPTION-IMPLIES-DESCRIPTION PROPOSITION) ((SUB DESCRIPTION) (SUPER DESCRIPTION) (DONTUPDATE? BOOLEAN)))", ((cpp_function_code)(&assertDescriptionImpliesDescription)), NULL);
    defineFunctionObject("ASSERT-IMPLIES-LINK-BETWEEN-RELATIONS", "(DEFUN (ASSERT-IMPLIES-LINK-BETWEEN-RELATIONS PROPOSITION) ((SUB NAMED-DESCRIPTION) (SUPER NAMED-DESCRIPTION)))", ((cpp_function_code)(&assertImpliesLinkBetweenRelations)), NULL);
    defineFunctionObject("FINALIZE-SUPERRELATION-LINKS", "(DEFUN FINALIZE-SUPERRELATION-LINKS ((SELF RELATION)))", ((cpp_function_code)(&finalizeSuperrelationLinks)), NULL);
    defineFunctionObject("UPDATE-ISA-PROPOSITION", "(DEFUN (UPDATE-ISA-PROPOSITION PROPOSITION) ((SELF OBJECT) (TYPE TYPE) (UPDATEMODE KEYWORD)) :PUBLIC? TRUE)", ((cpp_function_code)(&updateIsaProposition)), NULL);
    defineFunctionObject("ASSERT-ISA-PROPOSITION", "(DEFUN (ASSERT-ISA-PROPOSITION PROPOSITION) ((SELF OBJECT) (TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&assertIsaProposition)), NULL);
    defineFunctionObject("RETRACT-ISA-PROPOSITION", "(DEFUN (RETRACT-ISA-PROPOSITION PROPOSITION) ((SELF OBJECT) (TYPE TYPE)) :PUBLIC? TRUE)", ((cpp_function_code)(&retractIsaProposition)), NULL);
    defineFunctionObject("UPDATE-PROPERTY", "(DEFUN (UPDATE-PROPERTY PROPOSITION) ((SELF LOGIC-OBJECT) (SURROGATE SURROGATE) (UPDATEMODE KEYWORD)) :PUBLIC? TRUE)", ((cpp_function_code)(&updateProperty)), NULL);
    defineFunctionObject("ASSERT-PROPERTY", "(DEFUN (ASSERT-PROPERTY PROPOSITION) ((INSTANCE LOGIC-OBJECT) (PROPERTY SURROGATE)) :PUBLIC? TRUE)", ((cpp_function_code)(&assertProperty)), NULL);
    defineFunctionObject("RETRACT-PROPERTY", "(DEFUN (RETRACT-PROPERTY PROPOSITION) ((INSTANCE LOGIC-OBJECT) (PROPERTY SURROGATE)) :PUBLIC? TRUE)", ((cpp_function_code)(&retractProperty)), NULL);
    defineFunctionObject("UPDATE-TUPLE", "(DEFUN (UPDATE-TUPLE PROPOSITION) ((RELATION SURROGATE) (ARGUMENTS (CONS OF OBJECT)) (UPDATEMODE KEYWORD)) :PUBLIC? TRUE)", ((cpp_function_code)(&updateTuple)), NULL);
    defineFunctionObject("ASSERT-TUPLE", "(DEFUN (ASSERT-TUPLE PROPOSITION) ((RELATION SURROGATE) (ARGUMENTS (CONS OF OBJECT))) :PUBLIC? TRUE)", ((cpp_function_code)(&assertTuple)), NULL);
    defineFunctionObject("UPDATE-BINARY-VALUE", "(DEFUN (UPDATE-BINARY-VALUE PROPOSITION) ((RELATION SURROGATE) (SELF OBJECT) (VALUE OBJECT) (UPDATEMODE KEYWORD)) :PUBLIC? TRUE)", ((cpp_function_code)(&updateBinaryValue)), NULL);
    defineFunctionObject("ASSERT-BINARY-VALUE", "(DEFUN (ASSERT-BINARY-VALUE PROPOSITION) ((RELATION SURROGATE) (SELF OBJECT) (VALUE OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&assertBinaryValue)), NULL);
    defineFunctionObject("RETRACT-BINARY-VALUE", "(DEFUN (RETRACT-BINARY-VALUE PROPOSITION) ((RELATION SURROGATE) (SELF OBJECT) (VALUE OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&retractBinaryValue)), NULL);
    defineFunctionObject("ASSERT-MEMBER-OF-PROPOSITION", "(DEFUN (ASSERT-MEMBER-OF-PROPOSITION PROPOSITION) ((SELF OBJECT) (COLLECTION OBJECT)) :PUBLIC? TRUE)", ((cpp_function_code)(&assertMemberOfProposition)), NULL);
    defineFunctionObject("CREATE-LOGIC-INSTANCE", "(DEFUN (CREATE-LOGIC-INSTANCE OBJECT) ((NAME SURROGATE) (TYPE SURROGATE)) :PUBLIC? TRUE)", ((cpp_function_code)(&createLogicInstance)), NULL);
    defineFunctionObject("FIND-EXACT-DUPLICATE-PROPOSITION", "(DEFUN (FIND-EXACT-DUPLICATE-PROPOSITION PROPOSITION) ((SELF PROPOSITION)))", ((cpp_function_code)(&findExactDuplicateProposition)), NULL);
    defineFunctionObject("FIND-DUPLICATE-PROPOSITION", "(DEFUN (FIND-DUPLICATE-PROPOSITION PROPOSITION) ((SELF PROPOSITION)))", ((cpp_function_code)(&findDuplicateProposition)), NULL);
    defineFunctionObject("FASTEN-DOWN-ONE-PROPOSITION", "(DEFUN (FASTEN-DOWN-ONE-PROPOSITION PROPOSITION) ((SELF PROPOSITION) (DONTCHECKFORDUPLICATES? BOOLEAN)))", ((cpp_function_code)(&fastenDownOneProposition)), NULL);
    defineFunctionObject("HELP-FASTEN-DOWN-PROPOSITIONS", "(DEFUN (HELP-FASTEN-DOWN-PROPOSITIONS PROPOSITION) ((SELF PROPOSITION) (DONTCHECKFORDUPLICATES? BOOLEAN)))", ((cpp_function_code)(&helpFastenDownPropositions)), NULL);
    defineFunctionObject("RECURSIVELY-FASTEN-DOWN-PROPOSITIONS", "(DEFUN (RECURSIVELY-FASTEN-DOWN-PROPOSITIONS PROPOSITION) ((SELF PROPOSITION) (DONTCHECKFORDUPLICATES? BOOLEAN)))", ((cpp_function_code)(&recursivelyFastenDownPropositions)), NULL);
    defineFunctionObject("UNFASTEN-PROPOSITION", "(DEFUN UNFASTEN-PROPOSITION ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&unfastenProposition)), NULL);
    defineFunctionObject("INSTANTIATE-UNDEFINED-SURROGATES", "(DEFUN INSTANTIATE-UNDEFINED-SURROGATES ((SELF PROPOSITION)))", ((cpp_function_code)(&instantiateUndefinedSurrogates)), NULL);
    defineFunctionObject("BUILD-TOP-LEVEL-PROPOSITION", "(DEFUN (BUILD-TOP-LEVEL-PROPOSITION PROPOSITION) ((TREE OBJECT) (TRUEASSERTION? BOOLEAN)))", ((cpp_function_code)(&buildTopLevelProposition)), NULL);
    defineFunctionObject("BUILD-PROPOSITION", "(DEFUN (BUILD-PROPOSITION OBJECT) ((TREE OBJECT)))", ((cpp_function_code)(&buildProposition)), NULL);
    defineFunctionObject("SURROGATE.BUILD-PROPOSITION", "(DEFUN (SURROGATE.BUILD-PROPOSITION PROPOSITION) ((SELF SURROGATE)))", ((cpp_function_code)(&surrogateDbuildProposition)), NULL);
    defineFunctionObject("SYMBOL.BUILD-PROPOSITION", "(DEFUN (SYMBOL.BUILD-PROPOSITION OBJECT) ((SYMBOL SYMBOL)))", ((cpp_function_code)(&symbolDbuildProposition)), NULL);
    defineFunctionObject("CONS.BUILD-PROPOSITION", "(DEFUN (CONS.BUILD-PROPOSITION PROPOSITION) ((TREE CONS)))", ((cpp_function_code)(&consDbuildProposition)), NULL);
    defineFunctionObject("VERIFY-NUMBER-OF-PROPOSITION-ARGUMENTS", "(DEFUN VERIFY-NUMBER-OF-PROPOSITION-ARGUMENTS ((TREE CONS) (CORRECTNUMBER INTEGER)))", ((cpp_function_code)(&verifyNumberOfPropositionArguments)), NULL);
    defineMethodObject("(DEFMETHOD (DESCRIPTION-SURROGATE TYPE) ((SELF DESCRIPTION)))", ((cpp_method_code)(&Description::descriptionSurrogate)), ((cpp_method_code)(NULL)));
    defineFunctionObject("BUILD-ISA-PROPOSITION", "(DEFUN (BUILD-ISA-PROPOSITION PROPOSITION) ((TREE CONS)))", ((cpp_function_code)(&buildIsaProposition)), NULL);
    defineFunctionObject("BUILD-MEMBER-OF-PROPOSITION", "(DEFUN (BUILD-MEMBER-OF-PROPOSITION PROPOSITION) ((TREE CONS)))", ((cpp_function_code)(&buildMemberOfProposition)), NULL);
    defineFunctionObject("MEMBER-OF-PROPOSITION?", "(DEFUN (MEMBER-OF-PROPOSITION? BOOLEAN) ((PROPOSITION PROPOSITION)) :GLOBALLY-INLINE? TRUE (RETURN (EQL? (OPERATOR PROPOSITION) /PL-KERNEL-KB/@MEMBER-OF)))", ((cpp_function_code)(&memberOfPropositionP)), NULL);
    defineFunctionObject("SUBSET-OF-PROPOSITION?", "(DEFUN (SUBSET-OF-PROPOSITION? BOOLEAN) ((PROPOSITION PROPOSITION)) :GLOBALLY-INLINE? TRUE (RETURN (EQL? (OPERATOR PROPOSITION) /PL-KERNEL-KB/@SUBSET-OF)))", ((cpp_function_code)(&subsetOfPropositionP)), NULL);
    defineFunctionObject("SIMPLE-SUBRELATION-PROPOSITION?", "(DEFUN (SIMPLE-SUBRELATION-PROPOSITION? BOOLEAN) ((PROPOSITION PROPOSITION)) :PUBLIC? TRUE)", ((cpp_function_code)(&simpleSubrelationPropositionP)), NULL);
    defineFunctionObject("EXTRACT-ANNOTATION", "(DEFUN (EXTRACT-ANNOTATION OBJECT OBJECT) ((TREE OBJECT) (OPTION KEYWORD)))", ((cpp_function_code)(&extractAnnotation)), NULL);
    defineFunctionObject("LOOKUP-ANNOTATION", "(DEFUN (LOOKUP-ANNOTATION OBJECT) ((PROPOSITION PROPOSITION) (KEY KEYWORD)))", ((cpp_function_code)(&lookupAnnotation)), NULL);
    defineFunctionObject("INSERT-ANNOTATION", "(DEFUN INSERT-ANNOTATION ((PROPOSITION PROPOSITION) (KEY KEYWORD) (VALUE OBJECT)))", ((cpp_function_code)(&insertAnnotation)), NULL);
    defineFunctionObject("REMOVE-ANNOTATION", "(DEFUN REMOVE-ANNOTATION ((PROPOSITION PROPOSITION) (KEY KEYWORD)))", ((cpp_function_code)(&removeAnnotation)), NULL);
    defineFunctionObject("BUILD-ANNOTATED-PROPOSITION", "(DEFUN (BUILD-ANNOTATED-PROPOSITION PROPOSITION) ((TREE CONS)))", ((cpp_function_code)(&buildAnnotatedProposition)), NULL);
    defineFunctionObject("FINALIZE-PROPOSITION-ANNOTATIONS", "(DEFUN FINALIZE-PROPOSITION-ANNOTATIONS ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&finalizePropositionAnnotations)), NULL);
  }
}

void helpStartupPropositions11() {
  {
    defineFunctionObject("CLEAR-PROPOSITION-ANNOTATIONS", "(DEFUN CLEAR-PROPOSITION-ANNOTATIONS ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&clearPropositionAnnotations)), NULL);
    defineFunctionObject("FUNCTION-INDUCED-EXISTS?", "(DEFUN (FUNCTION-INDUCED-EXISTS? BOOLEAN) ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&functionInducedExistsP)), NULL);
    defineFunctionObject("PREDICATE-OF-FUNCTION-INDUCED-EXISTS", "(DEFUN (PREDICATE-OF-FUNCTION-INDUCED-EXISTS PROPOSITION) ((EXISTSPROPOSITION PROPOSITION)))", ((cpp_function_code)(&predicateOfFunctionInducedExists)), NULL);
    defineFunctionObject("EMBED-NEGATION-WITHIN-FUNCTION-INDUCED-EXISTS", "(DEFUN (EMBED-NEGATION-WITHIN-FUNCTION-INDUCED-EXISTS PROPOSITION) ((EXISTSPROPOSITION PROPOSITION)))", ((cpp_function_code)(&embedNegationWithinFunctionInducedExists)), NULL);
    defineFunctionObject("BUILD-AND-OR-NOT-PROPOSITION", "(DEFUN (BUILD-AND-OR-NOT-PROPOSITION PROPOSITION) ((TREE CONS)))", ((cpp_function_code)(&buildAndOrNotProposition)), NULL);
    defineFunctionObject("FUNCTIONAL-TERM?", "(DEFUN (FUNCTIONAL-TERM? BOOLEAN) ((TERM OBJECT)))", ((cpp_function_code)(&functionalTermP)), NULL);
    defineFunctionObject("CLIP-VALUE?", "(DEFUN (CLIP-VALUE? BOOLEAN) ((TERM1 OBJECT) (TERM2 OBJECT)))", ((cpp_function_code)(&clipValueP)), NULL);
    defineFunctionObject("CREATE-EQUIVALENCE-PROPOSITION", "(DEFUN (CREATE-EQUIVALENCE-PROPOSITION PROPOSITION) ((TERM1 OBJECT) (TERM2 OBJECT)))", ((cpp_function_code)(&createEquivalenceProposition)), NULL);
    defineFunctionObject("BUILD-EQUIVALENCE-PROPOSITION", "(DEFUN (BUILD-EQUIVALENCE-PROPOSITION PROPOSITION) ((TREE CONS)))", ((cpp_function_code)(&buildEquivalenceProposition)), NULL);
    defineFunctionObject("BUILD-CONJUNCTION-OF-EQUALITY-PROPOSITIONS", "(DEFUN (BUILD-CONJUNCTION-OF-EQUALITY-PROPOSITIONS PROPOSITION) ((REMAININGTERMS CONS) (EQUIVALENCES (CONS OF PROPOSITION))))", ((cpp_function_code)(&buildConjunctionOfEqualityPropositions)), NULL);
    defineFunctionObject("BUILD-FAIL-PROPOSITION", "(DEFUN (BUILD-FAIL-PROPOSITION PROPOSITION) ((TREE CONS)))", ((cpp_function_code)(&buildFailProposition)), NULL);
    defineFunctionObject("BUILD-COLLECT-INTO-PROPOSITION", "(DEFUN (BUILD-COLLECT-INTO-PROPOSITION PROPOSITION) ((TREE CONS)))", ((cpp_function_code)(&buildCollectIntoProposition)), NULL);
    defineFunctionObject("HOLDS-OPERATOR?", "(DEFUN (HOLDS-OPERATOR? BOOLEAN) ((OPERATOR OBJECT)))", ((cpp_function_code)(&holdsOperatorP)), NULL);
    defineFunctionObject("VERIFY-EVALUABLE-OPERATOR", "(DEFUN VERIFY-EVALUABLE-OPERATOR ((OPERATOR OBJECT) (KIND KEYWORD)))", ((cpp_function_code)(&verifyEvaluableOperator)), NULL);
    defineFunctionObject("BUILD-HOLDS-PROPOSITION", "(DEFUN (BUILD-HOLDS-PROPOSITION PROPOSITION) ((TREE CONS)))", ((cpp_function_code)(&buildHoldsProposition)), NULL);
    defineFunctionObject("MOST-GENERAL-EQUIVALENT-SLOTREF", "(DEFUN (MOST-GENERAL-EQUIVALENT-SLOTREF SURROGATE) ((SURROGATE SURROGATE)))", ((cpp_function_code)(&mostGeneralEquivalentSlotref)), NULL);
    defineFunctionObject("POLYMORPHIC-RELATION?", "(DEFUN (POLYMORPHIC-RELATION? BOOLEAN) ((SELF RELATION)))", ((cpp_function_code)(&polymorphicRelationP)), NULL);
    defineFunctionObject("NON-POLYMORPHIC-PREDICATE?", "(DEFUN (NON-POLYMORPHIC-PREDICATE? BOOLEAN) ((SELF LOGIC-OBJECT)))", ((cpp_function_code)(&nonPolymorphicPredicateP)), NULL);
    defineFunctionObject("EVALUATE-PREDICATE", "(DEFUN (EVALUATE-PREDICATE LOGIC-OBJECT) ((NAME GENERALIZED-SYMBOL) (FIRSTARGUMENT OBJECT)))", ((cpp_function_code)(&evaluatePredicate)), NULL);
    defineFunctionObject("PROPOSITIONAL-ARGUMENT?", "(DEFUN (PROPOSITIONAL-ARGUMENT? BOOLEAN) ((ARGUMENT OBJECT)))", ((cpp_function_code)(&propositionalArgumentP)), NULL);
    defineFunctionObject("EVALUATE-TYPED-ARGUMENT", "(DEFUN (EVALUATE-TYPED-ARGUMENT OBJECT) ((ARGUMENT OBJECT) (TYPE SURROGATE)))", ((cpp_function_code)(&evaluateTypedArgument)), NULL);
    defineFunctionObject("EVALUATE-FIRST-ARGUMENT", "(DEFUN (EVALUATE-FIRST-ARGUMENT OBJECT) ((ARGUMENT OBJECT) (RELATIONNAME SYMBOL)))", ((cpp_function_code)(&evaluateFirstArgument)), NULL);
    defineFunctionObject("EVALUATE-REMAINING-ARGUMENTS", "(DEFUN (EVALUATE-REMAINING-ARGUMENTS CONS) ((PREDICATEVALUE LOGIC-OBJECT) (ARGUMENTS CONS)))", ((cpp_function_code)(&evaluateRemainingArguments)), NULL);
    defineFunctionObject("CLASS-OPERATOR?", "(DEFUN (CLASS-OPERATOR? BOOLEAN) ((OPERATOR GENERALIZED-SYMBOL)))", ((cpp_function_code)(&classOperatorP)), NULL);
    defineFunctionObject("BUILD-PREDICATE-PROPOSITION", "(DEFUN (BUILD-PREDICATE-PROPOSITION PROPOSITION) ((TREE CONS)))", ((cpp_function_code)(&buildPredicateProposition)), NULL);
    defineFunctionObject("CHECK-TYPE?", "(DEFUN (CHECK-TYPE? BOOLEAN OBJECT) ((SELF OBJECT) (TYPE SURROGATE) (SHALLOW? BOOLEAN)))", ((cpp_function_code)(&checkTypeP)), NULL);
    defineFunctionObject("COERCE-INCOMPATIBLE-VALUE", "(DEFUN (COERCE-INCOMPATIBLE-VALUE OBJECT) ((VALUE OBJECT) (TYPE SURROGATE)))", ((cpp_function_code)(&coerceIncompatibleValue)), NULL);
    defineFunctionObject("COERCE-QUANTITY", "(DEFUN (COERCE-QUANTITY OBJECT) ((VALUE OBJECT) (QUANTITYTYPEREF SURROGATE)))", ((cpp_function_code)(&coerceQuantity)), NULL);
    defineFunctionObject("COERCE-DATE", "(DEFUN (COERCE-DATE OBJECT) ((VALUE OBJECT) (DATETYPEREF SURROGATE)))", ((cpp_function_code)(&coerceDate)), NULL);
    defineFunctionObject("VERIFY-ONE-ARGUMENT-TYPE", "(DEFUN VERIFY-ONE-ARGUMENT-TYPE ((ARG OBJECT) (TYPE SURROGATE) (PROPOSITION PROPOSITION) (DESCRIPTION NAMED-DESCRIPTION)))", ((cpp_function_code)(&verifyOneArgumentType)), NULL);
    defineFunctionObject("VERIFY-ARGUMENT-TYPES-AND-COUNT", "(DEFUN VERIFY-ARGUMENT-TYPES-AND-COUNT ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&verifyArgumentTypesAndCount)), NULL);
    defineFunctionObject("HANDLE-ARGUMENT-TYPE-VIOLATION", "(DEFUN HANDLE-ARGUMENT-TYPE-VIOLATION ((PROPOSITION PROPOSITION) (ARG OBJECT) (REQUIREDTYPE SURROGATE)))", ((cpp_function_code)(&handleArgumentTypeViolation)), NULL);
    defineFunctionObject("HANDLE-ARITY-VIOLATION", "(DEFUN HANDLE-ARITY-VIOLATION ((PROPOSITION PROPOSITION) (REQUIREDARITY INTEGER)))", ((cpp_function_code)(&handleArityViolation)), NULL);
    defineFunctionObject("POST-TO-CHECK-TYPES-AGENDA", "(DEFUN POST-TO-CHECK-TYPES-AGENDA ((SELF PROPOSITION)))", ((cpp_function_code)(&postToCheckTypesAgenda)), NULL);
    defineFunctionObject("PROCESS-CHECK-TYPES-AGENDA", "(DEFUN PROCESS-CHECK-TYPES-AGENDA ())", ((cpp_function_code)(&processCheckTypesAgenda)), NULL);
    defineFunctionObject("PROCESS-DEFINITIONS", "(DEFUN PROCESS-DEFINITIONS () :PUBLIC? TRUE :COMMAND? TRUE :DOCUMENTATION \"Finish processing all definitions and assertions that have\nbeen evaluated/loaded since that last call to 'process-definitions'.\nPowerLoom defers complete processing of definitions to make it easier\nto support cyclic definitions.  Following finalization of definitions,\nthis call performs semantic validation of any assertions evaluated since\nthe last call to 'process-definitions'.  PowerLoom calls this function\ninternally before each query; the primary reason to call it explicitly\nis to force the production of any diagnostic information that results from\nthe processing and validation.\")", ((cpp_function_code)(&processDefinitions)), NULL);
    defineFunctionObject("BUILD-EXISTS-PROPOSITION", "(DEFUN (BUILD-EXISTS-PROPOSITION PROPOSITION) ((TREE CONS)))", ((cpp_function_code)(&buildExistsProposition)), NULL);
    defineFunctionObject("TOP-LEVEL-EXISTS-PROPOSITION?", "(DEFUN (TOP-LEVEL-EXISTS-PROPOSITION? BOOLEAN) ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&topLevelExistsPropositionP)), NULL);
    defineFunctionObject("COLLECT-SKOLEMIZED-EXISTS-VARIABLES", "(DEFUN (COLLECT-SKOLEMIZED-EXISTS-VARIABLES (VECTOR OF SKOLEM)) ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&collectSkolemizedExistsVariables)), NULL);
    defineFunctionObject("HELP-COLLECT-SKOLEMIZED-EXISTS-VARIABLES?", "(DEFUN (HELP-COLLECT-SKOLEMIZED-EXISTS-VARIABLES? BOOLEAN) ((PROPOSITION PROPOSITION) (RESULT (LIST OF SKOLEM))))", ((cpp_function_code)(&helpCollectSkolemizedExistsVariablesP)), NULL);
    defineFunctionObject("BUILD-FORALL-PROPOSITION", "(DEFUN (BUILD-FORALL-PROPOSITION PROPOSITION) ((TREE CONS)))", ((cpp_function_code)(&buildForallProposition)), NULL);
    defineFunctionObject("VERIFY-FORALL-PROPOSITIONS", "(DEFUN VERIFY-FORALL-PROPOSITIONS ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&verifyForallPropositions)), NULL);
    defineFunctionObject("MAYBE-SUBSTITUTE-SURROGATE-ARGUMENT", "(DEFUN (MAYBE-SUBSTITUTE-SURROGATE-ARGUMENT OBJECT) ((ARG1 OBJECT) (ARG2 OBJECT)))", ((cpp_function_code)(&maybeSubstituteSurrogateArgument)), NULL);
    defineFunctionObject("CREATE-IMPLIES-PROPOSITION", "(DEFUN (CREATE-IMPLIES-PROPOSITION PROPOSITION) ((TAIL OBJECT) (HEAD OBJECT)))", ((cpp_function_code)(&createImpliesProposition)), NULL);
    defineFunctionObject("BUILD-IMPLIES-PROPOSITION", "(DEFUN (BUILD-IMPLIES-PROPOSITION PROPOSITION) ((TREE CONS)))", ((cpp_function_code)(&buildImpliesProposition)), NULL);
    defineFunctionObject("FIND-DUPLICATE-FUNCTION-PROPOSITION", "(DEFUN (FIND-DUPLICATE-FUNCTION-PROPOSITION PROPOSITION) ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&findDuplicateFunctionProposition)), NULL);
    defineFunctionObject("UNWRAP-WRAPPED-TYPE", "(DEFUN (UNWRAP-WRAPPED-TYPE TYPE) ((TYPE TYPE)))", ((cpp_function_code)(&unwrapWrappedType)), NULL);
    defineFunctionObject("BEGIN-CREATING-FUNCTION-PROPOSITION", "(DEFUN (BEGIN-CREATING-FUNCTION-PROPOSITION PROPOSITION) ((SURROGATE GENERALIZED-SYMBOL) (INPUTARGUMENTS CONS)))", ((cpp_function_code)(&beginCreatingFunctionProposition)), NULL);
    defineFunctionObject("FINISH-CREATING-FUNCTION-PROPOSITION", "(DEFUN (FINISH-CREATING-FUNCTION-PROPOSITION PROPOSITION) ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&finishCreatingFunctionProposition)), NULL);
    defineFunctionObject("FIND-OR-CREATE-FUNCTION-PROPOSITION", "(DEFUN (FIND-OR-CREATE-FUNCTION-PROPOSITION PROPOSITION) ((PREDICATE GENERALIZED-SYMBOL) (INPUTARGUMENTS CONS)))", ((cpp_function_code)(&findOrCreateFunctionProposition)), NULL);
    defineFunctionObject("CREATE-FUNCTION-PROPOSITION", "(DEFUN (CREATE-FUNCTION-PROPOSITION PROPOSITION) ((PREDICATE GENERALIZED-SYMBOL) (INPUTARGUMENTS CONS)))", ((cpp_function_code)(&createFunctionProposition)), NULL);
    defineFunctionObject("EVALUATE-FUNCTION-TERM", "(DEFUN (EVALUATE-FUNCTION-TERM OBJECT) ((TREE CONS)))", ((cpp_function_code)(&evaluateFunctionTerm)), NULL);
    defineFunctionObject("EXTENSIONAL-INDIVIDUAL?", "(DEFUN (EXTENSIONAL-INDIVIDUAL? BOOLEAN) ((INDIVIDUAL OBJECT)))", ((cpp_function_code)(&extensionalIndividualP)), NULL);
    defineFunctionObject("FUNCTION-WITH-DEFINED-VALUE?", "(DEFUN (FUNCTION-WITH-DEFINED-VALUE? BOOLEAN) ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&functionWithDefinedValueP)), NULL);
    defineFunctionObject("CREATE-ENUMERATED-SET", "(DEFUN (CREATE-ENUMERATED-SET SKOLEM) ((SET LIST)))", ((cpp_function_code)(&createEnumeratedSet)), NULL);
    defineFunctionObject("CREATE-LOGICAL-LIST", "(DEFUN (CREATE-LOGICAL-LIST SKOLEM) ((LIST LIST)))", ((cpp_function_code)(&createLogicalList)), NULL);
    defineFunctionObject("LOGICAL-COLLECTION?", "(DEFUN (LOGICAL-COLLECTION? BOOLEAN) ((SELF OBJECT)))", ((cpp_function_code)(&logicalCollectionP)), NULL);
    defineFunctionObject("SKOLEM-DEFINED-BY-OPERATOR?", "(DEFUN (SKOLEM-DEFINED-BY-OPERATOR? BOOLEAN) ((SELF OBJECT) (OPERATOR SURROGATE)) :GLOBALLY-INLINE? TRUE (TYPECASE SELF (SKOLEM (LET ((DEFININGPROPOSITION (DEFINING-PROPOSITION SELF))) (RETURN (AND (DEFINED? DEFININGPROPOSITION) (EQL? (OPERATOR DEFININGPROPOSITION) OPERATOR))))) (OTHERWISE (RETURN FALSE))))", ((cpp_function_code)(&skolemDefinedByOperatorP)), NULL);
    defineFunctionObject("ENUMERATED-SET?", "(DEFUN (ENUMERATED-SET? BOOLEAN) ((SELF OBJECT)))", ((cpp_function_code)(&enumeratedSetP)), NULL);
    defineFunctionObject("ENUMERATED-LIST?", "(DEFUN (ENUMERATED-LIST? BOOLEAN) ((SELF OBJECT)))", ((cpp_function_code)(&enumeratedListP)), NULL);
  }
}

void helpStartupPropositions12() {
  {
    defineFunctionObject("REMOVE-TRUE-AND-FALSE", "(DEFUN (REMOVE-TRUE-AND-FALSE OBJECT BOOLEAN) ((TREE CONS)))", ((cpp_function_code)(&removeTrueAndFalse)), NULL);
    defineFunctionObject("FLATTEN-NESTED-OPERATORS", "(DEFUN (FLATTEN-NESTED-OPERATORS OBJECT BOOLEAN) ((TREE CONS)))", ((cpp_function_code)(&flattenNestedOperators)), NULL);
    defineFunctionObject("CANONICALIZE-AND-OR-TREE", "(DEFUN (CANONICALIZE-AND-OR-TREE OBJECT) ((TREE CONS)))", ((cpp_function_code)(&canonicalizeAndOrTree)), NULL);
    defineFunctionObject("CANONICALIZE-NEGATION-TREE", "(DEFUN (CANONICALIZE-NEGATION-TREE OBJECT) ((TREE CONS)))", ((cpp_function_code)(&canonicalizeNegationTree)), NULL);
    defineFunctionObject("QUANTIFIER-SYMBOL?", "(DEFUN (QUANTIFIER-SYMBOL? BOOLEAN) ((SELF SYMBOL)))", ((cpp_function_code)(&quantifierSymbolP)), NULL);
    defineFunctionObject("ALL-VARIABLES-GOAL?", "(DEFUN (ALL-VARIABLES-GOAL? BOOLEAN) ((TREE OBJECT)))", ((cpp_function_code)(&allVariablesGoalP)), NULL);
    defineFunctionObject("EQUAL-ARGUMENTS?", "(DEFUN (EQUAL-ARGUMENTS? BOOLEAN) ((GOAL1 CONS) (GOAL2 CONS)))", ((cpp_function_code)(&equalArgumentsP)), NULL);
    defineFunctionObject("EXPLODE-FORALL-TREE", "(DEFUN (EXPLODE-FORALL-TREE CONS) ((TREE CONS)))", ((cpp_function_code)(&explodeForallTree)), NULL);
    defineFunctionObject("CANONICALIZE-QUANTIFICATION-TREE", "(DEFUN (CANONICALIZE-QUANTIFICATION-TREE CONS) ((TREE CONS)))", ((cpp_function_code)(&canonicalizeQuantificationTree)), NULL);
    defineFunctionObject("CONS.CANONICALIZE-PROPOSITION-TREE", "(DEFUN (CONS.CANONICALIZE-PROPOSITION-TREE OBJECT) ((TREE CONS)))", ((cpp_function_code)(&consDcanonicalizePropositionTree)), NULL);
    defineFunctionObject("CANONICALIZE-PROPOSITION-TREE", "(DEFUN (CANONICALIZE-PROPOSITION-TREE OBJECT) ((TREE OBJECT)))", ((cpp_function_code)(&canonicalizePropositionTree)), NULL);
    defineFunctionObject("ERASE-PROPOSITION", "(DEFUN ERASE-PROPOSITION ((PROPOSITION PROPOSITION)))", ((cpp_function_code)(&eraseProposition)), NULL);
    defineFunctionObject("UPDATE-EQUIVALENCE-PROPOSITION", "(DEFUN (UPDATE-EQUIVALENCE-PROPOSITION PROPOSITION) ((PROPOSITION PROPOSITION) (UPDATEMODE KEYWORD)))", ((cpp_function_code)(&updateEquivalenceProposition)), NULL);
    defineFunctionObject("SKOLEMIZE-EXISTS-PROPOSITION", "(DEFUN (SKOLEMIZE-EXISTS-PROPOSITION PROPOSITION) ((EXISTSPROPOSITION PROPOSITION)))", ((cpp_function_code)(&skolemizeExistsProposition)), NULL);
    defineFunctionObject("HELP-UPDATE-TOP-LEVEL-PROPOSITION", "(DEFUN (HELP-UPDATE-TOP-LEVEL-PROPOSITION (CONS OF PROPOSITION)) ((PROPOSITION PROPOSITION) (UPDATEMODE KEYWORD)))", ((cpp_function_code)(&helpUpdateTopLevelProposition)), NULL);
    defineFunctionObject("INVERT-UPDATE-MODE", "(DEFUN (INVERT-UPDATE-MODE KEYWORD) ((UPDATEMODE KEYWORD)))", ((cpp_function_code)(&invertUpdateMode)), NULL);
    defineFunctionObject("DEFAULTIFY-UPDATE-MODE", "(DEFUN (DEFAULTIFY-UPDATE-MODE KEYWORD) ((UPDATEMODE KEYWORD)))", ((cpp_function_code)(&defaultifyUpdateMode)), NULL);
    defineFunctionObject("UPDATE-TOP-LEVEL-PROPOSITION", "(DEFUN (UPDATE-TOP-LEVEL-PROPOSITION (CONS OF PROPOSITION)) ((TREE OBJECT) (UPDATEMODE KEYWORD)))", ((cpp_function_code)(&updateTopLevelProposition)), NULL);
    defineFunctionObject("STANDARDIZE-PROPOSITION-TREE", "(DEFUN (STANDARDIZE-PROPOSITION-TREE OBJECT) ((TREE OBJECT)))", ((cpp_function_code)(&standardizePropositionTree)), NULL);
    defineFunctionObject("UPDATE-PROPOSITION", "(DEFUN (UPDATE-PROPOSITION (CONS OF PROPOSITION)) ((TREE OBJECT) (UPDATEMODE KEYWORD)))", ((cpp_function_code)(&updateProposition)), NULL);
    defineFunctionObject("BUILD-TOP-LEVEL-TERM", "(DEFUN (BUILD-TOP-LEVEL-TERM OBJECT) ((TREE OBJECT)))", ((cpp_function_code)(&buildTopLevelTerm)), NULL);
    defineFunctionObject("CONCEIVE-TERM", "(DEFUN (CONCEIVE-TERM OBJECT) ((TREE OBJECT)) :DOCUMENTATION \"`tree' is a term expression (a string or an s-expression),\nor is a class reference (a symbol or surrogate).  Return a (possibly \nnewly-conceived) term representing the internalized representation of that term.\" :PUBLIC? TRUE :COMMAND? TRUE :EVALUATE-ARGUMENTS? TRUE :LISP-MACRO? FALSE)", ((cpp_function_code)(&conceiveTerm)), NULL);
    defineFunctionObject("CONCEIVE-SENTENCE", "(DEFUN (CONCEIVE-SENTENCE OBJECT) ((TREE OBJECT)))", ((cpp_function_code)(&conceiveSentence)), NULL);
    defineFunctionObject("CONCEIVE-FORMULA", "(DEFUN (CONCEIVE-FORMULA OBJECT) ((TREE OBJECT)))", ((cpp_function_code)(&conceiveFormula)), NULL);
    defineFunctionObject("SMART-UPDATE-PROPOSITION", "(DEFUN (SMART-UPDATE-PROPOSITION OBJECT) ((TREE OBJECT) (UPDATEMODE KEYWORD)))", ((cpp_function_code)(&smartUpdateProposition)), NULL);
    defineFunctionObject("ASSERT", "(DEFUN (ASSERT OBJECT) ((PROPOSITION PARSE-TREE)) :DOCUMENTATION \"Assert the truth of `proposition'.  Return the asserted proposition object.\nKIF example:  \\\"(assert (happy Fred))\\\" asserts that Fred is indeed happy.\nNote that for this assertion to succeed, the relation `happy' must already\nbe defined.  If the constant `Fred' has not yet been created, it is\nautomatically created as a side-effect of calling `assert'.\" :PUBLIC? TRUE :COMMAND? TRUE :EVALUATE-ARGUMENTS? FALSE)", ((cpp_function_code)(&asserT)), NULL);
    defineFunctionObject("PRESUME", "(DEFUN (PRESUME OBJECT) ((PROPOSITION PARSE-TREE)) :DOCUMENTATION \"Presume the default truth of `proposition'.  Return the presumed\nproposition object.  KIF example:  \\\"(presume (happy Fred))\\\" states that Fred\nis most probably happy.  Note, that for this to succeed, the relation\n`happy' must already be defined (see `assert').\" :PUBLIC? TRUE :COMMAND? TRUE :EVALUATE-ARGUMENTS? FALSE)", ((cpp_function_code)(&presume)), NULL);
    defineFunctionObject("RETRACT", "(DEFUN (RETRACT OBJECT) ((PROPOSITION PARSE-TREE)) :DOCUMENTATION \"Retract the truth of `proposition'.  Return the retracted proposition\nobject.  KIF example:  \\\"(retract (happy Fred))\\\" retracts that Fred is\nhappy.  Note that for this assertion to succeed, the relation `happy' must already\nbe defined.  If the constant `Fred' has not yet been created, it is\nautomatically created as a side-effect of calling `retract'.\" :PUBLIC? TRUE :COMMAND? TRUE :EVALUATE-ARGUMENTS? FALSE)", ((cpp_function_code)(&retract)), NULL);
    defineFunctionObject("DENY", "(DEFUN (DENY OBJECT) ((PROPOSITION PARSE-TREE)) :DOCUMENTATION \"Assert the falsity of `proposition'.  Return the asserted proposition\nobject.  KIF example:  \\\"(deny (happy Fred))\\\" asserts that Fred is not happy,\nwhich could have been done equivalently by \\\"(assert (not (happy Fred)))\\\".\nNote, that for this to succeed, the relation `happy' must already be defined\n(see `assert').\" :PUBLIC? TRUE :COMMAND? TRUE :EVALUATE-ARGUMENTS? FALSE)", ((cpp_function_code)(&deny)), NULL);
    defineFunctionObject("CONCEIVE", "(DEFUN (CONCEIVE OBJECT) ((FORMULA PARSE-TREE)) :DOCUMENTATION \"Guess whether 'formula' represents a term or a sentence/proposition.\nIf we are not sure, assume its a proposition.\nIf its, a term, return its internal representation.  If a proposition,\nconstruct a proposition for 'formula' without asserting its truth value.\nReturn the conceived proposition object.  KIF example: \\\"(conceive (happy Fred))\\\"\nbuilds the proposition expressing that Fred is happy without explictly asserting\nor denying it.  Note, that for this to succeed, the relation `happy' must already\nbe defined (see `assert').  If the logic constant `Fred' has not yet been\ncreated, it is automatically created as a side-effect of calling `conceive'.\" :PUBLIC? TRUE :COMMAND? TRUE :EVALUATE-ARGUMENTS? FALSE)", ((cpp_function_code)(&conceive)), NULL);
    defineFunctionObject("UNASSERT", "(DEFUN (UNASSERT OBJECT) ((PROPOSITION PARSE-TREE)) :DOCUMENTATION \"Retract the truth, falsity or inconsistency of `proposition'.  This is a\nmore general version of `retract' that also handles falsity.  For example, if\nwe assert the proposition \\\"(not (sad Fred))\\\", and then execute the statement\n\\\"(unassert (sad Fred))\\\", the truth value of the proposition \\\"(sad Fred)\\\"\nwill be set to UNKNOWN.  If we had called `retract' in place of `unassert',\nthe proposition \\\"(sad Fred)\\\" would remain set to FALSE.   Note that for this\nunassertion to succeed, the logic constant `Fred' and the relation `sad' must\nalready be defined.\" :PUBLIC? TRUE :COMMAND? TRUE :EVALUATE-ARGUMENTS? FALSE)", ((cpp_function_code)(&unassert)), NULL);
    defineFunctionObject("UNASSERT-PROPOSITION", "(DEFUN UNASSERT-PROPOSITION ((SELF PROPOSITION)) :DOCUMENTATION \"Retract the truth, falsity or inconsistency of the proposition `self'.\" :PUBLIC? TRUE)", ((cpp_function_code)(&unassertProposition)), NULL);
    defineMethodObject("(DEFMETHOD (DELETED? BOOLEAN) ((SELF PROPOSITION)))", ((cpp_method_code)(&Proposition::deletedP)), ((cpp_method_code)(NULL)));
    defineMethodObject("(DEFMETHOD (DELETED?-SETTER BOOLEAN) ((SELF PROPOSITION) (VALUE BOOLEAN)))", ((cpp_method_code)(&Proposition::deletedPSetter)), ((cpp_method_code)(NULL)));
    defineFunctionObject("DELETED-PROPOSITION?", "(DEFUN (DELETED-PROPOSITION? BOOLEAN) ((SELF PROPOSITION)))", ((cpp_function_code)(&deletedPropositionP)), NULL);
    defineFunctionObject("DESTROY-PROPOSITION", "(DEFUN (DESTROY-PROPOSITION PROPOSITION) ((PROPOSITION PROPOSITION)) :DOCUMENTATION \"Retract and destroy the proposition 'proposition'.\nRecursively destroy all propositions that reference 'proposition'.\nAlso, destroy all satellite propositions of 'proposition'.\")", ((cpp_function_code)(&destroyProposition)), NULL);
    defineFunctionObject("DESTROY-LOGIC-INSTANCE", "(DEFUN DESTROY-LOGIC-INSTANCE ((SELF OBJECT)))", ((cpp_function_code)(&destroyLogicInstance)), NULL);
    defineFunctionObject("DESTROY-TERM", "(DEFUN DESTROY-TERM ((SELF LOGIC-OBJECT)) :DOCUMENTATION \"Destroy all propositions that reference 'self',\nand mark it as 'deleted?', thereby making it invisible within class\nextensions.  Unlink descriptions from native relations.\" :PUBLIC? TRUE)", ((cpp_function_code)(&destroyTerm)), NULL);
    defineFunctionObject("DESTROY-INSTANCE", "(DEFUN DESTROY-INSTANCE ((SELF OBJECT)) :PUBLIC? TRUE :DOCUMENTATION \"Destroy all propositions that reference 'self',\nand mark it as 'deleted?', thereby making it invisible within class\nextensions.\")", ((cpp_function_code)(&destroyInstance)), NULL);
    defineFunctionObject("DESTROY-OBJECT", "(DEFUN DESTROY-OBJECT ((SELF OBJECT)) :PUBLIC? TRUE :DOCUMENTATION \"Destroy `self' which can be a term or a proposition.  Destroy all\npropositions that reference 'self' and mark it as 'deleted?' (thereby\nmaking it invisible within class extensions).\")", ((cpp_function_code)(&destroyObject)), NULL);
    defineFunctionObject("DESTROY", "(DEFUN (DESTROY OBJECT) ((OBJECTSPEC PARSE-TREE)) :PUBLIC? TRUE :COMMAND? TRUE :EVALUATE-ARGUMENTS? FALSE :DOCUMENTATION \"Find an object or proposition as specified by `objectSpec', and destroy all\npropositions and indices that reference it.  `objectSpec' must be a name or\na parse tree that evaluates to a proposition.  Return the deleted object, or\nNULL if no matching object was found.\")", ((cpp_function_code)(&destroy)), NULL);
    defineFunctionObject("RELATION-IN-MODULE?", "(DEFUN (RELATION-IN-MODULE? BOOLEAN) ((SELF NAMED-DESCRIPTION)))", ((cpp_function_code)(&relationInModuleP)), NULL);
    defineFunctionObject("INTRODUCE-MODULE", "(DEFUN INTRODUCE-MODULE ((MODULE MODULE)) :PUBLIC? TRUE)", ((cpp_function_code)(&introduceModule)), NULL);
    defineFunctionObject("CLEAR-LOGIC-MODULE-HOOK", "(DEFUN CLEAR-LOGIC-MODULE-HOOK ((MODULE MODULE)))", ((cpp_function_code)(&clearLogicModuleHook)), NULL);
    defineFunctionObject("CLEAR-INSTANCES", "(DEFUN CLEAR-INSTANCES (|&REST| (NAME NAME)) :PUBLIC? TRUE :COMMAND? TRUE :EVALUATE-ARGUMENTS? FALSE :DOCUMENTATION \"Destroy all instances belonging to module `name' or any of its children.\nLeave meta-objects, e.g., concepts and relations, alone.\nIf no `name' is supplied, the current module will be cleared after\nconfirming with the user.\")", ((cpp_function_code)(&clearInstances)), ((cpp_function_code)(&clearInstancesEvaluatorWrapper)));
    defineFunctionObject("DO-CLEAR-INSTANCES", "(DEFUN DO-CLEAR-INSTANCES ((MODULE MODULE)) :DOCUMENTATION \"Function version of `clear-instances' that evaluates its argument.\" :PUBLIC? TRUE)", ((cpp_function_code)(&doClearInstances)), NULL);
    defineFunctionObject("DESTROY-LOGIC-CONTEXT-HOOK", "(DEFUN DESTROY-LOGIC-CONTEXT-HOOK ((SELF CONTEXT)))", ((cpp_function_code)(&destroyLogicContextHook)), NULL);
    defineFunctionObject("CONS-LESS-THAN?", "(DEFUN (CONS-LESS-THAN? BOOLEAN) ((O1 CONS) (O2 CONS)))", ((cpp_function_code)(&consLessThanP)), NULL);
    defineFunctionObject("SAFE-STRING-LESS?", "(DEFUN (SAFE-STRING-LESS? BOOLEAN) ((S1 STRING) (S2 STRING)))", ((cpp_function_code)(&safeStringLessP)), NULL);
    defineFunctionObject("OBJECT-NAME-LESS-THAN?", "(DEFUN (OBJECT-NAME-LESS-THAN? BOOLEAN) ((O1 LOGIC-OBJECT) (O2 LOGIC-OBJECT)))", ((cpp_function_code)(&objectNameLessThanP)), NULL);
    defineFunctionObject("MODULE-NAME-LESS-THAN?", "(DEFUN (MODULE-NAME-LESS-THAN? BOOLEAN) ((M1 CONTEXT) (M2 CONTEXT)))", ((cpp_function_code)(&moduleNameLessThanP)), NULL);
    defineFunctionObject("PROPOSITION-LESS-THAN?", "(DEFUN (PROPOSITION-LESS-THAN? BOOLEAN) ((P1 PROPOSITION) (P2 PROPOSITION)))", ((cpp_function_code)(&propositionLessThanP)), NULL);
    defineFunctionObject("PROPOSITION-HEAD-SORT-STRING", "(DEFUN (PROPOSITION-HEAD-SORT-STRING STRING) ((P PROPOSITION)))", ((cpp_function_code)(&propositionHeadSortString)), NULL);
    defineFunctionObject("OBJECT-SORT-STRING", "(DEFUN (OBJECT-SORT-STRING STRING) ((O OBJECT)))", ((cpp_function_code)(&objectSortString)), NULL);
    defineFunctionObject("LOGIC-FORM-LESS?", "(DEFUN (LOGIC-FORM-LESS? BOOLEAN) ((O1 OBJECT) (O2 OBJECT)) :PUBLIC? TRUE :DOCUMENTATION \"A sorting predicate for objects `o1' and `o2' that can appear in logical\nforms.  Performs a combined numeric and lexocographic sort that accounts\nfor lists, collections and propositions.  Numbers precede all other values,\n`null' follows all other values.\")", ((cpp_function_code)(&logicFormLessP)), NULL);
    defineFunctionObject("STARTUP-PROPOSITIONS", "(DEFUN STARTUP-PROPOSITIONS () :PUBLIC? TRUE)", ((cpp_function_code)(&startupPropositions)), NULL);
    { MethodSlot* function = lookupFunction(SYM_PROPOSITIONS_LOGIC_STARTUP_PROPOSITIONS);

      setDynamicSlotValue(function->dynamicSlots, SYM_PROPOSITIONS_STELLA_METHOD_STARTUP_CLASSNAME, wrapString("_Startup-Propositions"), NULL_STRING_WRAPPER);
    }
  }
}

void startupPropositions() {
  { 
    BIND_STELLA_SPECIAL(oMODULEo, Module*, getStellaModule("/LOGIC", oSTARTUP_TIME_PHASEo > 1));
    BIND_STELLA_SPECIAL(oCONTEXTo, Context*, oMODULEo.get());
    if (currentStartupTimePhaseP(2)) {
      helpStartupPropositions1();
      helpStartupPropositions2();
      helpStartupPropositions3();
      helpStartupPropositions4();
      helpStartupPropositions5();
    }
    if (currentStartupTimePhaseP(4)) {
      helpStartupPropositions6();
    }
    if (currentStartupTimePhaseP(5)) {
      helpStartupPropositions7();
    }
    if (currentStartupTimePhaseP(6)) {
      finalizeClasses();
    }
    if (currentStartupTimePhaseP(7)) {
      helpStartupPropositions8();
      helpStartupPropositions9();
      helpStartupPropositions10();
      helpStartupPropositions11();
      helpStartupPropositions12();
    }
    if (currentStartupTimePhaseP(8)) {
      finalizeSlots();
      cleanupUnfinalizedClasses();
    }
    if (currentStartupTimePhaseP(9)) {
      defineStellaGlobalVariableFromStringifiedSource("(DEFCONSTANT ZERO-VARIABLES-VECTOR VARIABLES-VECTOR (NEW VARIABLES-VECTOR :ARRAY-SIZE 0) :DOCUMENTATION \"Save space by structure-sharing zero-length variable vectors.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *PARTIAL-SUPPORT-COUNTER* INTEGER 0)");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *OPERATOR-NAME-TO-SURROGATE-TABLE* (PROPERTY-LIST OF KEYWORD SURROGATE) (NEW PROPERTY-LIST :THE-PLIST (BQUOTE (:AND /PL-KERNEL-KB/@AND :OR /PL-KERNEL-KB/@OR :NOT /PL-KERNEL-KB/@NOT :FORALL /PL-KERNEL-KB/@FORALL :EXISTS /PL-KERNEL-KB/@EXISTS :EQUIVALENT /PL-KERNEL-KB/@EQUIVALENT :FAIL /PL-KERNEL-KB/@FAIL :COLLECT-INTO /PL-KERNEL-KB/@COLLECT-INTO :IMPLIES /PL-KERNEL-KB/@SUBSET-OF))) :DOCUMENTATION \"Maps names of KIF operators to relational surrogates.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *LOGIC-MODULE* MODULE (GET-STELLA-MODULE \"LOGIC\" TRUE) :PUBLIC? TRUE)");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *PL-KERNEL-MODULE* MODULE (GET-STELLA-MODULE \"PL-KERNEL\" TRUE) :PUBLIC? TRUE)");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *NOW-TIMESTAMP* TIMESTAMP 0 :DOCUMENTATION \"The NOW time stamp is incremented whenever a series\nof one or more updates is followed by a query.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *LAST-KB-ACTION* KEYWORD :UPDATE-PROPOSITION :DOCUMENTATION \"Records whether the last KB access was a query or\nan update.  Used to determine when to increment the NOW time stamp\ncounter.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL BOTTOM LOGIC-OBJECT NULL :PUBLIC? TRUE :DOCUMENTATION \"The undefined individual.  Denotes the non-existence of\nan individual in whatever slot it occupies.\")");
      BOTTOM = createSkolem(NULL, SYM_PROPOSITIONS_LOGIC_BOTTOM);
      defineStellaGlobalVariableFromStringifiedSource("(DEFSPECIAL *EVALUATIONMODE* KEYWORD :EXTENSIONAL-ASSERTION :DOCUMENTATION \"Indicates the context for evaluating a proposition.  One\nof :DESCRIPTION, :INTENSIONAL-ASSERTION, or :EXTENSIONAL-ASSERTION.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFSPECIAL *CLIPPINGENABLED?* BOOLEAN TRUE :DOCUMENTATION \"When enabled, slot-value assertions can be retracted\nby later conflicting assertions.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFSPECIAL *NATURALDEDUCTIONMODE?* BOOLEAN TRUE :DOCUMENTATION \"When enabled, blocks normalizations that significantly\nchange the behavior of inference rules.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *CONTEXT-DEPENDENT-SEARCH-MODE?* BOOLEAN FALSE :DOCUMENTATION \"Signals that we are performing search across multiple\ncontexts.  Used to disable retraction from collections, since that increases\nthe overhead of the context mechanism.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFSPECIAL *COMPUTEDQUERY?* BOOLEAN FALSE :DOCUMENTATION \"Used to signal 'ground-value-of' that it can\ncall 'bound-to' safely.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *SKOLEM-ID-COUNTER* INTEGER 0)");
      defineStellaGlobalVariableFromStringifiedSource("(DEFSPECIAL *VARIABLEIDCOUNTER* INTEGER 0)");
      defineStellaGlobalVariableFromStringifiedSource("(DEFSPECIAL *FREESKOLEMS* CONS NULL :DOCUMENTATION \"Cons-list of top-level existentially-quantified skolems.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFCONSTANT TRUE-TRUTH-VALUE TRUTH-VALUE (NEW TRUTH-VALUE :POLARITY :TRUE :STRENGTH :STRICT :POSITIVE-SCORE 1.0) :DOCUMENTATION \"Value representing TRUE.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFCONSTANT FALSE-TRUTH-VALUE TRUTH-VALUE (NEW TRUTH-VALUE :POLARITY :FALSE :STRENGTH :STRICT :POSITIVE-SCORE -1.0) :DOCUMENTATION \"Value representing FALSE.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFCONSTANT DEFAULT-TRUE-TRUTH-VALUE TRUTH-VALUE (NEW TRUTH-VALUE :POLARITY :TRUE :STRENGTH :DEFAULT :POSITIVE-SCORE 0.8) :DOCUMENTATION \"Value representing DEFAULT-TRUE.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFCONSTANT DEFAULT-FALSE-TRUTH-VALUE TRUTH-VALUE (NEW TRUTH-VALUE :POLARITY :FALSE :STRENGTH :DEFAULT :POSITIVE-SCORE -0.8) :DOCUMENTATION \"Value representing DEFAULT-FALSE.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFCONSTANT UNKNOWN-TRUTH-VALUE TRUTH-VALUE (NEW TRUTH-VALUE :POLARITY :UNKNOWN) :DOCUMENTATION \"Value representing UNKNOWN.  Needed for those cases\nwhere we need a non-NULL truth value to represents UNKNOWN.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFCONSTANT INCONSISTENT-TRUTH-VALUE TRUTH-VALUE (NEW TRUTH-VALUE :POLARITY :INCONSISTENT :STRENGTH :STRICT) :DOCUMENTATION \"Value representing a contradiction.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFSPECIAL *SUPPRESSUNTYPEDTYPEERROR?* BOOLEAN FALSE :DOCUMENTATION \"Used by 'safe-logical-type' to ask for a type\nwithout signalling an error if none exists.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *EQUIVALENCE-COLLECTIONS?* BOOLEAN TRUE :DOCUMENTATION \"Experiment with equality reasoning on collections.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFCONSTANT TRUE-PROPOSITION PROPOSITION (NEW PROPOSITION :KIND :CONSTANT :OPERATOR @TRUE :ARGUMENTS (NEW ARGUMENTS-VECTOR :ARRAY-SIZE 0) :TRUTH-VALUE TRUE-TRUTH-VALUE))");
      defineStellaGlobalVariableFromStringifiedSource("(DEFCONSTANT FALSE-PROPOSITION PROPOSITION (NEW PROPOSITION :KIND :CONSTANT :OPERATOR @FALSE :ARGUMENTS (NEW ARGUMENTS-VECTOR :ARRAY-SIZE 0) :TRUTH-VALUE FALSE-TRUTH-VALUE))");
      SGT_PROPOSITIONS_STELLA_TRUE->surrogateValue = TRUE_PROPOSITION;
      SGT_PROPOSITIONS_STELLA_FALSE->surrogateValue = FALSE_PROPOSITION;
      defineStellaGlobalVariableFromStringifiedSource("(DEFCONSTANT ANONYMOUS-VARIABLE-NAME SYMBOL (QUOTE ?) :DOCUMENTATION \"Variables with name 'ANONYMOUS-VARIABLE-NAME' are considered\nanonymous, and are not assumed to be identical to any other variable also named\n'ANONYMOUS-VARIABLE-NAME'.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *MISSING-KEY-VALUE-LIST* KEY-VALUE-LIST (NEW KEY-VALUE-LIST) :DOCUMENTATION \"Represents a key-value list that should never be used.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFSPECIAL *AUTOMATICINSTANCETABLE* (KEY-VALUE-LIST OF SYMBOL LOGIC-OBJECT) *MISSING-KEY-VALUE-LIST* :DOCUMENTATION \"Used by 'evaluate-automatic-instance' to\nrecord current bindings of automatic instance symbols.\")");
      addHook(oREDEFINE_RELATION_HOOKSo, SYM_PROPOSITIONS_LOGIC_TRANSFER_LOGIC_INFORMATION_FROM_RELATION_HOOK);
      defineStellaGlobalVariableFromStringifiedSource("(DEFSPECIAL *DEFAULTCREATIONTYPE* SURROGATE NULL)");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *STRINGIFIED-PROPOSITION-INDEX* (STRING-HASH-TABLE OF STRING LIST) (NEW STRING-HASH-TABLE) :DOCUMENTATION \"Contains a table of propositions, indexed by their\nstringifications.  Tricky: Propositions in two\ndifferent modules could have the same stringification but be different.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFSPECIAL *DONT-CHECK-FOR-DUPLICATE-PROPOSITIONS?* BOOLEAN FALSE :DOCUMENTATION \"If TRUE never check for the existence of duplicate\npropositions when building a new proposition.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFSPECIAL *VISITEDUNFASTENEDDEFININGPROPOSITIONS* LIST NULL :DOCUMENTATION \"Used by 'recursively-fasten-down-propositions'.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *AUTO-COERCE-PROPOSITIONAL-ARGUMENTS?* BOOLEAN FALSE :DOCUMENTATION \"If TRUE, automatically coerce propositional arguments of a\nproposition, even if the corresponding argument type of the hosting relation\ndoesn't indicate that.\" :PUBLIC? TRUE)");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *TYPE-CHECK-POLICY* KEYWORD :AUTOMATICALLY-FIX-TYPE-VIOLATIONS :DOCUMENTATION \"Three policies are implemented:\n   :AUTOMATICALLY-FIX-TYPE-VIOLATIONS asserts missing types to fix type\n       violations (default),\n   :REPORT-TYPE-VIOLATIONS complains about missing or incorrect types, and\n   :IGNORE-TYPE-VIOLATIONS which disables all type checking.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFSPECIAL *TYPECHECKMODE* KEYWORD :POST-TYPE-VIOLATIONS :DOCUMENTATION \"Controls the behavior of the type-checking\nroutines in the event that a type-check fails.  Options are:\n  :POST-TYPE-VIOLATIONS              push offending proposition onto queue,\n  :REPORT-TYPE-VIOLATIONS            print occasions of failed type checks,\n  :AUTOMATICALLY-FIX-TYPE-VIOLATIONS assert missing types on propositions, and\n  :IGNORE-TYPE-VIOLATIONS            don't perform any type checking at all.\")");
      defineStellaGlobalVariableFromStringifiedSource("(DEFGLOBAL *CHECK-TYPES-AGENDA* (KEY-VALUE-LIST OF PROPOSITION CHECK-TYPES-RECORD) (NEW KEY-VALUE-LIST) :DOCUMENTATION \"List of propositions that have failed a type check,\nbut might pass once finalization is complete.\")");
      registerNativeName(SYM_PROPOSITIONS_STELLA_ASSERT, KWD_PROPOSITIONS_CPP, KWD_PROPOSITIONS_FUNCTION);
      addHook(oDEFINE_MODULE_HOOKSo, SYM_PROPOSITIONS_LOGIC_INTRODUCE_MODULE);
      addHook(oCLEAR_MODULE_HOOKSo, SYM_PROPOSITIONS_LOGIC_CLEAR_LOGIC_MODULE_HOOK);
      addHook(oDESTROY_CONTEXT_HOOKSo, SYM_PROPOSITIONS_LOGIC_DESTROY_LOGIC_CONTEXT_HOOK);
    }
  }
}

Symbol* SYM_PROPOSITIONS_LOGIC_PROPOSITIONS_INDEX = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_INSTANCES_INDEX = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_BACKLINKS_INDEX = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_DEPENDENT_PROPOSITIONS_LIST = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_DEPENDENT_ISA_PROPOSITIONS_LIST = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_PREDICATE_PROPOSITIONS_TABLE = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_LOGIC_OBJECT = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_DEPENDENT_DESCRIPTIONS = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_BADp = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_VARIABLE_VALUE_INVERSE = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_SURROGATE_VALUE_INVERSE = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_OBJECT_STRINGIFIED_SOURCE = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_OBJECT_ORIGINATED_PROPOSITIONS = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_CS_VALUE = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_DELETED_OBJECTp = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_LOGIC_THING = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_PATTERN_VARIABLE = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_SKOLEM = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_SKOLEM_TYPE = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_SKOLEM_NAME = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_VARIABLE_VALUE = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_DEFINING_PROPOSITION = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_BOUND_TO_OFFSET = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_VARIABLE_NAME = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_OBJECT_VECTOR = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_ARGUMENTS_VECTOR = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_VARIABLES_VECTOR = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_FORWARD_CHAINING_INDEX = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_FORWARD_GOAL = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_MASTER_RULE = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_IO_VARIABLES = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_QUERY_BODY = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_INPUT_BINDINGS = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_CACHE_ID = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_CONSEQUENT = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_DESCRIPTION = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_FORWARD_CHAINING_INDICES = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_DEFERRED_CONTRAPOSITIVESp = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_PROPOSITION = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_INTERNAL_VARIABLES = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_QUERY_PATTERNS = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_RELATIVE_COMPLEMENT = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_COMPLEMENT_DESCRIPTION = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_NAMED_DESCRIPTION = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_NATIVE_RELATION_BACK_POINTER = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_IO_VARIABLE_NAMES = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_IO_VARIABLE_TYPES = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_EXTENSION = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_PARTIAL_MATCH_SCORE = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_SET_OF_SUPPORT = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_TIMESTAMP = NULL;

Keyword* KWD_PROPOSITIONS_UNKNOWN = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_TRUTH_VALUE = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_POLARITY = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_STRENGTH = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_POSITIVE_SCORE = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_PARTIAL_SUPPORT = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_FACT = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_AXIOM = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_SCORE = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_ID = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_ARGUMENT_SCORES = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_PROPOSITION = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_VARIABLE_TYPEp = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_SATELLITE_PROPOSITIONS = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_UNFASTENEDp = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_WEIGHT_DELTA = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_SUPPORT = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_KIND = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_TRUTH_VALUE = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_ARGUMENTS = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_OPERATOR = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_RELATIONREF = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_DEPENDENT_PROPOSITIONS = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_PROPOSITION_ORIGINATED_PROPOSITIONS = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_PROPOSITION_STRINGIFIED_SOURCE = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_INTEGER_LOGIC_WRAPPER = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_FLOAT_LOGIC_WRAPPER = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_STRING_LOGIC_WRAPPER = NULL;

Keyword* KWD_PROPOSITIONS_AND = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_AND = NULL;

Keyword* KWD_PROPOSITIONS_OR = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_OR = NULL;

Keyword* KWD_PROPOSITIONS_NOT = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_NOT = NULL;

Keyword* KWD_PROPOSITIONS_FORALL = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_FORALL = NULL;

Keyword* KWD_PROPOSITIONS_EXISTS = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_EXISTS = NULL;

Keyword* KWD_PROPOSITIONS_EQUIVALENT = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_EQUIVALENT = NULL;

Keyword* KWD_PROPOSITIONS_FAIL = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_FAIL = NULL;

Keyword* KWD_PROPOSITIONS_COLLECT_INTO = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_COLLECT_INTO = NULL;

Keyword* KWD_PROPOSITIONS_IMPLIES = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_SUBSET_OF = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_FUNCTION = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_PREDICATE = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_ISA = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_DESCRIPTIVEp = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_CODE_ONLYp = NULL;

Keyword* KWD_PROPOSITIONS_ERROR = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_LOCALLY_CONCEIVED_PROPOSITIONS_INTERNAL = NULL;

Keyword* KWD_PROPOSITIONS_PAGING = NULL;

Keyword* KWD_PROPOSITIONS_CONTEXT_PROPOSITIONS = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_LOCALLY_CONCEIVED_INSTANCES_INTERNAL = NULL;

Keyword* KWD_PROPOSITIONS_CONTEXT_INSTANCES = NULL;

Keyword* KWD_PROPOSITIONS_UPDATE_PROPOSITION = NULL;

Keyword* KWD_PROPOSITIONS_EXECUTE_QUERY = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_BOTTOM = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_LITERAL = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_F_LOGICAL_SUBTYPE_OF_LITERALp_MEMO_TABLE_000 = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_CLASS = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_LITERAL_WRAPPER = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_BOOLEAN = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_BOOLEAN_WRAPPER = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_VARIABLE_ARITY = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_COLLECTION = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_RELATION_COMPUTATION = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_THING = NULL;

Keyword* KWD_PROPOSITIONS_EXTENSIONAL_ASSERTION = NULL;

Keyword* KWD_PROPOSITIONS_DESCRIPTION = NULL;

Keyword* KWD_PROPOSITIONS_RETRACT_TRUE = NULL;

Keyword* KWD_PROPOSITIONS_FUNCTION = NULL;

Keyword* KWD_PROPOSITIONS_ISA = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_GOES_TRUE_DEMON = NULL;

Keyword* KWD_PROPOSITIONS_PREDICATE = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_GOES_FALSE_DEMON = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_GOES_UNKNOWN_DEMON = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_MASTER_PROPOSITION = NULL;

Keyword* KWD_PROPOSITIONS_ASSERT_TRUE = NULL;

Keyword* KWD_PROPOSITIONS_PRESUME_TRUE = NULL;

Keyword* KWD_PROPOSITIONS_ASSERT_FALSE = NULL;

Keyword* KWD_PROPOSITIONS_PRESUME_FALSE = NULL;

Keyword* KWD_PROPOSITIONS_RETRACT_FALSE = NULL;

Keyword* KWD_PROPOSITIONS_ASSERT = NULL;

Keyword* KWD_PROPOSITIONS_RETRACT_INCONSISTENT = NULL;

Keyword* KWD_PROPOSITIONS_RETRACT = NULL;

Keyword* KWD_PROPOSITIONS_CONCEIVE = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_HIDDEN_RELATION = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_HIDDEN_INSTANCE = NULL;

Keyword* KWD_PROPOSITIONS_KB_UPDATE = NULL;

Keyword* KWD_PROPOSITIONS_META_KB_UPDATE = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_STORAGE_SLOT = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_SURROGATE = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_RELATION = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_DESCRIPTION = NULL;

Keyword* KWD_PROPOSITIONS_WARNING = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_EQUIVALENT_VALUE = NULL;

Keyword* KWD_PROPOSITIONS_TRUE = NULL;

Keyword* KWD_PROPOSITIONS_STRICT = NULL;

Keyword* KWD_PROPOSITIONS_FALSE = NULL;

Keyword* KWD_PROPOSITIONS_DEFAULT = NULL;

Keyword* KWD_PROPOSITIONS_INCONSISTENT = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_META_INFERENCE_CACHE = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_TRUTH_MAINTAINED_INFERENCE_CACHE = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_IOTAp = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_CLASS = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_RELATION = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_SET = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_SET = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_INTERVAL_CACHE = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_ge = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_CONFLICTING_DEFAULT_VALUES = NULL;

Keyword* KWD_PROPOSITIONS_CONSTANT = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_TRUE = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_FALSE = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_CONS = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_SETOFALL = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_THE_ONLY = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_KAPPA = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_CREATE = NULL;

Keyword* KWD_PROPOSITIONS_INTENSIONAL_ASSERTION = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_p = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_CLASS_EXTENSION = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_TRANSFER_LOGIC_INFORMATION_FROM_RELATION_HOOK = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_SUBRELATION_LINKp = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_MEMBER_OF = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_ANNOTATIONS = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_STRING_WRAPPER = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_CONSTANT = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_TRUE = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_FALSE = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_SYMBOL = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_e = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_AND = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_OR = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_NOT = NULL;

Symbol* SYM_PROPOSITIONS_PL_KERNEL_KB_MEMBER_OF = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_IMPLIES = NULL;

Symbol* SYM_PROPOSITIONS_PL_KERNEL_KB_SUBSET_OF = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_EXISTS = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_FORALL = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_FAIL = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_COLLECT_INTO = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_ABOUT = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_SLOT = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_GENERALIZED_SYMBOL = NULL;

Keyword* KWD_PROPOSITIONS_DIRECTION = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_FORWARD_ONLYp = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_BACKWARD_ONLYp = NULL;

Keyword* KWD_PROPOSITIONS_WEIGHT = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_WEIGHT = NULL;

Keyword* KWD_PROPOSITIONS_CONFIDENCE_LEVEL = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_SETOF = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_LISTOF = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_EQUIVALENT = NULL;

Symbol* SYM_PROPOSITIONS_PL_KERNEL_KB_HOLDS = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_FUNCTION = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_HOLDS = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_TABLE = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_POLYMORPHIC = NULL;

Symbol* SYM_PROPOSITIONS_PL_KERNEL_KB_VALUE = NULL;

Symbol* SYM_PROPOSITIONS_PL_KERNEL_KB_CUT = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_VALUE = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_QUANTITY = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_DATE = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_FLOAT = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_INTEGER = NULL;

Keyword* KWD_PROPOSITIONS_AUTOMATICALLY_FIX_TYPE_VIOLATIONS = NULL;

Keyword* KWD_PROPOSITIONS_POST_TYPE_VIOLATIONS = NULL;

Keyword* KWD_PROPOSITIONS_IGNORE_TYPE_VIOLATIONS = NULL;

Keyword* KWD_PROPOSITIONS_REPORT_TYPE_VIOLATIONS = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_ASSERTED_BY_TYPE_CHECKERp = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_CHECK_TYPES_RECORD = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_PARENT_PROPOSITION = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_PARENT_DESCRIPTION = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_MODULE = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_TOTAL = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_SETOF = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_LISTOF = NULL;

Surrogate* SGT_PROPOSITIONS_PL_KERNEL_KB_LIST = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_FALS = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_SOME = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_LAMBDA = NULL;

Symbol* SYM_PROPOSITIONS_PL_KERNEL_KB_SETOF = NULL;

Symbol* SYM_PROPOSITIONS_PL_KERNEL_KB_LISTOF = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_BAGOF = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_ASSERT = NULL;

Keyword* KWD_PROPOSITIONS_CPP = NULL;

Keyword* KWD_PROPOSITIONS_DELETED = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_MODULE = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_INTRODUCE_MODULE = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_CLEAR_LOGIC_MODULE_HOOK = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_MONOTONICp = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_WORLD = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_DESTROY_LOGIC_CONTEXT_HOOK = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_MUTABLE_STRING_WRAPPER = NULL;

Surrogate* SGT_PROPOSITIONS_STELLA_CONTEXT = NULL;

Surrogate* SGT_PROPOSITIONS_LOGIC_COMPUTED_PROCEDURE = NULL;

Symbol* SYM_PROPOSITIONS_LOGIC_STARTUP_PROPOSITIONS = NULL;

Symbol* SYM_PROPOSITIONS_STELLA_METHOD_STARTUP_CLASSNAME = NULL;

} // end of namespace logic
